/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2009 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-coordgen
 *
 * FIXME:Describe coordgen here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! coordgen ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <string.h>
#include "gstcoordgen.h"

GST_DEBUG_CATEGORY_STATIC (gst_coordgen_debug);
#define GST_CAT_DEFAULT gst_coordgen_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory_coord = GST_STATIC_PAD_TEMPLATE ("coord",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("coordinates-list")
    );




GST_BOILERPLATE (GstCoordgen, gst_coordgen, GstElement,
    GST_TYPE_ELEMENT);

static void gst_coordgen_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_coordgen_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps *
gst_coordgen_get_caps (GstPad *pad);
static gboolean gst_coordgen_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_coordgen_chain (GstPad * pad, GstBuffer * buf);
static gboolean
gst_coordgen_event (GstPad  *pad, GstEvent * event);
static void delete_region(GstCoordgen * filter,gint64 xstart, gint64 ystart,gint64 xend,gint64 yend);
static void create_region(GstCoordgen * filter,gint64 x, gint64 y);
/* GObject vmethod implementations */

static void
gst_coordgen_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "Coordgen",
    "Gst-Plugins-Robotix",
    "Generate Vector of coordinates with each passing buffer",
    "Hannes Bistry <<hannesb@gmx.de>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the coordgen's class */
static void
gst_coordgen_class_init (GstCoordgenClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_coordgen_set_property;
  gobject_class->get_property = gst_coordgen_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_coordgen_init (GstCoordgen * filter,
    GstCoordgenClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_coordgen_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_coordgen_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_coordgen_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_coordgen_get_caps));
  gst_pad_set_event_function (filter->srcpad,GST_DEBUG_FUNCPTR(gst_coordgen_event));

  filter->srccoord = gst_pad_new_from_static_template (&src_factory_coord, "coord");
  gst_pad_set_getcaps_function (filter->srccoord,gst_coordgen_get_caps);


  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srccoord);
  filter->silent = FALSE;
  filter->minbox=50;
}

static void
gst_coordgen_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstCoordgen *filter = GST_COORDGEN (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_coordgen_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstCoordgen *filter = GST_COORDGEN (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */
static GstCaps *
gst_coordgen_get_caps (GstPad *pad){
  g_print("Start get caps!!\n");
  GstCoordgen *filter;

  filter = GST_COORDGEN (GST_OBJECT_PARENT (pad));

  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  

}
/* this function handles the link with other elements */
static gboolean
gst_coordgen_set_caps (GstPad * pad, GstCaps * caps)
{
  GstCoordgen *filter;
  filter = GST_COORDGEN (gst_pad_get_parent (pad));
  GstPad *otherpad;
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);//get datatype of caps
  if (strstr(mime,"video")){//get borders
    gst_structure_get_int(structure,"width",&(filter->iwidth));
    gst_structure_get_int(structure,"height",&(filter->iheight)); 
    }




  filter = GST_COORDGEN (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  return gst_pad_set_caps (otherpad, caps);
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_coordgen_chain (GstPad * pad, GstBuffer * buf)
{
  GstCoordgen *filter;
  GstFlowReturn ret1,ret2;
  filter = GST_COORDGEN (GST_OBJECT_PARENT (pad));
  GstBuffer *outbuf;
  GList *temp;
  //if (filter->silent == FALSE)
    //g_print ("I'm plugged, therefore I'm in.\n");
  filter->numregions=g_list_length(filter->coordinates);
  outbuf=gst_buffer_new_and_alloc(filter->numregions*sizeof(struct coordinates));
  outbuf->timestamp=buf->timestamp;
  outbuf->duration=buf->duration;
  outbuf->offset=buf->offset;
  outbuf->offset_end=buf->offset_end;
  GstCaps *tempcaps=gst_caps_copy (gst_pad_get_pad_template_caps(filter->srccoord));
  gst_buffer_set_caps(outbuf,tempcaps);
  gst_caps_unref(tempcaps);
  temp=filter->coordinates;
  int count=0;
  while (temp){
    ((struct coordinates*)(outbuf->data))[count].x1=((struct coordinates*)(temp->data))->x1;
    ((struct coordinates*)(outbuf->data))[count].x2=((struct coordinates*)(temp->data))->x2;
    ((struct coordinates*)(outbuf->data))[count].y1=((struct coordinates*)(temp->data))->y1;
    ((struct coordinates*)(outbuf->data))[count].y2=((struct coordinates*)(temp->data))->y2;
    temp=g_list_next(temp);
    count++;
  }
  ret1=gst_pad_push (filter->srccoord, outbuf);
  /* just push out the incoming buffer without touching it */
  ret2=gst_pad_push (filter->srcpad, buf);

  if (ret1!=GST_FLOW_OK) return ret1;
  else return ret2;
}

static gboolean
gst_coordgen_event (GstPad  *pad, GstEvent * event)
{
  GstCoordgen *filter;
  gboolean ret=TRUE;
  int button;

  filter = GST_COORDGEN (gst_pad_get_parent (pad));

  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_NAVIGATION:
      g_print("Event Navigation! Event named %s, Structure Named %s, Content %s\n",gst_event_type_get_name(GST_EVENT_TYPE(event)),gst_structure_get_name(  gst_event_get_structure(event)),gst_structure_to_string(gst_event_get_structure(event)));
      g_print("Eventstring: %s",g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")));
      if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"mouse-button-press")){
        button=g_value_get_int(gst_structure_get_value(gst_event_get_structure(event),"button"));
        if (button==1){
	  filter->x_pending=(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_x"));
	  filter->y_pending=(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y"));
	  filter->state_create=TRUE;
	}
	if (button==3){
	delete_region(filter,(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_x")),(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y")),(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_x")),(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y")));
	}
      }
      if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"mouse-button-release")){
        button=g_value_get_int(gst_structure_get_value(gst_event_get_structure(event),"button"));
	if (button==1){
	  create_region(filter,(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_x")),(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y")));
	  filter->state_create=FALSE;
	}
      }
      gst_pad_push_event (filter->sinkpad, event);
      break;
    default:
      g_print("Other event\n");
      ret = gst_pad_push_event (filter->sinkpad, event);
      break;
  }
  
  gst_object_unref (filter);
  return ret;
}

static void delete_region(GstCoordgen * filter,gint64 xstart, gint64 ystart,gint64 xend,gint64 yend){
  g_print("Deleting Region from x %lld y %lld to x2 %lld y2 %lld.\n",xstart,ystart,xend,yend);
  GList *temp;
  temp=filter->coordinates;
  while (temp){
   if ((((struct coordinates*)(temp->data))->x1<xend)&&(((struct coordinates*)(temp->data))->x2>xstart)&&(((struct coordinates*)(temp->data))->y1<yend)&&(((struct coordinates*)(temp->data))->y2>ystart)){//deleting this
	g_print("Deleting Element from x %f y %f to x2 %f y2 %f.\n",((struct coordinates*)(temp->data))->x1,((struct coordinates*)(temp->data))->y1,((struct coordinates*)(temp->data))->x2,((struct coordinates*)(temp->data))->y2);
	filter->coordinates=g_list_remove(filter->coordinates,temp->data);
	temp=filter->coordinates;
	}
   else temp=g_list_next(temp);
  }
  return;
}
static void create_region(GstCoordgen * filter,gint64 x, gint64 y){
  g_print("Creating Region from x %lld y %lld to x2 %lld y2 %lld.\n",filter->x_pending,filter->y_pending,x,y);
  gint64 x1,y1,x2,y2;
//order variables and check size
  if (x>filter->x_pending){
	x1=filter->x_pending;
	x2=x;
  }
  else{
	x2=filter->x_pending;
	x1=x;	
  }
  if (y>filter->y_pending){
	y1=filter->y_pending;
	y2=y;
  }
  else{
	y2=filter->y_pending;
	y1=y;	
  }
  x2+= (1+x2-x1)%4;
  y2+= (1+y2-y1)%4;
  if (x1<0) x1=0; else if ((filter->iwidth>0)&&(filter->iwidth<x1)) x1=filter->iwidth-1;
  if (x2<0) x2=0; else if ((filter->iwidth>0)&&(filter->iwidth<x2)) x2=filter->iwidth-1;
  if (y1<0) y1=0; else if ((filter->iheight>0)&&(filter->iheight<y1)) y1=filter->iheight-1;
  if (y2<0) y2=0; else if ((filter->iheight>0)&&(filter->iheight<y2)) y2=filter->iheight-1;
  x2-= (1+x2-x1)%4;
  y2-= (1+y2-y1)%4;
  if (((x2-x1)<filter->minbox)||((y2-y1)<filter->minbox)) return;
  delete_region(filter,x1,y1,x2,y2);
  struct coordinates* tempcoord;
  tempcoord=g_malloc(sizeof(struct coordinates));
  tempcoord->x1=x1;
  tempcoord->y1=y1; 
  tempcoord->x2=x2;
  tempcoord->y2=y2;
  filter->coordinates=g_list_append(filter->coordinates,(gpointer)tempcoord);
  g_print("Finished appending\n");
  return;
}
/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
coordgen_init (GstPlugin * coordgen)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template coordgen' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_coordgen_debug, "coordgen",
      0, "Template coordgen");

  return gst_element_register (coordgen, "coordgen", GST_RANK_NONE,
      GST_TYPE_COORDGEN);
}

/* gstreamer looks for this structure to register coordgens
 *
 * exchange the string 'Template coordgen' with your coordgen description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "coordgen",
    "Template coordgen",
    coordgen_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
