/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2012 Hannes <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-dummyvalve
 *
 * FIXME:Describe dummyvalve here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! dummyvalve ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */
#define MAXSTRING 10000

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
//#include <gtk/gtk.h>
#include <gio/gio.h>
#include <gst/gst.h>
//#include <gdk/gdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gstdummyvalve.h"

GST_DEBUG_CATEGORY_STATIC (gst_dummyvalve_debug);
#define GST_CAT_DEFAULT gst_dummyvalve_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_SPORT,
  PROP_HOSTNAME,
  PROP_DROP
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
/*
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );
*/
GST_BOILERPLATE (Gstdummyvalve, gst_dummyvalve, GstElement,
    GST_TYPE_ELEMENT);

static void gst_dummyvalve_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_dummyvalve_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstStateChangeReturn gst_dummyvalve_change_state (GstElement *element, GstStateChange transition);


static gboolean gst_dummyvalve_set_caps (GstPad * pad, GstCaps * caps);
//static GstFlowReturn gst_dummyvalve_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_dummyvalve_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "dummyvalve",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes <<user@hostname.org>>");
/*
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));*/
}

/* initialize the dummyvalve's class */
static void
gst_dummyvalve_class_init (GstdummyvalveClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_dummyvalve_set_property;
  gobject_class->get_property = gst_dummyvalve_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SPORT,
      g_param_spec_int ("sendport", "Sendport", "To what port should we send Packets",0,65535,4445, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_HOSTNAME,
      g_param_spec_string ("host", "Host", "Host adress of GUI-Program",
          "127.0.0.1", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_DROP,
      g_param_spec_boolean ("drop", "Drop", "Tell the remote - propexport instance that the valve with the same name should drop",
          FALSE, G_PARAM_READWRITE));

  gstelement_class->change_state = gst_dummyvalve_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_dummyvalve_init (Gstdummyvalve * filter,
    GstdummyvalveClass * gclass)
{
  /*filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_dummyvalve_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_dummyvalve_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);*/
  filter->silent = TRUE;
  filter->remoteadress=g_strdup("127.0.0.1");
  filter->sport=4445;
  filter->firsttime=TRUE;
  filter->ssock=NULL;
  filter->drop=FALSE;
}

static void
gst_dummyvalve_sendstate(Gstdummyvalve *filter){
    GError *err = NULL;
    if (filter->firsttime){




    //now set up send socket...
    filter->ssock = g_socket_new(G_SOCKET_FAMILY_IPV6,
	                    G_SOCKET_TYPE_DATAGRAM,
	                    G_SOCKET_PROTOCOL_UDP,
	                    &err);
    filter->saddr=NULL;
    filter->saddr=g_inet_socket_address_new(g_inet_address_new_from_string(filter->remoteadress),filter->sport);
    //g_socket_connect(varsp->ssock, varsp->saddr, NULL, NULL);
    //g_socket_bind(filter->ssock,G_SOCKET_ADDRESS(filter->saddr),TRUE,&err);
    } 
   //now sending

  int toalloc=strlen(GST_ELEMENT_NAME(filter))+10;
  gchar* tosend=g_malloc0(toalloc);
  if (filter->drop)sprintf(tosend,"%s;drop;1",GST_ELEMENT_NAME(filter));
  else sprintf(tosend,"%s;drop;0",GST_ELEMENT_NAME(filter));
  int numsent;
  numsent=g_socket_send_to(filter->ssock,filter->saddr,tosend,strlen(tosend),NULL,&err);
  if (numsent!=strlen(tosend))g_print ("Error sending !\n");
  filter->firsttime=FALSE;
  return;
}
 
static void
gst_dummyvalve_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstdummyvalve *filter = GST_DUMMYVALVE (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_SPORT:
      filter->sport = g_value_get_int (value);
      break;
    case PROP_HOSTNAME:
      if (filter->remoteadress)g_free (filter->remoteadress);
      filter->remoteadress = g_value_dup_string (value);
      filter->firsttime=TRUE;
      break;
    case PROP_DROP:{
      gboolean old=filter->drop;
      filter->drop=g_value_get_boolean (value);   
      if (old!=filter->drop)gst_dummyvalve_sendstate(filter);
      break;
      }
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_dummyvalve_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstdummyvalve *filter = GST_DUMMYVALVE (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_SPORT:
      g_value_set_int (value, filter->sport);
      break;
    case PROP_HOSTNAME:
      g_value_set_string (value, filter->remoteadress);
      break;
    case PROP_DROP:
      g_value_set_boolean (value, filter->drop);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/*
static GstFlowReturn
gst_dummyvalve_chain (GstPad * pad, GstBuffer * buf)
{
  Gstdummyvalve *filter;

  filter = GST_DUMMYVALVE (GST_OBJECT_PARENT (pad));
  gchar* tosend;

  if (filter->firsttime){
    GError *err = NULL;
    filter->rsock = g_socket_new(G_SOCKET_FAMILY_IPV6,
	                    G_SOCKET_TYPE_DATAGRAM,
	                    G_SOCKET_PROTOCOL_UDP,
	                    &err);
    g_assert(err == NULL);
    filter->raddr=g_inet_socket_address_new(g_inet_address_new_any(G_SOCKET_FAMILY_IPV6),filter->rport);
    g_socket_bind(filter->rsock,G_SOCKET_ADDRESS(filter->raddr),TRUE,&err);
    g_assert(err == NULL);

    int fd = g_socket_get_fd(filter->rsock);
    filter->rchannel = g_io_channel_unix_new(fd);
    filter->source = g_io_add_watch(filter->rchannel, G_IO_IN,(GIOFunc) rcvcallback, filter);
    g_print("Set up source %d\n",filter->source);
    g_io_channel_unref(filter->rchannel);




    //now set up send socket...
    filter->ssock = g_socket_new(G_SOCKET_FAMILY_IPV6,
	                    G_SOCKET_TYPE_DATAGRAM,
	                    G_SOCKET_PROTOCOL_UDP,
	                    &err);
    filter->saddr=NULL;
    filter->saddr=g_inet_socket_address_new(g_inet_address_new_from_string(filter->remoteadress),filter->sport);
    //g_socket_connect(varsp->ssock, varsp->saddr, NULL, NULL);
    //g_socket_bind(filter->ssock,G_SOCKET_ADDRESS(filter->saddr),TRUE,&err);
    filter->uidesc=setuppropertylist(filter);
    tosend=g_strdup(filter->uidesc);
  } 

  //Generate and send property
  else tosend=chupdate(filter,FALSE);
  if (strlen(tosend)){
    GError* err;
    int numsent=0;
    g_print ("List_length: %d Tosend: %s\n",g_list_length(filter->propertylist),tosend);
    numsent=g_socket_send_to(filter->ssock,filter->saddr,tosend,strlen(tosend),NULL,&err);
    if (numsent!=strlen(tosend))g_print ("Error sending !\n");
  }
  g_free(tosend);




  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");
  filter->firsttime=FALSE;
  return gst_pad_push (filter->srcpad, buf);
}
*/


static GstStateChangeReturn
gst_dummyvalve_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstdummyvalve *filter;
  GError* err;
  filter =  GST_DUMMYVALVE(element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    default:
      break;
  }

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      g_print ("Propexport Dropping Sockets.\n");
      if (filter->ssock)g_socket_close(filter->ssock,&err);
      filter->ssock=0;
      filter->firsttime=TRUE;
      break;
    default:
      break;
  }
  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

 return ret;
}




/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
dummyvalve_init (GstPlugin * dummyvalve)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template dummyvalve' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_dummyvalve_debug, "dummyvalve",
      0, "Template dummyvalve");

  return gst_element_register (dummyvalve, "dummyvalve", GST_RANK_NONE,
      GST_TYPE_DUMMYVALVE);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstdummyvalve"
#endif

/* gstreamer looks for this structure to register dummyvalves
 *
 * exchange the string 'Template dummyvalve' with your dummyvalve description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "dummyvalve",
    "Template dummyvalve",
    dummyvalve_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
