/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2012 Hannes <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-propexport
 *
 * FIXME:Describe propexport here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! propexport ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */
#define MAXSTRING 10000

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
//#include <gtk/gtk.h>
#include <gio/gio.h>
#include <gst/gst.h>
#include <gdk/gdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gstpropexport.h"

GST_DEBUG_CATEGORY_STATIC (gst_propexport_debug);
#define GST_CAT_DEFAULT gst_propexport_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_SPORT,
  PROP_RPORT,
  PROP_INTERVAL,
  PROP_HOSTNAME,
  PROP_DESC
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gstpropexport, gst_propexport, GstElement,
    GST_TYPE_ELEMENT);

static void gst_propexport_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_propexport_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstStateChangeReturn gst_propexport_change_state (GstElement *element, GstStateChange transition);


static gboolean gst_propexport_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_propexport_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_propexport_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "propexport",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the propexport's class */
static void
gst_propexport_class_init (GstpropexportClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_propexport_set_property;
  gobject_class->get_property = gst_propexport_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SPORT,
      g_param_spec_int ("sendport", "Sendport", "To what port should we send Packets",0,65535,4444, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_RPORT,
      g_param_spec_int ("rcvport", "Rcvport", "To what port should we listen",0,65535,4445, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_INTERVAL,
      g_param_spec_int ("interval", "Interval", "Minimum interval (ms) between checking the states of properties",-1,G_MAXINT,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_HOSTNAME,
      g_param_spec_string ("host", "Host", "Host adress of GUI-Program",
          "127.0.0.1", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_DESC,
      g_param_spec_string ("exportdesc", "Exportdesc", "Export description, Elementname;Propertyname;Elementname2;Pr...",
          NULL, G_PARAM_READWRITE)); 

  gstelement_class->change_state = gst_propexport_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_propexport_init (Gstpropexport * filter,
    GstpropexportClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_propexport_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_propexport_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->remoteadress=g_strdup("127.0.0.1");
  //filter->exportdsc=g_strdup("videotestsrc0;pattern;videotestsrc0;is-live");
  filter->exportdsc=NULL;
  filter->rport=4445;
  filter->sport=4444;
  filter->interval=-1;
  filter->firsttime=TRUE;
  filter->propertylist=NULL;
  filter->width=-1;
  filter->height=-1;
  filter->rsock=NULL;
  filter->ssock=NULL;
  filter->source=0;

}
 
static void
gst_propexport_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstpropexport *filter = GST_PROPEXPORT (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_SPORT:
      filter->sport = g_value_get_int (value);
      break;
    case PROP_RPORT:
      filter->rport = g_value_get_int (value);
      break;
    case PROP_INTERVAL:
      filter->interval = g_value_get_int (value);
      break;
    case PROP_DESC:
      if (filter->exportdsc)g_free (filter->exportdsc);
      filter->exportdsc = g_value_dup_string (value);
      filter->firsttime=TRUE;
      break;
    case PROP_HOSTNAME:
      if (filter->remoteadress)g_free (filter->remoteadress);
      filter->remoteadress = g_value_dup_string (value);
      filter->firsttime=TRUE;
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_propexport_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstpropexport *filter = GST_PROPEXPORT (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_SPORT:
      g_value_set_int (value, filter->sport);
      break;
    case PROP_RPORT:
      g_value_set_int (value, filter->rport);
      break;
    case PROP_INTERVAL:
      g_value_set_int (value, filter->interval);
      break;
    case PROP_HOSTNAME:
      g_value_set_string (value, filter->remoteadress);
      break;
    case PROP_DESC:
      g_value_set_string (value, filter->exportdsc);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static gchar* setuppropertylist(Gstpropexport *filter){
  char* temp=filter->exportdsc;//the char that we use to parse
  gchar* outputstring=g_malloc0(MAXSTRING);
  if (filter->firsttime)strcat(outputstring,"UI;");//User interface
  gboolean valid=1;
  gboolean finished=0;
  while (valid && !finished && temp && strlen(temp)){
    gdouble definedmin;
    gdouble definedmax;
    gboolean defined=FALSE;
    struct property* thisproperty=g_malloc0(sizeof(struct property));
    char* thiselement=g_malloc0(MAXSTRING);
    char* thispropertyname=g_malloc0(MAXSTRING);
    char* thistempstr=g_malloc0(MAXSTRING);
    temp=deserialize_string(thiselement,temp);
    temp=deserialize_string(thispropertyname,temp);  
    if(*temp==91){
      temp=deserialize_double(&definedmin,temp+1);
      temp=deserialize_double(&definedmax,temp);
      defined=TRUE;
      temp++;
    } 
    else g_print("Dont do user definition because rest starts with %u is %s\n",(unsigned int)*temp,temp); 
    if (strlen(thispropertyname)){
      thisproperty->elpointer=gst_bin_get_by_name_recurse_up (GST_BIN(GST_ELEMENT_PARENT(filter)),thiselement);/*
      if (!thisproperty->elpointer){
	GstBin* binbin=GST_BIN(GST_ELEMENT_PARENT(filter));
	if (!binbin) g_print("Bin Failed");
	GstElementFactory *factory=gst_element_factory_find(thiselement);
	if (!factory) g_print("Factory Failed");
	if (!gst_element_factory_get_element_type)g_print("Type Failed");
	thisproperty->elpointer=gst_bin_get_by_interface(GST_BIN(GST_ELEMENT_PARENT(filter)),gst_element_factory_get_element_type(gst_element_factory_find(thiselement)));
	if (thisproperty->elpointer)g_print("and no element:(");
	sleep(10);  funzt nicht der scheiß
      }*/
      if (thisproperty->elpointer)     
	thisproperty->spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(thisproperty->elpointer)),thispropertyname);
      thisproperty->elname=g_strdup(thiselement);
      thisproperty->propertyname=g_strdup(thispropertyname);
      thisproperty->alias=g_strdup_printf("%s:%s",thiselement,thispropertyname);
      if (thisproperty->spec==NULL){
      }
      else if ((thisproperty->spec->value_type==G_TYPE_BOOLEAN)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){
        gboolean tempbool;
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&tempbool,NULL);
	thisproperty->type=BOOLEAN;
	thisproperty->active=tempbool;
	if (tempbool)sprintf(thistempstr,"%s:%s;BOOLEAN;1;",thiselement,thispropertyname);
	else sprintf(thistempstr,"%s:%s;BOOLEAN;0;",thiselement,thispropertyname);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_BOOLEAN)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){
        gboolean tempbool;
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&tempbool,NULL);
	thisproperty->type=BOOLEAN_RO;
	thisproperty->active=tempbool;
	if (tempbool)sprintf(thistempstr,"%s:%s;BOOLEAN_RO;1;",thiselement,thispropertyname);
	else sprintf(thistempstr,"%s:%s;BOOLEAN_RO;0;",thiselement,thispropertyname);
      }					
      else if ((thisproperty->spec->value_type==G_TYPE_INT)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	int min,max,cur;					
	min=G_PARAM_SPEC_INT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT(thisproperty->spec)->maximum;
	thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(int)definedmin;
	if (defined)max=(int)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%d;%d;%d;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_INT)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	int min,max,cur;					
	min=G_PARAM_SPEC_INT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT(thisproperty->spec)->maximum;
	thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(int)definedmin;
	if (defined)max=(int)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%d;%d;%d;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	unsigned int min,max,cur;					
	min=G_PARAM_SPEC_UINT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT(thisproperty->spec)->maximum;
	thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(unsigned int)definedmin;
	if (defined)max=(unsigned int)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%u;%u;%u;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	unsigned int min,max,cur;					
	min=G_PARAM_SPEC_UINT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT(thisproperty->spec)->maximum;
	thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(unsigned int)definedmin;
	if (defined)max=(unsigned int)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%u;%u;%u;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_INT64)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gint64 min,max,cur;					
	min=G_PARAM_SPEC_INT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT64(thisproperty->spec)->maximum;
	thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(gint64)definedmin;
	if (defined)max=(gint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%lld;%lld;%lld;",thiselement,thispropertyname,(long long int)min,(long long int)max,(long long int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_INT64)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gint64 min,max,cur;					
	min=G_PARAM_SPEC_INT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT64(thisproperty->spec)->maximum;
	thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(gint64)definedmin;
	if (defined)max=(gint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%lld;%lld;%lld;",thiselement,thispropertyname,(long long int)min,(long long int)max,(long long int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT64)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	guint64 min,max,cur;					
	min=G_PARAM_SPEC_UINT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT64(thisproperty->spec)->maximum;
	thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(guint64)definedmin;
	if (defined)max=(guint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%llu;%llu;%llu;",thiselement,thispropertyname,(long long unsigned int)min,(long long unsigned int)max,(long long unsigned int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT64)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	guint64 min,max,cur;					
	min=G_PARAM_SPEC_UINT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT64(thisproperty->spec)->maximum;
	thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(guint64)definedmin;
	if (defined)max=(guint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%llu;%llu;%llu;",thiselement,thispropertyname,(long long unsigned int)min,(long long unsigned int)max,(long long unsigned int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_DOUBLE)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gdouble min,max,cur;					
	min=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->maximum;
	thisproperty->type=FLOAT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=definedmin;
	if (defined)max=definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_DOUBLE)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gdouble min,max,cur;					
	min=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->maximum;
	thisproperty->type=FLOAT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=definedmin;
	if (defined)max=definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT_RO;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_FLOAT)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gfloat min,max,cur;					
	min=G_PARAM_SPEC_FLOAT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_FLOAT(thisproperty->spec)->maximum;
	thisproperty->type=FLOAT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=(float)definedmin;
	if (defined)max=(float)definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_FLOAT)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gfloat min,max,cur;					
	min=G_PARAM_SPEC_FLOAT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_FLOAT(thisproperty->spec)->maximum;
	thisproperty->type=FLOAT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=(float)definedmin;
	if (defined)max=(float)definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT_RO;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
	
      else if (G_IS_PARAM_SPEC_ENUM (thisproperty->spec)){
	gint i;
        gchar *str=NULL;
        GEnumClass *class = G_ENUM_CLASS (g_type_class_ref (thisproperty->spec->value_type));
        GValue value = { 0, };	
        g_value_init (&value, thisproperty->spec->value_type);
	g_object_get_property(G_OBJECT(thisproperty->elpointer),thispropertyname,&value);
	int current=g_value_get_enum(&value);
	sprintf(thistempstr,"%s:%s;MENU;%d;",thiselement,thispropertyname,class->n_values);
	for (i = 0; i < class->n_values; i++) {
          GEnumValue *value = &class->values[i];
	  strcat(thistempstr,value->value_nick);
	  strcat(thistempstr,";");
	  if (i==current)str=g_strdup(value->value_nick);
        }
	strcat(thistempstr,str);
	strcat(thistempstr,";");
	thisproperty->charval=str;
        g_type_class_unref(class);
        g_value_unset(&value);
      }
    }
    if (strlen(thistempstr)){
	strcat(outputstring,thistempstr);
        filter->propertylist=g_list_append(filter->propertylist,thisproperty);
    }
    else g_free(thisproperty);
    g_free(thistempstr);
    g_free(thiselement);
    g_free(thispropertyname);

  }
 return outputstring;
}

static gchar* chupdate(Gstpropexport *filter,gboolean sendall){
  gchar* outputstring=g_malloc0(MAXSTRING);
  GList *templ=filter->propertylist;
  struct property* thisproperty;
  while (templ){
    char* thistempstr=g_malloc0(MAXSTRING);
    thisproperty=(struct property*)templ->data;
    if (!filter->silent)g_print("Checking listelement %s for updates\n",thisproperty->elname);
    if (thisproperty->spec->value_type==G_TYPE_BOOLEAN){
        gboolean tempbool;
	g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&tempbool,NULL);
        if (!(thisproperty->active==tempbool)||sendall){
          if (tempbool)sprintf(thistempstr,"%s:%s;1;",thisproperty->elname,thisproperty->propertyname);
 	  else sprintf(thistempstr,"%s:%s;0;",thisproperty->elname,thisproperty->propertyname);
	  thisproperty->active=tempbool;
        }
      }
    else if (thisproperty->spec->value_type==G_TYPE_INT){						
      int cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%d;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnumi=cur;
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_UINT){						
      unsigned int cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%u;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnumi=cur;
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_INT64){						
      gint64 cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%lld;",thisproperty->elname,thisproperty->propertyname,(long long int)cur);
	thisproperty->currentnumi=cur;
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_UINT64){						
      guint64 cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%llu;",thisproperty->elname,thisproperty->propertyname,(long long unsigned int)cur);
	thisproperty->currentnumi=cur;
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_DOUBLE){						
      gdouble cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnum==cur)||sendall){
	//g_print("Resending double, because old %f, new %f\n",thisproperty->currentnum,cur);
        sprintf(thistempstr,"%s:%s;%f;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnum=cur;
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_FLOAT){						
      gfloat cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!((float)thisproperty->currentnum==cur)||sendall){
	//g_print("Resending float, because old %f, new %f\n",thisproperty->currentnum,cur);
        sprintf(thistempstr,"%s:%s;%f;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnum=cur;
      }
    }
    else if (G_IS_PARAM_SPEC_ENUM (thisproperty->spec)){	
      gint i;
      gchar *str=NULL;
      GEnumClass *class = G_ENUM_CLASS (g_type_class_ref (thisproperty->spec->value_type));
      GValue value = { 0, };	
      g_value_init (&value, thisproperty->spec->value_type);
      g_object_get_property(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&value);
      int current=g_value_get_enum(&value);
      for (i = 0; i < class->n_values; i++) {
          GEnumValue *value = &class->values[i];
	  if (i==current)str=g_strdup(value->value_nick);
        }
      if (!strcmp(str,thisproperty->charval)){
     	if (sendall)sprintf(thistempstr,"%s:%s;%s;",thisproperty->elname,thisproperty->propertyname,str);
	g_free(str);
      }
      else{
	g_free(thisproperty->charval);
	thisproperty->charval=str;
	sprintf(thistempstr,"%s:%s;%s;",thisproperty->elname,thisproperty->propertyname,str);
      }
      g_type_class_unref(class);
      g_value_unset(&value);
    }
    if (strlen(thistempstr))strcat(outputstring,thistempstr);
    templ=g_list_next(templ);
    g_free(thistempstr);
  }
  return outputstring;
}


/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_propexport_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstpropexport *filter;
  GstPad *otherpad;

  filter = GST_PROPEXPORT (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);
  
  GstStructure *structure = gst_caps_get_structure (caps, 0);  
  if (!gst_structure_get_int(structure,"width",&(filter->width)))filter->width=-1 ;
  if (!gst_structure_get_int(structure,"height",&(filter->height)))filter->height=-1 ;
  return gst_pad_set_caps (otherpad, caps);
}

/* chain function
 * this function does the actual processing
 */
struct property* luproperty(GList* list,gchar* name){
  if (!name) return NULL;
  GList* temp=list;
  struct property* result=NULL;
  while(temp){
    if(!strcmp(((struct property*)(temp->data))->alias,name))result=(struct property*)(temp->data);
    temp=g_list_next(temp);
  }
  return result;
}


static gboolean chupdatew(Gstpropexport *filter,gchar* description){
  gboolean finished=FALSE,valid=TRUE;
  gchar* temp=description;
  struct property* thisproperty;
  while(!finished&&valid){  
    char* thiscontrolname=g_malloc0(MAXSTRING);
    temp=deserialize_string(thiscontrolname,temp);
    thisproperty=luproperty(filter->propertylist,thiscontrolname);
    GParamSpec* spec=NULL;
    GstElement* elpointer=NULL;
    gchar* propertyname=NULL;
    //g_print("Now updating eventually %s with poiner %p\n",thiscontrolname,thisproperty);
    if (!thisproperty){//it seems to be an unknown , then maybe its an element
	GstElement* elpointer=gst_bin_get_by_name_recurse_up (GST_BIN(GST_ELEMENT_PARENT(filter)),thiscontrolname);
	if (elpointer){
		propertyname=g_malloc0(MAXSTRING);
		temp=deserialize_string(propertyname,temp);
		spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(elpointer)),propertyname);
	}
    }
    else{
      spec=thisproperty->spec;
      elpointer=thisproperty->elpointer;
      propertyname=thisproperty->propertyname;
    }
    if (!spec) finished=TRUE;//could not find it anywhere
    else if (spec->value_type==G_TYPE_BOOLEAN){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gboolean)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_INT){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gint)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_UINT){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(guint)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_INT64){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gint64)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_UINT64){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(guint64)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_DOUBLE){
	gdouble cur;
	temp=deserialize_double(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gdouble)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_FLOAT){
	gdouble cur;
	temp=deserialize_double(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gfloat)cur,NULL);
    }
    else if (G_IS_PARAM_SPEC_ENUM (spec)){
	int i;
	gchar* selection=g_malloc0(MAXSTRING);
	temp=deserialize_string(selection,temp);	
	GEnumClass *class = G_ENUM_CLASS (g_type_class_ref (spec->value_type));
	gint selpointer=0;
        for (i = 0; i < class->n_values; i++) {
          GEnumValue *value = &class->values[i];
	  if (!strcmp(selection,value->value_nick))selpointer=value->value;
        }
        //g_print("Updating Enum to %d\n",selpointer);
	g_object_set(G_OBJECT(elpointer),propertyname,selpointer,NULL);
	g_free(selection);
	g_type_class_unref(class);
    }
    if (!thisproperty){
      if (propertyname)g_free(propertyname);//then we have malloced it
      if (elpointer)gst_object_unref(GST_OBJECT (elpointer)); 
    }
    if (!temp)finished=TRUE;   
    g_free(thiscontrolname); 
  }
  return TRUE;
}

void genbrevent(Gstpropexport * filter, gchar* desc){
//parse description
  int buttoni;
  gint64 button,startx_raw,starty_raw,endx_raw,endy_raw,width_raw,height_raw;
  gdouble startx,starty,endx,endy;
  gchar* temp=desc;
  temp=deserialize_int64(&button,temp);  
  temp=deserialize_int64(&endx_raw,temp);  
  temp=deserialize_int64(&endy_raw,temp);
  temp=deserialize_int64(&startx_raw,temp);
  temp=deserialize_int64(&starty_raw,temp);
  temp=deserialize_int64(&width_raw,temp);
  temp=deserialize_int64(&height_raw,temp);
  //calculate accurate positions..
  buttoni=(int)button;
  if (42){//not scaled, just placed in the middle
    if ((filter->width>-1)&&(width_raw>(gdouble)filter->width)) startx=(gdouble)startx_raw-(0.5*(width_raw-(gdouble)filter->width));
    else startx=(gdouble)startx_raw;
    if ((filter->height>-1)&&(height_raw>(gdouble)filter->height))starty=(gdouble)starty_raw-(0.5*(height_raw-(gdouble)filter->height));
    else starty=(gdouble)starty_raw;
    endx=startx+(endx_raw-startx_raw);
    endy=starty+(endy_raw-starty_raw);
  }
  else if (0){//scaled, changed aspect ratio
    if ((filter->width>-1)){
      startx=((gdouble)startx_raw/(gdouble)width_raw)*(gdouble)filter->width;
      endx=((gdouble)endx_raw/(gdouble)width_raw)*(gdouble)filter->width; 
      }     
    else{ 
      startx=(gdouble)startx_raw;
      endx=(gdouble)endx_raw;   
      } 
    if ((filter->height>-1)){
      starty=((gdouble)starty_raw/(gdouble)height_raw)*(gdouble)filter->height;
      endy=((gdouble)endy_raw/(gdouble)height_raw)*(gdouble)filter->height;  
      }    
    else{ 
      starty=(gdouble)starty_raw;
      endy=(gdouble)endy_raw;
      }    
  }
  else if (0){//scaled, original aspect ratio
  }
  //generate pressed event
  GstStructure *structure1 =gst_structure_new(  "application/x-gst-navigation",
                            "event",    G_TYPE_STRING,  "mouse-button-press",
                            "pointer_x",      G_TYPE_DOUBLE, startx,
                            "pointer_y",      G_TYPE_DOUBLE, starty,
                            "button",      G_TYPE_INT, buttoni,
                             NULL);
  GstEvent* event1=gst_event_new_navigation(structure1);
  gst_pad_push_event(filter->sinkpad,event1);
  //generate released event
  GstStructure *structure2 =gst_structure_new(  "application/x-gst-navigation",
                            "event",    G_TYPE_STRING,  "mouse-button-release",
                            "pointer_x",      G_TYPE_DOUBLE, endx,
                            "pointer_y",      G_TYPE_DOUBLE, endy,
                            "button",      G_TYPE_INT, buttoni,
                             NULL);
  GstEvent* event2=gst_event_new_navigation(structure2);
  gst_pad_push_event(filter->sinkpad,event2);
  return;
}

void genkeyevent(Gstpropexport * filter, gchar* desc){
  gint64 key;
  deserialize_int64(&key,desc);  
  //*ks=(gchar)key;
  GstStructure *structure =gst_structure_new(  "application/x-gst-navigation",
                            "event",    G_TYPE_STRING,  "key-press",
                            //"key",      G_TYPE_STRING,  ks,
			     "key",      G_TYPE_STRING,  gdk_keyval_name((uint)key),
                             NULL);
  GstEvent* event=gst_event_new_navigation(structure);
  gst_pad_push_event(filter->sinkpad,event);
}


gboolean rcvcallback (GIOChannel *source,GIOCondition condition,gpointer data){
  if (!GST_IS_PROPEXPORT(data))return FALSE;
  Gstpropexport *filter=GST_PROPEXPORT(data);
  if(!filter->silent)g_print("Propexport Received something\n");
  GError *err = NULL;
  gchar* recv=g_malloc0(MAXSTRING);
  gsize bytes_read=0;
  //g_io_channel_read_chars(source,recv,MAXSTRING,&bytes_read,&err);
  GSocketAddress *recfrom;
  bytes_read=g_socket_receive_from (filter->rsock,&recfrom,recv,MAXSTRING,NULL,&err);
  g_object_unref(recfrom);
  if(!filter->silent)g_print("Received %d bytes: %s\n",bytes_read,recv);  
  if (recv==strstr(recv,"EVENTBR"))genbrevent(filter,recv);
  else if (recv==strstr(recv,"EVENTK"))genkeyevent(filter,recv);
  else chupdatew(filter,recv);
  /*if (!varsp->saddr){
    varsp->saddr=recfrom;
    gint checkport=0;
    g_print("Will Set port: %d\n",checkport);
    g_object_set(G_OBJECT(varsp->saddr),"port",varsp->sport,NULL);
    g_print("Will Checkport: %d\n",checkport);
    g_object_get(G_OBJECT(varsp->saddr),"port",&checkport,NULL);
    g_print("Checkport: %d\n",checkport);
    g_socket_connect(varsp->ssock, varsp->saddr, NULL, NULL);
  }*/
  g_free(recv);
  return TRUE;
}


static GstFlowReturn
gst_propexport_chain (GstPad * pad, GstBuffer * buf)
{
  Gstpropexport *filter;

  filter = GST_PROPEXPORT (GST_OBJECT_PARENT (pad));
  gchar* tosend;

  if (filter->firsttime){
    GError *err = NULL;
    filter->rsock = g_socket_new(G_SOCKET_FAMILY_IPV6,
	                    G_SOCKET_TYPE_DATAGRAM,
	                    G_SOCKET_PROTOCOL_UDP,
	                    &err);
    g_assert(err == NULL);
    filter->raddr=g_inet_socket_address_new(g_inet_address_new_any(G_SOCKET_FAMILY_IPV6),filter->rport);
    g_socket_bind(filter->rsock,G_SOCKET_ADDRESS(filter->raddr),TRUE,&err);
    g_assert(err == NULL);

    int fd = g_socket_get_fd(filter->rsock);
    filter->rchannel = g_io_channel_unix_new(fd);
    filter->source = g_io_add_watch(filter->rchannel, G_IO_IN,(GIOFunc) rcvcallback, filter);
    g_print("Set up source %d\n",filter->source);
    g_io_channel_unref(filter->rchannel);




    //now set up send socket...
    filter->ssock = g_socket_new(G_SOCKET_FAMILY_IPV6,
	                    G_SOCKET_TYPE_DATAGRAM,
	                    G_SOCKET_PROTOCOL_UDP,
	                    &err);
    filter->saddr=NULL;
    filter->saddr=g_inet_socket_address_new(g_inet_address_new_from_string(filter->remoteadress),filter->sport);
    //g_socket_connect(varsp->ssock, varsp->saddr, NULL, NULL);
    //g_socket_bind(filter->ssock,G_SOCKET_ADDRESS(filter->saddr),TRUE,&err);
    filter->uidesc=setuppropertylist(filter);
    tosend=g_strdup(filter->uidesc);
  } 

  //Generate and send property
  else tosend=chupdate(filter,FALSE);
  if (strlen(tosend)){
    GError* err;
    int numsent=0;
    g_print ("List_length: %d Tosend: %s\n",g_list_length(filter->propertylist),tosend);
    numsent=g_socket_send_to(filter->ssock,filter->saddr,tosend,strlen(tosend),NULL,&err);
    if (numsent!=strlen(tosend))g_print ("Error sending !\n");
  }
  g_free(tosend);




  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");
  filter->firsttime=FALSE;
  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, buf);
}

static GstStateChangeReturn
gst_propexport_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstpropexport *filter;
  GError* err;
  filter =  GST_PROPEXPORT(element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    default:
      break;
  }

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      g_print ("Propexport Dropping Sockets.\n");
      if (filter->source) g_source_remove (filter->source);
      filter->source=0;
      if (filter->rsock)g_socket_close(filter->rsock,&err);
      filter->rsock=0;
      if (filter->ssock)g_socket_close(filter->ssock,&err);
      filter->ssock=0;
      filter->firsttime=TRUE;
      //g_io_channel_close (filter->rchannel);
      //g_io_channel_unref(filter->rchannel);
      break;
    default:
      break;
  }
  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

 return ret;
}




/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
propexport_init (GstPlugin * propexport)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template propexport' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_propexport_debug, "propexport",
      0, "Template propexport");

  return gst_element_register (propexport, "propexport", GST_RANK_NONE,
      GST_TYPE_PROPEXPORT);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstpropexport"
#endif

/* gstreamer looks for this structure to register propexports
 *
 * exchange the string 'Template propexport' with your propexport description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "propexport",
    "Template propexport",
    propexport_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
