/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2012 Hannes <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_PROPEXPORT_H__
#define __GST_PROPEXPORT_H__

#include <gst/gst.h>

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_PROPEXPORT \
  (gst_propexport_get_type())
#define GST_PROPEXPORT(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_PROPEXPORT,Gstpropexport))
#define GST_PROPEXPORT_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_PROPEXPORT,GstpropexportClass))
#define GST_IS_PROPEXPORT(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_PROPEXPORT))
#define GST_IS_PROPEXPORT_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_PROPEXPORT))

typedef struct _Gstpropexport      Gstpropexport;
typedef struct _GstpropexportClass GstpropexportClass;

enum lutype{
INT,
INT_RO,
FLOAT,
FLOAT_RO,
TEXT,
TEXT_RO,
BOOLEAN,
BOOLEAN_RO,
MENU
};

struct property{
 GstElement* elpointer;
 gchar* elname;//name of gstreamer element
 gchar* propertyname;//name of the property
 gchar* alias;//alias, for later use  - maybe
 GParamSpec* spec;//pointer to spec
 enum lutype type;//
 gdouble currentnum,min,max;
 gint64 currentnumi,mini,maxi;
 gchar* charval;
 gboolean active;//active for gboolean
 gboolean pending;//whether currently a change occurs;
};

char* deserialize_uint64 (guint64* result,const char * s)//gives end of char
{
  if (!s)return NULL;
  while (!isdigit(*s)&&(*s!=0))s++;
  if (*s==0){
    //*result=NAN;
    return NULL;
  }
  char *end;
  *result = g_ascii_strtoull (s, &end, 0);
  return end;
}
char* deserialize_int64 (gint64* result,const char * s)//gives end of char
{
  if (!s)return NULL;
  while (!isdigit(*s)&&(*s!=0)&&(*s!=45))s++;
  if (*s==0){
    //*result=NAN;
    return NULL;
  }
  char *end;
  *result = g_ascii_strtoll (s, &end, 0);
  return end;
}

char* deserialize_double (gdouble* result,const char * s)//gives end of char
{
  if (!s)return NULL;
  while (!isdigit(*s)&&(*s!=0)&&(*s!=45))s++;
  if (*s==0){
    //*result=NAN;
    return NULL;
  }
  char *end;
  //*result = g_ascii_strtod (s, &end);
  *result = strtod (s, &end);
  return end;
}

char* deserialize_string (char* result,const char * s)//gives end of char
{
  if (!s)return NULL;
  printf("String to parse %s\n",s);
  while (!isalpha(*s)&&(*s!=0)) s++;//segfault ;)
  //printf("String after cutting semicolons %s\n",s);
  if (*s == 0)return NULL;
  char *end=strchr(s,';');
  char *end2=strchr(s,'[');
  char *end3=strchr(s,0);
  if (!end)end=end3;
  if (!end2)end2=end3; 
  if (end>end2)end=end2; 
  //printf("Copy %d",(end-s));
  if ((!end)||((end-s)>MAXSTRING))return NULL;
  //g_print ("Copy Startpointer %p , Endpointer %p, target %p %d bytes\n",s,end,result,(end-s)); 
  memcpy(result,s,(end-s));
  *(result+(end-s))=0;//letzter wird endcharackter
  return end;
}




struct _Gstpropexport
{
  GstElement element;

  GstPad *sinkpad, *srcpad;

  gboolean silent;


  GList *propertylist;//list of properties we watch and export
  gchar* exportdsc;//export description as a CSV
  gchar* uidesc;
  gboolean firsttime;//is it the first-time to send sth 
  gchar* remoteadress;
  int interval;

 GSocket *rsock;
 GSocketAddress * raddr;//adress
 gint rport;//recvport
 int rfd;//send file descriptor for fd
 GIOChannel* rchannel; //send channel

 GSocket *ssock;
 GSocketAddress * saddr;//adress
 gint sport;//sendport, normally recvport +1
 gint width,height;

 guint source;
};

struct _GstpropexportClass 
{
  GstElementClass parent_class;
};

GType gst_propexport_get_type (void);

G_END_DECLS

#endif /* __GST_PROPEXPORT_H__ */
