/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-cl
 *
 * FIXME:Describe cl here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! cl ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>

#include <errno.h>
#include <assert.h>
#include <string.h>

#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif
#include "gstcl.h"

GST_DEBUG_CATEGORY_STATIC (gst_cl_debug);
#define GST_CAT_DEFAULT gst_cl_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_KERNELFILE,
  PROP_CLCPU,
  PROP_WGS
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_GRAY8)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_GRAY8)
    );

GST_BOILERPLATE (Gstcl, gst_cl, GstElement,
    GST_TYPE_ELEMENT);

static void gst_cl_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_cl_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_cl_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_cl_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_cl_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "cl",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the cl's class */
static void
gst_cl_class_init (GstclClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_cl_set_property;
  gobject_class->get_property = gst_cl_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_KERNELFILE,
      g_param_spec_string ("kernel", "Kernel", "Filename of kernel file",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_CLCPU,
      g_param_spec_boolean ("clcpu", "Clcpu", "Whether to do OpenCL based CPU processing?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_WGS,
      g_param_spec_int ("fixlocal", "Fixlocal", "Whether to fix local workgoup size to value? -1 choose automatically",
          -1,G_MAXINT,-1, G_PARAM_READWRITE));
}

static char *load_program_source(const char *filename)
{
    struct stat statbuf;
    FILE        *fh;
    char        *source;
	
    fh = fopen(filename, "r");
    if (fh == 0)
        return 0;
	
    stat(filename, &statbuf);
    source = (char *) malloc(statbuf.st_size + 1);
    fread(source, statbuf.st_size, 1, fh);
    source[statbuf.st_size] = '\0';
	
    return source;
}

static void gst_cl_release_cl(Gstcl * filter){ //release all the CL related stuff
  	clReleaseMemObject(filter->input);
	clReleaseMemObject(filter->output);
	clReleaseProgram(filter->program);
	clReleaseKernel(filter->kernel);
	clReleaseCommandQueue(filter->commands);
	clReleaseContext(filter->context);
}

static gboolean gst_cl_init_cl(Gstcl * filter){//initialize or update all the CL related stuff
        g_print("Init CL stuff.\n");
	//sleep(1);
        if (!(filter->width*filter->height)) return FALSE;//dont do sth until it has been initialized
	if (filter->context)gst_cl_release_cl(filter);//delete everything if already used
	int gpu,err;
	if (!filter->clcpu)gpu=1;
	else gpu=0;
        filter->ret = clGetPlatformIDs(1, &(filter->platform_id), &(filter->ret_num_platforms));
	if((filter->ret=clGetDeviceIDs(filter->platform_id, gpu ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &(filter->device_id), NULL)) != CL_SUCCESS){
		g_print("Failed to create a device group! Errorcode %d",err);//for CPU this message is normal
		return FALSE;
	}	
	// Create a compute context 
	if(!(filter->context = clCreateContext(0, 1, &(filter->device_id), NULL, NULL, &err))){
		g_print("Failed to create a compute context! Errorcode %d",err);
		return FALSE;
	}
	// Create a command commands
	//
	if(!(filter->commands = clCreateCommandQueue(filter->context, filter->device_id, 0, &err))){
		g_print("Failed to create a command commands! Errorcode %d",err);
		return FALSE;		
	}
	// Load Program source
	//
	char *source = load_program_source(filter->kernelfile);
        if(!source)
        {
          printf("Error: Failed to load compute program from file!\n");
          return FALSE;    
        }
	// Create the compute program from the source buffer
	//
	if(!(filter->program = clCreateProgramWithSource(filter->context, 1, (const char **) & source, NULL, &err))){
	  g_print("Failed to create compute program! Errorcode %d\n",err);
 	  free(source);
	  return FALSE;
	}
 	free(source);
	// Build the program executable
	//
	#ifdef NOMAC
    	char *flags = "-cl-fast-relaxed-math";
	#else
    	char *flags = "-cl-fast-relaxed-math -DMAC";
	#endif
	err = clBuildProgram(filter->program, 0, NULL, flags, NULL, NULL);
 	//free(flags);	
	if (err != CL_SUCCESS)
	{
		size_t len;
		char buffer[2048];
		
		g_print("Error: Failed to build program executable!\n");
		clGetProgramBuildInfo(filter->program, filter->device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
		g_print("%s\n", buffer);
		return FALSE;
	}
	
	// Create the compute kernel in the program we wish to run
	//
	filter->kernel = clCreateKernel(filter->program, "sobel", &err);
	if (!filter->kernel || err != CL_SUCCESS){
		g_print("Failed to create compute kernel,  Errorcode %d\n",err);
	}

	//getting timestamp
	//gettimeofday(&start,NULL);
	// Create the input and output arrays in device memory for our calculation
	//
        printf("Creating Buffers and upload Timestamp=0 ms!\n");
	filter->input = clCreateBuffer(filter->context,  CL_MEM_READ_ONLY, filter->width*filter->height , NULL, NULL);
	filter->output = clCreateBuffer(filter->context, CL_MEM_WRITE_ONLY, filter->width*filter->height, NULL, NULL);
	if (!filter->input || !filter->output){
		g_print("Failed to allocate device memory!");
		return FALSE;
	}
	return TRUE; 
}
/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_cl_init (Gstcl * filter,
    GstclClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_cl_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_cl_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->ready= FALSE;
  filter->platform_id = NULL;
  filter->context= NULL;
  filter->caps= NULL;
  filter->clcpu=0;
  filter->fixlocal=-1;
  //g_print("gst_cl_init.\n");
  //sleep(1);
}

static void
gst_cl_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstcl *filter = GST_CL (object);
  g_print("gst_cl_set_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_KERNELFILE:
      g_free (filter->kernelfile);
      filter->kernelfile = g_value_dup_string (value);
      filter->ready=gst_cl_init_cl(filter);
      break;
    case PROP_CLCPU:
      filter->clcpu = g_value_get_boolean (value);
      break;
    case PROP_WGS:
      filter->fixlocal = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_cl_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstcl *filter = GST_CL (object);
  g_print("gst_cl_get_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_KERNELFILE:
      g_value_set_string (value, filter->kernelfile);
      break;
    case PROP_CLCPU:
      g_value_set_boolean (value, filter->clcpu);
      break;
    case PROP_WGS:
      g_value_set_boolean (value, filter->fixlocal);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_cl_set_caps (GstPad * pad, GstCaps * caps)
{
  //g_print("gst_cl_set_caps.\n");
  Gstcl *filter;
  GstPad *otherpad;
  int oldwidth,oldheight;
  gboolean ret;
  
  filter = GST_CL (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  oldwidth=filter->width;
  oldheight=filter->height;

  GstStructure *structure=gst_caps_get_structure(caps,0);
  if (!gst_structure_get_int(structure,"width",&filter->width)) return FALSE;
  if (!gst_structure_get_int(structure,"height",&filter->height)) return FALSE;
  gst_video_parse_caps_framerate      (caps,&filter->fps_n,&filter->fps_d);

  if (filter->caps)gst_caps_unref(filter->caps);
  filter->caps=gst_caps_new_simple ("video/x-raw-gray",
      		"width", G_TYPE_INT, filter->width,
      		"height", G_TYPE_INT, filter->height,
      		"bpp",G_TYPE_INT , 8,
      		"depth",G_TYPE_INT , 8,
		"framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
		NULL);

  if ((oldwidth*oldheight)!=(filter->width*filter->height)){
    filter->ready=gst_cl_init_cl(filter);
  }
  ret=gst_pad_set_caps (otherpad, caps);
  if (!ret) g_print ("GstCL Warning Set Caps will return false.\n");
  return ret;
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_cl_chain (GstPad * pad, GstBuffer * buf)
{
  Gstcl *filter;
  filter = GST_CL (GST_OBJECT_PARENT (pad));
  int err;
  size_t global;                      // global domain size for our calculation
  size_t local;  
  GstFlowReturn ret;
  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");
    //sleep(1);

  //check wether we are ready
  if (!filter->ready){
    g_print ("GstCL not ready for some reason/problem with kernel?.\n");
    return GST_FLOW_ERROR;
  }
  if (buf->size!=(filter->width*filter->height)){
    g_print ("GstCL seems to have allocated the wrong buffersize!?!.\n");
    return GST_FLOW_ERROR;
  }
  //upload and launch

  // Write our data set into the input array in device memory 
  //
  err = clEnqueueWriteBuffer(filter->commands, filter->input, CL_TRUE, 0, buf->size, buf->data, 0, NULL, NULL);
  if (err != CL_SUCCESS)
  {
	printf("Error: Failed to write to source array!\n");
	return GST_FLOW_ERROR;
  }
	
  // Set the arguments to our compute kernel
  //
  err = 0;
  err  = clSetKernelArg(filter->kernel, 0, sizeof(cl_mem), &(filter->input));
  err |= clSetKernelArg(filter->kernel, 1, sizeof(cl_mem), &(filter->output));
  err |= clSetKernelArg(filter->kernel, 2, sizeof(unsigned int), &(filter->width));
  err |= clSetKernelArg(filter->kernel, 3, sizeof(unsigned int), &(filter->height));
  if (err != CL_SUCCESS)
  {
	printf("Error: Failed to set kernel arguments! %d\n", err);
	return GST_FLOW_ERROR;
  }

  err = clGetKernelWorkGroupInfo(filter->kernel, filter->device_id, CL_KERNEL_WORK_GROUP_SIZE, sizeof(int) * 4 + sizeof(float) * 2, &local, NULL);
  if (err != CL_SUCCESS)
  {
    printf("Error: Failed to retrieve kernel work group info! %d\n", err);
    return GST_FLOW_ERROR;
  }
  global = filter->width*filter->height;
  if (!local)local = 64;
  if (filter->fixlocal>0)local = filter->fixlocal;
  err = clEnqueueNDRangeKernel(filter->commands, filter->kernel, 1, NULL, &global, &local, 0, NULL, NULL);
  if (err)
  {
    printf("Failed to execute kernel!, code %d %s global %d local %d \n",err,print_cl_errstring(err),global,local);
    return GST_FLOW_ERROR;
  }
//inbetween stuff
  //alloc new buffer during gpu is working:)
  GstBuffer *out;
  ret=gst_pad_alloc_buffer_and_set_caps   (filter->srcpad,
                                                         buf->offset,
                                                         filter->width*filter->height,
                                                         filter->caps,
                                                         &out);
  if (ret!=GST_FLOW_OK){
    g_print ("GSTCL: Error allocating buffer.\n");
    return ret;
  }
  out->timestamp=buf->timestamp;
  out->offset_end=buf->offset_end;
  out->duration=buf->duration;
  gst_buffer_unref(buf);
  //sync and download
  // Wait for the command commands to get serviced before reading back results
//
  clFinish(filter->commands);
  err = clEnqueueReadBuffer(filter->commands, filter->output, CL_TRUE, 0, filter->width*filter->height, out->data, 0, NULL, NULL );  
  if (err != CL_SUCCESS)
  {
    printf("Error: Failed to read output array! code %d, %s\n", err,print_cl_errstring(err));
    exit(1);
  }
  //g_print ("Pushing buffer in 1 s.\n");
  //sleep(1);
  /* just push out the incoming buffer without touching it */
  //ret=gst_pad_push (filter->srcpad, out);
  ret=gst_pad_push (filter->srcpad, out);
  //g_print ("returning %d in 5 s.\n",ret);
  return ret;
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
cl_init (GstPlugin * cl)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template cl' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_cl_debug, "cl",
      0, "Template cl");

  return gst_element_register (cl, "cl", GST_RANK_NONE,
      GST_TYPE_CL);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstcl"
#endif

/* gstreamer looks for this structure to register cls
 *
 * exchange the string 'Template cl' with your cl description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "cl",
    "Template cl",
    cl_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
