/*
 * GStreamer
 * Copyright 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-plugin
 *
 * <refsect2>
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch -v -m audiotestsrc ! plugin ! fakesink silent=TRUE
 * </programlisting>
 * </para>
 * </refsect2>
 */

//This plugin is supposed to connect gstreamer framework to opencv-library


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <unistd.h>
#include <math.h>

#include <opencv/cv.h>
#include "gstcv.h"

#ifndef NAN
#define NAN (0.0/0.0)
#endif

GST_DEBUG_CATEGORY_STATIC (gst_cv_debug);
#define GST_CAT_DEFAULT gst_cv_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_SILENT,
  ARG_FORCEFORMAT,
  ARG_OPMODE,
  ARG_PEERALLOC,
  ARG_CANNYMASKSIZE,
  ARG_PARDOUBLE1,
  ARG_PARDOUBLE2,
  ARG_SCALEUP,
  ARG_PARINT1,
  ARG_PARINT2,
  ARG_PARSTRING1
};

// enum
// {
//   CANNY,
//   SOBEL,
//   MATRIX
// };



// static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
//     GST_PAD_SINK,
//     GST_PAD_ALWAYS,
//     GST_STATIC_CAPS ("ANY")
//     );

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-rgb, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ]")
    );


static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-rgb, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ]")
    );

static GstStaticPadTemplate src_factory_coord = GST_STATIC_PAD_TEMPLATE ("src-coord",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("coordinates-list")
    );

// static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
//     GST_PAD_SRC,
//     GST_PAD_ALWAYS,
//     GST_STATIC_CAPS ("ANY")
//     );

GST_BOILERPLATE (GstCV, gst_cv, GstElement,
    GST_TYPE_ELEMENT);

static void gst_cv_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_cv_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_cv_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps * gst_cv_getcaps (GstPad *pad);
static gboolean gst_cv_set_caps2 (GstPad * pad, GstCaps * caps){return TRUE;}
static GstCaps * gst_cv_getcaps2 (GstPad *pad){return gst_caps_copy(gst_pad_get_pad_template_caps(pad));}
static GstFlowReturn gst_cv_chain (GstPad * pad, GstBuffer * buf);
static gboolean little_endian;
#define GST_TYPE_OP_MODE (gst_type_op_mode ())
static GType
gst_type_op_mode (void)
{
  static GType op_mode = 0;
  static const GEnumValue functions[] = {
    {CV_CANNY, "Canny-Line detector, works on greyscale images", "canny"},
    {CV_CSP, "Colorspace converter based on csp-library", "csp"},
    {CV_MATRIX, "Matrix Operator over each Pixel", "matrix"},
    {CV_CALCOPTICALFLOWBASE, "Calculates optical flow using differences between two frames", "calcopticalflowbase"},
    {CV_CALCOPTICALFLOWHS, "Calculates optical flow using Horn & Schnuck algorithm (not implemented)", "calcopticalflowhs"},
    {CV_CALCOPTICALFLOWLK, "Calculates optical flow using Lukas & Kanade algorithm (not implemented)", "calcopticalflowlk"},
    {CV_CALCOPTICALFLOWBM, "Calculates optical flow using block matching (not implemented)", "calcopticalflowbm"},
    {CV_CALCOPTICALFLOWPYRLK, "Calculates optical flow using Lukas & Kanade algorithm in Pyramids(not implemented)", "calcopticalflowpyrlk"},
    {CV_SMOOTH_BLUR_NO_SCALE, "Smoothen image with blurring method without scaling", "smoothblurnoscale"},
    {CV_SMOOTH_BLUR, "Smoothen image with blurring method", "smoothblur"},
    {CV_SMOOTH_GAUSSIAN, "Smoothen image with gaussian method", "smoothgaussian"},
    {CV_SMOOTH_MEDIAN, "Smoothen image with median method", "smoothmedian"},
    {CV_SMOOTH_BILATERAL, "Smoothen image with bilateral method", "smoothbilateral"},
    {CV_HAAR, "Haar Detection of features", "haar"},
    {CV_HOUGH_CIRCLE, "Circle detection using hough", "hough_circle"},
    {CV_LAPLACE, "Laplace filtering", "laplace"},
    {CV_SOBEL, "Sobel filtering", "sobel"},
    {0, NULL, NULL}
  };

  if (!op_mode) {
    op_mode =
        g_enum_register_static ("GstOpMode", functions);
  }
  return op_mode;
}






static void
gst_cv_base_init (gpointer gclass)
{

  static GstElementDetails element_details = {
    "GstCV",
    "Gst-Plugins-Robotix",
    "Encapsulates several OpenCV functions",
    "Hannes Bistry <bistry@informatik.uni-hamburg.de>"
  };
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory_coord));
  gst_element_class_set_details (element_class, &element_details);
}

/* initialize the plugin's class */
static void
gst_cv_class_init (GstCVClass * klass)
{
  GObjectClass *gobject_class;
  //GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  //gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_cv_set_property;
  gobject_class->get_property = gst_cv_get_property;

  g_object_class_install_property (gobject_class, ARG_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_PEERALLOC,
      g_param_spec_boolean ("peeralloc", "Peeralloc", "Use buffer from Downstream Element ?",
          FALSE, (GParamFlags)G_PARAM_READWRITE));
//Own properties
  g_object_class_install_property (gobject_class, ARG_FORCEFORMAT,
      g_param_spec_boolean ("forceformat", "forceformat", "Wether this Plugin should try to adjust colorspc?",
          FALSE, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_OPMODE,
      g_param_spec_enum ("opmode", "Opmode",
          "What filter should this plugin use?", GST_TYPE_OP_MODE,
          CV_CANNY, (GParamFlags)G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, ARG_CANNYMASKSIZE,
      g_param_spec_int ("cannymasksize", "Cannymasksize",
          "Mask size of filter before Canny operation?", 3,7,
          3, (GParamFlags)G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, ARG_PARDOUBLE1,
      g_param_spec_double ("pardouble1", "Pardouble1",
          "First double Value i.e. Low threshold of Canny filter", (double)G_MINDOUBLE,(double)G_MAXDOUBLE,
          1.0, (GParamFlags)G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, ARG_PARDOUBLE2,
      g_param_spec_double ("pardouble2", "Pardouble2",
          "Second double Value i.e. High threshold of Canny filter", (double)G_MINDOUBLE,(double)G_MAXDOUBLE,
          1.0, (GParamFlags)G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, ARG_SCALEUP,
      g_param_spec_double ("scaleup", "Scaleup",
          "Scale up coordinates of srcpad", (double)G_MINDOUBLE,(double)G_MAXDOUBLE,
          1.0, (GParamFlags)G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, ARG_PARINT1,
      g_param_spec_int ("parint1", "Parint1",
          "General Purpose int parameter, i.e. window size of operation", G_MININT,G_MAXINT,
          3, (GParamFlags)G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, ARG_PARINT2,
      g_param_spec_int ("parint2", "Parint2",
          "General Purpose int parameter, i.e. window size of operation", G_MININT,G_MAXINT, 
          3, (GParamFlags)G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, ARG_PARSTRING1,
      g_param_spec_string ("parstring1", "Parstring1",
          "General Purpose string parameter, i.e. filename of Haar-Classifier", NULL, (GParamFlags)G_PARAM_READWRITE));

}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_cv_init (GstCV * filter,
    GstCVClass * gclass)
{
  int x=1;//for endian test, as testing is not supportet while crosscompiling
  GstElementClass *klass = GST_ELEMENT_GET_CLASS (filter);
  filter->cvcspcodes=gst_structure_new("cvcspcodes",
"CV_BGR2BGRA",G_TYPE_INT,0,
"CV_RGB2RGBA",G_TYPE_INT,0, 

"CV_BGRA2BGR",G_TYPE_INT,1,
"CV_RGBA2RGB",G_TYPE_INT,1,
"CV_BGR2RGBA",G_TYPE_INT,2,
"CV_RGB2BGRA",G_TYPE_INT,2,

"CV_RGBA2BGR",G_TYPE_INT,3,
"CV_BGRA2RGB",G_TYPE_INT,3,

"CV_BGR2RGB",G_TYPE_INT,4,
"CV_RGB2BGR",G_TYPE_INT,4,

"CV_BGRA2RGBA",G_TYPE_INT,5,
"CV_RGBA2BGRA",G_TYPE_INT,5,

"CV_BGR2GRAY",G_TYPE_INT,6,
"CV_RGB2GRAY",G_TYPE_INT,7,
"CV_GRAY2BGR",G_TYPE_INT,8,
"CV_GRAY2RGB",G_TYPE_INT,8,
"CV_GRAY2BGRA",G_TYPE_INT,9,
"CV_GRAY2RGBA",G_TYPE_INT,9,
"CV_BGRA2GRAY",G_TYPE_INT,10,
"CV_RGBA2GRAY",G_TYPE_INT,11,

"CV_BGR2BGR565",G_TYPE_INT,  12,
"CV_RGB2BGR565",G_TYPE_INT,  13,
"CV_BGR5652BGR",G_TYPE_INT,  14,
"CV_BGR5652RGB",G_TYPE_INT,  15,
"CV_BGRA2BGR565",G_TYPE_INT, 16,
"CV_RGBA2BGR565",G_TYPE_INT, 17,
"CV_BGR5652BGRA",G_TYPE_INT, 18,
"CV_BGR5652RGBA",G_TYPE_INT, 19,

"CV_GRAY2BGR565",G_TYPE_INT, 20,
"CV_BGR5652GRAY",G_TYPE_INT, 21,

"CV_BGR2BGR555",G_TYPE_INT,  22,
"CV_RGB2BGR555",G_TYPE_INT,  23,
"CV_BGR5552BGR",G_TYPE_INT,  24,
"CV_BGR5552RGB",G_TYPE_INT,  25,
"CV_BGRA2BGR555",G_TYPE_INT, 26,
"CV_RGBA2BGR555",G_TYPE_INT, 27,
"CV_BGR5552BGRA",G_TYPE_INT, 28,
"CV_BGR5552RGBA",G_TYPE_INT, 29,

"CV_GRAY2BGR555",G_TYPE_INT, 30,
"CV_BGR5552GRAY",G_TYPE_INT, 31,

"CV_BGR2XYZ",G_TYPE_INT,     32,
"CV_RGB2XYZ",G_TYPE_INT,     33,
"CV_XYZ2BGR",G_TYPE_INT,     34,
"CV_XYZ2RGB",G_TYPE_INT,     35,

"CV_BGR2YCrCb",G_TYPE_INT,   36,
"CV_RGB2YCrCb",G_TYPE_INT,   37,
"CV_YCrCb2BGR",G_TYPE_INT,   38,
"CV_YCrCb2RGB",G_TYPE_INT,   39,

"CV_BGR2HSV",G_TYPE_INT,     40,
"CV_RGB2HSV",G_TYPE_INT,     41,

"CV_BGR2Lab",G_TYPE_INT,     44,
"CV_RGB2Lab",G_TYPE_INT,     45,

"CV_BayerBG2BGR",G_TYPE_INT, 46,
"CV_BayerGB2BGR",G_TYPE_INT, 47,
"CV_BayerRG2BGR",G_TYPE_INT, 48,
"CV_BayerGR2BGR",G_TYPE_INT, 49,

"CV_BayerBG2RGB",G_TYPE_INT, 48,
"CV_BayerGB2RGB",G_TYPE_INT, 49,
"CV_BayerRG2RGB",G_TYPE_INT, 46,
"CV_BayerGR2RGB",G_TYPE_INT, 47,

"CV_BGR2Luv",G_TYPE_INT,     50,
"CV_RGB2Luv",G_TYPE_INT,     51,
"CV_BGR2HLS",G_TYPE_INT,     52,
"CV_RGB2HLS",G_TYPE_INT,     53,

"CV_HSV2BGR",G_TYPE_INT,     54,
"CV_HSV2RGB",G_TYPE_INT,     55,

"CV_Lab2BGR",G_TYPE_INT,     56,
"CV_Lab2RGB",G_TYPE_INT,     57,
"CV_Luv2BGR",G_TYPE_INT,     58,
"CV_Luv2RGB",G_TYPE_INT,     59,
"CV_HLS2BGR",G_TYPE_INT,     60,
"CV_HLS2RGB",G_TYPE_INT,     61,
NULL);
  filter->sinkpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "sink"), "sink");
  gst_pad_set_setcaps_function (filter->sinkpad, gst_cv_set_caps);
//   gst_pad_set_setcaps_function (filter->sinkpad,
//                                 GST_DEBUG_FUNCPTR(gst_cv_set_caps));
/*  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));*/
  gst_pad_set_getcaps_function (filter->sinkpad,gst_cv_getcaps);

  filter->srcpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "src"), "src");

//   gst_pad_set_getcaps_function (filter->srcpad,
//                                 GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
//   gst_pad_set_getcaps_function (filter->srcpad,
//                                 GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_getcaps_function (filter->srcpad,gst_cv_getcaps);
  gst_pad_set_setcaps_function (filter->srcpad,gst_cv_set_caps);

  filter->srccoord =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "src-coord"), "src-coord");

  gst_pad_set_getcaps_function (filter->srccoord,gst_cv_getcaps2);
  gst_pad_set_setcaps_function (filter->srccoord,gst_cv_set_caps2);


  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srccoord);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);

  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_cv_chain));
  filter->silent = TRUE;
  filter->cannymasksize=3;
  filter->pardouble1=1.0;
  filter->pardouble2=1.0;
  filter->scaleup=1.0;
  filter->parint1=3;
  filter->parint2=3;
  filter->cascade=0;
  filter->lastbuf=NULL;
  little_endian=(gboolean) (*(char*)&x); 
}

static void
gst_cv_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstCV *filter = GST_CV (object);

  switch (prop_id) {
    case ARG_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case ARG_FORCEFORMAT:
      filter->forceformat = g_value_get_boolean (value);
      break;
    case ARG_OPMODE:
      if (GST_STATE(filter)==GST_STATE_NULL){
      	filter->opmode=(GstOpMode)g_value_get_enum(value);
	}
      else printf("Hey, what are you doing there??? Trying to force a crash\n");
      break;
    case ARG_PEERALLOC:
      filter->peeralloc = g_value_get_boolean (value);
      break;
    case ARG_CANNYMASKSIZE:
	if ((g_value_get_int (value)==3)||(g_value_get_int (value)==5)||(g_value_get_int (value)==7))
      		filter->cannymasksize = g_value_get_int (value);
	else g_print("Warning, cannymasksize invalid");
      break;
    case ARG_PARDOUBLE1:
      filter->pardouble1 = g_value_get_double (value);
      break;
    case ARG_PARDOUBLE2:
      filter->pardouble2 = g_value_get_double (value);
      break;
    case ARG_SCALEUP:
      filter->scaleup = g_value_get_double (value);
      break;
    case ARG_PARINT1:
      filter->parint1 = g_value_get_int (value);
      break;
    case ARG_PARINT2:
      filter->parint2 = g_value_get_int (value);
      break;
    case ARG_PARSTRING1:
      g_print("Will set parstring now!!\n");
      filter->parstring1=g_strdup(g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_cv_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstCV *filter = GST_CV (object);

  switch (prop_id) {
    case ARG_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case ARG_FORCEFORMAT:
      g_value_set_boolean (value, filter->forceformat);
      break;
    case ARG_OPMODE:
      g_value_set_enum(value, filter->opmode);
      break;
    case ARG_PEERALLOC:
      g_value_set_boolean(value,filter->peeralloc);
      break;
    case ARG_CANNYMASKSIZE:
      g_value_set_int(value,filter->cannymasksize);
      break;
    case ARG_PARDOUBLE1:
      g_value_set_double (value,filter->pardouble1);
      break;
    case ARG_PARDOUBLE2:
      g_value_set_double (value,filter->pardouble2);
      break;
    case ARG_SCALEUP:
      g_value_set_double (value,filter->scaleup);
      break;
    case ARG_PARINT1:
      g_value_set_int(value,filter->parint1);
      break;
    case ARG_PARINT2:
      g_value_set_int(value,filter->parint2);
      break;
    case ARG_PARSTRING1:
      g_value_set_string (value, filter->parstring1);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}






static gboolean
gst_cv_parse_caps (GstPad * pad, GstCaps * caps){//copy framerate, width and height
  GstCV *filter;
  //GstPad *otherpad;

  filter = GST_CV (gst_pad_get_parent (pad));
  //otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  GstStructure *structure = gst_caps_get_structure (caps, 0);
  if (pad == filter->sinkpad){
	if (!gst_structure_get_int(structure,"width",&(filter->imagesize.width))){
		g_print ("Width not set!");
		return FALSE;
	}
	if (!gst_structure_get_int(structure,"height",&(filter->imagesize.height))){
		g_print ("Height not set!");
		return FALSE;
	}
  }
  if (pad == filter->srcpad){
	if (!gst_structure_get_int(structure,"width",&(filter->outimagesize.width))){
		g_print ("Width not set!");
		return FALSE;
	}
	if (!gst_structure_get_int(structure,"height",&(filter->outimagesize.height))){
		g_print ("Height not set!");
		return FALSE;
	}
  }  
  if ((filter->imagesize.width)%4){
    g_print ("OpenCV is buggy and supports only width values that are multiples of 4 !\n");
    return FALSE;
    }
  const GValue *framerate;
  framerate = gst_structure_get_value(structure, "framerate");
  if (framerate) {
  filter->rate_numerator = gst_value_get_fraction_numerator (framerate);
  filter->rate_denominator = gst_value_get_fraction_denominator (framerate);
  } else{ 
	g_print ("No framerate specified!");
	return FALSE;
  }
  return TRUE;
}


//source out to auxfunctions-file
static gboolean acceptonechannel(GstPad * pad, GstCaps * caps){
  GstCV *filter;
  GstPad *otherpad;

  filter = GST_CV (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);  
if (0 == strcmp (mime, "video/x-raw-yuv")){ 
	guint32 fourccval;
	if (!gst_structure_has_field(structure,"format")){
		g_print ("No format field");
		return FALSE;
	}  
	if (!gst_structure_get_fourcc(structure,"format",&fourccval)){
		g_print ("FourCC could not be read");
		return FALSE;
	}
	if (!fourccval==GST_MAKE_FOURCC ('Y', '8', '0', '0')){
		g_print ("Wrong format suggested: %" GST_FOURCC_FORMAT "\n", GST_FOURCC_ARGS (fourccval));
		return FALSE;
	}
	filter->estchannels=1;
	//now test if our neighbour likes it
	//if (!gst_pad_set_caps (otherpad, caps)) {
	//	g_print ("Failed, next element does not support Y800!");
	//	return FALSE;
	//}   //CANNY will not change format
	//configure ourself
	//this is commented out due to a bug in ffmpegcolorspace that accepts Y800

	if (pad == filter->sinkpad){//foreward caps negotiation
		if (!gst_cv_parse_caps(pad,caps)) return FALSE; 
		GstCaps *peer,*resultcaps;
		//GstStructure* peerstructure;
		
		peer=gst_pad_get_allowed_caps(filter->srcpad);
		//test
		resultcaps=gst_caps_intersect(caps,peer);
		if (!filter->silent){
		  gchar * cts1=gst_caps_to_string(resultcaps); 
		  g_print ("%s: caps Intersection: %s\n",GST_ELEMENT_NAME(filter),cts1);
		  g_free(cts1);
		}

		if (!gst_pad_set_caps (otherpad,resultcaps)){
			g_print ("%s: Peer does not accept own caps",GST_ELEMENT_NAME(filter));
			return FALSE;
		}
	}
	else if (pad == filter->srcpad){
		if (!gst_cv_parse_caps(pad,caps)) return FALSE;
		if ((filter->imagesize.width==filter->outimagesize.width)&&(filter->imagesize.height==filter->outimagesize.height))
		return TRUE;
		else {
			g_print ("Source Format: %d x %d, dest. reneg. format: %d x %d !",filter->imagesize.width,filter->imagesize.height,filter->outimagesize.width,filter->outimagesize.height);
			return FALSE;
		}
	}
	//	g_print ("Failed, next element does not support Y800!");
	//	return FALSE;
	
	if (!gst_cv_parse_caps(pad,caps)) return FALSE; 
	if (!gst_cv_parse_caps(otherpad,caps)) return FALSE; //same input and outpuf format for canny
	g_print ("Caps negotiation success!");
	return TRUE;
}
else if (0 == strcmp (mime, "video/x-raw-gray")){
	filter->estchannels=1;
	if (pad == filter->sinkpad){
		int temp;
		if (!gst_structure_get_int(structure,"depth",&temp)) return FALSE;
		if (temp!=8) return FALSE;
		if (!gst_structure_get_int(structure,"bpp",&temp)) return FALSE;
		if (temp!=8) return FALSE;
		filter->inbpp=8;
		if (!gst_cv_parse_caps(pad,caps)) return FALSE; 
		if (!(gst_pad_set_caps (otherpad, caps))) {g_print ("Failed, next element does not support Gray!");return FALSE;}
			//same input and outpuf format for canny
		g_print ("Caps negotiation success!");
		return TRUE;
	}
	else if(pad == filter->srcpad){
		if (!gst_cv_parse_caps(pad,caps)) return FALSE;
		if ((filter->imagesize.width==filter->outimagesize.width)&&(filter->imagesize.height==filter->outimagesize.height))
		return TRUE;
		else {
			g_print ("Source Format: %d x %d, dest. reneg. format: %d x %d !",filter->imagesize.width,filter->imagesize.height,filter->outimagesize.width,filter->outimagesize.height);
			return FALSE;
		}
	}
}
else {			
	g_print ("Wrong mimetype %s provided, we only support %s",mime, "video/x-raw-yuv");
	return FALSE;
}
g_print ("Caps negotiation success!");
return TRUE;
}

static gboolean acceptrgb(GstPad * pad, GstCaps * caps){
  GstCV *filter;
  GstPad *otherpad;

  filter = GST_CV (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);  
	if (0 == strcmp (mime, "video/x-raw-rgb")){
		int temp;
		if (!gst_structure_get_int(structure,"depth",&temp)) return FALSE;
		if (temp!=24) return FALSE;
		//filter->indepth=8;
		if (!gst_structure_get_int(structure,"bpp",&temp)) return FALSE;
		if ((temp!=24)&&(temp!=32)) return FALSE;
		filter->inbpp=temp;

		if (!gst_cv_parse_caps(pad,caps)) return FALSE; 
		if (!(gst_pad_set_caps (otherpad, caps))) {g_print ("Failed, next element does not support RGB!");return FALSE;}
			//same input and outpuf format for canny
		g_print ("Caps negotiation success with RGB!");
		return TRUE;
	}
  return FALSE;
}


/* GstElement vmethod implementations */
/* this function handles the link with other elements */
static gboolean
gst_cv_set_caps (GstPad * pad, GstCaps * caps)
{
  GstCV *filter;
  GstPad *otherpad;

  filter = GST_CV (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  if (!filter->silent){
    if (pad == filter->srcpad){
      gchar * cts=gst_caps_to_string(caps); 
      g_print ("%s: Das Zielelement will am Format drehen: Caps: %s\n",GST_ELEMENT_NAME(filter),cts);
      g_free(cts);
    }
    if (pad == filter->sinkpad){
      gchar * cts=gst_caps_to_string(caps);
      g_print ("%s: Das Quellelement will am Format drehen: Caps %s\n",GST_ELEMENT_NAME(filter),cts);
      g_free(cts);
    }
  }
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);//get datatype of caps

  switch (filter->opmode){
	case CV_CANNY://test if we like it
		return acceptonechannel(pad, caps);			
	  break;
	case CV_CALCOPTICALFLOWBASE:
	case CV_CALCOPTICALFLOWLK:
	case CV_CALCOPTICALFLOWHS:
	case CV_HOUGH_CIRCLE:
	case CV_LAPLACE:
	case CV_SOBEL:
		//if (showgrey)
		return acceptonechannel(pad, caps);
	  break;
        case CV_SMOOTH_BLUR_NO_SCALE:
        case CV_SMOOTH_BLUR:
        case CV_SMOOTH_GAUSSIAN:
        case CV_SMOOTH_MEDIAN:
        case CV_SMOOTH_BILATERAL:
	case CV_HAAR:
		if (pad == filter->sinkpad){
			int temp;
			if (0 == strcmp (mime, "video/x-raw-yuv")){
				temp=8;//TODO this is quite dirty but should work as only Y800 will be accepted
			}
			else gst_structure_get_int(structure,"bpp",&temp);
			filter->estchannels=temp/8;
			filter->inbpp=temp;
			g_print ("Set channels to %d...\n",filter->estchannels);
		}	
		if(FALSE==acceptonechannel(pad, caps)){
				return acceptrgb(pad, caps);
			}
		else return TRUE;
	break;
	case CV_CSP:	
		if (0 == strcmp (mime, "video/x-raw-yuv")){
			guint32 fourccval;
			if (!gst_structure_has_field(structure,"format")){
				g_print ("No format field");
				return FALSE;
			}  
			if (!gst_structure_get_fourcc(structure,"format",&fourccval)){
				g_print ("FourCC could not be read");
				return FALSE;
			}
			if (!fourccval==GST_MAKE_FOURCC ('Y', '8', '0', '0')){
				g_print ("Wrong format suggested: %" GST_FOURCC_FORMAT "\n", GST_FOURCC_ARGS (fourccval));
				return FALSE;
			}
			if (pad == filter->sinkpad){
				strcpy(filter->sinkc,"GRAY");
			}
			if (pad == filter->srcpad){
				strcpy(filter->srcc,"GRAY");
			}
			if (gst_caps_is_fixed(caps)){
				if (!gst_cv_parse_caps(pad,caps)) return FALSE; 
			}
		}
		else if (0 == strcmp (mime, "video/x-raw-gray")){
			int temp;
			if (!gst_structure_get_int(structure,"depth",&temp)) return FALSE;
			if (temp!=8) return FALSE;
			if (!gst_structure_get_int(structure,"bpp",&temp)) return FALSE;
			if (temp!=8) return FALSE;
			if (!gst_cv_parse_caps(pad,caps)) return FALSE;
			if (pad == filter->sinkpad){
				strcpy(filter->sinkc,"GRAY");
			}
			if (pad == filter->srcpad){
				strcpy(filter->srcc,"GRAY");
			}
		}
		else if (0 == strcmp (mime, "video/x-raw-rgb")){
			gchar tempchar[10];
			//Bitanordnung herausfinden
			int depth,bpp,red,blue;
			if (!gst_structure_get_int(structure,"red_mask",&red)){g_print ("No color mask given!"); return FALSE;}
			if (!gst_structure_get_int(structure,"blue_mask",&blue)){g_print ("No color mask given!"); return FALSE;}
			if (*((guint32*)&red)>*((guint32*)&blue)) strcpy(tempchar,"RGB");
			else strcpy(tempchar,"BGR");

			//depth USW
			if (!gst_structure_get_int(structure,"bpp",&bpp)){g_print ("No bpp given!"); return FALSE;}
			if (!gst_structure_get_int(structure,"depth",&depth)){g_print ("No depth given!"); return FALSE;}

			if ((depth==15)||(depth==16)) {
				if (bpp !=16){
					g_print ("Bpp does not match depth (bpp is not 16!)");
					return FALSE;
				}
				//filter->depth=depth;
				//filter->bpp=bpp;
				if (depth==15) strcpy(&tempchar[3],"555");
				else strcpy(&tempchar[3],"565");
			}
			else if (depth==24){
				if ((bpp!=24)&&(bpp!=32)){g_print ("For depth==24 bpp needs to be 24 or 32!"); return FALSE;}
				//filter->depth=depth;
				//filter->bpp=bpp;
				if (bpp==32) strcpy(&tempchar[3],"A");//Alpha-Channel
			}
			else return FALSE;
			if (pad == filter->sinkpad){
				strcpy(filter->sinkc,tempchar);
			}
			if (pad == filter->srcpad){
				strcpy(filter->srcc,tempchar);
			}
			if (gst_caps_is_fixed(caps)){
				if (!gst_cv_parse_caps(pad,caps)){g_print ("GstCV could not parse caps!"); return FALSE; }
			}
			}
		else if (0 == strcmp (mime, "video/x-raw-bayer")){
			g_print ("Not implemented yet");
			}
		else{
			g_print ("MIMEtype Unknown");
			return FALSE;
		}
		
		if (pad == filter->sinkpad){//start foreward caps negotiation
			if (!gst_cv_parse_caps(pad,caps)) return FALSE;
			if (!gst_structure_get_int(structure,"bpp",&filter->inbpp)){
				filter->inbpp=8;
			}
			GstCaps *peer;
			GstStructure* peerstructure;
			
			peer=gst_pad_get_allowed_caps(filter->srcpad);
			if (!filter->silent){
			  gchar* cts=gst_caps_to_string(peer);
			  g_print ("Downstream caps negotiation: %s\n\n",cts);
			  g_free(cts);
			}
			gboolean result;
			result=FALSE;
			int i;
			i=0;
			while ((i<gst_caps_get_size (peer))&&(!result)){
				peerstructure=gst_caps_get_structure(peer,0);
				gst_structure_fixate_field_nearest_int (peerstructure, "width", filter->imagesize.width);
				gst_structure_fixate_field_nearest_int (peerstructure, "height", filter->imagesize.height);
				gst_structure_fixate_field_nearest_fraction (peerstructure, "framerate", filter->rate_numerator, filter->rate_denominator);
				result=gst_pad_set_caps (otherpad, gst_caps_new_full (gst_structure_copy(peerstructure),NULL));
				i++;
			}			
			return result;
		}
		else{			
			if (!gst_structure_get_int(structure,"bpp",&filter->outbpp)){
			filter->outbpp=8;
			}			
		}
		//check whether both sides are available
		if ((filter->sinkc[0]!=0)&&(filter->srcc[0]!=0)){
			sprintf(filter->cspchar,"CV_%s2%s",filter->sinkc,filter->srcc);
			g_print ("Call for Transcoder: %s",filter->cspchar);
			if (!gst_structure_has_field(filter->cvcspcodes,filter->cspchar)){
				g_print ("Sorry, Transcoder option unknown: %s",filter->cspchar);
				return FALSE;
			}
		}
		else{ 
			if (filter->sinkc[0]!=0) g_print("Sinkchar undefined");
			if (filter->srcc[0]!=0) g_print("Sourcechar undefined");	
		}
		g_print ("Caps negotiation succedet.");
		return TRUE;
	  break;
	case CV_MATRIX://test if we like it
		return acceptonechannel(pad, caps);			
	  break;

	default:
	  g_print ("OPmode Unknown");
	  return FALSE;
  }
  return FALSE;
}


static GstCaps *
gst_cv_getcaps (GstPad *pad)
{
  g_print ("GetCaps aufgerufen\n");
  GstCV *filter;

  filter = GST_CV (gst_pad_get_parent (pad));
  GstCaps * caps;

  if (pad == filter->sinkpad){
    g_print ("Das Quellelement macht getcaps\n");
  }
  if (pad == filter->srcpad){
  g_print ("Das Zielelement macht getcaps\n");
  }
  switch (filter->opmode){
	case CV_CANNY://all the same
	case CV_CALCOPTICALFLOWBASE:
	case CV_CALCOPTICALFLOWHS:
  	case CV_CALCOPTICALFLOWLK:
  	case CV_CALCOPTICALFLOWBM:
  	case CV_CALCOPTICALFLOWPYRLK:
	case CV_HOUGH_CIRCLE:
	  caps = gst_caps_new_full( 
		gst_structure_new ("video/x-raw-yuv",
		"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('Y', '8', '0', '0'),
      		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
      		NULL),
		gst_structure_new ("video/x-raw-gray",
		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
		"depth", G_TYPE_INT, 8,
		"bpp", G_TYPE_INT, 8,
		"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
		NULL),
		NULL);
	  return caps;
	  break;
	case CV_CSP:
	   caps = gst_caps_new_full (
			gst_structure_new ("video/x-raw-gray",
			"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
			"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
			"depth", G_TYPE_INT, 8,
			"bpp", G_TYPE_INT, 8,
			"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
			NULL),
      			gst_structure_new ("video/x-raw-yuv",
			"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('Y', '8', '0', '0'),
      			"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
			 NULL),
      			gst_structure_new ("video/x-raw-rgb",
      			"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
			 NULL),
      			gst_structure_new ("video/x-raw-bayer",
      			"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
			 NULL),
      			NULL);
	  return caps;
	  break;
        case CV_SMOOTH_BLUR_NO_SCALE:
        case CV_SMOOTH_BLUR:
        case CV_SMOOTH_GAUSSIAN:
        case CV_SMOOTH_MEDIAN:
        case CV_SMOOTH_BILATERAL:
        case CV_LAPLACE:
	case CV_SOBEL:
	case CV_HAAR:
	  g_print ("Get Caps on Smoothing with %d...\n",filter->estchannels);	
	  if (filter->estchannels==1){
	  caps = gst_caps_new_full( 
		gst_structure_new ("video/x-raw-yuv",
		"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('Y', '8', '0', '0'),
      		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
      		NULL),
		gst_structure_new ("video/x-raw-gray",
		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
		"depth", G_TYPE_INT, 8,
		"bpp", G_TYPE_INT, 8,
		"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
		NULL),
		NULL);
	  }
	  else if ((filter->estchannels==3)||(filter->estchannels==4)){
	  caps = gst_caps_new_full (	
		gst_structure_new ("video/x-raw-rgb",
      		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
		NULL),
		NULL);
	  }
	  else{//not specified yet	
	  	caps = gst_caps_new_full (
			gst_structure_new ("video/x-raw-gray",
			"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
			"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
			"depth", G_TYPE_INT, 8,
			"bpp", G_TYPE_INT, 8,
			"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
			NULL),
      			gst_structure_new ("video/x-raw-yuv",
			"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('Y', '8', '0', '0'),
      			"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
			 NULL),
      			gst_structure_new ("video/x-raw-rgb",
      			"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      			"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
			 NULL),
			NULL);
	  	}
	  return caps;	
	  break;
	default:
	  g_print ("Das was draufsteht eben...\n");
	  caps=gst_caps_copy(gst_pad_get_pad_template_caps (pad));
	  return caps;
  	}
g_print ("Warning!! This point should not be reached\n");
caps=gst_caps_copy(gst_pad_get_pad_template_caps (pad));
return caps;
}



/* chain function
 * this function does the actual processing
 */

static GstFlowReturn
gst_cv_chain (GstPad * pad, GstBuffer * buf)
{
  GstCV *filter;
  GstFlowReturn ret=GST_FLOW_OK;
  filter = GST_CV (GST_OBJECT_PARENT (pad));

  //if (filter->silent == FALSE)
  //g_print ("I'm plugged, therefore I'm in.\n");

  IplImage *process, *process2 , *ipout, *ipout2;		
  GstBuffer* outbuf;
  int i;//only for loops
  switch (filter->opmode){
	case CV_CANNY:
	case CV_LAPLACE:
	case CV_SOBEL:
		process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,1 );
		process->imageData=(char*)buf->data;
		if (filter->peeralloc){
			if (GST_FLOW_OK !=gst_pad_alloc_buffer_and_set_caps (filter->srcpad,
			GST_BUFFER_OFFSET_NONE, (filter->imagesize.width*filter->imagesize.height), GST_PAD_CAPS (filter->srcpad),
			&outbuf)){ 
				g_print ("Warning buffer corrupt.\n");
				outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height);
				gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
			}
		}
		else{
			outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height);
			gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
		}
		outbuf->timestamp=buf->timestamp;
		outbuf->duration=buf->duration;
		outbuf->offset=buf->offset;
		outbuf->offset_end=buf->offset_end;
		ipout=cvCreateImageHeader( filter->imagesize, IPL_DEPTH_8U,1);
		//g_print ("Allocated %d.\n",(filter->imagesize.width*filter->imagesize.height));
		ipout->imageData=(char*)outbuf->data;
		if (filter->opmode==CV_CANNY) cvCanny(process,ipout,filter->pardouble1,filter->pardouble2,filter->cannymasksize); 
		else if (filter->opmode==CV_LAPLACE){
		  IplImage* middle = cvCreateImage(cvSize(filter->imagesize.width,filter->imagesize.height), IPL_DEPTH_16S, 1);
		  cvLaplace(process, middle, filter->cannymasksize);
		  cvConvertScaleAbs( middle, ipout, 1, 0 );
		  cvReleaseImage(&middle );
		}
		else if (filter->opmode==CV_SOBEL){
		  IplImage* middles = cvCreateImage(cvSize(filter->imagesize.width,filter->imagesize.height), IPL_DEPTH_16S, 1);
		  cvSobel(process,middles,1,1, 3 );
		  cvConvertScaleAbs( middles, ipout, 1, 0 );
		  cvReleaseImage(&middles );		
		}
		if (filter->trash){
			gst_buffer_unref(buf);
		}
		else{//return unmodified buffer
		//gst_pad_push (filter->notouchpad, outbuf);
			gst_buffer_unref(buf);
		}
		//g_print ("Sending\n");
		cvReleaseImageHeader(&process);
		cvReleaseImageHeader(&ipout);
		return gst_pad_push (filter->srcpad, outbuf);
		//eventuell zwischenvariablen erzeugen und Format konvertieren
	   break;
	case CV_CSP://adjust Colorspace like required in CAPS
		process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,(filter->inbpp/8) );
		process->imageData=(char*)buf->data;
		if (filter->peeralloc){
			if (GST_FLOW_OK !=gst_pad_alloc_buffer_and_set_caps (filter->srcpad,
			GST_BUFFER_OFFSET_NONE, (filter->imagesize.width*filter->imagesize.height*(filter->outbpp/8)), GST_PAD_CAPS (filter->srcpad),
			&outbuf)){ 
				g_print ("Warning buffer corrupt.\n");
				outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height*(filter->outbpp/8));
				gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
			}
		}
		else{
			outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height*(filter->outbpp/8));
			gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
		}
		//copy some caps stuff
		outbuf->timestamp=buf->timestamp;
		outbuf->duration=buf->duration;
		outbuf->offset=buf->offset;
		outbuf->offset_end=buf->offset_end;
		ipout=cvCreateImageHeader( filter->imagesize, IPL_DEPTH_8U,(filter->outbpp/8));
		//g_print ("Allocated %d. and will copy from a %d Byte each pixel buffer\n",(filter->imagesize.width*filter->imagesize.height*(filter->outbpp/8)),(filter->inbpp/8));
		ipout->imageData=(char*)outbuf->data;
		int temp;
		if (!gst_structure_get_int(filter->cvcspcodes,filter->cspchar,&temp)){
			g_print ("Warning, no opmode.\n");	
			return (GstFlowReturn)FALSE;
		}
		/*if (1){
		  int i;
		  g_print ("Touching inbuffer. size %d\n",buf->size);
		  for (i=0;i<buf->size;i++) buf->data[i]=0;
		  g_print ("Touching outbuffer. size %d\n",outbuf->size);
		  for (i=0;i<outbuf->size;i++) outbuf->data[i]=0;
		}*/
		cvCvtColor(process,ipout, temp); 
		//g_print ("Survived color conversion\n");
		if (filter->trash){
			gst_buffer_unref(buf);
		}
		else{//return unmodified buffer
		//gst_pad_push (filter->notouchpad, outbuf);
			gst_buffer_unref(buf);
		}
		//g_print ("Sending\n");
		cvReleaseImageHeader(&process);
		cvReleaseImageHeader(&ipout);
		return gst_pad_push (filter->srcpad, outbuf);
	   break;
	case CV_CALCOPTICALFLOWBASE:
	case CV_CALCOPTICALFLOWLK:
	case CV_CALCOPTICALFLOWHS:
		;
		//g_print ("Im alive 1.\n");

		if (filter->lastbuf!=NULL){
			process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,1);
			process->imageData=(char*)buf->data;
			process2=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,1 );
			process2->imageData=(char*)filter->lastbuf->data;
			//g_print ("Im alive 2.\n");		
			if (filter->peeralloc){
				if (GST_FLOW_OK !=gst_pad_alloc_buffer_and_set_caps (filter->srcpad,
				GST_BUFFER_OFFSET_NONE, (filter->imagesize.width*filter->imagesize.height), GST_PAD_CAPS (filter->srcpad),&outbuf)){ 
					g_print ("Warning buffer corrupt.\n");
					outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height);
					gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
				}
			}
			else{
				outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height);
				gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
			}
			//g_print ("Im alive 3.\n");
			ipout=cvCreateImageHeader( filter->imagesize, IPL_DEPTH_32F,1);
			//g_print ("Allocated %d.\n",(filter->imagesize.width*filter->imagesize.height));
			ipout->imageData=(char*)filter->resdata1;
			ipout2=cvCreateImageHeader( filter->imagesize, IPL_DEPTH_32F,1);
			ipout2->imageData=(char*)filter->resdata2;
			if(filter->opmode==CV_CALCOPTICALFLOWBASE){
			  int temp, maxchange;
			  maxchange=0;
			  for (temp=0; temp<(filter->imagesize.width*filter->imagesize.height);temp++){
				outbuf->data[temp]=(unsigned char)abs(((int)buf->data[temp]-(int)filter->lastbuf->data[temp]));
				//if (outbuf->data[temp]>maxchange)maxchange=outbuf->data[temp];
			  }
			  g_print ("Noise/Change is %d .\n",maxchange);
			}

			int temp;
			double maxchange, changesum;
			changesum=maxchange=0;

			if(filter->opmode==CV_CALCOPTICALFLOWLK){
				cvCalcOpticalFlowLK( process2, process, filter->win_size, ipout, ipout2);
				for (temp=0; temp<(filter->imagesize.width*filter->imagesize.height);temp++){
				  if (filter->resdata1[temp]>40)outbuf->data[temp]=255;
				  else if (filter->resdata1[temp]<-40)outbuf->data[temp]=0;
				  else outbuf->data[temp]=127+(char)((filter->resdata1[temp]/40.0)*128);
				  //statistic
				  if (filter->resdata1[temp]>maxchange)maxchange=filter->resdata1[temp];
				  changesum+=abs(filter->resdata1[temp])+abs(filter->resdata2[temp]);
				}
			}
			else if (filter->opmode==CV_CALCOPTICALFLOWHS){
				cvCalcOpticalFlowHS(  process2, process, filter->use_previous,
                          ipout, ipout2, filter->lambda,filter->criteria );
				filter->use_previous=1;
				for (temp=0; temp<(filter->imagesize.width*filter->imagesize.height);temp++){
				  if (filter->resdata1[temp]>40)outbuf->data[temp]=255;
				  else if (filter->resdata1[temp]<-40)outbuf->data[temp]=0;
				  else outbuf->data[temp]=127+(char)((filter->resdata1[temp]/40.0)*128);
				  //statistic
				  if (filter->resdata1[temp]>maxchange)maxchange=filter->resdata1[temp];
				  changesum+=abs(filter->resdata1[temp])+abs(filter->resdata2[temp]);
				}
			}


			changesum=changesum/(2.0*(double)(filter->imagesize.width*filter->imagesize.height));//normalize
			filter->pardouble1=changesum;
			filter->pardouble2=maxchange;
			//send positions
			if((TRUE)&&(gst_pad_is_linked(filter->srccoord))){
				GList *temp=NULL,*free;
				GstBuffer* posbuf;
				int numregions;
				numregions=listregions (&temp,filter->resdata1,4,4,filter->imagesize.width ,filter->imagesize.height,filter->scaleup,3.0,10.0);
				posbuf=gst_buffer_new_and_alloc(numregions*sizeof(struct coordinates));
  				posbuf->timestamp=buf->timestamp;
  				posbuf->duration=buf->duration;
  				posbuf->offset=buf->offset;
  				posbuf->offset_end=buf->offset_end;
  				int count=0;
				free=temp;
				while (temp){
				  //for big endian machines
				  if (!little_endian){
				  ((struct coordinates*)(posbuf->data))[count].x1=DoubleSwap(((struct coordinates*)(temp->data))->x1);
				  ((struct coordinates*)(posbuf->data))[count].x2=DoubleSwap(((struct coordinates*)(temp->data))->x2-(float)((1+(int)((struct coordinates*)(temp->data))->x2-(int)((struct coordinates*)(temp->data))->x1)%4));
				  ((struct coordinates*)(posbuf->data))[count].y1=DoubleSwap(((struct coordinates*)(temp->data))->y1);
				  ((struct coordinates*)(posbuf->data))[count].y2=DoubleSwap(((struct coordinates*)(temp->data))->y2-(float)((1+(int)((struct coordinates*)(temp->data))->y2-(int)((struct coordinates*)(temp->data))->y1)%4));
				  }

				  else{
			          ((struct coordinates*)(posbuf->data))[count].x1=((struct coordinates*)(temp->data))->x1;
				  ((struct coordinates*)(posbuf->data))[count].x2=((struct coordinates*)(temp->data))->x2-(float)((1+(int)((struct coordinates*)(temp->data))->x2-(int)((struct coordinates*)(temp->data))->x1)%4);
				  ((struct coordinates*)(posbuf->data))[count].y1=((struct coordinates*)(temp->data))->y1;
				  ((struct coordinates*)(posbuf->data))[count].y2=((struct coordinates*)(temp->data))->y2-(float)((1+(int)((struct coordinates*)(temp->data))->y2-(int)((struct coordinates*)(temp->data))->y1)%4);
				  }
				  //g_print("Motiondetection coordinates x1: %f x2: %f y1: %f y2:%f\n",((struct coordinates*)(posbuf->data))[count].x1,((struct coordinates*)(posbuf->data))[count].x2,((struct coordinates*)(posbuf->data))[count].y1,((struct coordinates*)(posbuf->data))[count].y2);
				  g_free(temp->data);
				  temp=g_list_next(temp);
				  count++;
				}
				GstFlowReturn ret1;
				g_list_free(free);
				GstCaps *tempcaps=gst_caps_copy(gst_pad_get_pad_template_caps(filter->srccoord));
  				gst_buffer_set_caps(posbuf,tempcaps);
				gst_caps_unref(tempcaps);
				g_print ("Pushing buffer on src with fields: %d",numregions);
				ret1=gst_pad_push (filter->srccoord, posbuf);
				if (ret1!=GST_FLOW_OK) return ret1;
			}
			//
			g_print ("Im alive after FlowLK Max shifted Pixel=%f. Average:%f Method Nr. %d\n",maxchange,changesum,filter->opmode);
			outbuf->timestamp=buf->timestamp;
			outbuf->duration=buf->duration;
			outbuf->offset=buf->offset;
			outbuf->offset_end=buf->offset_end;
			gst_buffer_unref(filter->lastbuf);
			filter->lastbuf=buf;
			cvReleaseImageHeader(&process);
			cvReleaseImageHeader(&process2);
			cvReleaseImageHeader(&ipout);
			cvReleaseImageHeader(&ipout2);
			return gst_pad_push (filter->srcpad, outbuf);
		}
		else{
			filter->lastbuf=buf;
			filter->resdata1=malloc((filter->imagesize.width*filter->imagesize.height)*sizeof(float));
			filter->resdata2=malloc((filter->imagesize.width*filter->imagesize.height)*sizeof(float));
			filter->win_size.width=filter->parint1;
			filter->win_size.height=filter->parint2;
			filter->criteria=cvTermCriteria( 1, 10, 0.0 ); 
			filter->use_previous=0;
			filter->lambda=1.0;
			g_print ("Prerolling.\n");
			return GST_FLOW_OK;
		}
	break;
        case CV_SMOOTH_BLUR_NO_SCALE:
        case CV_SMOOTH_BLUR:
        case CV_SMOOTH_GAUSSIAN:
        case CV_SMOOTH_MEDIAN:
        case CV_SMOOTH_BILATERAL:
        case CV_HAAR:	
	case CV_HOUGH_CIRCLE:
		process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,(filter->inbpp/8));
		process->imageData=(char*)buf->data;
		if (filter->peeralloc){
			if (GST_FLOW_OK !=gst_pad_alloc_buffer_and_set_caps (filter->srcpad,
			GST_BUFFER_OFFSET_NONE, (filter->imagesize.width*filter->imagesize.height*(filter->inbpp/8)), GST_PAD_CAPS (filter->srcpad),&outbuf)){ 
				g_print ("Warning buffer corrupt.\n");
				outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height*(filter->inbpp/8));
				gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
			}
		}
		else{
			outbuf=gst_buffer_new_and_alloc (filter->imagesize.width*filter->imagesize.height*(filter->inbpp/8));
			gst_buffer_set_caps (outbuf, GST_PAD_CAPS (filter->srcpad));
		}
		ipout=cvCreateImageHeader( filter->imagesize, IPL_DEPTH_8U,(filter->inbpp/8));
		ipout->imageData=(char*)outbuf->data;
		
		if(filter->opmode==CV_SMOOTH_BLUR_NO_SCALE){
			cvSmooth( process, ipout, CV_BLUR_NO_SCALE,
               			filter->parint1, filter->parint2, filter->pardouble1,filter->pardouble2);
			}
		else if (filter->opmode==CV_SMOOTH_BLUR){
			cvSmooth( process, ipout, CV_BLUR,
               			filter->parint1, filter->parint2, filter->pardouble1 ,filter->pardouble2);
			}
		else if (filter->opmode==CV_SMOOTH_GAUSSIAN){
			cvSmooth( process, ipout, CV_GAUSSIAN,
               			filter->parint1, filter->parint2, filter->pardouble1 ,filter->pardouble2);
			}
		else if (filter->opmode==CV_SMOOTH_MEDIAN){
			cvSmooth( process, ipout, CV_MEDIAN,
               			filter->parint1, filter->parint2, filter->pardouble1 ,filter->pardouble2);
			}
		else if (filter->opmode==CV_SMOOTH_BILATERAL){
			cvSmooth( process, ipout, CV_BILATERAL,
               			filter->parint1, filter->parint2, filter->pardouble1 ,filter->pardouble2);
			}
		else if (filter->opmode==CV_HAAR){
			g_print ("Entering Haar.\n");
			if(!filter->cascade){//laden
				g_print ("Loading Cascade.\n");
				filter->cascade = (CvHaarClassifierCascade*)cvLoad( filter->parstring1, 0, 0, 0 );
				if(!filter->cascade){
					g_print ("Warning: could not load cascade.\n");
					return GST_FLOW_CUSTOM_ERROR;
					}
			}
			CvMemStorage* storage = cvCreateMemStorage(0);
			cvClearMemStorage( storage );
			g_print ("Start Detection Adresses process %p cascade %p storage %p and datafield %p calculated size %d.\n",process, filter->cascade, storage,buf->data,(filter->imagesize.width*filter->imagesize.height*(filter->inbpp/8)));
			g_print ("Width %d, Height %d, Bpp %d \n",filter->imagesize.width,filter->imagesize.height,filter->inbpp);

			CvSeq* faces = cvHaarDetectObjects( process, filter->cascade, storage,
                                            filter->pardouble1, filter->parint1, CV_HAAR_DO_CANNY_PRUNING,
                                            cvSize(30, 30)
#if (CV_MAJOR_VERSION >= 2) && (CV_MINOR_VERSION >= 2)
					   ,cvSize(filter->imagesize.width,filter->imagesize.height) 
#endif
);
			//cvReleaseMemStorage( storage );
			g_print ("Finished detection, start copy and show.\n");
			if (gst_pad_is_linked(filter->srccoord)){
				GstBuffer *outcoord;
				if (faces) outcoord=gst_buffer_new_and_alloc(sizeof(struct coordinates)*(faces->total));
				else outcoord=gst_buffer_new_and_alloc(0);
				for( i = 0; i < (faces ? faces->total : 0); i++ ){
				  CvRect* r = (CvRect*)cvGetSeqElem( faces, i );
				  gint x1,x2,width;
				  width=(int)(filter->scaleup*r->width);
				  x1=(int)(filter->scaleup*r->x);
				  x2=x1+width-((width+1)%4);
				  if (!little_endian){
				  ((struct coordinates*)(outcoord->data))[i].x1=DoubleSwap((double)x1);
				  ((struct coordinates*)(outcoord->data))[i].x2=DoubleSwap((double)x2);
				  ((struct coordinates*)(outcoord->data))[i].y1=DoubleSwap(filter->scaleup*r->y);
				  ((struct coordinates*)(outcoord->data))[i].y2=DoubleSwap((filter->scaleup*r->y)+(filter->scaleup*r->height));
				  }
				  else{
				  ((struct coordinates*)(outcoord->data))[i].x1=(double)x1;
				  ((struct coordinates*)(outcoord->data))[i].x2=(double)x2;
				  ((struct coordinates*)(outcoord->data))[i].y1=(double)(int)(filter->scaleup*r->y);
				  ((struct coordinates*)(outcoord->data))[i].y2=(double)(int)((filter->scaleup*r->y)+(filter->scaleup*r->height));
				  }
				  //g_print("appending face structure x1: %d, x2: %d\n",(int)((struct coordinates*)(outcoord->data))[i].x1,(int)((struct coordinates*)(outcoord->data))[i].x2);


				}

                                gst_buffer_set_caps(outcoord,gst_caps_new_simple ("coordinates-list",
				"width", G_TYPE_INT, filter->imagesize.width,
      				"height", G_TYPE_INT,filter->imagesize.height ,
				NULL));

  				outcoord->timestamp=buf->timestamp;
  				outcoord->duration=buf->duration;
  				outcoord->offset=buf->offset;
  				outcoord->offset_end=buf->offset_end;
				g_print("Pushing srccoord buffer in facemode with timestamp %lu\n",outcoord->offset);
				ret=gst_pad_push (filter->srccoord, outcoord);
				if (ret!=GST_FLOW_OK)g_print("Warning!! GstCV pushing Buffer went wrong\n");	
			}
			memcpy(outbuf->data,buf->data,(filter->imagesize.width*filter->imagesize.height*(filter->inbpp/8)));
			for( i = 0; i < (faces ? faces->total : 0); i++ )
				{
				CvRect* r = (CvRect*)cvGetSeqElem( faces, i );
				//CvPoint center;
				//int radius;
				//center.x = cvRound((r->x + r->width*0.5));
				//center.y = cvRound((r->y + r->height*0.5));
				//radius = cvRound((r->width + r->height)*0.25);
				if (i==1) {filter->roix1=r->x;filter->roiy1=r->y;filter->roix2=r->x+r->width;filter->roix2=r->y+r->height;}
				//cvCircle( ipout, center, radius, CV_RGB ( 255, 255, 255 ), 3, 8, 0 );
				//cvRectangle(ipout, cvPoint(r->x,r->y), cvPoint(r->x+ r->width,r->y + r->height), cvScalar(255,255,255), 3);
				cvRectangle(ipout, cvPoint(r->x,r->y), cvPoint(r->x+ r->width,r->y + r->height),CV_RGB ( 255, 255, 255 ) , 3,8,0);
				}
			//cvClearMemStorage( storage );
			cvReleaseMemStorage( &storage );				
			}
  		
		else if (filter->opmode==CV_HOUGH_CIRCLE){
    		    CvMemStorage* storage = cvCreateMemStorage(0);
		    CvSeq* circles=NULL;
		    circles = cvHoughCircles(process, storage, CV_HOUGH_GRADIENT, 1,2*filter->parint1 , filter->pardouble1,filter->pardouble2 , filter->parint1,filter->parint2);
    		    int i;
		    if (gst_pad_is_linked(filter->srccoord)){
			GstBuffer *outcoord;
			//check because of bug in opencv
			int validcircles=0;
			for( i = 0; i < (circles ? circles->total : 0); i++ ){
			  float* p = (float*)cvGetSeqElem( circles, i );
			  if ((p[2]>=(double)filter->parint1)&&(p[2]<=(double)filter->parint2))validcircles++;
			}
			//g_print("Validcircles is %d\n",validcircles);
			if (validcircles>0) outcoord=gst_buffer_new_and_alloc(sizeof(struct coordinates)*validcircles);
			else{
			  outcoord=gst_buffer_new_and_alloc(sizeof(struct coordinates));
 			  ((struct coordinates*)(outcoord->data))[0].x1=NAN;
			  ((struct coordinates*)(outcoord->data))[0].y1=NAN;
			  ((struct coordinates*)(outcoord->data))[0].x2=NAN;
			  ((struct coordinates*)(outcoord->data))[0].y2=NAN; 
			}
			int v=0;
			for( i = 0; i < (circles ? circles->total : 0); i++ ){
         	                  float* p = (float*)cvGetSeqElem( circles, i );
				  double x,y,radius;
				  radius=filter->scaleup*p[2];
				  x=filter->scaleup*p[0];
				  y=filter->scaleup*p[1];
				  if ((p[2]>=(double)filter->parint1)&&(p[2]<=(double)filter->parint2)){
				    if (!little_endian){
				    ((struct coordinates*)(outcoord->data))[v].x1=DoubleSwap(x-radius);
				    ((struct coordinates*)(outcoord->data))[v].y1=DoubleSwap(y-radius);
				    ((struct coordinates*)(outcoord->data))[v].x2=DoubleSwap(x+radius);
				    ((struct coordinates*)(outcoord->data))[v].y2=DoubleSwap(y+radius);
				    }
				    else{
				    ((struct coordinates*)(outcoord->data))[v].x1=(double)(x-radius);
				    ((struct coordinates*)(outcoord->data))[v].y1=(double)(y-radius);
				    ((struct coordinates*)(outcoord->data))[v].x2=(double)(x+radius);
				    ((struct coordinates*)(outcoord->data))[v].y2=(double)(y+radius);
				    }
				    v++;
				  }
				}

                                gst_buffer_set_caps(outcoord,gst_caps_new_simple ("coordinates-list",
				"width", G_TYPE_INT, filter->imagesize.width,
      				"height", G_TYPE_INT,filter->imagesize.height ,
				NULL));


  			outcoord->timestamp=buf->timestamp;
  			outcoord->duration=buf->duration;
  			outcoord->offset=buf->offset;
  			outcoord->offset_end=buf->offset_end;
			g_print("Pushing srccoord buffer in circlemode with timestamp %lu\n",outcoord->offset);
			ret=gst_pad_push (filter->srccoord, outcoord);
			if (ret!=GST_FLOW_OK)g_print("Warning!! GstCV pushing Buffer went wrong\n");	
		    }
		if (gst_pad_is_linked(filter->srcpad)){
		//if(0){
		   memcpy(outbuf->data,buf->data,(filter->imagesize.width*filter->imagesize.height*(filter->inbpp/8)));
	           for (i = 0; i <(circles ? circles->total : 0); i++) 
    		    {
         	      float* p = (float*)cvGetSeqElem( circles, i );
		      if ((p[2]>=(double)filter->parint1)&&(p[2]<=(double)filter->parint2)){
         	        cvCircle( ipout, cvPoint(cvRound(p[0]),cvRound(p[1])), 
                        3, CV_RGB(0,0,0), -1, 8, 0 );
                        cvCircle( ipout, cvPoint(cvRound(p[0]),cvRound(p[1])), 
                        cvRound(p[2]), CV_RGB(255,255,255), 3, 8, 0 );
		      }
    		    }
		  }
		  cvReleaseMemStorage( &storage );
		}
		if (!filter->silent){
		  gchar* cts=gst_caps_to_string(outbuf->caps);
		  g_print ("Im alive after CV Operation %d. Pushing buffer with %s\n",filter->opmode,cts);
		  g_free(cts);
		}
		outbuf->timestamp=buf->timestamp;
		outbuf->duration=buf->duration;
		outbuf->offset=buf->offset;
		outbuf->offset_end=buf->offset_end;
		gst_buffer_unref(buf);
		cvReleaseImageHeader(&process);
		cvReleaseImageHeader(&ipout);
		if (gst_pad_is_linked(filter->srcpad))return gst_pad_push (filter->srcpad, outbuf);
		else{
		  gst_buffer_unref(outbuf);
		  return ret;
		}
	break;


	default:
	g_print ("Mode not working.\n");
  }
  g_print ("Normally i should not reach this point.\n");
  usleep(10000000);
  return (GstFlowReturn) gst_pad_push (filter->srcpad, buf);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features
 *
 * exchange the string 'plugin' with your elemnt name
 */
static gboolean
plugin_init (GstPlugin * plugin)
{
  /* exchange the strings 'plugin' and 'Template plugin' with your
   * plugin name and description */
  GST_DEBUG_CATEGORY_INIT (gst_cv_debug, "gstcv",
      0, "Use OpenCV inside Gst");

  return gst_element_register (plugin, "gstcv",
      GST_RANK_NONE, GST_TYPE_CV);
}

/* this is the structure that gstreamer looks for to register plugins
 *
 * exchange the strings 'plugin' and 'Template plugin' with you plugin name and
 * description
 */
GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "gstcv",
    "Use OpenCV inside Gst",
    plugin_init, VERSION, "LGPL", "GStreamer", "http://gstreamer.net/")
