/*
 * GStreamer
 * Copyright 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_CV_H__
#define __GST_CV_H__

#include <gst/gst.h>

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_CV \
  (gst_cv_get_type())
#define GST_CV(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_CV,GstCV))
#define GST_CV_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_CV,GstCVClass))
#define GST_IS_CV(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_CV))
#define GST_IS_CV_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_CV))

typedef struct _GstCV      GstCV;
typedef struct _GstCVClass GstCVClass;

typedef enum {
  CV_CANNY,
  CV_CSP,
  CV_MATRIX,
  CV_CALCOPTICALFLOWBASE,
  CV_CALCOPTICALFLOWHS,
  CV_CALCOPTICALFLOWLK,
  CV_CALCOPTICALFLOWBM,
  CV_CALCOPTICALFLOWPYRLK,
  CV_SMOOTH_BLUR_NO_SCALE,
  CV_SMOOTH_BLUR,
  CV_SMOOTH_GAUSSIAN,
  CV_SMOOTH_MEDIAN,
  CV_SMOOTH_BILATERAL,
  CV_HAAR,
  CV_HOUGH_CIRCLE,
  CV_LAPLACE,
  CV_SOBEL
} GstOpMode;

struct coordinates
{
  gdouble x1;
  gdouble y1;
  gdouble x2;
  gdouble y2;
};



struct _GstCV
{
  GstElement element;

  GstPad *sinkpad, *srcpad, *srccoord;
  gchar* parstring1;

  gboolean silent;
  GstOpMode opmode;//what function to carry out
  int inbpp,outbpp,indepth;//parameters of image;
  int estchannels;//number of estimated channels, will reduce output of getcaps
  //int outdepth,outbpp;//dimensions of outoutimage;
  gboolean forceformat;//wether transcode colorspace on demand 
  gboolean cspadjust, show, trash, peeralloc;//whether system needs to adjust colorspace or whether its correct 
//show-wether system changes format to RGB on output pad, trash: 1 modifies or trashes input-buffer, 0 provides it on trashpad
  gdouble pardouble1,pardouble2,scaleup;
  gint cannymasksize,parint1,parint2;
  gint roix1,roiy1,roix2,roiy2;
  CvSize imagesize,outimagesize;
  CvHaarClassifierCascade* cascade;  
  GstStructure* cvcspcodes;
  gchar srcc[10],sinkc[10], cspchar[15];//input for operation modes
  gint rate_numerator,rate_denominator;

//internal variables for motion detection
  GstBuffer* lastbuf;
  float *resdata1,*resdata2;
  CvSize win_size;
  int use_previous;
  CvTermCriteria criteria;
  float lambda;


};

struct _GstCVClass 
{
  GstElementClass parent_class;
};

GType gst_cv_get_type (void);

G_END_DECLS

gdouble DoubleSwap( double f )
{
  union
  {
    double f;
    unsigned char b[8];
  } dat1, dat2;

  dat1.f = f;
  dat2.b[0] = dat1.b[7];
  dat2.b[1] = dat1.b[6];
  dat2.b[2] = dat1.b[5];
  dat2.b[3] = dat1.b[4];
  dat2.b[4] = dat1.b[3];
  dat2.b[5] = dat1.b[2];
  dat2.b[6] = dat1.b[1];
  dat2.b[7] = dat1.b[0];
  return dat2.f;
}



int patchregionlist (GList *list,int width, int height,int orgwidth,int orgheight, double scaleup){
  GList *temp;
  temp=list;
  while(temp){
    ((struct coordinates*)(temp->data))->x1=((struct coordinates*)(temp->data))->x1*((scaleup*(double)orgwidth)/width);
    ((struct coordinates*)(temp->data))->x2=((1+((struct coordinates*)(temp->data))->x2)*((scaleup*(double)orgwidth)/width))-1;
    ((struct coordinates*)(temp->data))->y1=((struct coordinates*)(temp->data))->y1*((scaleup*(double)orgheight)/height);
    ((struct coordinates*)(temp->data))->y2=((1+((struct coordinates*)(temp->data))->y2)*((scaleup*(double)orgheight)/height))-1;
g_print("Patched coordinates x1: %f x2: %f y1: %f y2:%f Scaleup: %f\n",((struct coordinates*)(temp->data))->x1,((struct coordinates*)(temp->data))->x2,((struct coordinates*)(temp->data))->y1,((struct coordinates*)(temp->data))->y2,scaleup);
    temp=g_list_next(temp);
  }
  return 0;
};

int listregions (GList **list,float *indata,int width, int height, int orgwidth,int orgheight,double scaleup, double threshmin, double threshmax){
  float* array;
  int* array_decision;
  int x,y,xorg,yorg,i;
  int x1,x2,y1,y2;
  float thresh=0.0;
  array=g_malloc(width*height*sizeof(float));
  array_decision=g_malloc0(width*height*sizeof(int));
  //g_print("List_regions\n");
  for (x=0;x<width;x++){
    for (y=0;y<height;y++){
      //g_print("List_regions x:%d y:%d\n",x,y);
      float temp=0.0;
      float tempmax=0.0,tempsum=0.0;
      //calculate value
	for (yorg=y*(orgheight/height);yorg<((y+1)*(orgheight/height));yorg++){
	  //g_print("Outer yorg loop %d, limit %d\n",yorg,(y*((orgheight/height)+1)));
          for (xorg=x*(orgwidth/width);xorg<((x+1)*(orgwidth/width));xorg++){
	    //g_print("List_regions x:%d y:%d  xorg: %d, yorg: %d\n",x,y,xorg,yorg);
	    temp=abs(indata[yorg*orgwidth+xorg]);
	    tempsum+=temp;
	    if (temp>tempmax) tempmax=temp;
          }
	}
      //g_print("List_regions, survived array\n");
      array[y*width+x]=0.5*tempmax+(0.5*(tempsum/((orgheight/height)*((orgwidth/width)))));
      if (array[y*width+x]>(2*thresh))thresh=array[y*width+x]/2.0;
    }
  }
//Schwellwert festlegen(<1.0 =ignore, >5 include, in between  )
  if (thresh<threshmin) thresh=threshmin;
  else if (thresh>threshmax) thresh=threshmax;
      //g_print("List_regions, setting thresh to %f\n",thresh);
//check if each will be includet
  int number=0;
  gboolean fail_right=FALSE,fail_down=FALSE;
  for (y=0;y<height;y++){
    for (x=0;x<width;x++){
      if ((array_decision[y*width+x]==0)&&(array[y*width+x]>thresh)){
	//g_print("Including region x=%d y=%d with decision array = %d and motion value = %f\n",x,y,array_decision[y*width+x],array[y*width+x]);
	number++;
	x1=x2=x;
	y1=y2=y;
	array_decision[(y*width)+x]=number;
	while (!(fail_right&&fail_down)){
	  //check if right side belongs to us
	  if ((x2+1)>=width)fail_right=TRUE;
	  else{
	    for (i=y1;i<=y2;i++){
	      fail_right=fail_right||(array[i*width+x2+1]<thresh);
	    }
	    if (!fail_right){//then enlarge region
		x2++;
		for (i=y1;i<=y2;i++){
		  //g_print("setting %d to number %d\n",(i*width+x2),number);
		  array_decision[i*width+x2]=number;
		}
	    }
	  }
	  //check if down side belongs to us
	  if ((y2+1)>=height)fail_down=TRUE;
	  else{
	    for (i=x1;i<=x2;i++){
	      fail_down=fail_down||(array[((y2+1)*width)+i]<thresh);
	    }
	    if (!fail_down){//then enlarge region
		y2++;
		for (i=x1;i<=x2;i++){
		  //g_print("setting %d to number %d\n",(y2*width+i),number);
		  array_decision[y2*width+i]=number;
		}
	    }
	  }	  
	}
	//now the region is extendet prpopperly
	struct coordinates* tempcoord;
  	tempcoord=g_malloc(sizeof(struct coordinates));
  	tempcoord->x1=x1;
  	tempcoord->y1=y1; 
  	tempcoord->x2=x2;
  	tempcoord->y2=y2;
	g_print("appending region with x1=%d, y1=%d, x2=%d, y2=%d\n",x1,y1,x2,y2);
  	*list=g_list_append(*list,(gpointer)tempcoord);
      }
      else g_print("Not including region x=%d y=%d with decision array = %d and motion value = %f\n",x,y,array_decision[y*width+x],array[y*width+x]);
    }
  }
  patchregionlist (*list,width,height,orgwidth,orgheight,scaleup);
  g_free(array);
  g_free(array_decision);
  return number;
};


#endif /* __GST_CV_H__ */
