/* GStreamer
 * Copyright (C) 2006 David A. Schleef <ds@schleef.org>
 *
 * gstinotifysrc.c:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_MULTIFILESRC_H__
#define __GST_MULTIFILESRC_H__

#include <gst/gst.h>
#include <gst/base/gstpushsrc.h>

G_BEGIN_DECLS

#define GST_TYPE_INOTIFYSRC \
  (gst_inotifysrc_get_type())
#define GST_INOTIFYSRC(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_INOTIFYSRC,Gstinotifysrc))
#define GST_INOTIFYSRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_INOTIFYSRC,GstinotifysrcClass))
#define GST_IS_INOTIFYSRC(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_INOTIFYSRC))
#define GST_IS_INOTIFYSRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_INOTIFYSRC))

#define EVENT_SIZE  ( sizeof (struct inotify_event) )
#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )

typedef struct _Gstinotifysrc Gstinotifysrc;
typedef struct _GstinotifysrcClass GstinotifysrcClass;

struct _Gstinotifysrc
{
  GstPushSrc parent;

  gchar *dirname;
  int index;
  int i,length;//running through event buffer, length of current event return
  int fd,wd,childpid;
  int offset;
  char buffer[EVENT_BUF_LEN];
  char filename[1024];
  char *commandeach, *commandonce,*excludestr;//commands to start once and everytime, substring that shall be excluded"~"
  gboolean sigusr,deletefile;//send sigusr to child each run? delete sent files?
  GstCaps *caps;
  int timeout,sleep;//timeout for select calls
  gboolean successful_read;
  fd_set rfds;
  struct timeval tv;

};

struct _GstinotifysrcClass
{
  GstPushSrcClass parent_class;
};

GType gst_inotifysrc_get_type (void);

G_END_DECLS

#endif /* __GST_MULTIFILESRC_H__ */
