/* GStreamer
 * Copyright (C) 2006 Edward Hervey <edward@fluendo.com>
 *
 * gstmultisync.h:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_MULTISYNC_H__
#define __GST_MULTISYNC_H__

#include <gst/gst.h>
#include <gst/base/gstdataqueue.h>

G_BEGIN_DECLS

#define GST_TYPE_MULTISYNC \
  (gst_multisync_get_type())
#define GST_MULTISYNC(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_MULTISYNC,Gstmultisync))
#define GST_MULTISYNC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_MULTISYNC,GstmultisyncClass))
#define GST_IS_MULTISYNC(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_MULTISYNC))
#define GST_IS_MULTISYNC_CLASS(obj) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_MULTISYNC))

typedef struct _Gstmultisync Gstmultisync;
typedef struct _GstmultisyncClass GstmultisyncClass;

typedef enum {
  SYNC_NONE,
  SYNC_OFFSET,
  SYNC_OFFSET_ORDERED,
  SYNC_SORT
} MultiQueueOpMode;

/**
 * Gstmultisync:
 *
 * Opaque #Gstmultisync structure.
 */
struct _Gstmultisync {
  GstElement element;

  /* number of queues */
  guint	nbqueues;

  /* The list of individual queues */
  GList *queues;
  guint32 queues_cookie;

  GstDataQueueSize  max_size, extra_size;

  guint32  counter;	/* incoming object counter */
  guint32  highid;	/* contains highest id of last outputted object */

  GMutex * qlock;	/* Global queue lock (vs object lock or individual */
			/* queues lock). Protects nbqueues, queues, global */
			/* GstmultisyncSize, counter and highid */

  MultiQueueOpMode opmode;	/*Defines type of synchronization*/
  GMutex * synclock;	/* Protects those variables relevant to sync operations*/
  GMutex * sendlock;    /* will be taken within synclock to ensure sending in order*/
  guint64  drop;	/* Dorp any buffers < this value */
  guint64  nextsend;    /* Offset of buffer that is allowed to send next */
  guint64  sent;    	/* Offset of buffer that has completely been sent */
  gpointer nextsq;      /* Singlequeue that is allowed to send next */
  guint64  timeout ;    /* In sync-mode, time that we wait for a buffer */
  gboolean flushing;    /* for sync-modes, sometimes it is better not to g_cond_wait */

  gint nextnotlinked;	/* ID of the next queue not linked (-1 : none) */

  gint numwaiting;	/* number of not-linked pads waiting */
};

struct _GstmultisyncClass {
  GstElementClass parent_class;

  /* signals emitted when ALL queues are either full or empty */
  void (*underrun)	(Gstmultisync *queue);
  void (*overrun)	(Gstmultisync *queue);
};

GType gst_multisync_get_type (void);

G_END_DECLS


#endif /* __GST_MULTISYNC_H__ */
