/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2010 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-multivis
 *
 * FIXME:Describe multivis here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! multivis ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <cv.h>
#include <cxcore.h>

#include <sys/time.h>
#include <ctype.h>
#include "gstmultivis.h"

GST_DEBUG_CATEGORY_STATIC (gst_multivis_debug);
#define GST_CAT_DEFAULT gst_multivis_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_THICKNESS,
  PROP_FONTSIZE,
  PROP_MAINFOV,
  PROP_OVERLAY,
  PROP_ORDEREDYAW,
  PROP_ORDEREDPITCH,
  PROP_ORDEREDHFOV,
  PROP_ORDEREDFOCUS,
  PROP_CURRENTYAW,
  PROP_CURRENTPITCH,
  PROP_CURRENTHFOV,
  PROP_CURRENTFOCUS,
  PROP_DRAG,
  PROP_YAWDISPOFFSET,
  PROP_ALWAYSDISPLAY,
  PROP_MANUALFOCUS,
  PROP_RET,
  PROP_FIXZ,
  PROP_SNAPSHOTNEXT,
  PROP_OBJECTNEXT,
  PROP_ALWAYSGRASP
};

/* The following function will generate an output image .
 *
 * describe the real formats here.
 */
IplImage* RequestRegion(IplImage* pipsrc,IplImage* main, int* row, int* roh ,int* rrx1,int* rry1, int* rrx2, int* rry2,float piphfov,float bighfov,float yaw, float pitch ,int* retval, float* reqhfov,gboolean maponly,float** map){//source PIPimage, Main image, just for width and height info, requested output witdh/height - set to actual, rrx1-rry2 selected region in main image(overwritten if necessaty), field of view of pip image, field of view of main image, yaw/pitch of pip towards big(center is 0,0), retval for error messages.
g_print("Current PIP configuration: hfov: %f, yaw %f, pitch %f. \n",piphfov,yaw,pitch);
gboolean left=FALSE;
if ((main->width-*rrx2)<(*rrx1))left=TRUE;

float pipaspectratio=(float)pipsrc->width/(float)pipsrc->height;
float pipvfov=(piphfov/pipaspectratio);
float availableborderleft=yaw-(piphfov/2.0);
float availableborderright=yaw+(piphfov/2.0);
float availableborderhigh=pitch+((piphfov/pipaspectratio)/2.0);
float availableborderlow=pitch-((piphfov/pipaspectratio)/2.0);

float mainaspectratio=(float)main->width/(float)main->height;
float bigvfov=(bighfov/mainaspectratio);
float requestedborderleft=((float)*rrx1/(float)(main->width)*bighfov)-(bighfov/2.0);
float requestedborderright=((float)*rrx2/(float)(main->width)*bighfov)-(bighfov/2.0);
float requestedborderhigh=-((float)*rry1/(float)(main->height)*bigvfov)+(bigvfov/2.0);
float requestedborderlow=-((float)*rry2/(float)(main->height)*bigvfov)+(bigvfov/2.0);

//determining requested horizontal field of view
float neccesaryhfieldofview=requestedborderright-requestedborderleft;
//what field of view does the vfieldofview imply
float neccesaryhfieldofviewfromv=(requestedborderhigh-requestedborderlow)*pipaspectratio;
if (neccesaryhfieldofview>neccesaryhfieldofviewfromv)*reqhfov=neccesaryhfieldofview;
else *reqhfov=neccesaryhfieldofviewfromv;

      g_print("RequestRegion Info, returning NULL requestedborderleft %f,requestedborderright %f, requestedborderhigh %f requestedborderlow %f\n",requestedborderleft,requestedborderright,requestedborderhigh,requestedborderlow);
      g_print("RequestRegion Info, returning NULL availableborderleft %f,availableborderright %f, availableborderhigh %f availableborderlow %f\n",availableborderleft,availableborderright,availableborderhigh,availableborderlow);

if ((requestedborderleft>=availableborderright)||
    (requestedborderright<=availableborderleft)||
    (requestedborderhigh<=availableborderlow)||
    (requestedborderlow>=availableborderhigh)){
      g_print("RequestRegion Warning, returning NULL requestedborderleft %f,requestedborderright %f, requestedborderhigh %f requestedborderlow %f\n",requestedborderleft,requestedborderright,requestedborderhigh,requestedborderlow);
      g_print("RequestRegion Warning, returning NULL availableborderleft %f,availableborderright %f, availableborderhigh %f availableborderlow %f\n",availableborderleft,availableborderright,availableborderhigh,availableborderlow);
      //sleep(10);
      *row=0;
      *roh=0;
      return NULL;//completely out of frame
    }
if ((requestedborderleft<availableborderleft)||
    (requestedborderright>availableborderright)||
    (requestedborderhigh>availableborderhigh)||
    (requestedborderlow<availableborderlow)){
      g_print("RequestRegion Aberation requestedborderleft %f,requestedborderright %f, requestedborderhigh %f requestedborderlow %f",requestedborderleft,requestedborderright,requestedborderhigh,requestedborderlow);
      g_print(", availableborderleft %f,availableborderright %f, availableborderhigh %f availableborderlow %f\n",availableborderleft,availableborderright,availableborderhigh,availableborderlow);
}
int roix1,roiy1,roix2,roiy2;
if (requestedborderleft>=availableborderleft){//everything OK, calc pixel
     roix1=(int)round(((requestedborderleft-availableborderleft)/piphfov)*(float)pipsrc->width);
   }
else{
     roix1=0;
     *rrx1=(int)round(((availableborderleft+(bighfov/2.0))/bighfov)*(float)main->width);
} 
if (requestedborderright<=availableborderright){//everything OK, calc pixel
     roix2=(int)round(((requestedborderright-availableborderleft)/piphfov)*(float)pipsrc->width);
   }
else{
     roix2=pipsrc->width-1;
     *rrx2=(int)round(((availableborderright+(bighfov/2.0))/bighfov)*(float)main->width);
} 
if (requestedborderhigh<=availableborderhigh){//everything OK, calc pixel
     roiy1=(int)round(((availableborderhigh-requestedborderhigh)/pipvfov)*(float)pipsrc->height);
   }
else{
     roiy1=0;
     *rry1=main->height-(int)round(((availableborderhigh+(bigvfov/2.0))/bigvfov)*(float)main->height);
} 
if (requestedborderlow>=availableborderlow){//everything OK, calc pixel
     roiy2=(int)round(((availableborderhigh-requestedborderlow)/pipvfov)*(float)pipsrc->height);
   }
else{
     roiy2=pipsrc->height-1;
     *rry2=main->height-(int)round(((availableborderlow+(bigvfov/2.0))/bigvfov)*(float)main->height);
} 
g_print("The calculated ROI is x1: %d, y1:%d x2:%d y2:%d. \n",roix1,roiy1,roix2,roiy2);
if ((roix2<=roix1)||(roiy2<=roiy1)){
  g_print("Therefore it is null");
  return NULL;
}
//code path for create region only
if (maponly){
//determine left or right
  int srcwidth=(roix2-roix1);
  int srcheight=(roiy2-roiy1);
  int outwidth=roix2-roix1;
  int outheight=roiy2-roiy1;//normally only as big als available
  if (((roix2-roix1)>*row)||((roiy2-roiy1)>*roh)){
    float outratio=(float)outwidth/(float)outheight;
    if (outwidth>*row){
      outwidth=*row;
      outheight=(int)((float)outwidth/outratio);
    }
    if (outheight>*roh){
      outheight=*roh;
      outwidth=(int)((float)outheight*outratio);
    }
  }
  *row=outwidth;
  *roh=outheight;
  int outx=5;
  if (!left)outx=*rrx2+5;
  int outy=main->height-5-outheight;
  float *outmap=g_malloc(2*sizeof(float)*main->width*main->height);
  int x,y;
  //int bpp=2*sizeof(float);
  for (y=0;y<main->height;y++)for (x=0;x<main->width;x++){
    if ((y<outy)||(y>main->height-5)||(x<outx)||(x>outx+outwidth)){
      outmap[y*main->width+2*x]=-1;
      outmap[y*main->width+2*x+1]=-1;
    }
    else{
      outmap[y*main->width+2*x]=(float)roix1+(((float)(x-outx)/(float)outwidth)*(float)srcwidth);
      outmap[y*main->width+2*x+1]=(float)roiy1+(((float)(y-outy)/(float)outheight)*(float)srcheight);
    }
  }
  *map=outmap;
  return NULL;
}
cvSetImageROI(pipsrc, cvRect(roix1,roiy1,(roix2-roix1),(roiy2-roiy1)));
IplImage* output;
if (((roix2-roix1)<=*row)&&((roiy2-roiy1)<=*roh)){
   g_print("Therefore direct copy ROI. \n");
  //output=cvCreateImage(cvGetSize(pipsrc),
  output=cvCreateImage(cvSize((roix2-roix1),(roiy2-roiy1)),
                               pipsrc->depth,
                               pipsrc->nChannels); 
  cvCopy(pipsrc, output, NULL); 
  *row=(roix2-roix1);
  *roh=(roiy2-roiy1);
}
else{
  g_print("Therefore scaling down ROI. \n");  
  int outwidth=roix2-roix1;
  int outheight=roiy2-roiy1;
  float outratio=(float)outwidth/(float)outheight;
  if (outwidth>*row){
    outwidth=*row;
    outheight=(int)((float)outwidth/outratio);
  }
  if (outheight>*roh){
    outheight=*roh;
    outwidth=(int)((float)outheight*outratio);
  }
  *row=outwidth;
  *roh=outheight;
  output=cvCreateImage(cvSize(outwidth,outheight), pipsrc->depth,pipsrc->nChannels);
  cvResize(pipsrc, output, CV_INTER_LINEAR );
}
g_print("Returning Acwidth %d Acheight %d. \n",*row,*roh);
//
return output;
}


/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
//GST_STATIC_CAPS("video/x-raw-rgb, bpp=32, depth=24, endianness=4321, red_mask=65280, green_mask=16711680, blue_mask=-16777216")
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "video/x-raw-rgb, "
      "width=(int) [ 1, MAX ],"
      "height=(int) [ 1, MAX ],"
      "framerate = (fraction) [ 0, MAX ], " 
      "red_mask=65280,"
      "green_mask=16711680," 
      "blue_mask=-16777216,"
      "bpp = (int) 32, "
      "depth = (int) 24"
    )
    );

static GstStaticPadTemplate sink_factory_sift = GST_STATIC_PAD_TEMPLATE ("sink-sift",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("sift/raw,"
        "framerate = (fraction) [ 0, MAX ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];"
	"sift/lowe;"
	"sift/oxf;"
	"sift/lowelight,"
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];")
    );




static GstStaticPadTemplate sink_factory_all = GST_STATIC_PAD_TEMPLATE ("sink-all",
    GST_PAD_SINK,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS (
      "uber-coordinates;"
      "uber-coordinates-list;"
      "coordinates-list"
    )
    );



static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "video/x-raw-rgb, "
      "width=(int) [ 1, MAX ],"
      "height=(int) [ 1, MAX ],"
      "framerate = (fraction) [ 0, MAX ], " 
      "red_mask=65280,"
      "green_mask=16711680," 
      "blue_mask=-16777216,"
      "bpp = (int) 32, "
      "depth = (int) 24"
    )
    );

static GstStaticPadTemplate src_camconfig = GST_STATIC_PAD_TEMPLATE ("srccamconfig",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("camconfig")
    );

static GstStaticPadTemplate sink_camconfig = GST_STATIC_PAD_TEMPLATE ("sinkcamconfig",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("camconfig")
    );

static GstStaticPadTemplate src_grasp = GST_STATIC_PAD_TEMPLATE ("srcgrasp",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("grasp")
    );

static GstStaticPadTemplate src_map = GST_STATIC_PAD_TEMPLATE ("srcmap",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
	"map/float,"
        "width=(int) [ 1, MAX ],"
        "height=(int) [ 1, MAX ]"
    )
    );

GST_BOILERPLATE (Gstmultivis, gst_multivis, GstElement,
    GST_TYPE_ELEMENT);

static void gst_multivis_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_multivis_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstPad *gst_multivis_request_new_pad (GstElement * element,
    GstPadTemplate * temp, const gchar * name);
static void gst_multivis_release_pad (GstElement * element, GstPad * pad);

static gboolean gst_multivis_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps * gst_multivis_get_caps (GstPad *pad);
static GstFlowReturn gst_multivis_chain (GstPad * pad, GstBuffer * buf);
static GstStateChangeReturn
gst_multivis_change_state (GstElement *element, GstStateChange transition);
static gboolean
gst_multivis_event (GstPad  *pad, GstEvent * event);
/* GObject vmethod implementations */

static void
gst_multivis_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "multivis",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory_sift));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory_all));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_camconfig));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_camconfig));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_map));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_grasp));


}

/* initialize the multivis's class */
static void
gst_multivis_class_init (GstmultivisClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_multivis_set_property;
  gobject_class->get_property = gst_multivis_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, PROP_THICKNESS,
      g_param_spec_int ("thickness", "Thickness", "thickness of font",
          1,10,1, G_PARAM_READWRITE)); 
   g_object_class_install_property (gobject_class, PROP_FONTSIZE,
      g_param_spec_int ("fontsize", "fontsize", "size of font",
          1,100,10, G_PARAM_READWRITE));   
   g_object_class_install_property (gobject_class, PROP_MAINFOV,
      g_param_spec_float ("mainhfov", "Mainhfov", "Horizontal field of view of the main image",
          G_MINFLOAT,360,360, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_OVERLAY,
      g_param_spec_boolean ("pipoverlay", "Pipoverlay", "Directly overlay Pip Image, otherwise generates map",
          TRUE, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, PROP_ORDEREDYAW,
      g_param_spec_float ("yaw", "Yaw", "Yaw angle for PTZ camera (ordered)",
          -180,180,0, G_PARAM_READWRITE));  
   g_object_class_install_property (gobject_class, PROP_ORDEREDPITCH,
      g_param_spec_float ("pitch", "Pitch", "Pitch angle for PTZ camera (ordered)",
          -90,90,0, G_PARAM_READWRITE));  
   g_object_class_install_property (gobject_class, PROP_ORDEREDHFOV,
      g_param_spec_float ("hfov", "Hfov", "Horizontal field of view of the pip image (ordered)",
          G_MINFLOAT,360,360, G_PARAM_READWRITE)); 
   g_object_class_install_property (gobject_class, PROP_ORDEREDFOCUS,
      g_param_spec_float ("focus", "Focus", "Focus for manual focus operation (ordered)",
          -1,G_MAXFLOAT,60, G_PARAM_READWRITE));  
   g_object_class_install_property (gobject_class, PROP_CURRENTYAW,
      g_param_spec_float ("currentyaw", "Currentyaw", "Yaw angle for PTZ camera (current)",
          -180,180,0, G_PARAM_READABLE));  
   g_object_class_install_property (gobject_class, PROP_CURRENTPITCH,
      g_param_spec_float ("currentpitch", "Currentpitch", "Pitch angle for PTZ camera (current)",
          -90,90,0, G_PARAM_READABLE));  
   g_object_class_install_property (gobject_class, PROP_CURRENTHFOV,
      g_param_spec_float ("currenthfov", "Currenthfov", "Horizontal field of view of the pip image (current)",G_MINFLOAT,360,360, G_PARAM_READABLE));  
   g_object_class_install_property (gobject_class, PROP_CURRENTFOCUS,
      g_param_spec_float ("currentfocus", "Currentfocus", "Current falue of focus",-1,G_MAXFLOAT,60, G_PARAM_READABLE));  
  g_object_class_install_property (gobject_class, PROP_DRAG,
      g_param_spec_boolean ("drag", "Drag", "Drag image to change the displayed region",
          FALSE, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, PROP_YAWDISPOFFSET,
      g_param_spec_int64 ("yawdisplayoffset", "Yawdisplayoffset", "Yaw display offset in pixel",
          0,G_MAXINT64,0, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_ALWAYSDISPLAY,
      g_param_spec_boolean ("alwaysdisplay", "Alwaysdispaly", "Always overlay PIP image at fixed position (can be dragged)",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MANUALFOCUS,
      g_param_spec_boolean ("manualfocus", "Manualfocus", "Manually operate focus",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_RET,
      g_param_spec_boolean ("return", "Return", "Yaan PTU returns home ?",FALSE, G_PARAM_WRITABLE));
   g_object_class_install_property (gobject_class, PROP_FIXZ,
      g_param_spec_int ("fixz", "Fixz", "Fix the Z distance of object to camera, instead of detecting",
          -1,G_MAXINT,-1, G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_SNAPSHOTNEXT,
      g_param_spec_boolean ("snapshot", "Snapshot", "Set true to take snapshot of next frame",FALSE, G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_OBJECTNEXT,
      g_param_spec_boolean ("objectshot", "Objectshot", "Set true to take snapshot of next selected object",FALSE, G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_ALWAYSGRASP,
      g_param_spec_boolean ("alwaysgrasp", "Alwaysgrasp", "Grasp all SIFT-Detected Objects",FALSE, G_PARAM_READWRITE));

  gstelement_class->request_new_pad =
      GST_DEBUG_FUNCPTR (gst_multivis_request_new_pad);
  gstelement_class->release_pad =
      GST_DEBUG_FUNCPTR (gst_multivis_release_pad);
  gstelement_class->change_state = gst_multivis_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_multivis_init (Gstmultivis * filter,
    GstmultivisClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_multivis_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_multivis_chain));
  gst_pad_set_event_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));


  filter->sinkcamconfig = gst_pad_new_from_static_template (&sink_camconfig, "sinkcamconfig");
  gst_pad_set_setcaps_function (filter->sinkcamconfig,
                                GST_DEBUG_FUNCPTR(gst_multivis_set_caps));
  gst_pad_set_getcaps_function (filter->sinkcamconfig,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_chain_function (filter->sinkcamconfig,
                              GST_DEBUG_FUNCPTR(gst_multivis_chain));
  gst_pad_set_event_function (filter->sinkcamconfig,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));




  filter->sinksift = gst_pad_new_from_static_template (&sink_factory_sift, "sink-sift");
  gst_pad_set_setcaps_function (filter->sinksift,
                                GST_DEBUG_FUNCPTR(gst_multivis_set_caps));
  gst_pad_set_getcaps_function (filter->sinksift,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_chain_function (filter->sinksift,
                              GST_DEBUG_FUNCPTR(gst_multivis_chain));
  gst_pad_set_event_function (filter->sinksift,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));
   
  
  filter->sinkpip = gst_pad_new_from_static_template (&sink_factory, "sink-pip");
  gst_pad_set_setcaps_function (filter->sinkpip,
                                GST_DEBUG_FUNCPTR(gst_multivis_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpip,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_chain_function (filter->sinkpip,
                              GST_DEBUG_FUNCPTR(gst_multivis_chain));
  gst_pad_set_event_function (filter->sinkpip,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));
  
  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_event_function (filter->srcpad,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));

  filter->srccamconfig = gst_pad_new_from_static_template (&src_camconfig, "srccamconfig");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));

  filter->srcsnapshot = gst_pad_new_from_static_template (&src_factory, "srcsnapshot");
  gst_pad_set_getcaps_function (filter->srcsnapshot,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_event_function (filter->srcsnapshot,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));

  filter->srcselection = gst_pad_new_from_static_template (&src_factory, "srcselection");
  gst_pad_set_getcaps_function (filter->srcselection,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_event_function (filter->srcselection,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));

  filter->srcmap = gst_pad_new_from_static_template (&src_map, "srcmap");
  gst_pad_set_getcaps_function (filter->srcmap,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_event_function (filter->srcmap,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));


  filter->srcgrasp = gst_pad_new_from_static_template (&src_grasp, "srcgrasp");
  gst_pad_set_getcaps_function (filter->srcgrasp,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_event_function (filter->srcgrasp,
                              GST_DEBUG_FUNCPTR(gst_multivis_event));


  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpip);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinksift);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srccamconfig);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkcamconfig);

  gst_element_add_pad (GST_ELEMENT (filter), filter->srcselection);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcsnapshot);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcmap);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcgrasp);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);

  filter->objectbuf=NULL;
  filter->silent = TRUE;
  filter->mutex=g_mutex_new();
  filter->siftlatency=malloc(sizeof(struct avuint64));
  filter->siftavg=malloc(sizeof(struct avuint64));
  filter->siftbetween=malloc(sizeof(struct avuint64));  
  filter->objlatency=malloc(sizeof(struct avuint64));
  filter->objavg=malloc(sizeof(struct avuint64));
  filter->objbetween=malloc(sizeof(struct avuint64));
  filter->mainlatency=malloc(sizeof(struct avuint64));
  filter->mainbetween=malloc(sizeof(struct avuint64)); 
  initavuint64(filter->siftlatency,64, 6);
  initavuint64(filter->siftavg,256, 8);  
  initavuint64(filter->siftbetween,64, 6);
  initavuint64(filter->objlatency,64, 6);
  initavuint64(filter->objavg,256, 8);  
  initavuint64(filter->objbetween,64, 6);
  initavuint64(filter->mainlatency,256, 8);  
  initavuint64(filter->mainbetween,256, 8);  
  filter->timeout=15;
  filter->thickness=2;
  filter->fontsize=1.0;
  filter->lastbuffer=G_MAXUINT64;
  filter->firstbufts=G_MAXUINT64;
  filter->starttime=G_MAXUINT64;
  cvInitFont( &filter->font, CV_FONT_VECTOR0, filter->fontsize*1.0,
                 filter->fontsize*1.0, 0,filter->thickness, 0 );
  //PIP-stuff
  filter->pipoverlay=1;//default is on
  //filter->mainfocal=1.7;
  //filter->pipfocalmin=12;
  //filter->pipfocalmax=120;
  filter->minbox=10;
  g_print("Multivis: Activating resendcamconfig for PTUCONTROL\n");
  filter->resendcamconfig=TRUE;
  filter->mainhfov=360;
  filter->currentlyorderedhfov=filter->mainhfov;
  filter->currentconf=NULL;
  filter->ccbuf=NULL;

  filter->selectionnext=FALSE;
  filter->snapshotnext=FALSE;

  filter->numcsent=0;
  filter->numcrecv=0;

  filter->resendmap=TRUE;
  filter->acheight=0;
  filter->acwidth=0;

  filter->alwaysdisplay=FALSE;

  filter->asx=5;
  filter->asy=5;
  filter->asscale=1;

  filter->orderedyaw=0.0;
  filter->orderedpitch=0.0;

  filter->graspmode=FALSE;
  filter->graspx=-1;
  filter->graspy=-1;

  filter->docc=FALSE;
  filter->fixz=-1;

  filter->padsets=NULL;
  filter->alwaysgrasp=FALSE;
}




static void
gst_multivis_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstmultivis *filter = GST_MULTIVIS (object);
  g_mutex_lock(filter->mutex); 
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_THICKNESS:
      filter->thickness = g_value_get_int(value);
      break; 
    case PROP_FONTSIZE:
      filter->fontsize = (gdouble)g_value_get_int(value)/20.0;
      break;
    case PROP_MAINFOV:
      filter->mainhfov = g_value_get_float(value);
      break; 
    case PROP_OVERLAY:
      filter->pipoverlay = g_value_get_boolean (value);
      break;
    case PROP_ORDEREDYAW:
      filter->orderedyaw = g_value_get_float(value);
      filter->resendcamconfig=TRUE;
      break;  
    case PROP_ORDEREDPITCH:
      filter->orderedpitch = g_value_get_float(value);
      filter->resendcamconfig=TRUE;
      break;
    case PROP_ORDEREDHFOV:
      filter->currentlyorderedhfov = g_value_get_float(value);
      filter->resendcamconfig=TRUE;
      break;  
    case PROP_ORDEREDFOCUS:     
      filter->orderedfocus= g_value_get_float(value);
      if (filter->manualfocus)filter->resendcamconfig=TRUE;
      break;  
    case PROP_DRAG:
      filter->drag = g_value_get_boolean (value);
      break; 
    case PROP_YAWDISPOFFSET:
      filter->yawoffset = g_value_get_int64(value);
      if (filter->imagesize.width>0)filter->yawoffset = filter->yawoffset % filter->imagesize.width;
      break;
    case PROP_ALWAYSDISPLAY:
      filter->alwaysdisplay = g_value_get_boolean (value);
      break; 
    case PROP_MANUALFOCUS:
      {
      gboolean newfocus=g_value_get_boolean (value);
      if ((newfocus)&&(!filter->manualfocus))filter->resendcamconfig=TRUE;
      if ((!newfocus)&&(filter->manualfocus))filter->resendcamconfig=TRUE;
      filter->manualfocus=newfocus;
      }
      break;
    case PROP_RET:
      filter->docc = g_value_get_boolean (value);
      break;
    case PROP_FIXZ:
      filter->fixz =g_value_get_int (value);
      break; 
    case PROP_SNAPSHOTNEXT:
      filter->snapshotnext = g_value_get_boolean (value);
      break;
    case PROP_OBJECTNEXT:
      filter->selectionnext = g_value_get_boolean (value);
      break;
    case PROP_ALWAYSGRASP:
      filter->alwaysgrasp=g_value_get_boolean (value);
      break; 
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }

  cvInitFont( &filter->font, CV_FONT_VECTOR0, filter->fontsize*1.0,
                 filter->fontsize*1.0, 0,filter->thickness, 0 ); 
  g_mutex_unlock(filter->mutex);   
}

static void
gst_multivis_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstmultivis *filter = GST_MULTIVIS (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_THICKNESS:
      g_value_set_int (value, filter->thickness);
      break;  
    case PROP_FONTSIZE:
      g_value_set_int (value, (int)(filter->fontsize*20));
      break;     
    case PROP_MAINFOV:
      g_value_set_float (value, filter->mainhfov); 
      break;  
    case PROP_OVERLAY:
       g_value_set_boolean (value, filter->pipoverlay);
      break; 
    case PROP_ORDEREDYAW:
      g_value_set_float (value, filter->orderedyaw); 
      break;       
    case PROP_ORDEREDPITCH:
      g_value_set_float (value, filter->orderedpitch); 
      break; 
    case PROP_ORDEREDHFOV:
      g_value_set_float (value, filter->currentlyorderedhfov); 
      break; 
    case PROP_ORDEREDFOCUS:
      g_value_set_float (value, filter->orderedfocus); 
      break;
    case PROP_CURRENTYAW:
      g_print("CURRENTYAW\n");
      if (filter->currentconf)
          g_value_set_float (value, filter->currentconf->yaw);
      else g_value_set_float (value, 0);
      break; 
    case PROP_CURRENTPITCH:
      if (filter->currentconf)
          g_value_set_float (value, filter->currentconf->pitch);
      else g_value_set_float (value, 0);
      break; 
    case PROP_CURRENTHFOV:
      if (filter->currentconf)
          g_value_set_float (value, filter->currentconf->hfov);
      else g_value_set_float (value, 0);
      break; 
    case PROP_CURRENTFOCUS:
      if (filter->currentconf)
          g_value_set_float (value, filter->currentconf->focus);
      else g_value_set_float (value, -1);
      break; 
    case PROP_DRAG:
       g_value_set_boolean (value, filter->drag);
      break; 
    case PROP_YAWDISPOFFSET:
      g_value_set_int64 (value, filter->yawoffset); 
      break; 
    case PROP_ALWAYSDISPLAY:
       g_value_set_boolean (value, filter->alwaysdisplay);
      break; 
    case PROP_MANUALFOCUS:
       g_value_set_boolean (value, filter->manualfocus);
      break; 
    case PROP_FIXZ:
       g_value_set_int (value, filter->fixz);
      break; 
    case PROP_ALWAYSGRASP:
       g_value_set_boolean (value, filter->alwaysgrasp);
      break; 
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_multivis_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstmultivis *filter;
  //GstPad *otherpad;
  filter = GST_MULTIVIS (gst_pad_get_parent (pad));
  gchar* cts=gst_caps_to_string (caps);
  if (!filter->silent)g_print("Multivis set caps: %s and trying to get mutex!!\n",cts);
  g_free(cts);


  gboolean ret=TRUE;
  g_mutex_lock(filter->mutex);
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);  
  cts=gst_caps_to_string (caps);
  if (!filter->silent)g_print("Multivis set caps: %s!!\n",cts);
  g_free(cts);
  if (pad == filter->sinkcamconfig)ret=TRUE;
  else if (pad == filter->sinksift){
    if (!gst_structure_get_int(structure,"width",&(filter->siftwidth))) filter->siftwidth=-1;
    if (!gst_structure_get_int(structure,"height",&(filter->siftheight))) filter->siftheight=-1;
  }
  else if (pad == filter->sinkpad){//foreward negotiation, save size and bpp
    if (!filter->silent)g_print("Multivis caps sink 1!!\n");
    if (strcmp (mime, "video/x-raw-rgb")) ret=FALSE;
    if (!filter->silent)g_print("Multivis caps caps sink 1!!\n");
    if (!gst_structure_get_int(structure,"width",&(filter->imagesize.width))) ret=FALSE;
    if (!filter->silent)g_print("Multivis caps set caps sink 1!!\n");
    if (!gst_structure_get_int(structure,"height",&(filter->imagesize.height))) ret=FALSE;
    if (!filter->silent)g_print("Multivis caps set caps sink 1!!\n");
    if (!gst_structure_get_int(structure,"bpp",&(filter->bpp))) ret=FALSE;
    gst_video_parse_caps_framerate(caps,&filter->fps_n,&filter->fps_d);
    if(ret){
      ret=gst_pad_set_caps(filter->srcpad,caps);
    }
  }
  else if (pad == filter->sinkpip){//foreward negotiation, save size and bpp
    if (!filter->silent)g_print("Multivis caps sink PIP!!\n");
    if (strcmp (mime, "video/x-raw-rgb")) ret=FALSE;
    if (!gst_structure_get_int(structure,"width",&(filter->pipimagesize.width))) ret=FALSE;
    if (!filter->silent)g_print("Multivis caps set caps PIPsink!!\n");
    if (!gst_structure_get_int(structure,"height",&(filter->pipimagesize.height))) ret=FALSE;
    if (!filter->silent)g_print("Multivis caps set caps PIPsink!!\n");
    //if (!gst_structure_get_int(structure,"bpp",&(filter->bpp))) ret=FALSE;
  }
  else{
    if (!filter->silent)g_print("Multivis searching for Request Pad due to SetCaps Call!!\n");    
    struct padset* thispadset=NULL;
    GList *templ;
    templ=filter->padsets;
    while((!thispadset)&&(templ)){
      if(((struct padset*)templ->data)->sinkpad==pad)thispadset=(struct padset*)templ->data;
      templ=g_list_next(templ);
    }
    if(thispadset){
      if (!filter->silent)g_print("Marker found one of the padsets\n");
      if (!strcmp (mime, "uber-coordinates"))thispadset->mode=0;
      if (!strcmp (mime, "uber-coordinates-list"))thispadset->mode=1;
      if (!strcmp (mime, "coordinates-list"))thispadset->mode=2;
      if (!gst_structure_get_int(structure,"width",&(thispadset->width)))thispadset->width=-1;
      if (!gst_structure_get_int(structure,"height",&(thispadset->height))) thispadset->height=-1;
      thispadset->name=gst_element_get_name(gst_pad_get_parent_element(gst_pad_get_peer(pad))); 
      if (!thispadset->name)g_print("Multivis warning, could not get name for padset!!\n");    
    }
  }

  g_mutex_unlock(filter->mutex);
  gst_object_unref (filter);

  return ret;
}

static GstCaps *
gst_multivis_get_caps (GstPad *pad){
  //if (!filter->silent)g_print("Multivis: Start get caps!!\n");
  GstCaps * ret;
  //Gstmultivis *filter;
  //filter = GST_MULTIVIS (GST_OBJECT_PARENT (pad));
  ret=gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  //g_print("Multivis, PadName %s, returning %s",gst_pad_get_name(pad),gst_caps_to_string(ret));
  return ret;
}

static GstPad *gst_multivis_request_new_pad (GstElement * element,
    GstPadTemplate * temp, const gchar * name){
    gchar* tmp;
  Gstmultivis *filter;
  struct padset* temppadset;
  temppadset=malloc(sizeof(struct padset));
  temppadset->buffer=NULL;
  temppadset->name=NULL;
  filter = GST_MULTIVIS (element);
  g_mutex_lock(filter->mutex);
  if (!filter->silent)g_print("Multivis: New pad with name %s requested\n",name);

  tmp = g_strdup_printf ("sink-all%d", g_list_length(filter->padsets));
  temppadset->sinkpad = gst_pad_new_from_static_template (&sink_factory_all, tmp);
  gst_pad_set_setcaps_function (temppadset->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_multivis_set_caps));
  gst_pad_set_getcaps_function (temppadset->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_multivis_get_caps));
  gst_pad_set_chain_function (temppadset->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_multivis_chain));  
  
  g_free (tmp);

  

  filter->padsets=g_list_append(filter->padsets,temppadset);
  
  gst_element_add_pad (GST_ELEMENT (filter), temppadset->sinkpad);

  temppadset->latency=malloc(sizeof(struct avuint64));
  temppadset->avg=malloc(sizeof(struct avuint64));
  temppadset->between=malloc(sizeof(struct avuint64));  
  initavuint64(temppadset->latency,64, 6);
  initavuint64(temppadset->avg,256, 8);  
  initavuint64(temppadset->between,64, 6);
  
  g_mutex_unlock(filter->mutex);
  return temppadset->sinkpad;
}


static void gst_multivis_release_pad (GstElement * element, GstPad * pad){
  
}


//append object buffer
static int
gst_multivis_appendobject (Gstmultivis *filter,struct padset* thispadset,GstBuffer * buf){ 
 //gboolean found=FALSE;
 GstBuffer *prior=NULL;
 GList* temp=filter->objectbuf;
 while ((temp)&&(!prior)){
   if (!strcmp(((struct uber_coordinates*)(((GstBuffer*)(temp->data))->data))->name,((struct uber_coordinates*)buf->data)->name)) prior=temp->data;
   else temp=g_list_next(temp);
 }
 if (prior){
   if (prior->timestamp>buf->timestamp)gst_buffer_unref(buf);
   else{
     avuint64app(filter->objbetween,buf->timestamp-prior->timestamp );
     avuint64app(thispadset->between,buf->timestamp-prior->timestamp );
     buf->duration=buf->timestamp-prior->timestamp;//
     gst_buffer_unref(prior);
     temp->data=buf;//replace   
   }
 }
 else{
   filter->objectbuf=g_list_append(filter->objectbuf,buf);
 }
 if (!filter->silent)g_print("MultiVIS Obj pad append after %d and pointer is %p\n",g_list_length(filter->objectbuf),filter->objectbuf);
 return TRUE; 
}

static int
gst_multivis_sweepobject (Gstmultivis *filter){//remove all object informations older than threshold
 struct timeval now;
 guint64 timenow;
 timenow=((guint64)now.tv_sec*1000000000ULL)+((guint64)now.tv_usec*1000ULL);  
 GList* temp=filter->objectbuf;
 while (temp){
   if (timenow-((GstBuffer*)(temp->data))->timestamp>(filter->timeout*1000000000ULL)){
     //filter->objectbuf=g_list_delete_link(filter->objectbuf,temp);
     //temp=filter->objectbuf;
     temp=g_list_next(temp);
   }
   else temp=g_list_next(temp);
 }
 return TRUE; 
}
/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_multivis_chain (GstPad * pad, GstBuffer * buf)
{
  Gstmultivis *filter;
  //g_print("GStreamer Multivis triggered, locking mutex\n");
  filter = GST_MULTIVIS (GST_OBJECT_PARENT (pad));
  g_mutex_lock(filter->mutex);
  //g_print("GStreamer Multivis triggered, locked mutex\n");
  if (filter->firstbufts==G_MAXUINT64)filter->firstbufts=buf->timestamp;
  struct timeval now;
  guint64 timenow;
  gettimeofday(&now, NULL);
  int i,n;
  if (!filter->silent)g_print("whats up with the time??? %d   %d  \n",(int)now.tv_sec,(int)now.tv_usec);
  timenow=(((guint64)(now.tv_sec))*1000000000ULL)+(((guint64)(now.tv_usec))*1000ULL);
  if (filter->starttime==G_MAXUINT64)filter->starttime=timenow;  
  if (filter->firstbufts<1000000000UL){
	timenow=timenow-filter->starttime;
	if (timenow<buf->timestamp)timenow=buf->timestamp;
  }
  if (!filter->silent)g_print("GStreamer Multivis, vorher bin ich denn blöd??? Firstbufts= %llu Timenow %llu, Buffer Timestamp=%llu\n",(long long unsigned int)filter->firstbufts,(long long unsigned int)timenow,(long long unsigned int)buf->timestamp);
  if (!filter->silent)g_print("GStreamer Multivis, nachher bin ich denn blöd??? Firstbufts= %llu Timenow %llu\n",(long long unsigned int)filter->firstbufts,(long long unsigned int)timenow);
  
  //g_print("GStreamer Marker Got Mutex\n");
  gint ankery=3+filter->fontsize*25;
  if (pad == filter->sinkcamconfig){
    if (!filter->silent)g_print("Camconfigpad\n");
    if (filter->ccbuf)gst_buffer_unref(filter->ccbuf);
    filter->ccbuf=buf;
    filter->currentconf=(struct camconfig*)buf->data;
    swapcamconfig(filter->currentconf);
    g_mutex_unlock(filter->mutex);
    filter->numcrecv++;
    filter->resendmap=TRUE;
    return GST_FLOW_OK;    
  }
  else if (pad == filter->sinksift){
        //g_print("Sift Pad\n");
	if (filter->siftbuf){
	  if (filter->siftbuf->timestamp>buf->timestamp)gst_buffer_unref(buf);
	  else{
	    avuint64app(filter->siftbetween,buf->timestamp-filter->siftbuf->timestamp );
	    gst_buffer_unref(filter->siftbuf);
	    filter->siftbuf=buf;
	  }
	}
	else filter->siftbuf=buf;
	avuint64app(filter->siftlatency,timenow-filter->siftbuf->timestamp);
	g_mutex_unlock(filter->mutex);
	return GST_FLOW_OK;
  }
  else if (pad == filter->sinkpip){
        if (!filter->silent)g_print("MultiVIS PIP pad \n");
	if (filter->pipbuf)gst_buffer_unref(filter->pipbuf);
	filter->pipbuf=buf;
        g_mutex_unlock(filter->mutex);
	return GST_FLOW_OK;
  }
  else if (pad!=filter->sinkpad){
    if (!filter->silent)g_print("MultiVIS one of the Sink-All pad soisses\n");
    struct padset* thispadset=NULL;
    GList *templ=filter->padsets;
    while((!thispadset)&&(templ)){
      if(((struct padset*)templ->data)->sinkpad==pad)thispadset=(struct padset*)templ->data;
      templ=g_list_next(templ);
    }
    if(thispadset){
      if ((thispadset->mode==1)||(thispadset->mode==2)){//if we have a list mode just store the buffer and update timestamp
	if (thispadset->buffer){
	  if (thispadset->buffer->timestamp>buf->timestamp)gst_buffer_unref(buf);//if new one is newer
	  else{
	    avuint64app(thispadset->between,buf->timestamp-thispadset->buffer->timestamp );
	    gst_buffer_unref(thispadset->buffer);
	    thispadset->buffer=buf;
	  }
	}
	else thispadset->buffer=buf;
	avuint64app(thispadset->latency,timenow-thispadset->buffer->timestamp);	
	g_mutex_unlock(filter->mutex);
	return GST_FLOW_OK;
      }
      else{//
	if (!filter->silent)g_print("MultiVIS Universal pad detected object\n");
        avuint64app(thispadset->latency,timenow-buf->timestamp);
        avuint64app(filter->objlatency,timenow-buf->timestamp);
        gst_multivis_appendobject (filter,thispadset,buf);
        g_mutex_unlock(filter->mutex);
	return GST_FLOW_OK;
      } 	
    }
  }
  if (42) buf=gst_buffer_make_writable(buf);

  //kick too old buffers
  if ((filter->siftbuf)&&(timenow-filter->siftbuf->timestamp>(filter->timeout*1000000000UL))){
     if (!filter->silent)g_print("Killing Siftbuf because too old Timenow %llu   Buffer Timestamp: %llu Difference %llu\n",(long long unsigned int)timenow,(long long unsigned int)filter->siftbuf->timestamp,(long long unsigned int)(timenow-filter->siftbuf->timestamp));
     gst_buffer_unref(filter->siftbuf);
     filter->siftbuf=NULL;
   }  
  if ((filter->pipbuf)&&(filter->pipbuf->timestamp<timenow)&&(timenow-filter->pipbuf->timestamp>(filter->timeout*1000000000UL))){
     if (!filter->silent)g_print("Killing Pipbuf because too old Timenow %llu   Buffer Timestamp: %llu Difference %llu\n",(long long unsigned int)timenow,(long long unsigned int)filter->pipbuf->timestamp,(long long unsigned int)(timenow-filter->pipbuf->timestamp));
     gst_buffer_unref(filter->pipbuf);
     filter->pipbuf=NULL;
   }    
  gst_multivis_sweepobject(filter);
  //





  IplImage *process;
  process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,(filter->bpp/8));
  process->imageData=(char*)buf->data;//we only manipulate data about to be sent  
  //main latency
  avuint64app(filter->mainlatency,timenow-buf->timestamp);
  if (filter->lastbuffer!=G_MAXUINT64)avuint64app(filter->mainbetween,buf->timestamp-filter->lastbuffer);
  filter->lastbuffer=buf->timestamp;

  if (filter->siftbuf){
    struct feature* feat;
    feat=(struct feature*)filter->siftbuf->data;
    n=filter->siftbuf->size/sizeof(struct feature);
    if (!filter->silent)g_print("GSTMULTIVIS Running sift. Width=%d Height=%d , Features: %d\n",process->width,process->height,n);
    int w,h,cnt=0;
    for (i=0;i<n;i++){	
	if (filter->siftwidth!=-1)w=(feat[i].x*(double)process->width)/(double)filter->siftwidth;
  	else w=feat[i].x;
	if (filter->siftheight!=-1)h=(feat[i].y*(double)process->height)/(double)filter->siftheight;
  	else h=feat[i].y;
  	if ((w>0)&&(w<process->width)&&(h>0)&&(h<process->height)){
		cvCircle(process,cvPoint(w,h), 5, CV_RGB(255,255,0),3,8,0);
		cnt++;
	}
    }
    if (!filter->silent)g_print("GSTMULTIVIS printed %d features\n",cnt);
    avuint64app(filter->siftavg,timenow-filter->siftbuf->timestamp);
    sprintf(filter->text,"Sift Cur: %llu , Avg: %llu Lat: %llu , FPS: %4.2f",(long long unsigned int)((timenow-filter->siftbuf->timestamp)/1000000ULL),getavuint64(filter->siftavg)/1000000ULL,getavuint64(filter->siftlatency)/1000000ULL,1000000000.0/getavuint64(filter->siftbetween));
    if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0));

    ankery=ankery+(filter->fontsize*30);  
  }
  struct padset* thispadset=NULL;
  GList *templ=filter->padsets;
  while(templ){
    thispadset=(struct padset*)templ->data;
    if (!filter->silent)g_print("Multivis: Checking padset. Name: %s, mode %d\n",thispadset->name,thispadset->mode);
    if (!thispadset->name)thispadset->name=gst_element_get_name(gst_pad_get_parent_element(gst_pad_get_peer(thispadset->sinkpad))); 
    if (!thispadset->name)g_print("Multivis warning, could not get name for padset!!\n");   
    //dont do something for ubercoordinates mode=0 here as it is handeled in the next
    if ((thispadset->mode==0)){
      if (thispadset->name){
        sprintf(filter->text,"%s Lat: %llu , FPS: %4.2f",thispadset->name,(long long unsigned int)getavuint64(thispadset->latency)/1000000ULL,(1000000000.0/getavuint64(thispadset->between)));
        if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0));
        ankery=ankery+(filter->fontsize*30);
      }   
    }
    if ((thispadset->buffer)&&(thispadset->mode==1)){
      n=thispadset->buffer->size/sizeof(struct uber_coordinates);
      for( i = 0; i < n; i++ ){
        struct uber_coordinates* sh=(struct uber_coordinates*)(thispadset->buffer->data+i*(sizeof(struct uber_coordinates)));
	swapuber_coordinates(sh);
        if (sh->a22>0.0){
	if (!filter->silent)g_print("Object %s is valid\n",sh->name);
	int ix1,ix2,iy1,iy2;
	if (sh->imagewidth>0) ix1=((double)sh->x1*(double)process->width)/(double)sh->imagewidth;
   	else ix1=sh->x1;
	if (sh->imagewidth>0) ix2=((double)sh->x2*(double)process->width)/(double)sh->imagewidth;
   	else ix2=sh->x2;
	if (sh->imageheight>0) iy1=((double)sh->y1*(double)process->height)/(double)sh->imageheight;
   	else iy1=sh->y1;
	if (sh->imageheight>0) iy2=((double)sh->y2*(double)process->height)/(double)sh->imageheight;
   	else iy2=sh->y2;

        cvRectangle(process, cvPoint(ix1,iy1), cvPoint(ix2,iy2),CV_RGB ( 255, 255, 255 ) , 3,8,0);
	cvPutText (process, sh->name, cvPoint ((int)sh->x1,
          (int)sh->y2), &(filter->font), cvScalar (255,
          255, 255, 0));
	
	}
      }
      avuint64app(thispadset->avg,timenow-thispadset->buffer->timestamp);
      if (thispadset->name){
        sprintf(filter->text,"%s Cur: %llu , Avg: %llu Lat: %llu , FPS: %4.2f",thispadset->name,(long long unsigned int)(timenow-thispadset->buffer->timestamp)/1000000ULL,(long long unsigned int)getavuint64(thispadset->avg)/1000000ULL,(long long unsigned int)getavuint64(thispadset->latency)/1000000ULL,(1000000000.0/getavuint64(thispadset->between)));
        if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0));
        ankery=ankery+(filter->fontsize*30);
      }    
    }
    if ((thispadset->buffer)&&(thispadset->mode==2)){
      //g_print("Multivis: Checking  %s for Valid coordinates, size of buffer %lu. \n",thispadset->name,thispadset->buffer->size);
      n=thispadset->buffer->size/sizeof(struct coordinates);
      struct coordinates* temp;
      for( i = 0; i < n; i++ )
      {
      temp=&(((struct coordinates*)thispadset->buffer->data)[i]);
        swap_coordinates(temp);
	int ix1,ix2,iy1,iy2;
	if (thispadset->width>0) ix1=((double)temp->x1*(double)process->width)/(double)thispadset->width;
   	else ix1=temp->x1;
	if (thispadset->width>0) ix2=((double)temp->x2*(double)process->width)/(double)thispadset->width;
   	else ix2=temp->x2;
	if (thispadset->height>0) iy1=((double)temp->y1*(double)process->height)/(double)thispadset->height;
   	else iy1=temp->y1;
	if (thispadset->height>0) iy2=((double)temp->y2*(double)process->height)/(double)thispadset->height;
   	else iy2=temp->y2;
      cvRectangle(process, cvPoint(ix1,iy1), cvPoint(ix2,iy2),CV_RGB ( 255, 255, 255 ) , 3,8,0);
      if (filter->fontsize>1.1)cvPutText (process, thispadset->name, cvPoint (ix1,
          (iy2+(filter->fontsize*25))), &(filter->font), cvScalar (255,
          255, 255, 0));
	if (filter->graspx>-1){
	  if ((filter->graspx>ix1)&&(filter->graspx<ix2)&&(filter->graspy>iy1)&&(filter->graspy<iy2)){
	   if (gst_pad_is_linked(filter->srcgrasp)){
	     GstBuffer* outgrasp=gst_buffer_new_and_alloc(sizeof(struct grasp));
    	     struct grasp* thisgrasp=(struct grasp*)outgrasp->data;
	     gst_buffer_set_caps(outgrasp,gst_caps_copy(gst_pad_get_pad_template_caps(filter->srcgrasp)));
	     float kartx,karty,kartz;
	     float angx,angy;
	     float smallwidth=75.0;
             //get betterfromfilename
	     gint64 bfromname=-1;
             deserialize_int64 (&bfromname,thispadset->name);
	     if (bfromname>0)smallwidth=(float)bfromname;
	     //float neuebreitegrad=((float)(ix2-ix1)/(float)process->width)*filter->mainhfov;
	     //else kartz=(float)smallwidth/(float)(2.0*tan(neuebreitegrad*3.14159/360.0));
             float gb=(float)smallwidth*((float)process->width/(float)(ix2-ix1));
	     if (filter->fixz>-1)kartz=filter->fixz;
             else kartz=(gb/2.0)/tan(filter->mainhfov*(3.14159/360.0));
	     angx=((((ix1+ix2)/2.0)-(process->width/2.0))/((float)process->width))*filter->mainhfov;
	     kartx=tan((3.14159*angx)/180)*kartz;
	     angy=((((iy1+iy2)/2.0)-(process->height/2.0))/((float)process->width))*filter->mainhfov;
	     karty=tan((3.14159*angy)/180)*kartz;

	     thisgrasp->x=(int)kartx;
	     thisgrasp->y=(int)karty;
	     thisgrasp->z=(int)kartz;
    	     outgrasp->timestamp=buf->timestamp;
             outgrasp->offset=buf->offset;
	     if (!filter->silent)g_print("Multivis sending grasping Buffer!\n");
	     gst_pad_push(filter->srcgrasp,outgrasp);
	     }
	   filter->graspx=-1;
	   filter->graspy=-1;
	   }
	}
      } 
      //  
      avuint64app(thispadset->avg,timenow-thispadset->buffer->timestamp);
      if (thispadset->name){
        sprintf(filter->text,"%s Cur: %llu , Avg: %llu Lat: %llu , FPS: %4.2f",thispadset->name,(long long unsigned int)((timenow-thispadset->buffer->timestamp)/1000000ULL),getavuint64(thispadset->avg)/1000000ULL,getavuint64(thispadset->latency)/1000000ULL,1000000000.0/getavuint64(thispadset->between));
        if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0));
	ankery=ankery+(filter->fontsize*30);  
      }
    }
    templ=g_list_next(templ);
  }
  //now all the single objects
  if (g_list_length(filter->objectbuf)){
    if (!filter->silent)g_print("Multivis: Currently having %d Objects in our list.\n",g_list_length(filter->objectbuf));
    GList* temp=filter->objectbuf;
    while (temp){
      struct uber_coordinates* sh=(struct uber_coordinates*)((GstBuffer*)temp->data)->data;
      swapuber_coordinates(sh);
      avuint64app(filter->objavg,timenow-((GstBuffer*)temp->data)->timestamp);
      guint64 toage=timenow-((GstBuffer*)temp->data)->timestamp;//this object age
      if (sh->a22>0.0){
	if (!filter->silent)g_print("Object %s is valid\n",sh->name);

	int ix1,ix2,iy1,iy2;
	if (sh->imagewidth>0) ix1=((double)sh->x1*(double)process->width)/(double)sh->imagewidth;
   	else ix1=sh->x1;
	if (sh->imagewidth>0) ix2=((double)sh->x2*(double)process->width)/(double)sh->imagewidth;
   	else ix2=sh->x2;
	if (sh->imageheight>0) iy1=((double)sh->y1*(double)process->height)/(double)sh->imageheight;
   	else iy1=sh->y1;
	if (sh->imageheight>0) iy2=((double)sh->y2*(double)process->height)/(double)sh->imageheight;
   	else iy2=sh->y2;


        cvRectangle(process, cvPoint(ix1,iy1), cvPoint(ix2,iy2),CV_RGB ( 255, 255, 255 ) , 3,8,0); 	   sprintf(filter->text,"%s Age %llu",sh->name,(long long unsigned int)toage/1000000ULL);
	//cvRectangle(process
	if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint ((int)ix1,
          ((int)iy2+(filter->fontsize*25))), &(filter->font), cvScalar (255,
          255, 255, 0)); 
	//debug matrix
	/*
        sprintf(filter->text,"%.2f   %.2f   %.2f",sh->a00,sh->a01,sh->a02);
	cvPutText (process, filter->text, cvPoint ((int)sh->x1,
          (int)sh->y2+filter->fontsize*55), &(filter->font), cvScalar (255,
          255, 255, 0));

        sprintf(filter->text,"%.2f   %.2f   %.2f",sh->a10,sh->a11,sh->a12);
	cvPutText (process, filter->text, cvPoint ((int)sh->x1,
          (int)sh->y2+filter->fontsize*80), &(filter->font), cvScalar (255,
          255, 255, 0));

        sprintf(filter->text,"%.2f   %.2f   %.2f",sh->a20,sh->a21,sh->a22);
	cvPutText (process, filter->text, cvPoint ((int)sh->x1,
          (int)sh->y2+filter->fontsize*105), &(filter->font), cvScalar (255,
          255, 255, 0));
          */
	//calculate distance: 
	
	float kartx,karty,kartz;
	float angx,angy;
	float smallwidth=75;
        //get betterfromfilename
	gint64 bfromname=-1;
        deserialize_int64 (&bfromname,sh->name);
	if (bfromname>0)smallwidth=(float)bfromname;
	//mainhfov
	//neue Breite errechnen:
	float neuebreite=((float)sh->mwidth)*((float)sh->scale);
        
	//float neuebreitegrad=(neuebreite/(float)sh->imagewidth)*filter->mainhfov;
        //g_print("neuebreite %.2f, X1=%.2f X2=%.2f Y1=%.2f Y2=%.2f smallwidth=%lld neuebreitegrad=%.2f\n",neuebreite,sh->x1,sh->x2,sh->y1,sh->y2,(long long int)bfromname,neuebreitegrad);
	//if (filter->fixz>-1)kartz=filter->fixz;
	//else 
        //this was bs
        //kartz=(float)smallwidth/(float)(2.0*tan((neuebreitegrad*3.14159)/360.0));
        float gb=(float)smallwidth*(sh->imagewidth/neuebreite);
        kartz=(gb/2.0)/tan(filter->mainhfov*(3.14159/360.0));
	angx=((((sh->x1+sh->x2)/2.0)-(sh->imagewidth/2.0))/((float)sh->imagewidth))*filter->mainhfov;
	kartx=tan((3.14159*angx)/180)*kartz;
	angy=((((sh->y1+sh->y2)/2.0)-(sh->imageheight/2.0))/((float)sh->imagewidth))*filter->mainhfov;
	karty=tan((3.14159*angy)/180)*kartz;
        int yaw,tilt,roll;
	yaw=(180.0/3.14)*atan2(sh->a10,sh->a00);
	//tilt=(180.0/3.14)*atan2(-sh->a20,sqrt(((sh->a10)*(sh->a10))+((sh->a00)*(sh->a00))));
	//tilt=(180.0/3.14)*atan2(sh->a00,sqrt(((sh->a10)*(sh->a10))+((sh->a00)*(sh->a00))));
	//tilt=(180.0/3.14)*atan2(sh->a00,sh->a11);
	//roll=(180.0/3.14)*atan2(sh->a21,sh->a22);
	//mwidth 
	//imagewidth
	//
        //sprintf(filter->text,"Scale   %.2f Ab %.2f Nbg %.2f Nb %.2f Z %.2f cm",(float)sh->scale,(float)sh->mwidth,(float)neuebreitegrad,(float)neuebreite,(float)kartz/10.0);
        sprintf(filter->text,"X %.2f Y %.2f Z %.2f cm, yaw %d tilt %d pan %d degrees scale=%.2f",(float)kartx/10.0,(float)karty/10.0,(float)kartz/10.0,yaw,tilt,roll,sh->scale);
	if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (ix1,(iy2+filter->fontsize*55)), &(filter->font), cvScalar (255,
          255, 255, 0));
	if ((filter->alwaysgrasp)||(filter->graspx>-1)){
	  if ((filter->alwaysgrasp)||((filter->graspx>ix1)&&(filter->graspx<ix2)&&(filter->graspy>iy1)&&(filter->graspy<iy2))){
	   if (gst_pad_is_linked(filter->srcgrasp)){
	     GstBuffer* outgrasp=gst_buffer_new_and_alloc(sizeof(struct grasp));
    	     struct grasp* thisgrasp=(struct grasp*)outgrasp->data;
	     gst_buffer_set_caps(outgrasp,gst_caps_copy(gst_pad_get_pad_template_caps(filter->srcgrasp)));
	     thisgrasp->x=(int)kartx;
	     thisgrasp->y=(int)karty;
	     thisgrasp->z=(int)kartz;
	     thisgrasp->yaw=yaw;
	     thisgrasp->tilt=tilt;
	     thisgrasp->pan=roll;
    	     outgrasp->timestamp=buf->timestamp;
             outgrasp->offset=buf->offset;
	     if (!filter->silent)g_print("Multivis sending grasping Buffer!\n");
	     gst_pad_push(filter->srcgrasp,outgrasp);
	     }
	   filter->graspx=-1;
	   filter->graspy=-1;
	   }
	}	
	//rox=
	//(cvmGet(&invA,0,0)*((gdouble)current->width)+cvmGet(&invA,0,1)*((float)current->height)+cvmGet(&invA,0,2))/bufscale;
      }
      temp=g_list_next(temp);
    }
  }




  //now the ROI:

  if (!filter->alwaysdisplay&&filter->pipactive)cvRectangle(process, cvPoint(filter->seloriginx,filter->seloriginy),cvPoint(filter->seloriginx+filter->selwidth,filter->seloriginy+filter->selheight),CV_RGB ( 255, 255, 0 ) , 2,8,0); //do rectangle everytime
  if (!filter->alwaysdisplay&&filter->pipactive&&filter->pipbuf){
    int ax1=filter->seloriginx,ay1=filter->seloriginy,ax2=filter->seloriginx+filter->selwidth,ay2=filter->seloriginy+filter->selheight;
    IplImage *pipimage;
    pipimage=cvCreateImageHeader(filter->pipimagesize, IPL_DEPTH_8U,(filter->bpp/8));
    if (filter->pipbuf&&!filter->pipoverlay&&filter->resendmap){
      filter->acwidth=filter->ovwidth;
      filter->acheight=filter->ovheight;
      float *outmap;
      if (filter->currentconf) RequestRegion(pipimage,process, &filter->acwidth,&filter->acheight,&ax1,&ay1,&ax2,&ay2,filter->currentconf->hfov,filter->mainhfov,filter->currentconf->yaw,filter->currentconf->pitch,NULL,&filter->currentlyorderedhfov,1,&outmap);
      else RequestRegion(pipimage,process, &filter->acwidth,&filter->acheight,&ax1,&ay1,&ax2,&ay2,filter->mainhfov,filter->mainhfov,0,0,NULL,&filter->currentlyorderedhfov,1,&outmap);
      GstBuffer* bufmap=gst_buffer_new();
      gst_buffer_set_data(bufmap,(unsigned char*) outmap,(2*sizeof(float)*process->width*process->height));
      GST_BUFFER_MALLOCDATA(bufmap)= (unsigned char*)outmap;
      //gst_buffer_set_caps(bufmap,gst_caps_copy(gst_pad_get_pad_template_caps(filter->srcmap)));
      //GstStructure *structure=gst_caps_get_structure(bufmap->caps,0);
      //if (!structure) g_print("Multivis: Scheisse!\n");
      //gst_structure_fixate_field_nearest_int (structure,"width",process->width);
      //gst_structure_fixate_field_nearest_int (structure,"height",process->height);
      //gst_structure_set(structure,"height",G_TYPE_INT,process->height,NULL);//set size of the map
      //gst_structure_set(structure,"width",G_TYPE_INT,process->width,NULL);
      gst_buffer_set_caps(bufmap,gst_caps_new_simple ("map/float",
      		"width", G_TYPE_INT, process->width,
      		"height", G_TYPE_INT, process->height,
		NULL));
      bufmap->timestamp=buf->timestamp;
      bufmap->offset=buf->offset;
      filter->resendmap=FALSE;
      gst_pad_push(filter->srcmap,bufmap);
    }
    //or now overlay the image...
    if (filter->pipbuf&&filter->pipoverlay){
      filter->acwidth=filter->ovwidth;
      filter->acheight=filter->ovheight;
    pipimage->imageData=(char*)filter->pipbuf->data;//we only manipulate data about to be sent
    IplImage* middle; 
    if (filter->currentconf) middle=RequestRegion(pipimage,process, &filter->acwidth,&filter->acheight,&ax1,&ay1,&ax2,&ay2,filter->currentconf->hfov,filter->mainhfov,filter->currentconf->yaw,filter->currentconf->pitch,NULL,&filter->currentlyorderedhfov,0,NULL);
      else middle=RequestRegion(pipimage,process, &filter->acwidth,&filter->acheight,&ax1,&ay1,&ax2,&ay2,filter->mainhfov,filter->mainhfov,0,0,NULL,&filter->currentlyorderedhfov,0,NULL);
      if (middle){
        //new code
      int h;
      int xstart=5;
      int ystart=process->height-middle->height-5;
      //if (filter->seloriginx<filter->ovoriginx)xstart=process->width-filter->acwidth-5;
      if (filter->seloriginx<filter->ovoriginx)xstart=filter->ovoriginx+5;
      else xstart=filter->seloriginx-filter->acwidth-5;
      if (xstart<0)xstart=5;
      for (h=ystart;h<(ystart+middle->height);h++){
        memcpy(process->imageData+(filter->bpp/8)*(h*process->width+xstart),middle->imageData+(filter->bpp/8)*((h-ystart)*middle->width),(middle->width*(filter->bpp/8)));
        }      

        cvReleaseImage(&middle);
      }
    }
    cvReleaseImageHeader(&pipimage);
    //paint the boxes
    if ((filter->acwidth>0)&&(filter->acheight>0)){
      if ((filter->seloriginx!=ax1)||(filter->seloriginy!=ay1)||
	((filter->seloriginx+filter->selwidth)!=ax2)||
	((filter->seloriginy+filter->selheight)!=ay2)){
	  if (!filter->silent)g_print("Aberation of ROI desired: x:%d, y:%d, width:%d,height:%d, and overlay frame will be displayed to: x:%d, y:%d, width:%d, height:%d, Sent: %d Received %d\n",
	filter->seloriginx,filter->seloriginy,filter->selwidth,filter->selheight,
	ax1,ay1,ax2-ax1,ay2-ay1,filter->numcsent,filter->numcrecv);
	cvRectangle(process, cvPoint(ax1,ay1),cvPoint(ax2,ay2),CV_RGB ( 255, 0, 0 ) , 2,8,0);
	}
        int xstart=5;
        int ystart=process->height-filter->acheight-5;	
	if (!filter->silent)g_print("Acwidth=%d Acheight=%d\n",filter->acwidth,filter->acheight);
      if (filter->seloriginx<filter->ovoriginx){//overlay is right of selection
	//xstart=process->width-filter->acwidth-5;
        xstart=filter->ovoriginx+5;
        cvLine(process, cvPoint(ax2,ay1), cvPoint (xstart,ystart), cvScalar (255,255, 0, 0), 1, 8, 0);
        cvLine(process, cvPoint(ax2,ay2), cvPoint (xstart,process->height-5), cvScalar (255,255, 0, 0), 1, 8, 0);
      }
      else{
        xstart=filter->seloriginx-filter->acwidth-5;
        if (xstart<0)xstart=5;
        cvLine(process, cvPoint(filter->seloriginx-5,ystart), cvPoint (ax1,ay1), cvScalar (255,255, 0, 0), 1, 8, 0);
        cvLine(process, cvPoint(filter->seloriginx-5,process->height-5), cvPoint (ax1,ay2), cvScalar (255,255, 0, 0), 1, 8, 0);
      }
      cvRectangle(process, cvPoint(xstart,ystart),cvPoint(xstart+filter->acwidth,ystart+filter->acheight),CV_RGB ( 255, 255, 0 ) , 2,8,0); 
        if (!filter->silent)g_print("I will now overlay the ROI with width %d and height %d buffer\n",filter->acwidth,filter->acheight);
    }
    float yaw=(((2.0*(float)filter->seloriginx)+(float)filter->selwidth)/(2.0*(float)filter->imagesize.width))*filter->mainhfov-(filter->mainhfov/2.0);
    float pitch=(0.5-((2.0*(float)filter->seloriginy+(float)filter->selheight)/(2.0*(float)filter->imagesize.height)))*((filter->mainhfov*(float)filter->imagesize.height)/(float)filter->imagesize.width);
    float actualyaw=-1;
    float actualpitch=-1;
    int limit=0;
    if (filter->currentconf){
      actualyaw=filter->currentconf->yaw;
      actualpitch=filter->currentconf->pitch;
      limit=filter->currentconf->limit;
    }
    if (limit==1) sprintf(filter->text,"PTZ Limit: %llu Orderedyaw : %4.1f OrderedPitch %4.1f, Actualyaw %4.1f, ActualPitch %4.1f",(timenow-filter->pipbuf->timestamp)/1000000ULL,yaw,pitch,actualyaw,actualpitch); 
    else sprintf(filter->text,"Pip Cur: %llu Orderedyaw : %4.1f OrderedPitch %4.1f, Actualyaw %4.1f, ActualPitch %4.1f",(timenow>filter->pipbuf->timestamp) ? ((timenow-filter->pipbuf->timestamp)/1000000ULL): 0,yaw,pitch,actualyaw,actualpitch); 
    if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0));  
    ankery=ankery+(filter->fontsize*30);  
  }
//
  if (filter->yawoffset){
    GstBuffer* yawbuf=gst_buffer_new_and_alloc(buf->size);
    gst_buffer_set_caps(yawbuf,buf->caps);
    yawbuf->timestamp=buf->timestamp;
    yawbuf->offset=buf->offset;
    yawbuf->offset_end=buf->offset_end;
    yawbuf->duration=buf->duration;
    int h;
    for (h=0;h<filter->imagesize.height;h++){
      memcpy((yawbuf->data+(h*4*filter->imagesize.width)),(buf->data+(h*4*filter->imagesize.width)+(4*(filter->imagesize.width-filter->yawoffset))),4*filter->yawoffset);
      memcpy((yawbuf->data+(h*4*filter->imagesize.width)+(4*filter->yawoffset)),(buf->data+(h*4*filter->imagesize.width)),(4*(filter->imagesize.width-filter->yawoffset)));
    }
    gst_buffer_unref(buf);
    buf=yawbuf;
    process->imageData=(char*)buf->data;
  }


  if (filter->alwaysdisplay){
    if(!filter->pipbuf){
      if (filter->fontsize>1.1)cvPutText (process, "No PTZ image available", cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0));
      ankery=ankery+(filter->fontsize*30); 
    }
    else{
      IplImage *pipimage;
      pipimage=cvCreateImageHeader(filter->pipimagesize, IPL_DEPTH_8U,(filter->bpp/8));
      pipimage->imageData=(char*)filter->pipbuf->data;
      int asx2,asy2;
      asx2=filter->asx+(filter->asscale*filter->pipimagesize.width);
      if (asx2>process->width){
	filter->asscale=(float)(process->width-filter->asx)/(float)filter->pipimagesize.width;
        asx2=filter->asx+(filter->asscale*filter->pipimagesize.width);
      }
      asy2=filter->asy+(filter->asscale*filter->pipimagesize.height);
      if (asy2>process->height){
	filter->asscale=(float)(process->height-filter->asy)/(float)filter->pipimagesize.height;
        asx2=filter->asx+(filter->asscale*filter->pipimagesize.width);
      }
      cvRectangle(process, cvPoint(filter->asx,filter->asy), cvPoint(asx2,asy2),CV_RGB ( 255, 255, 0 ) , 3,8,0); 
      if (filter->currentconf){
	//draw desired position
	
      }
      if (filter->pipoverlay){
        int outwidth=filter->asscale*pipimage->width;
        int outheight=filter->asscale*pipimage->height;
   	//g_print("Now doing alwaysoverlay %d %d\n",outwidth,outheight);
	IplImage* middle;
	gboolean tofree=FALSE;
        if ((filter->asscale>0.99)&&(filter->asscale<1.01))middle=pipimage;
        else{
          middle=cvCreateImage(cvSize(outwidth,outheight), pipimage->depth,pipimage->nChannels);
          cvResize(pipimage, middle, CV_INTER_LINEAR );
	  tofree=TRUE;
        }
	//copy the image
	int h;
        for (h=filter->asy;((h<(filter->asy+middle->height))&&(h<filter->imagesize.height));h++){
          memcpy(process->imageData+(filter->bpp/8)*(h*process->width+filter->asx),middle->imageData+(filter->bpp/8)*((h-filter->asy)*middle->width),(middle->width*(filter->bpp/8)));
        }
	if (tofree)cvReleaseImage(&middle);
      }
      else{
	//TODO, set mapping parameters to 
	}
      cvReleaseImageHeader(&pipimage);
    }
    //calculate the "actual position"
  }


  //

  sprintf(filter->text,"Main Cur: %llu Lat: %llu , FPS: %4.2f",(timenow-buf->timestamp)/1000000ULL,getavuint64(filter->mainlatency)/1000000ULL,1000000000.0/getavuint64(filter->mainbetween)); 
  if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0));  
  ankery=ankery+(filter->fontsize*30);    
  if (filter->objectbuf!=NULL){
  sprintf(filter->text,"Single Objects, Average Age: %llu Lat: %llu , FPS: %4.2f",getavuint64(filter->objavg)/1000000ULL,getavuint64(filter->objlatency)/1000000ULL,1000000000.0/getavuint64(filter->objbetween)); 
  if (filter->fontsize>1.1)cvPutText (process, filter->text, cvPoint (0,
          ankery), &(filter->font), cvScalar (255,
          255, 255, 0)); 

  }
//eventually send camconfig
  if ((filter->docc)&&(gst_pad_is_linked(filter->srccamconfig))){
    if (!filter->silent)g_print("Multivis sending cable unwinding buffer to PTUCONTROL\n");
    GstBuffer* outcamc=gst_buffer_new_and_alloc(sizeof(struct camconfig));
    struct camconfig* thisconfig=(struct camconfig*)outcamc->data;
    thisconfig->active=filter->pipactive||filter->alwaysdisplay;//wether the camera/PTU should be set to idle
    thisconfig->relative=TRUE;
    thisconfig->yaw=360.0;//turn left or right, 0 to 360 degrees
    thisconfig->pitch=0.0;//up and down, depending on input image, but inverse to the order of rows in the image
    thisconfig->roll=0.0;
    if(!thisconfig->active){
      thisconfig->yaw=0;
      thisconfig->pitch=0;
    }
    thisconfig->hastranslation=FALSE;
    thisconfig->x=0;//for future use
    thisconfig->y=0;//for future use
    thisconfig->z=0;//for future use
    thisconfig->shutter=-1;//shutter speed in ms, <0 is AUTO
    thisconfig->iris=-1; 
    thisconfig->focus=-1;
    swapcamconfig(thisconfig);
    gst_buffer_set_caps(outcamc,gst_caps_copy(gst_pad_get_pad_template_caps(filter->srccamconfig)));
    outcamc->timestamp=buf->timestamp;
    outcamc->offset=buf->offset;
    filter->docc=FALSE;
    gst_pad_push (filter->srccamconfig, outcamc);
  }
  if ((filter->resendcamconfig)&&(gst_pad_is_linked(filter->srccamconfig))){
    GstBuffer* outcamc=gst_buffer_new_and_alloc(sizeof(struct camconfig));
    struct camconfig* thisconfig=(struct camconfig*)outcamc->data;
    thisconfig->active=filter->pipactive||filter->alwaysdisplay;//wether the camera/PTU should be set to idle
    thisconfig->relative=FALSE;//we just drink absolut
    thisconfig->hfov=filter->currentlyorderedhfov;//ordered field of view determinated before
    if (thisconfig->hfov==0)thisconfig->hfov=filter->mainhfov;
    thisconfig->vfov=-1;   
    //g_print("Analyzing seloriginx Analysing term ((2*filter->seloriginx)+filter->selwidth): %f!\n",((2*filter->seloriginx)+filter->selwidth));
    thisconfig->yaw=filter->orderedyaw;//turn left or right, 0 to 360 degrees
    thisconfig->pitch=filter->orderedpitch;//up and down, depending on input image, but inverse to the order of rows in the image
    thisconfig->roll=0.0;
    if(!thisconfig->active){
      thisconfig->yaw=0;
      thisconfig->pitch=0;
    }
    thisconfig->hastranslation=FALSE;
    thisconfig->x=0;//for future use
    thisconfig->y=0;//for future use
    thisconfig->z=0;//for future use
    thisconfig->shutter=-1;//shutter speed in ms, <0 is AUTO
    thisconfig->iris=-1; 
    if (filter->manualfocus)thisconfig->focus=filter->orderedfocus;
    else thisconfig->focus=-1;
    swapcamconfig(thisconfig);
    gst_buffer_set_caps(outcamc,gst_caps_copy(gst_pad_get_pad_template_caps(filter->srccamconfig)));
    outcamc->timestamp=buf->timestamp;
    outcamc->offset=buf->offset;
    filter->resendcamconfig=FALSE;
    filter->numcsent++;
    if (!filter->silent)g_print("Multivis sending buffer to PTUCONTROL yaw %f pitch %f hfov % focus %f\n",thisconfig->yaw,thisconfig->pitch,thisconfig->hfov,thisconfig->focus);
    gst_pad_push (filter->srccamconfig, outcamc);
  }
  if ((filter->snapshotnext)&&(gst_pad_is_linked(filter->srcsnapshot))){
    if (gst_pad_push (filter->srcsnapshot, gst_buffer_copy(buf))!=GST_FLOW_OK){
      g_print("Multivis Warning, pushing snapshot went wrong!\n");
    }
    filter->snapshotnext=FALSE;
  }
  
  if ((filter->selectionnext)&&(filter->pipactive)){
    IplImage* selection=NULL;

    if (filter->pipbuf){
      int ax1=filter->seloriginx,ay1=filter->seloriginy,ax2=filter->seloriginx+filter->selwidth,ay2=filter->seloriginy+filter->selheight;
      IplImage *pipimage;
      pipimage=cvCreateImageHeader(filter->pipimagesize, IPL_DEPTH_8U,(filter->bpp/8));
      pipimage->imageData=(char*)filter->pipbuf->data;//we only manipulate data about to be sent
      int acwidth=G_MAXINT;
      int acheight=G_MAXINT;
      if (filter->currentconf) selection=RequestRegion(pipimage,process, &acwidth,&acheight,&ax1,&ay1,&ax2,&ay2,filter->currentconf->hfov,filter->mainhfov,filter->currentconf->yaw,filter->currentconf->pitch,NULL,&filter->currentlyorderedhfov,0,NULL);
      else selection=RequestRegion(pipimage,process,&acwidth,&acheight,&ax1,&ay1,&ax2,&ay2,filter->mainhfov,filter->mainhfov,0,0,NULL,&filter->currentlyorderedhfov,0,NULL);
      if (selection){
	if ((filter->seloriginx!=ax1)||(filter->seloriginy!=ay1)||
	((filter->seloriginx+filter->selwidth)!=ax2)||
	((filter->seloriginy+filter->selheight)!=ay2)){
	//does not match, throw away
	  cvReleaseImage(&selection);
	  selection=NULL;
	}
      }
    }
    if (!selection){
      if (!filter->silent)g_print("Taking selection from ROI!\n");
      int ax1=filter->seloriginx,ay1=filter->seloriginy,ax2=filter->seloriginx+filter->selwidth,ay2=filter->seloriginy+filter->selheight;
      cvSetImageROI(process, cvRect(ax1,ay1,(ax2-ax1),(ay2-ay1)));
      selection=cvCreateImage(cvSize((ax2-ax1),(ay2-ay1)),
                               process->depth,
                               process->nChannels); 
      cvCopy(process,selection , NULL); 
    }
  GstCaps *selcaps=gst_caps_copy(gst_buffer_get_caps(buf));
  GstStructure *structure=gst_caps_get_structure(selcaps,0);
  if (!structure) g_print("Multivis: Scheisse!\n");
  gst_structure_set(structure,"height",G_TYPE_INT,selection->height,NULL);
  gst_structure_set(structure,"width",G_TYPE_INT,selection->width,NULL);
  GstBuffer* bufselection=gst_buffer_new_and_alloc(selection->imageSize);
  gst_buffer_set_caps(bufselection,selcaps);
  memcpy(bufselection->data,selection->imageData,selection->imageSize);
  gst_caps_unref(selcaps);
  cvReleaseImage(&selection);  
  filter->selectionnext=FALSE;
  if (gst_pad_push (filter->srcselection, bufselection)!=GST_FLOW_OK){
      if (!filter->silent)g_print("Multivis Warning, pushing selection went wrong!\n");
    }
  }
  

  //g_print("Multivis end of cain, unlocking mutex!\n");
  
  //g_print("Multivis end of cain, unlocked mutex, push buffer soon!\n");
  gst_buffer_set_caps(buf,gst_caps_new_simple ("video/x-raw-rgb",
  "width", G_TYPE_INT, process->width,
  "height", G_TYPE_INT, process->height,
  "bpp",G_TYPE_INT,32, 
  "depth",G_TYPE_INT,24, 
  "endianness",G_TYPE_INT,4321,
  "red_mask",G_TYPE_INT,65280, 
  "green_mask",G_TYPE_INT,16711680, 
  "blue_mask",G_TYPE_INT,-16777216,
  "framerate", GST_TYPE_FRACTION, 18,1,
  NULL));
  cvReleaseImageHeader(&process);
  filter->graspx=-1;//reset grasp order
  filter->graspy=-1;

  g_mutex_unlock(filter->mutex);
  //gst_structure_remove_field(gst_caps_get_structure(buf->caps,0),"pixel-aspect-ratio");


  return gst_pad_push (filter->srcpad, buf);
}


static void create_region(Gstmultivis * filter,gint64 x, gint64 y){
  if (!filter->silent)g_print("Creating Region from x %d y %d to x2 %lld y2 %lld.\n",filter->seloriginx,filter->seloriginy,(long long int)x,(long long int)y);
  //if (filter->pipimagesize.height==0){
  //  g_print("Multivis Creating Region failed, because PIPpad not initialized!\n");
  //  return;
  //}
  int x1,y1,x2,y2;
//order variables and check size
  if (x>filter->seloriginx){
	x1=filter->seloriginx;
	x2=x;
  }
  else{
	x2=filter->seloriginx;
	x1=x;	
  }
  if (y>filter->seloriginy){
	y1=filter->seloriginy;
	y2=y;
  }
  else{
	y2=filter->seloriginy;
	y1=y;	
  }
  x2+= (1+x2-x1)%4;
  y2+= (1+y2-y1)%4;
  if (x1<0) x1=0; else if ((filter->imagesize.width>0)&&(filter->imagesize.width<x1)) x1=filter->imagesize.width-1;
  if (x2<0) x2=0; else if ((filter->imagesize.width>0)&&(filter->imagesize.width<x2)) x2=filter->imagesize.width-1;
  if (y1<0) y1=0; else if ((filter->imagesize.height>0)&&(filter->imagesize.height<y1)) y1=filter->imagesize.height-1;
  if (y2<0) y2=0; else if ((filter->imagesize.height>0)&&(filter->imagesize.height<y2)) y2=filter->imagesize.height-1;
  x2-= (1+x2-x1)%4;
  y2-= (1+y2-y1)%4;

  filter->seloriginx=x1;
  filter->selwidth=x2-x1;
  filter->seloriginy=y1; 
  filter->selheight=y2-y1; 
  //if ((filter->selwidth<filter->minbox)||(filter->selheight<filter->minbox)||((filter->selwidth*2)>(filter->imagesize.width))||((filter->selheight*2)>(filter->imagesize.height))){
  if ((filter->selwidth<filter->minbox)||(filter->selheight<filter->minbox)){
    if (!filter->silent)g_print("Multivis Creating Region failed!\n");
    //TODO activate suggestion, invalid region choice, as a zoom to something nearly as big as the image is not usefull
    return;
  }
  //struct coordinates* tempcoord;
  filter->pipactive=TRUE;
  //region for blending in detection
  int spaceleft=filter->seloriginx-10;
  int spaceright=filter->imagesize.width-x2-10;
  if (spaceleft>spaceright){
	filter->ovwidth=spaceleft;
	filter->ovoriginx=5;
  }
  else{
	filter->ovwidth=spaceright;
	//filter->ovoriginx=x2+5;
        filter->ovoriginx=filter->imagesize.width-spaceright-5;
  }
  filter->ovheight=filter->imagesize.height-10;

  if (!filter->silent)g_print("Multivis acticating PTUCONTROL camconfig!\n");
  filter->resendcamconfig=TRUE;
  filter->resendmap=TRUE;
  filter->orderedyaw=(((2.0*(float)filter->seloriginx)+(float)filter->selwidth)/(2.0*(float)filter->imagesize.width))*filter->mainhfov-(filter->mainhfov/2.0);//turn left or right, -180 to 180 degrees
 filter->orderedpitch=(0.5-((2.0*(float)filter->seloriginy+(float)filter->selheight)/(2.0*(float)filter->imagesize.height)))*((filter->mainhfov*(float)filter->imagesize.height)/(float)filter->imagesize.width);
  return;
}

static void execute_drag(Gstmultivis * filter,gint64 x, gint64 y){
  //determine coordinates
  int asx2,asy2;
  asx2=filter->asx+(filter->asscale*filter->pipimagesize.width);
  asy2=filter->asy+(filter->asscale*filter->pipimagesize.height);
  if ((filter->alwaysdisplay)&&(filter->seloriginx_raw>filter->asx)&&(filter->seloriginy_raw>filter->asy)&&(filter->seloriginx_raw<asx2)&&(filter->seloriginy_raw<asy2)){

    if ((filter->seloriginx_raw+10)>asx2){
      filter->asscale=filter->asscale*((float)((filter->relposx-filter->seloriginx_raw)+(filter->pipimagesize.width*filter->asscale))/(float)(filter->pipimagesize.width*filter->asscale));
      float minscale=50.0/(float)(filter->pipimagesize.width);
      if (filter->asscale<minscale)filter->asscale=minscale;
    }
    else{
      filter->asx=filter->asx+(filter->relposx-filter->seloriginx_raw);
      if (filter->asx>(filter->imagesize.width-50))filter->asx=filter->imagesize.width-50;
      if (filter->asx<0)filter->asx=0;
      filter->asy=filter->asy+(y-filter->seloriginy_tmp);
      if (filter->asy>(filter->imagesize.height-50))filter->asy=filter->imagesize.height-50;
      if (filter->asy<0)filter->asy=0;
    }
  }
  else{
    filter->yawoffset=((filter->yawoffset+(int)(x-filter->seloriginx_tmp)));
    while (filter->yawoffset<0)filter->yawoffset+=(filter->imagesize.width);
    filter->yawoffset=filter->yawoffset%(filter->imagesize.width);
  }
}

static gboolean
gst_multivis_event (GstPad  *pad, GstEvent * event)
{
  Gstmultivis *filter;
  filter = GST_MULTIVIS (gst_pad_get_parent (pad));
  gboolean ret=TRUE;
  int button=0;
  GstPad *otherpad;
  if ((pad==filter->srcpad)||(pad==filter->srcselection)||(pad==filter->srcsnapshot)||(pad==filter->srcmap))otherpad=filter->sinkpad;
  else otherpad=filter->srcpad;
  //if (pad==filter->srcpad) return gst_pad_push_event (filter->sinkpad, event);
  switch (GST_EVENT_TYPE (event)) {
    //case GST_EVENT_EOS:
      //if (pad == filter->sinkpad) { filter->stopped1=TRUE; g_print("MULTIVIS Event EOS from normal sinkpad!\n");}
      //if (pad == filter->sinksift) { filter->stopped2=TRUE;g_print("MULTIVIS  Event EOS from position sinkpad!\n");}
      //if (0) ret = gst_pad_push_event (filter->srcpad, event);
      //g_print("PIP Event EOS!\n");
     // break;
    case GST_EVENT_NAVIGATION:
      //g_print("Multivis: Event Navigation! Event named %s, Structure Named %s, Content %s\n",gst_event_type_get_name(GST_EVENT_TYPE(event)),gst_structure_get_name(  gst_event_get_structure(event)),gst_structure_to_string(gst_event_get_structure(event)));
      //g_print("Eventstring: %s",g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")));
      if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"mouse-button-press")){
        button=g_value_get_int(gst_structure_get_value(gst_event_get_structure(event),"button"));
        if (button==1){
	  if (!filter->drag)filter->pipactive=FALSE;//disable until finished
	  filter->seloriginx_raw=filter->seloriginx_tmp=(int)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_x"));
	  filter->seloriginx_tmp=filter->seloriginx_tmp-filter->yawoffset;
	  if (filter->seloriginx_tmp < 0)filter->seloriginx_tmp+=filter->imagesize.width;
	  filter->seloriginy_raw=filter->seloriginy_tmp=(int)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y"));
          if (!filter->drag){
	    filter->seloriginx=filter->seloriginx_tmp;
	    filter->seloriginy=filter->seloriginy_tmp;
	  }
	  if (filter->graspmode){
	    filter->graspmode=FALSE;
	    g_mutex_lock(filter->mutex);
	    filter->graspx=filter->seloriginx;
	    filter->graspy=filter->seloriginy;
	    g_mutex_unlock(filter->mutex);
	  }
	}
	if (button==3){
	  filter->pipactive=FALSE;//disable until finished
	  //g_print("Multivis: Activating resendcamconfig for PTUCONTROL\n");
	  filter->resendcamconfig=TRUE;
	}
      }
      if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"mouse-button-release")){
        button=g_value_get_int(gst_structure_get_value(gst_event_get_structure(event),"button"));
	if (button==1){
          g_mutex_lock(filter->mutex);
	  gint64 releaseposx=g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_x"));
          filter->relposx=releaseposx;
	  releaseposx=releaseposx-filter->yawoffset;
	  if (releaseposx < 0)releaseposx+=filter->imagesize.width;
	  if (!filter->drag)create_region(filter,(int)releaseposx,(int)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y")));
	  else execute_drag(filter,(int)releaseposx,(int)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y")));
	  g_mutex_unlock(filter->mutex);
	}
      }
      if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"key-press")){
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"s")){
	  //g_print("Enabling snapshot for next frame\n");
	  filter->snapshotnext=TRUE;			  
	}
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"o")){
	  //g_print("Enabling add to object database next frame\n");
	  filter->selectionnext=TRUE;	  
	}
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"g")){
	  //g_print("Enabling add to object database next frame\n");
	  filter->graspmode=TRUE;	  
	}
      }
      ret = gst_pad_push_event (otherpad, event);
      break;
    default:
      ret = gst_pad_push_event (otherpad, event);
      break;
  }
  
  gst_object_unref (filter);
  return ret;
}

static GstStateChangeReturn
gst_multivis_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  //Gstmultivis *filter =  GST_MULTIVIS(element);
  struct timeval now;
  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
	//g_print ("Start Playing now!!");
	
  	gettimeofday(&now, NULL);
      	//filter->starttime=((now.tv_sec*1000000000ULL)+(now.tv_usec*1000ULL));
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      break;
    default:
      break;
  }
 return ret;
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
multivis_init (GstPlugin * multivis)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template multivis' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_multivis_debug, "multivis",
      0, "Template multivis");

  return gst_element_register (multivis, "multivis", GST_RANK_NONE,
      GST_TYPE_MULTIVIS);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstmultivis"
#endif

/* gstreamer looks for this structure to register multiviss
 *
 * exchange the string 'Template multivis' with your multivis description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "multivis",
    "Template multivis",
    multivis_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
