/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2009 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-onedlocalizer
 *
 * FIXME:Describe onedlocalizer here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! onedlocalizer ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <unistd.h>
#include <math.h>

#include <gst/gst.h>
#include <string.h>
#include "gstonedlocalizer.h"

GST_DEBUG_CATEGORY_STATIC (gst_onedlocalizer_debug);
#define GST_CAT_DEFAULT gst_onedlocalizer_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_SCALE,
  PROP_DEV,
  PROP_OPW,
  PROP_MT
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ]")
    );

static GstStaticPadTemplate src_factory_pos = GST_STATIC_PAD_TEMPLATE ("srcpos",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ]")
    );

static GstStaticPadTemplate src_factory_res = GST_STATIC_PAD_TEMPLATE ("srcres",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ]")
    );

static GstStaticPadTemplate src_factory_asc = GST_STATIC_PAD_TEMPLATE ("srcasc",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("text")
    );


GST_BOILERPLATE (Gstonedlocalizer, gst_onedlocalizer, GstElement,
    GST_TYPE_ELEMENT);

static void gst_onedlocalizer_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_onedlocalizer_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps *
gst_onedlocalizer_get_caps (GstPad *pad);
static gboolean gst_onedlocalizer_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_onedlocalizer_chain (GstPad * pad, GstBuffer * buf);
static gboolean
gst_onedlocalizer_event (GstPad  *pad, GstEvent * event);
static GstFlowReturn gst_onedlocalizer_bufferalloc (GstPad * pad, guint64 offset,
    guint size, GstCaps * caps, GstBuffer ** buf);
/* GObject vmethod implementations */

static void
gst_onedlocalizer_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "Coordgen",
    "Gst-Plugins-Robotix",
    "Evaluation for Markov-Localization",
    "Hannes Bistry <<hannesb@gmx.de>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory_pos));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory_res));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the onedlocalizer's class */
static void
gst_onedlocalizer_class_init (GstonedlocalizerClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_onedlocalizer_set_property;
  gobject_class->get_property = gst_onedlocalizer_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SCALE,
      g_param_spec_double ("scale", "Scale", "Scale between current to model",
          0.0,(double)G_MAXDOUBLE,1.0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DEV,
      g_param_spec_double ("dev", "Dev", "Deviation for gaussian",
          0.0,(double)G_MAXDOUBLE,0.5, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OPW,
      g_param_spec_uint ("opw", "Opw", "Size of the gaussian matrix(1,3,5...)",
          1,9,1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MT,
      g_param_spec_uint ("matchtype", "Matchtype", "How will we match, 0=all or nothing, 1=linear distance, 2=quadratic distance, 3=qd with tolerance ",
          1,9,1, G_PARAM_READWRITE));  
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_onedlocalizer_init (Gstonedlocalizer * filter,
    GstonedlocalizerClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sinkcur");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_onedlocalizer_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_onedlocalizer_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_onedlocalizer_chain));
  gst_pad_set_bufferalloc_function (filter->sinkpad,GST_DEBUG_FUNCPTR(gst_onedlocalizer_bufferalloc));
  gst_pad_set_event_function (filter->sinkpad,GST_DEBUG_FUNCPTR(gst_onedlocalizer_event));

  filter->sinkmodel = gst_pad_new_from_static_template (&sink_factory, "sinkmod");
  gst_pad_set_setcaps_function (filter->sinkmodel,
                                GST_DEBUG_FUNCPTR(gst_onedlocalizer_set_caps));
  gst_pad_set_getcaps_function (filter->sinkmodel,
                                GST_DEBUG_FUNCPTR(gst_onedlocalizer_get_caps));
  gst_pad_set_chain_function (filter->sinkmodel,
                              GST_DEBUG_FUNCPTR(gst_onedlocalizer_chain));
  gst_pad_set_bufferalloc_function (filter->sinkmodel,GST_DEBUG_FUNCPTR(gst_onedlocalizer_bufferalloc));
  gst_pad_set_event_function (filter->sinkmodel,GST_DEBUG_FUNCPTR(gst_onedlocalizer_event));
  filter->srcpos = gst_pad_new_from_static_template (&src_factory_pos, "pos");
  gst_pad_set_getcaps_function (filter->srcpos,
                                GST_DEBUG_FUNCPTR(gst_onedlocalizer_get_caps));
  //gst_pad_set_getcaps_function (filter->srcpos,
  //                             GST_DEBUG_FUNCPTR(gst_onedlocalizer_get_caps));
  gst_pad_set_event_function (filter->srcpos,GST_DEBUG_FUNCPTR(gst_onedlocalizer_event));

  filter->srcres = gst_pad_new_from_static_template (&src_factory_res, "res");
  gst_pad_set_getcaps_function (filter->srcres,gst_onedlocalizer_get_caps);

  filter->srcasc = gst_pad_new_from_static_template (&src_factory_asc, "asc");
  gst_pad_set_getcaps_function (filter->srcasc,gst_onedlocalizer_get_caps);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkmodel);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpos);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcres);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcasc);  
  filter->silent = FALSE;
  filter->eventmutex=g_mutex_new();
  filter->factor=1.0;
  filter->scale=1.0;
  filter->dev=0.5;
  filter->posx=-1;
  filter->posy=-1;
  filter->matchtype=1;
  filter->opw=1;
  filter->current=NULL;
  filter->model=NULL;
}

static void
gst_onedlocalizer_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstonedlocalizer *filter = GST_ONEDLOCALIZER (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_SCALE:
      filter->scale = g_value_get_double (value);
      break;
    case PROP_DEV:
      filter->dev = g_value_get_double (value);
      break;
    case PROP_OPW:
      filter->opw = g_value_get_uint (value);
      break;
    case PROP_MT:
      filter->matchtype = g_value_get_uint (value);
      break;      
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_onedlocalizer_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstonedlocalizer *filter = GST_ONEDLOCALIZER (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_SCALE:
      g_value_set_double (value, filter->scale);
      break;
    case PROP_DEV:
      g_value_set_double (value, filter->dev);
      break;    
    case PROP_OPW:
      g_value_set_uint (value, filter->opw);
      break;
    case PROP_MT:
      g_value_set_uint (value, filter->matchtype);
      break;      
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

 static GstFlowReturn
gst_onedlocalizer_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf)
{
  //Gsttimestamper *filter;
  //GstFlowReturn result;
  g_print ("Bufferalloc has been caled with Caps: %s\n",gst_caps_to_string(caps));
  //filter = GST_TIMESTAMPER (GST_PAD_PARENT (pad));

  /* Forward to src pad, without setting caps on the src pad */
  //result = gst_pad_alloc_buffer (filter->srcpad, offset, size, caps, buf);
  *buf=gst_buffer_new_and_alloc(size);
  gst_buffer_set_caps(*buf,caps);
  (*buf)->offset=offset;

  return GST_FLOW_OK;
}


/* GstElement vmethod implementations */
static GstCaps *
gst_onedlocalizer_get_caps (GstPad *pad){
  g_print("Start get caps from %s, returning %s!!\n",gst_pad_get_name(pad),gst_caps_to_string(gst_pad_get_pad_template_caps(pad)));
  Gstonedlocalizer *filter;

  filter = GST_ONEDLOCALIZER (GST_OBJECT_PARENT (pad));
  
  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  

}
/* this function handles the link with other elements */
static gboolean
gst_onedlocalizer_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstonedlocalizer *filter;
  filter = GST_ONEDLOCALIZER (gst_pad_get_parent (pad));
  
  GstPad *otherpad;
  gboolean returnval;
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);//get datatype of caps
  if ((pad==filter->sinkpad)&&(strstr(mime,"video"))){//get borders
    gst_structure_get_int(structure,"width",&(filter->width));
    gst_structure_get_int(structure,"height",&(filter->height)); 
    otherpad=filter->srcpos;
    }
  if ((pad==filter->sinkmodel)&&(strstr(mime,"video"))){//get borders
    gst_structure_get_int(structure,"width",&(filter->modelwidth));
    gst_structure_get_int(structure,"height",&(filter->modelheight)); 
    otherpad=filter->srcres;
    }

  //otherpad = (pad == filter->srcpos) ? filter->sinkpad : filter->srcpos;
  gst_object_unref (filter);
  returnval=gst_pad_set_caps (otherpad, caps);
  g_print("Onedlocalizer Set_caps on Pad %s with caps %s will be set on pad %s\n",gst_pad_get_name(pad),gst_caps_to_string(caps),gst_pad_get_name(otherpad));
  if ((!otherpad)||(!returnval)){ 
	g_print("Onedlocalizer Set_caps will fail\n");
  	sleep(10);
  }
  return returnval;
}

inline gdouble gauss(gdouble x, gdouble y, gdouble dev){
  return ((1/(2*M_PI*dev*dev))*exp(-(x*x+y*y)/(2*dev*dev)));
}

inline void generate_matrix(gdouble *result,gdouble off_x, gdouble off_y, gdouble dev, int opw){
  int x,y,c;
  c=0;
  for (y=0-((opw-1)/2);y<=0+((opw-1)/2);y++)	
    for (x=0-((opw-1)/2);x<=0+((opw-1)/2);x++){
      result[c++]=gauss((x+off_x),(y+off_y),dev);
      g_print("Y:%d X:%d c:%d %lf",y,x,c-1,result[c-1]);
  }
}

void shiftmatrix(Gstonedlocalizer *filter,double dx,double dy,double dev, int opw){
  int x,y,c,dxint,dyint,x2,y2;
  gdouble *operator;
  gdouble  off_x,off_y;
  dxint=round(dx);
  dyint=round(dy);
  if ((opw==1)||(dev==0.0)){
    for (y=0;y<filter->modelheight;y++)for(x=0;x<filter->modelwidth;x++){
      if G_LIKELY(((y-dyint)>0)&&((y-dyint)<filter->modelheight)&&((x-dxint)>0)&&((x-dxint)<filter->modelwidth)){
        filter->estimationnew[y*filter->modelwidth+x]=filter->estimation[((y-dyint)*filter->modelwidth)+x-dxint];
      }
      else filter->estimationnew[y*filter->modelwidth+x]=0;
    }
  }
  else{
    operator=malloc(opw*opw*sizeof(gdouble));
    off_x=dxint-dx;
    off_y=dyint-dy;
    g_print("off_x: %f, dxint: %d dx: %f\n",off_x,dxint,dx);
    g_print("off_x: %f, dxint: %d dx: %f\n",off_x,dxint,dx);
    generate_matrix(operator,off_x,off_y,dev,opw);
    for (y=0;y<filter->modelheight;y++)for(x=0;x<filter->modelwidth;x++)filter->estimationnew[y*filter->modelwidth+x]=0;
    c=0;
    for (y=0;y<filter->modelheight;y++)for(x=0;x<filter->modelwidth;x++){
      c=0;
      for (y2=0-((opw-1)/2);y2<=0+((opw-1)/2);y2++)	
        for (x2=0-((opw-1)/2);x2<=0+((opw-1)/2);x2++){
          if G_LIKELY(((y+y2+dyint)>0)&&((y+y2+dyint)<filter->modelheight)&&((x+x2+dxint)>0)&&((x+x2+dxint)<filter->modelwidth)){
	  filter->estimationnew[((y+y2+dyint)*filter->modelwidth)+x+x2+dxint]+=operator[c++]*filter->estimation[y*filter->modelwidth+x];
	  }
          else c++;
	//
        //filter->estimationnew[y*filter->modelwidth+x]=filter->estimation((y-dyint)*filter->modelwidth)+x-dxint];
        }
      }
  free(operator);
    }
}


/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_onedlocalizer_chain (GstPad * pad, GstBuffer * buf)
{

  Gstonedlocalizer *filter;
  GstFlowReturn ret1,ret2;
  filter = GST_ONEDLOCALIZER (GST_OBJECT_PARENT (pad));
  g_print("Onedlocalizer pipeline with caps %s\n",gst_caps_to_string(buf->caps));
  g_mutex_lock(filter->eventmutex);
    if (pad==filter->sinkpad){
	if (!filter->current)filter->current=buf;
	else {
	  g_print("Onedlocalizer Current-Sinkpad goes sleeping\n");
          g_mutex_unlock(filter->eventmutex);
	  while(1) usleep(10000);
	}
    }
    else if (pad==filter->sinkmodel){
	if(!filter->model)filter->model=buf;
	else{
	  g_print("Onedlocalizer Model-Sinkpad goes sleeping\n");
          g_mutex_unlock(filter->eventmutex);
	  while(1) usleep(10000);
	}
  }
  
  if (!((filter->model)&&(filter->current))) {
	g_mutex_unlock(filter->eventmutex);
	return GST_FLOW_OK;
  }
  else g_mutex_unlock(filter->eventmutex);
  GstBuffer *outres,*outpos; 
  gboolean end=FALSE;
  //initialize variables
  filter->estimation=(gdouble*)malloc(filter->modelwidth*filter->modelheight*sizeof(gdouble));
  filter->estimationnew=(gdouble*)malloc(filter->modelwidth*filter->modelheight*sizeof(gdouble));
  filter->likelyhood=(gdouble*)malloc(filter->modelwidth*filter->modelheight*sizeof(gdouble));
  int x,y,nrresets;
  nrresets=0;
  gdouble sum;
  for (x=0;x<filter->modelwidth*filter->modelheight;x++){
    filter->estimation[x]=1.0/(double)(filter->modelwidth*filter->modelheight);
    filter->estimationnew[x]=1.0/(double)(filter->modelwidth*filter->modelheight);
    filter->likelyhood[x]=0.0;
  }  
  filter->eventpending=TRUE;
  filter->eventtype=-1;

  while (!end){
    //wait for valid input
    while (!filter->eventpending){
	usleep(10000);
    }
    g_mutex_lock(filter->eventmutex);


    filter->eventpending=FALSE;
    g_mutex_unlock(filter->eventmutex);
    if(filter->eventtype==-2){
	end=TRUE;	
    }
    //update states
    else if (filter->eventtype==-1){
	filter->posx=-1;
	filter->posy=-1;
 	nrresets++;
        for (y=0;x<filter->modelwidth*filter->modelheight;x++){
          filter->estimationnew[x]=1.0/(double)(filter->modelwidth*filter->modelheight);
           filter->likelyhood[x]=0.0;
        } 
    }
    else if(filter->eventtype==1){
 	//a-priori
        for (x=0;x<filter->modelwidth*filter->modelheight;x++){
          filter->estimationnew[x]=1.0/(double)(filter->modelwidth*filter->modelheight);
           filter->likelyhood[x]=0.0;
        } 
    }
    else if(filter->eventtype==8){//UP
 	//a-priori
	if (filter->posy>0)filter->posy--;
        //for (y=0;y<(filter->modelheight-1);y++)for(x=0;x<filter->modelwidth;x++){
        //  filter->estimationnew[y*filter->modelwidth+x]=filter->estimation[((y+1)*filter->modelwidth)+x];
	//}
	//y=filter->modelheight-1;
	//for(x=0;x<filter->modelwidth;x++)filter->estimationnew[y*filter->modelwidth+x]=0;
	shiftmatrix(filter,0,-1*(1/filter->scale),filter->dev,filter->opw);
    }
    else if(filter->eventtype==2){//Down
 	//a-priori
	if (filter->posy<filter->height)filter->posy++;
        //for (y=1;y<filter->modelheight;y++)for(x=0;x<filter->modelwidth;x++){
        //  filter->estimationnew[y*filter->modelwidth+x]=filter->estimation[((y-1)*filter->modelwidth)+x];
	//}
	//y=0;
	//for(x=0;x<filter->modelwidth;x++)filter->estimationnew[y*filter->modelwidth+x]=0;
	shiftmatrix(filter,0,1*(1/filter->scale),filter->dev,filter->opw);	
    }
    else if(filter->eventtype==4){//Left
 	//a-priori
	if (filter->posx>0)filter->posx--;
        //for (y=0;y<filter->modelheight;y++)for(x=0;x<filter->modelwidth-1;x++){
        //  filter->estimationnew[y*filter->modelwidth+x]=filter->estimation[(y*filter->modelwidth)+x+1];
	//}
	//x=filter->modelwidth-1;
	//for(y=0;y<filter->modelheight;y++)filter->estimationnew[y*filter->modelwidth+x]=0;
	shiftmatrix(filter,-1*(1/filter->scale),0,filter->dev,filter->opw);		
    }
    else if(filter->eventtype==6){//Right
 	//a-priori
	if (filter->posx<filter->width)filter->posx++;
        //for (y=0;y<filter->modelheight;y++)for(x=1;x<filter->modelwidth;x++){
        //  filter->estimationnew[y*filter->modelwidth+x]=filter->estimation[(y*filter->modelwidth)+x-1];
	//}
	//x=0;
	//for(y=0;y<filter->modelheight;y++)filter->estimationnew[y*filter->modelwidth+x]=0;
	shiftmatrix(filter,1*(1/filter->scale),0,filter->dev,filter->opw);		
    }
    //calculate likelyhood
    int likely=0;
    if ((filter->posx>-1)&&(filter->posy>-1)){
        for (x=0;x<filter->modelwidth*filter->modelheight;x++){
	  //if (buf->data[(filter->posy*filter->modelwidth+filter->posx)]==buf->data[x]){
	  if (filter->matchtype==0){
	    if (filter->current->data[(filter->posy*filter->width+filter->posx)]==filter->model->data[x])		
		{
	  	filter->likelyhood[x]=1;
	  	likely++;
	  	}
	  	else filter->likelyhood[x]=0;
	  }
	  else if(filter->matchtype==1)filter->likelyhood[x]=1.0-((gdouble)abs((gdouble)filter->current->data[(filter->posy*filter->width+filter->posx)]-(gdouble)filter->model->data[x])/255.0);
	  else if(filter->matchtype==2)filter->likelyhood[x]=1.0-(((gdouble)abs((gdouble)filter->current->data[(filter->posy*filter->width+filter->posx)]-(gdouble)filter->model->data[x])/255.0)*((gdouble)abs((gdouble)filter->current->data[(filter->posy*filter->width+filter->posx)]-(gdouble)filter->model->data[x])/255.0));
	  else if(filter->matchtype==3){
	    filter->likelyhood[x]=1.0-((gdouble)abs((gdouble)filter->current->data[(filter->posy*filter->width+filter->posx)]-(gdouble)filter->model->data[x])/255.0);
	    if (filter->likelyhood[x]>0.9)filter->likelyhood[x]=1.0;
	    else if (filter->likelyhood[x]<0.5)filter->likelyhood[x]=0.0;
	    //else filter->likelyhood[x]=(filter->likelyhood[x]-0.2)*1.6666666666666;
	  }
	}
    }
    //multiply and sum


    sum=0;
    for (x=0;x<filter->modelwidth*filter->modelheight;x++){
	filter->estimationnew[x]=filter->estimationnew[x]*filter->likelyhood[x];    
	sum+=filter->estimationnew[x];
    }
    //normalize matrix
    if (sum>0) for (x=0;x<filter->modelwidth*filter->modelheight;x++){
	filter->estimationnew[x]=filter->estimationnew[x]/sum;
    }
    else for (x=0;x<filter->modelwidth*filter->modelheight;x++){
    	  //g_print("Error, sum is ZERO\n");
          filter->estimationnew[x]=1.0/(double)(filter->modelwidth*filter->modelheight);
        } 
    //update estimation
    for (x=0;x<filter->modelwidth*filter->modelheight;x++){
          filter->estimation[x]=filter->estimationnew[x];
          filter->estimationnew[x]=0;
    } 
    
    if (gst_pad_is_linked (filter->srcasc)){
      GstBuffer * bufasc;
      int ascsize=filter->modelwidth*filter->modelheight*14;
      gchar* outasc;
      gint offset=0;
      int i,j;
      outasc=g_malloc(ascsize);//just guessing output size 
      for( i = filter->modelheight-1; i >= 0; i-- ){
        for( j = 0; j < filter->modelwidth; j++ )
	{
	  //g_print("Onedlocalizer printing %f , or in file %1.4f",filter->estimation[i*filter->modelwidth+j],filter->estimation[i*filter->modelwidth+j]);
          offset+=sprintf(outasc+offset,"%1.10f ",filter->estimation[i*filter->modelwidth+j] );
	}
	offset+=sprintf(outasc+offset,"\n" );
      }
      g_print ("Onedlocalizer Allocated %d, used %d \n",filter->modelwidth*filter->modelheight*14,offset);
      outasc=g_realloc(outasc,offset+1);
      bufasc=gst_buffer_new_and_alloc(0);
      gst_buffer_set_data(bufasc,(guint8 *)outasc,offset+1);
      GST_BUFFER_MALLOCDATA(bufasc)=(guint8 *)outasc;
      gst_pad_push (filter->srcasc, bufasc);
    }
    
    
    
    
    //generate position output image
    outpos=gst_buffer_copy(filter->current);
    //outpos=gst_buffer_copy(filter->model);
    gst_buffer_make_writable(outpos);
    // fix pixel-aspect-shit
    //GstCaps* newcaps;
    //newcaps=gst_caps_new_simple("");
    outpos->caps=gst_caps_make_writable(outpos->caps);  
    gst_structure_remove_field(gst_caps_get_structure(outpos->caps,0),"pixel-aspect-ratio");
    //
    outpos->offset=filter->offset;
    outpos->duration=-1;
    outpos->timestamp=-1;
    if ((filter->posx>-1)&&(filter->posy>-1)) outpos->data[(filter->posy*filter->width+filter->posx)]=255;
    //generate result output image
    gdouble min=1,max=0;
    //GstCaps* schrott;
    //schrott=gst_pad_get_caps_reffed(filter->srcpos);
    //if (!gst_caps_can_intersect (schrott, outpos->caps)) g_print("Onedlocalizer Outpos Scheisshaufen!!\n");


    outres=gst_buffer_copy(filter->model);
    gst_buffer_make_writable(outres);
    outres->offset=filter->offset;
    outres->duration=-1;
    outres->timestamp=-1;
    filter->offset++;
    outres->caps=gst_caps_make_writable(outres->caps);  
    gst_structure_remove_field(gst_caps_get_structure(outres->caps,0),"pixel-aspect-ratio");
    for (x=0;x<filter->modelwidth*filter->modelheight;x++){
	if(filter->estimation[x]>max)max=filter->estimation[x];
	if(filter->estimation[x]<min)min=filter->estimation[x];	
    }
    g_print("Onedlocalizer min=%f max=%f posx=%d posy=%d sum=%f and %d likely to be the same, and %d times resetted, width is %d, height is %d\n",min,max,filter->posx,filter->posy,sum,likely,nrresets,filter->width,filter->height);
    for (x=0;x<filter->modelwidth*filter->modelheight;x++){
	(outres->data[x])=(unsigned char)(255.0*((filter->estimation[x]-min)/(max-min)));
    }
    g_print("Onedlocalizer sending buffer to outpos with caps %s\n",gst_caps_to_string(outpos->caps));  
    ret1=gst_pad_push (filter->srcpos, outpos);
    if (ret1!=GST_FLOW_OK) {g_print("Onedlocalizer Outpos GST_FLOW_NOT_OK %d\n",ret1);
	sleep(100);
    }
    //just push out the incoming buffer without touching it 
    ret2=gst_pad_push (filter->srcres, outres); 
    if (ret2!=GST_FLOW_OK) g_print("Onedlocalizer Outres GST_FLOW_NOT_OK %d\n",ret2);   
    if ((ret1!=GST_FLOW_OK)||(ret2!=GST_FLOW_OK)){
	g_print("Onedlocalizer GST_FLOW_NOT_OK");
	}
  }
/*
  GList *temp;
  //if (filter->silent == FALSE)
    //g_print ("I'm plugged, therefore I'm in.\n");
  filter->numregions=g_list_length(filter->coordinates);
  outbuf=gst_buffer_new_and_alloc(filter->numregions*sizeof(struct coordinates));
  outbuf->timestamp=buf->timestamp;
  outbuf->duration=buf->duration;
  outbuf->offset=buf->offset;
  outbuf->offset_end=buf->offset_end;
  GstCaps *tempcaps=gst_caps_copy (gst_pad_get_pad_template_caps(filter->srccoord));
  gst_buffer_set_caps(outbuf,tempcaps);
  gst_caps_unref(tempcaps);
  temp=filter->coordinates;
  int count=0;
  while (temp){
    ((struct coordinates*)(outbuf->data))[count].x1=((struct coordinates*)(temp->data))->x1;
    ((struct coordinates*)(outbuf->data))[count].x2=((struct coordinates*)(temp->data))->x2;
    ((struct coordinates*)(outbuf->data))[count].y1=((struct coordinates*)(temp->data))->y1;
    ((struct coordinates*)(outbuf->data))[count].y2=((struct coordinates*)(temp->data))->y2;
    temp=g_list_next(temp);
    count++;
  }
  ret1=gst_pad_push (filter->srccoord, outbuf);
  //just push out the incoming buffer without touching it 
  ret2=gst_pad_push (filter->srcpad, buf);

  if (ret1!=GST_FLOW_OK) return ret1;
*/
  if (ret1!=GST_FLOW_OK) return ret1;
  else if (ret2!=GST_FLOW_OK) return ret2;
  else return GST_FLOW_OK;
}

static gboolean
gst_onedlocalizer_event (GstPad  *pad, GstEvent * event)
{
  Gstonedlocalizer *filter;
  gboolean ret=TRUE;
  int button;

  filter = GST_ONEDLOCALIZER (gst_pad_get_parent (pad));





  if (pad==filter->sinkpad){
	  switch (GST_EVENT_TYPE (event)) {
    		case GST_EVENT_EOS:
      		//if (pad == filter->sinkpad1) filter->stopped1=TRUE;
      		//if (pad == filter->sinkpad2) filter->stopped2=TRUE;
      		//if ((filter->stopped1)&&(filter->stopped2)) ret = gst_pad_push_event (filter->srcpad, event);
      		//g_print("Event EOS!\n");
      	  break;
         default:
	  gst_pad_push_event (filter->srcpos, event);
        break;
    }
	return TRUE;
  }
  if (pad==filter->sinkmodel){
    switch (GST_EVENT_TYPE (event)) {
    	 case GST_EVENT_EOS:
      	   break;
         default:
	   gst_pad_push_event (filter->srcres, event);
      	   break;
    }
    return TRUE;
  }
  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_NAVIGATION:
      while (filter->eventpending){
	usleep(100000);
      }
      //get event mutex
      g_mutex_lock(filter->eventmutex);
      g_print("Event Navigation! Event named %s, Structure Named %s, Content %s\n",gst_event_type_get_name(GST_EVENT_TYPE(event)),gst_structure_get_name(  gst_event_get_structure(event)),gst_structure_to_string(gst_event_get_structure(event)));
      g_print("Eventstring: %s",g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")));
      if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"mouse-button-press")){
        button=g_value_get_int(gst_structure_get_value(gst_event_get_structure(event),"button"));
        if (button==1){
             filter->posx=(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_x"));
	     filter->posy=(gint64)g_value_get_double(gst_structure_get_value(gst_event_get_structure(event),"pointer_y"));
	     filter->eventtype=1;
	     filter->eventpending=TRUE;
	}
	if (button==3){
	  filter->eventtype=-1;
	  filter->eventpending=TRUE;
	}	     
      }
      g_mutex_unlock(filter->eventmutex);
      if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"key-press")){
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"Up")){
	  filter->eventtype=8;
	  filter->eventpending=TRUE;	  
	}
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"Down"))   
	{
	  filter->eventtype=2;
	  filter->eventpending=TRUE;	  
	}
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"Left"))
	{
	  filter->eventtype=4;
	  filter->eventpending=TRUE;	  
	}
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"Right"))
	{
	  filter->eventtype=6;
	  filter->eventpending=TRUE;	  
	}
	if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"e"))
	{
	  filter->eventtype=-2;
	  filter->eventpending=TRUE;	  
	}
      }
      break;
    default:
      g_print("Other event\n");
      ret = gst_pad_push_event (filter->sinkpad, event);
      break;
  }
  
  gst_object_unref (filter);
  return ret;
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
onedlocalizer_init (GstPlugin * onedlocalizer)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template onedlocalizer' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_onedlocalizer_debug, "onedlocalizer",
      0, "Template onedlocalizer");

  return gst_element_register (onedlocalizer, "onedlocalizer", GST_RANK_NONE,
      GST_TYPE_ONEDLOCALIZER);
}

/* gstreamer looks for this structure to register onedlocalizers
 *
 * exchange the string 'Template onedlocalizer' with your onedlocalizer description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "onedlocalizer",
    "Template onedlocalizer",
    onedlocalizer_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
