/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-piptrans
 *
 * FIXME:Describe piptrans here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! piptrans ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <cv.h>
#include <cxcore.h>

#include "gstpiptrans.h"

GST_DEBUG_CATEGORY_STATIC (gst_piptranstrans_debug);
#define GST_CAT_DEFAULT gst_piptranstrans_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_TRIGGER,
  PROP_TRIGGER_SMALL,
  PROP_TRIGGER_POS,
  PROP_TRIGGER_EQUAL,
  PROP_SCALE,
  PROP_DIFF,
  PROP_JUSTTRANS,
  PROP_FWEOS,
  PROP_SIFTOFF_X,
  PROP_SIFTOFF_Y,
  PROP_AVG,
  PROP_DIFFF
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "video/x-raw-rgb, "
      "width=(int) [ 1, MAX ],"
      "height=(int) [ 1, MAX ],"
      "framerate = (fraction) [ 0, MAX ], " 
      "bpp = (int) 32, "
      "depth = (int) 24"
    )
    );

static GstStaticPadTemplate sink_factory_pos = GST_STATIC_PAD_TEMPLATE ("sink-pos",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "matrix-list"
    )
    );



// static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
//     GST_PAD_SRC,
//     GST_PAD_ALWAYS,
//     GST_STATIC_CAPS (
//       "video/x-raw-rgb, "
//       "width=(int) [ 1, MAX ],"
//       "height=(int) [ 1, MAX ],"
//       "framerate = (fraction) [ 0, MAX ], " 
//       "bpp = (int) 32, "
//       "depth = (int) 24"
//     )
//     );
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gstpiptranstrans, gst_piptranstrans, GstElement,
    GST_TYPE_ELEMENT);

static void gst_piptranstrans_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_piptranstrans_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstCaps * gst_piptranstrans_get_caps (GstPad *pad);
static gboolean gst_piptranstrans_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_piptranstrans_chain (GstPad * pad, GstBuffer * buf);
static GstFlowReturn gst_piptranstrans_chain2 (GstPad * pad, GstBuffer * buf);
static gboolean gst_piptranstrans_event (GstPad  *pad, GstEvent * event);
static GstStateChangeReturn
gst_piptranstrans_change_state (GstElement *element, GstStateChange transition);
/* GObject vmethod implementations */

static void
gst_piptranstrans_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "PIPTRANS",
    "Gst-Plugins-Robotix",
    "Warp one image to another according to transformation-matrix",
    "Hannes Bistry <<hannesb@gmx.de>>");


  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the piptrans's class */
static void
gst_piptranstrans_class_init (GstpiptranstransClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_piptranstrans_set_property;
  gobject_class->get_property = gst_piptranstrans_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER,
      g_param_spec_boolean ("trigger", "Trigger", "Trigger on new data on main Input ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER_SMALL,
      g_param_spec_boolean ("trigger_small", "Trigger_small", "Trigger on new data on small picture Input ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER_POS,
      g_param_spec_boolean ("trigger_pos", "Trigger_pos", "Trigger on new data on matrix vector ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER_EQUAL,
      g_param_spec_boolean ("trigger_equal", "Trigger_equal", "Trigger on equal frame numbers ?",
          TRUE, G_PARAM_READWRITE));
g_object_class_install_property (gobject_class, PROP_SCALE,
      g_param_spec_double ("scale", "Scale",
          "Wether the small/static image is scaled down", (double)G_MINDOUBLE,(double)G_MAXDOUBLE,
          1.0, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DIFF,
      g_param_spec_boolean ("diff", "Diff", "Whether to show diff image",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_JUSTTRANS,
      g_param_spec_boolean ("justtrans", "Justtrans", "Whether to do just the transformation",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_FWEOS,
      g_param_spec_boolean ("fweos", "FWEOS", "Whether to foreward EOS event",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SIFTOFF_X,
      g_param_spec_uint ("siftoffx", "Siftoffx",
          "X-Offset of Small Embedded picture", 0, G_MAXUINT, 0,
          G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SIFTOFF_Y,
      g_param_spec_uint ("siftoffy", "Siftoffy",
          "Y-Offset of Small Embedded picture", 0, G_MAXUINT, 0,
          G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_AVG,
      g_param_spec_boolean ("avg", "Avg", "Generate average of the two images",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DIFFF,
      g_param_spec_double ("difffactor", "Difffactor", "When building diff of two images, the higher the stronger, saturating",
          -G_MAXDOUBLE,G_MAXDOUBLE,0.5, G_PARAM_READWRITE)); 
  gstelement_class->change_state = gst_piptranstrans_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_piptranstrans_init (Gstpiptranstrans * filter,
    GstpiptranstransClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_piptranstrans_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_piptranstrans_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_piptranstrans_chain2));
  gst_pad_set_event_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_piptranstrans_event));

  filter->sinksmall = gst_pad_new_from_static_template (&sink_factory, "small");
  gst_pad_set_setcaps_function (filter->sinksmall,
                                GST_DEBUG_FUNCPTR(gst_piptranstrans_set_caps));
  gst_pad_set_getcaps_function (filter->sinksmall,
                                GST_DEBUG_FUNCPTR(gst_piptranstrans_get_caps));
  gst_pad_set_chain_function (filter->sinksmall,
                              GST_DEBUG_FUNCPTR(gst_piptranstrans_chain2));
  gst_pad_set_event_function (filter->sinksmall,
                              GST_DEBUG_FUNCPTR(gst_piptranstrans_event));

  filter->sinkpos = gst_pad_new_from_static_template (&sink_factory_pos, "pos");
  gst_pad_set_setcaps_function (filter->sinkpos,
                                GST_DEBUG_FUNCPTR(gst_piptranstrans_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpos,
                                GST_DEBUG_FUNCPTR(gst_piptranstrans_get_caps));
  gst_pad_set_chain_function (filter->sinkpos,
                              GST_DEBUG_FUNCPTR(gst_piptranstrans_chain2));
  gst_pad_set_event_function (filter->sinkpos,
                              GST_DEBUG_FUNCPTR(gst_piptranstrans_event));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_piptranstrans_get_caps));
  gst_pad_set_event_function (filter->srcpad,
                              GST_DEBUG_FUNCPTR(gst_piptranstrans_event));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinksmall);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpos);



  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->mutex=g_mutex_new();
  filter->trigger=TRUE;
  filter->trigger_small=TRUE;
  filter->trigger_pos=TRUE;
  filter->scale=1.0;
  filter->penalty=100000;
  filter->valve_buf=5;
  filter->off_x=0;
  filter->off_y=0;
  filter->diff = FALSE;
  filter->fweos = FALSE;
  filter->justtrans = FALSE;
  filter->avg=FALSE;
  filter->difff=0.5;
}

static void
gst_piptranstrans_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstpiptranstrans *filter = GST_PIPTRANS (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER:
      filter->trigger = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER_SMALL:
      filter->trigger_small = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER_POS:
      filter->trigger_pos = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER_EQUAL:
      filter->trigger_num = g_value_get_boolean (value);
      break;
    case PROP_SCALE:
      filter->scale = g_value_get_double (value);
      break;
    case PROP_DIFF:
      filter->diff = g_value_get_boolean (value);
      break;
    case PROP_JUSTTRANS:
      filter->justtrans = g_value_get_boolean (value);
      break;
    case PROP_FWEOS:
      filter->fweos = g_value_get_boolean (value);
      break;
    case PROP_SIFTOFF_X:
      filter->siftoff_x = g_value_get_uint (value);
      break;
    case PROP_SIFTOFF_Y:
      filter->siftoff_y = g_value_get_uint (value);
      break;
    case PROP_AVG:
      filter->avg = g_value_get_boolean (value);
      break;
    case PROP_DIFFF:
      filter->difff = g_value_get_double (value);
    break;    
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_piptranstrans_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstpiptranstrans *filter = GST_PIPTRANS (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_TRIGGER:
      g_value_set_boolean (value, filter->trigger);
      break;
    case PROP_TRIGGER_SMALL:
      g_value_set_boolean (value, filter->trigger_small);
      break;
    case PROP_TRIGGER_POS:
      g_value_set_boolean (value, filter->trigger_pos);
      break;
    case PROP_TRIGGER_EQUAL:
      g_value_set_boolean (value, filter->trigger_num);
      break;
    case PROP_SCALE:
      g_value_set_double (value, filter->scale);
      break;
    case PROP_DIFF:
      g_value_set_boolean (value, filter->diff);
      break;
    case PROP_JUSTTRANS:
      g_value_set_boolean (value, filter->justtrans);
      break;
    case PROP_FWEOS:
      g_value_set_boolean (value, filter->fweos);
      break;
    case PROP_SIFTOFF_X:
      g_value_set_uint (value, filter->siftoff_x);
      break;
    case PROP_SIFTOFF_Y:
      g_value_set_uint (value, filter->siftoff_y);
      break;
    case PROP_AVG:
      g_value_set_boolean (value, filter->avg);
      break;
    case PROP_DIFFF:
      g_value_set_double (value,filter->difff);
    break;        
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */










static GstCaps *
gst_piptranstrans_get_caps (GstPad *pad){
  g_print("Start get caps!!\n");
  Gstpiptranstrans *filter;

  filter = GST_PIPTRANS (GST_OBJECT_PARENT (pad));

  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  

}
/* this function handles the link with other elements */
static gboolean
gst_piptranstrans_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstpiptranstrans *filter;
  gboolean ret=TRUE;
  filter = GST_PIPTRANS (gst_pad_get_parent (pad));
  g_mutex_lock(filter->mutex);
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);  
  g_print("PIPTRANS set caps: %s!!\n",gst_caps_to_string (caps));

  if (pad == filter->sinkpad){//foreward negotiation, save size and bpp
    g_print("PIPTRANS set caps sink 1!!\n");
    if (strcmp (mime, "video/x-raw-rgb")) ret=FALSE;
    g_print("PIPTRANS set caps sink 1!!\n");
    if (!gst_structure_get_int(structure,"width",&(filter->w))) ret=FALSE;
    g_print("PIPTRANS set caps sink 1!!\n");
    if (!gst_structure_get_int(structure,"height",&(filter->h))) ret=FALSE;
    g_print("PIPTRANS set caps sink 1!!\n");
    if (!gst_structure_get_int(structure,"bpp",&(filter->bpp))) ret=FALSE;
    if (filter->smallneg){
      if (((filter->w)<(filter->ws))||((filter->h)<(filter->hs))||((filter->bpp)!=(filter->bppsmall))) ret=FALSE;
    }
    if(ret){
      filter->bigneg=TRUE;
      ret=gst_pad_set_caps(filter->srcpad,caps);
    }
  }
  else if (pad == filter->sinksmall){//chack resolution smaller than 
    g_print("PIPTRANS set caps sink 2!!\n");
    if (strcmp (mime, "video/x-raw-rgb")) ret=FALSE;
    g_print("PIPTRANS set caps sink 2!!\n");
    if (!gst_structure_get_int(structure,"width",&(filter->ws))) ret=FALSE;
    g_print("PIPTRANS set caps sink 2!!\n");
    if (!gst_structure_get_int(structure,"height",&(filter->hs))) ret=FALSE;
    filter->hs_neg=filter->hs;
    filter->ws_neg=filter->ws;
    if (!gst_structure_get_int(structure,"bpp",&(filter->bppsmall))) ret=FALSE;
    if (filter->bigneg){
      if (((filter->w)<(filter->ws))||((filter->h)<(filter->hs))||((filter->bpp)!=(filter->bppsmall))) ret=FALSE;
      if (!ret) g_print ("diagnose embedded picture: %d x %d bpp: %d, big picture %d x %d bpp %d \n",filter->ws,filter->hs,filter->bppsmall,filter->w,filter->h,filter->bpp);
    }
    if (ret)filter->smallneg=TRUE;
    //gst_caps_unref(caps); will lead to an error
  }
  else if (pad == filter->sinkpos){
  }
  else if (pad == filter->srcpad){//wird schon stimmen
  }
  g_mutex_unlock(filter->mutex);  
  gst_object_unref (filter);
  if (!ret) g_print("WARNING PIPTRANS will return false!!\n");
  //return gst_pad_set_caps (otherpad, caps);
  return ret;
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_piptranstrans_chain (GstPad * pad, GstBuffer * buf)
{
  Gstpiptranstrans *filter;

  filter = GST_PIPTRANS (GST_OBJECT_PARENT (pad));

  gboolean trigger;

  //if (filter->silent == FALSE)
  //  g_print ("Starting PIPTRANS main chain.\n");

  g_mutex_lock(filter->mutex);

  if (pad== filter->sinkpad){
    g_print ("PIPTRANS Main Sink pad.\n");
    trigger=filter->trigger;
      filter->offbig=buf->offset;
      if (filter->b1)gst_buffer_unref(filter->b1);
      filter->b1=buf;
      //filter->feat1=(struct feature*)buf->data;
      //filter->n1=(buf->size)/(sizeof(struct feature));
  }
  else if (pad == filter->sinksmall){
    g_print ("PIPTRANS Small Sink pad.\n");
    trigger=filter->trigger_small;
    filter->offsmall=buf->offset;
    if (filter->b2) gst_buffer_unref(filter->b2);
    filter->b2=buf;
    //filter->feat2=(struct feature*)buf->data;
    //filter->n2=(buf->size)/(sizeof(struct feature));
  }
  else if (pad == filter->sinkpos){
    g_print ("PIPTRANS Position Sink pad.\n");
    trigger=filter->trigger_pos;
    filter->offpos=buf->offset;
    if (filter->b3) gst_buffer_unref(filter->b3);
    filter->b3=buf;
  }

  if (trigger&&filter->trigger_num){
    g_print ("Position Big: %llu Small %llu Pos %llu.\n",filter->offbig,filter->offsmall,filter->offpos);
    if ((!filter->trigger_small||(filter->offbig==filter->offsmall))&&(!filter->trigger_pos||(filter->offpos==filter->offbig))) trigger=TRUE;
    else trigger=FALSE;
  }




  //decide to generate output?
  if ((filter->b1)&&(filter->b2)&&(trigger)){
    g_print ("Got triggered.\n");
//security check


//copy
    gint h;
    g_print("Will copy %d Sizes: w=%d ws=%d h=%d hs=%d\n",filter->ws*(filter->bpp/8),filter->w,filter->ws,filter->h,filter->hs);
    for (h=filter->siftoff_y;h<(filter->siftoff_y+filter->hs);h++){
      memcpy(filter->b1->data+(filter->bpp/8)*(h*filter->w+filter->siftoff_x),filter->b2->data+(filter->bpp/8)*((h-filter->siftoff_y)*filter->ws),(filter->ws*(filter->bpp/8)));
      }
    IplImage *process;
//TODO nächsten zwei zeilen ordentlich machen 
      process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,(filter->bpp/8));
      process->imageData=(char*)filter->b1->data;  
      filter->imagesize.height=filter->h;
      filter->imagesize.width=filter->w;
//paint lines
    if(filter->b3){



      CvPoint pt1, pt2;
      gint numlines;
      numlines=GST_BUFFER_SIZE(filter->b3)/sizeof(struct coordinates);
      g_print ("Will draw %d lines.\n",numlines);
      for (h=0; h<numlines;h++){
      pt1=cvPoint( cvRound(((struct coordinates*)(filter->b3->data))[h].x1),cvRound(((struct coordinates*)(filter->b3->data))[h].y1));
      pt2=cvPoint( cvRound(filter->siftoff_x+filter->scale*(((struct coordinates*)(filter->b3->data))[h].x2)), cvRound(filter->siftoff_y+filter->scale*(((struct coordinates*)(filter->b3->data))[h].y2)) );
      //TODO cvpoint nur einmal initialisieren, sonst Speicherleck
      g_print ("Will draw from x %d y %d to x2 %d y2 %d .\n",pt1.x,pt1.y,pt2.x,pt2.y);
      //if (650<((struct coordinates*)(filter->b3->data))[h].x1) cvLine( process, pt1, pt2, CV_RGB(255,0,255), 1, 8, 0 );
      //cvLine( process, pt1, pt2, CV_RGB(255,0,255), 1, 8, 0 );
	//myline
      if (42){
	int x1,x2,y1,y2,xi,yi,temp;
	unsigned char R=255,G=0,B=255;
	x1=pt1.x;
	x2=pt2.x;
	y1=pt1.y;
	y2=pt2.y;
	/*if (x1>x2){
	  temp=x1;
	  x1=x2;
	  x2=temp;
	}
	if (y1>y2){
	  temp=y1;
	  y1=y2;
	  y2=temp;
	}*/
	if (x1<x2)xi=x1; else xi=x2;
	if (y1<y2)yi=y1; else yi=y2;
	gdouble stg=(gdouble)(y2-y1)/(gdouble)(x2-x1);
	gdouble stgy=(gdouble)(x2-x1)/(gdouble)(y2-y1);
        //for(xi=xi;((xi<=x2)||(xi<=x1));xi++)g_print("Hululululu %d\n",xi);
	if (x1<x2)xi=x1; else xi=x2;
	for(;((xi<=x2)||(xi<=x1));xi++) {
	  for(;((yi<=y2)||(yi<=y1));yi++){
	    gdouble intensity=fabs((gdouble)yi-((gdouble)y1+((gdouble)(xi-x1)*stg)));
	    gdouble intensitx=fabs((gdouble)xi-((gdouble)x1+((gdouble)(yi-y1)*stgy)));
            //g_print("Checking X:%d and Y:%d intx: %f inty %f and if in too stupid x1 %d x2 %d\n",xi,yi,intensity,intensitx, x1,x2);
	    if ((intensity<0.5)||(intensitx<0.5)){
		unsigned char * base=filter->b1->data+((xi+(yi*filter->w))*(filter->bpp/8));
		/*
		*base=(unsigned char)((1.0-intensity)*((gdouble)(*base))+(intensity*(gdouble)B));//blau
		*(base+1)=(unsigned char)((1.0-intensity)*((gdouble)(*(base+1)))+(intensity*(gdouble)G));
		*(base+2)=(unsigned char)((1.0-intensity)*((gdouble)(*(base+2)))+(intensity*(gdouble)R));
		*(base+3)=0;*/
		
		*base=255;
		*(base+1)=0;
		*(base+2)=255;
		*(base+3)=0;		
	    }
          }
	if (y1<y2)yi=y1; else yi=y2;
	//g_print("Checking X inner loop:%d\n",xi);
	}
	//g_print("Checking X:%d\n",xi);
	
      }
    }
  }
//
    gst_buffer_ref(filter->b1);
    g_mutex_unlock(filter->mutex);
    cvReleaseImageHeader(&process);
    //sleep(10);
    //g_print ("PIPTRANS Pushing now!!!!!!.\n");
    GstFlowReturn ret=gst_pad_push (filter->srcpad, filter->b1);
    if (ret!=GST_FLOW_OK){
	g_print ("Flow not ok!!!!!!.\n");
        //sleep(10);
        g_print ("Flow not ok after sleep!!!!!!.\n");
    }
    //return gst_pad_push (filter->srcpad, filter->b1);
    return ret;	
  }
  else{
    g_mutex_unlock(filter->mutex);
    return GST_FLOW_OK;
  }
  /* just push out the incoming buffer without touching it */

}

static GstFlowReturn
gst_piptranstrans_chain2 (GstPad * pad, GstBuffer * buf){//synchronize one main image to many small and a matrix-list
  Gstpiptranstrans *filter;

  filter = GST_PIPTRANS (GST_OBJECT_PARENT (pad));
  //if (filter->sift) return gst_piptranstrans_chain (pad,buf);
  gboolean trigger,ready=FALSE,stop=FALSE;
  ready=FALSE;
  GList* temp;
  guint64 smallbuffers, delete, difference;//number of small pictures in big one, delete buffer numbers smaller than this
  gint i,h;
  g_print ("PIPTRANS trying to get lock!");
  g_mutex_lock(filter->mutex);
  delete=0;
  GstFlowReturn res=GST_FLOW_OK;
//Append
  if (pad== filter->sinkpad){
    filter->lbig=g_list_append(filter->lbig,buf);
    filter->offbignew=buf->offset;
    filter->offbigold=((GstBuffer*)filter->lbig->data)->offset;
    g_print ("PIPTRANS Main Sink pad. Number %llu Old %llu Length %d\n",filter->offbignew,filter->offbigold,g_list_length(filter->lbig));
    difference=filter->offbignew-filter->offbigold;
  }
  else if (pad == filter->sinksmall){
    filter->lsmall=g_list_append(filter->lsmall,buf);
    filter->offsmallnew=buf->offset;
    filter->offsmallold=((GstBuffer*)filter->lsmall->data)->offset;
    g_print ("PIPTRANS Small Sink pad. Number %llu Old %llu Length %d\n",filter->offsmallnew,filter->offsmallold,g_list_length(filter->lsmall));
    difference=filter->offsmallnew-filter->offsmallold;
  }
  else if (pad == filter->sinkpos){
    filter->lpos=g_list_append(filter->lpos,buf);
    filter->offposnew=buf->offset;
    filter->offposold=((GstBuffer*)filter->lpos->data)->offset;
    g_print ("PIPTRANS Position Sink pad. Number %llu Old %llu Length %d\n",filter->offposnew,filter->offposold,g_list_length(filter->lpos));
    difference=filter->offposnew-filter->offposold;
  }
//determine number of images to embed in big image

  if (filter->lpos) smallbuffers=(guint64)((GstBuffer*)filter->lpos->data)->size/(guint64)sizeof(struct transmat);
  //g_print("Will try to embed %llu buffers",smallbuffers);
//check if set is ready
  if((!filter->trigger_pos)&&(filter->lbig)&&((smallbuffers==0)||(filter->lsmall))&&(filter->lpos)){
    temp=filter->lsmall;
    for (i=smallbuffers;((temp)&&(!stop)&&(i>0));){
      if (((GstBuffer*)filter->lbig->data)->offset!=((GstBuffer*)temp->data)->offset) stop=TRUE;
      temp=g_list_next(temp);
      i--;
    }
    if ((i==0)&&(!stop)) ready=TRUE;//sucessfully passed
  }
  else if ((filter->lbig)&&((smallbuffers==0)||(filter->lsmall))&&(filter->lpos)){
    temp=filter->lsmall;
    //if (((GstBuffer*)filter->lbig->data)->offset!=((GstBuffer*)filter->lsmall->data)->offset) stop=TRUE;
    if (((GstBuffer*)filter->lbig->data)->offset!=((GstBuffer*)filter->lpos->data)->offset) stop=TRUE;
    for (i=smallbuffers;((temp)&&(!stop)&&(i>0));){
      if (((GstBuffer*)filter->lbig->data)->offset!=((GstBuffer*)temp->data)->offset) stop=TRUE;
      temp=g_list_next(temp);
      i--;
    }
    if ((i==0)&&(!stop)) ready=TRUE;//sucessfully passed
  }
//send and clear deprecated
  if (ready){
    g_print("ready, sending number %llu and there should be %llu buffers embedded\n",filter->offbigold,smallbuffers);
    filter->b1=(GstBuffer*)filter->lbig->data;
    temp=filter->lsmall;
    //create cv header
    IplImage *process;
    filter->imagesize.height=filter->h;
    filter->imagesize.width=filter->w;
    process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,(filter->bpp/8));
    process->imageData=(char*)filter->b1->data;  
    //embed each buffer

    for(i=0;i<smallbuffers;i++){

	//making small image
      filter->b2=(GstBuffer*)temp->data;
      gint widthsmall,heightsmall;
      GstStructure *structure=gst_caps_get_structure(filter->b2->caps,0);
      if (!gst_structure_get_int(structure,"width",&widthsmall)) return GST_FLOW_ERROR;
      if (!gst_structure_get_int(structure,"height",&heightsmall)) return GST_FLOW_ERROR;

      IplImage *currentsmall;
      filter->smallsize.height=heightsmall;
      filter->smallsize.width=widthsmall;
      currentsmall=cvCreateImageHeader(filter->smallsize, IPL_DEPTH_8U,(filter->bpp/8));
      currentsmall->imageData=(char*)filter->b2->data;  
	//making matrix


      filter->b3=(GstBuffer*)filter->lpos->data;

      //CvMat M = cvMat(3,3,CV_64FC1,((struct transmat*)(filter->b3->data))[i]);
      CvMat *M = cvCreateMat(3,3,CV_64FC1);
      cvmSet(M,0,0,((struct transmat*)(filter->b3->data))[i].a00);
      cvmSet(M,0,1,((struct transmat*)(filter->b3->data))[i].a01);
      cvmSet(M,0,2,((struct transmat*)(filter->b3->data))[i].a02);

      cvmSet(M,1,0,((struct transmat*)(filter->b3->data))[i].a10);
      cvmSet(M,1,1,((struct transmat*)(filter->b3->data))[i].a11);
      cvmSet(M,1,2,((struct transmat*)(filter->b3->data))[i].a12);

      cvmSet(M,2,0,((struct transmat*)(filter->b3->data))[i].a20);
      cvmSet(M,2,1,((struct transmat*)(filter->b3->data))[i].a21);
      cvmSet(M,2,2,((struct transmat*)(filter->b3->data))[i].a22);



	g_print("Warping\n");
      //cvWarpPerspective(currentsmall , process, M, 
	//		   CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS,
	//		   cvScalarAll( 0 ) );
/*
      cvWarpPerspective(currentsmall , process, M, 
			   CV_INTER_LINEAR,
			   cvScalarAll( 0 ) );

*/
      if (filter->diff){//difference 
        IplImage *xformed,*result;
        xformed = cvCreateImage( cvGetSize( process ), IPL_DEPTH_8U, 4 );
        result = cvCreateImage( cvGetSize( process ), IPL_DEPTH_8U, 4 );
	cvWarpPerspective(currentsmall , xformed, M, 
			   CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS,
			   cvScalarAll( 0 ) );
 	g_print("Making Diff\n");  
        double alpha=-filter->difff;
        double beta=filter->difff;
        //double alpha=1.0;
        //double beta=1.0;
        double gamma=128;
        cvAddWeighted(process,alpha,xformed ,beta,gamma, result); 
        int i=0;
	for (i=0;i<result->width*result->height;i++){
          //((unsigned char*)result->imageData)[4*i+2]=127;
	  //if (((unsigned char*)currentsmall->imageData)[4*i+3]!=255){
	  //  g_print("Erasing, because %d\n",(int)((unsigned char*)currentsmall->imageData)[4*i+3]);  
	  //  ((unsigned char*)result->imageData)[4*i+0]=0;
	  //  ((unsigned char*)result->imageData)[4*i+1]=0;
	  //  ((unsigned char*)result->imageData)[4*i+2]=0;
	  //}
	}
        //cvAbsDiff( process, xformed, result); 
 	g_print("Copy Back\n"); 
        cvCopy(result,process,0);
 	g_print("Finished Copy Back, releasing\n"); 
        cvReleaseImage(&xformed);
        cvReleaseImage(&result); 
      }


      else if (0){
        IplImage *xformed,*result;
        xformed = cvCreateImage( cvGetSize( process ), IPL_DEPTH_8U, 4 );
        result = cvCreateImage( cvGetSize( process ), IPL_DEPTH_8U, 4 );
	cvWarpPerspective(currentsmall , xformed, M, 
			   CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS,
			   cvScalarAll( 0 ) );
 	g_print("Making Diff\n");     
        cvAbsDiff( process, xformed, result); 
 	g_print("Copy Back\n"); 
        cvCopy(result,process,0);
 	g_print("Finished Copy Back, releasing\n"); 
        cvReleaseImage(&xformed);
        cvReleaseImage(&result);     
      }
      else if (filter->avg){
        IplImage *xformed,*result;
        xformed = cvCreateImage( cvGetSize( process ), IPL_DEPTH_8U, 4 );
        result = cvCreateImage( cvGetSize( process ), IPL_DEPTH_8U, 4 );
	cvWarpPerspective(currentsmall , xformed, M, 
			   CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS,
			   cvScalarAll( 0 ) );
 	g_print("Making Diff\n");  
        double alpha=0.5;
        double beta=0.5;
        //double alpha=1.0;
        //double beta=1.0;
        double gamma=0;
        cvAddWeighted(process,alpha,xformed ,beta,gamma, result); 
        //cvAbsDiff( process, xformed, result); 
 	g_print("Copy Back\n"); 
        cvCopy(result,process,0);
 	g_print("Finished Copy Back, releasing\n"); 
        cvReleaseImage(&xformed);
        cvReleaseImage(&result);     
      }
      else if (filter->justtrans){
        IplImage *xformed,*result;
        xformed = cvCreateImage( cvGetSize( process ), IPL_DEPTH_8U, 4 );
	cvWarpPerspective(currentsmall , xformed, M, 
			   CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS,
			   cvScalarAll( 0 ) );
 	g_print("Copy Back\n"); 
        cvCopy(xformed,process,0);
 	g_print("Finished Copy Back, releasing\n"); 
        cvReleaseImage(&xformed);
	}
      else{
        cvWarpPerspective(currentsmall , process, M, 
			   CV_INTER_LINEAR,
			   cvScalarAll( 0 ) );
      }
      cvReleaseMat(&M);
      cvReleaseImageHeader(&currentsmall);
      temp=g_list_next(temp);
    }
    cvReleaseImageHeader(&process);
    gst_buffer_ref((GstBuffer*)(filter->lbig->data));
    g_print("Gstpiptranstrans: Now pushing this buffer number %d !!\n",filter->offbigold);
    res=gst_pad_push (filter->srcpad, (GstBuffer*)(filter->lbig->data));
    delete=filter->offbigold;    
  }
//else determine & clear deprecated
  
    //determine delete threshold:

   //no matching frames has been found
   //if oldest main frame is older than all other frames
   if ((filter->offsmallnew>filter->offbigold)&&(filter->offposnew>filter->offbigold)&&(filter->offbigold>delete))delete=filter->offbigold;    //if position is older than newest, check if this is even smaller
   if ((filter->offbignew>filter->offposold)&&(filter->offsmallnew>filter->offposold)&&(filter->offposold>delete))delete=filter->offposold;
    //if small is deprecated..
   if ((filter->offbignew>filter->offsmallold)&&(filter->offposnew>filter->offsmallold)&&(filter->offsmallold>delete))delete=filter->offsmallold;
   if (delete==0){
	delete=filter->offbignew;
	if (filter->offsmallnew>delete)delete=filter->offsmallnew;
	if (filter->offposnew>delete)delete=filter->offposnew;
	if (delete>20)delete=delete-20;//for security reasons we delete everything thats older than 20
   }
   g_print("Will kick everything %llu or lower\n",delete);
   temp=filter->lbig;
   while ((temp)&&(filter->trigger)){
       if (((GstBuffer*)temp->data)->offset<=delete){
	gst_buffer_unref((GstBuffer*)temp->data);
	filter->lbig=g_list_remove(filter->lbig,temp->data);
	temp=filter->lbig;
       }
       else temp=g_list_next(temp);
    }
   temp=filter->lsmall;
   while ((temp)&&(filter->trigger_small)){
       if (((GstBuffer*)temp->data)->offset<=delete){
	gst_buffer_unref((GstBuffer*)temp->data);
	filter->lsmall=g_list_remove(filter->lsmall,temp->data);
	temp=filter->lsmall;
	}
       else temp=g_list_next(temp);
    }
   temp=filter->lpos;
   while ((temp)&&(filter->trigger_pos)){
       if (((GstBuffer*)temp->data)->offset<=delete){
	gst_buffer_unref((GstBuffer*)temp->data);
	filter->lpos=g_list_remove(filter->lpos,temp->data);
	temp=filter->lpos;
       }
       else temp=g_list_next(temp);
   }
  if (filter->lpos) filter->offposold=((GstBuffer*)filter->lpos->data)->offset; else filter->offposold=0;
  if (filter->lsmall) filter->offsmallold=((GstBuffer*)filter->lsmall->data)->offset; else filter->offsmallold=0;
  if (filter->lbig) filter->offbigold=((GstBuffer*)filter->lbig->data)->offset; else filter->offbigold=0;

  g_print ("After deleting Big %llu Small %llu Pos %llu\n",filter->offbigold,filter->offsmallold ,filter->offposold);

  g_mutex_unlock(filter->mutex);
  if (difference>filter->valve_buf) usleep(filter->penalty);
  return res;

}



static gboolean
gst_piptranstrans_event (GstPad  *pad, GstEvent * event)
{
  Gstpiptranstrans *filter;
  gboolean ret=TRUE;
  GstPad *otherpad;

  filter = GST_PIPTRANS (gst_pad_get_parent (pad));
  if (pad==filter->srcpad) return gst_pad_push_event (filter->sinkpad, event);
  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_EOS:
      if (pad == filter->sinkpad) { filter->stopped1=TRUE; g_print("PIPTRANS Event EOS from normal sinkpad!\n");}
      if (pad == filter->sinksmall){ filter->stopped2=TRUE;g_print("PIPTRANS Event EOS from small sinkpad!\n");}
      if (pad == filter->sinkpos) { filter->stopped3=TRUE;g_print("PIPTRANS Event EOS from position sinkpad!\n");}
      if ((filter->stopped1)&&(filter->stopped2)&&(filter->stopped3)&&(filter->fweos)) ret = gst_pad_push_event (filter->srcpad, event);
      //g_print("PIPTRANS Event EOS!\n");
      break;
    //case GST_EVENT_NAVIGATION:
    //  ret = gst_pad_push_event (filter->sinkpad, event);
    //  break;
    default:
      ret = gst_pad_push_event (filter->srcpad, event);
      break;
  }
  
  gst_object_unref (filter);
  return ret;
}

static GstStateChangeReturn
gst_piptranstrans_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstpiptranstrans *filter;
  filter = GST_PIPTRANS (element);
  //g_mutex_lock(filter->mutex);  
  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      filter->stopped1=FALSE;
      filter->stopped2=FALSE;
      filter->stopped3=FALSE;
      filter->offbigold=filter->offsmallold=filter->offposold=filter->offbignew=filter->offsmallnew=filter->offposnew=0;
      g_list_foreach(filter->lbig,(GFunc)gst_mini_object_unref,NULL);//bug gst_buffer_unref will not compile
      g_list_foreach(filter->lsmall,(GFunc)gst_mini_object_unref,NULL);
      g_list_foreach(filter->lpos,(GFunc)gst_mini_object_unref,NULL);
      g_list_free(filter->lbig);
      g_list_free(filter->lsmall);
      g_list_free(filter->lpos);
      filter->lbig=filter->lsmall=filter->lpos=NULL;
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      break;
    default:
      break;
  }
 //g_mutex_unlock(filter->mutex);
 return ret;
}



/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
piptrans_init (GstPlugin * piptrans)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template piptrans' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_piptranstrans_debug, "piptrans",
      0, "Template piptrans");

  return gst_element_register (piptrans, "piptrans", GST_RANK_NONE,
      GST_TYPE_PIPTRANS);
}

/* gstreamer looks for this structure to register piptranss
 *
 * exchange the string 'Template piptrans' with your piptrans description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "piptrans",
    "Template piptrans",
    piptrans_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
