

/*************    Attention    *****************

  **********************************************
  **********************************************
  **********************************************

* Modified by kunshan on Aug. 07, 2012.
* 1. Add a blind zone in horizontal area, PTU
*    can never access into this area.
* 2. rewrite relative functions.
* 3. just search "kunshan" to find the
     modification.

**********************************************
 **********************************************
***********************************************/




/**********************************************
* Header file for PTU YS3051 control
* Main functions:
*   PTU position control
*     int HorizontalRotate(int fd,double degree)
*     int VerticalRotate(int fd,double degree)
*   PTU position consult
*     int HorizontalConsult(int fd)
*     double HorizontalReceive(int fd)
*     int VerticalConsult(int fd)
*     double VerticalReceive(int fd)
* Any position is the absolute value in degree
***********************************************/
#ifndef _PTU_YS3051_H
#define _PTU_YS3051_H

///////// Important Macro definition /////////////
#define PTU_ADDR            0x01
#define DEVNAME             "/dev/ttyUSB0"
//#define DEVNAME             "/dev/ttyS0"
#define BAUDRATE            2400
///////// Important Macro definition END /////////

///////// Header file definition /////////////////
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>
///////// Header file definition END /////////////

///////// Regular Macro definition ///////////////
#define TRUE 1
#define FALSE 0

//////////  define the blind zone ///////////////
//////////  from 160 degree to 200 degree, this area can not be accessed into ////////////
#define minBlindZone 160
#define maxBlindzone 200

int  nspeed_arr[]={B38400,B19200,B9600,B4800,B2400,B1200,B300,B38400,B19200,B9600,B4800,B2400,B1200,B300,};
int  nname_arr[]={38400,19200,9600,4800,2400,1200,300,38400,19200,9600,4800,2400,1200,300,};
///////// Regular Macro definition END ///////////

///////// Serial Communication Functions /////////
// Serial Port Speed Configuration
void  nSetSpeed(int fd, int speed)
{
    int i;
    int status;
    struct termios Opt;
    tcgetattr(fd,&Opt);
    for (i=0;i<sizeof(nspeed_arr)/sizeof(int);i++)
    {
        if (speed==nname_arr[i])
        {
            tcflush(fd,TCIOFLUSH);
            cfsetispeed(&Opt,nspeed_arr[i]);
            cfsetospeed(&Opt,nspeed_arr[i]);
            status=tcsetattr(fd,TCSANOW,&Opt);
            if (status!=0)
                printf("set_speed [Failue]\n");
            return ;
        }
        tcflush(fd,TCIOFLUSH);
    }
}
// Serial Port Configuration
int  nSetParity(int fd, int databits, int stopbits, int parity)
{
    struct termios options;
    if (tcgetattr(fd,&options)!=0)
    {
        printf("set_Parity [Failue]\n");
        return (FALSE);
    }
    options.c_cflag&=~CSIZE;
    switch  (databits)
    {
        case   7 :
            options.c_cflag  |=  CS7;
        break ;
        case   8 :
            options.c_cflag  |=  CS8;
        break ;
        default :
            fprintf(stderr, "Unsupported data size\n");
        return  (FALSE);
    }
    switch  (parity)
    {
        case   'n' :
        case   'N' :
            options.c_cflag  &=   ~ PARENB;     /**/ /*  Clear parity enable  */
            options.c_iflag  &=   ~ INPCK;      /**/ /*  Enable parity checking  */
        break ;
        case   'o' :
        case   'O' :
            options.c_cflag  |=  (PARODD  |  PARENB);
            options.c_iflag  |=  INPCK;               /**/ /*  Disnable parity checking  */
        break ;
        case   'e' :
        case   'E' :
            options.c_cflag  |=  PARENB;       /**/ /*  Enable parity  */
            options.c_cflag  &=   ~ PARODD;
            options.c_iflag  |=  INPCK;        /**/ /*  Disnable parity checking  */
        break ;
        case   'S' :
        case   's' :   /**/ /* as no parity */
            options.c_cflag  &=   ~ PARENB;
            options.c_cflag  &=   ~ CSTOPB;
        break ;
        default :
            fprintf(stderr, "Unsupported parity\n");
            return  (FALSE);
    }
    switch  (stopbits)
    {
        case   1 :
            options.c_cflag  &=   ~ CSTOPB;
        break ;
        case   2 :
            options.c_cflag  |=  CSTOPB;
        break ;
        default :
            fprintf(stderr, "Unsupported stop bits\n");
        return  (FALSE);
    }

    if  (parity  !=   'n' || parity !='N')
        options.c_iflag  |=  INPCK;
    options.c_cc[VTIME]  =   150 ;  //  15 seconds
    options.c_cc[VMIN]  =   0 ;

    tcflush(fd,TCIFLUSH);  /**/ /*  Update the options and do it NOW  */
    if  (tcsetattr(fd,TCSANOW, & options)  !=   0 )
    {
        perror( "SetupSerial 3");
        return  (FALSE);
    }
    return  (TRUE);
}
// Open Serial Port
int nOpenDev(char *Dev)
{
    int  fd=open(Dev,O_RDWR);   // | O_NOCTTY | O_NDELAY

    if (-1==fd)
    {
        return -1;
    }
    else
    {
        return fd;
    }
}
// Initialize Serial Port
int Initialization_SerialPort()
{
    int fd=OpenDev(DEVNAME);
    if (fd>0)
    {
        printf("Serial Port '%s' is opened\n",DEVNAME);
        SetSpeed(fd,BAUDRATE);
    }
    else
    {
        printf("Can't Open Serial Port '%s'\n",DEVNAME);
        return -1;
    }
    if (SetParity(fd,8,1,'n')==-1)
    {
        printf("Set Parity Error\n");
        return -1 ;
    }
    return fd;
}
///////// Serial Communication Functions END /////

///////// Regular Functions //////////////////////
// Make a delay
void ndo_delay(long msec)
{
  struct timespec ts, rem;
  int retc;

  ts.tv_sec = msec / 1000;
  ts.tv_nsec = (msec % 1000) * 1000000;
  do
  {
    retc = nanosleep(&ts, &rem);
    ts = rem;
  } while (retc == -1);
}
// Make Check Sum
inline int nCheckSum(char *strCommOut)
{
    int sum=0;
    int i;
    for (i=1;i<6;i++)
    {
        sum+=strCommOut[i];
    }
    if (sum>0xFF)
        strCommOut[6]=sum%0x100;
    else
        strCommOut[6]=sum;
    return sum;
}
///////// Regular Functions END //////////////////

///////// PTU Control ////////////////////////////
// Make Sure Degree between 0-360
double ValidDegree(double degree)
{
    double r=degree;
    while (r>360) r-=360;
    while (r<0) r+=360;
            
    return r;
}

//************************** added by kunshan ************************************
// Make sure target horizontoal degree is not in blind zone.
// if it in, the target degree will be set to the boundry of blind zone
double ValidHorizontoalDegree(double degree)
{
    double s=degree;
    double TDegree;
    TDegree=ValidDegree(s);

    if ( TDegree >= minBlindZone && TDegree <= 180 )
	{
        TDegree = minBlindZone;
	} 

    else if ( TDegree > 180 && TDegree <= maxBlindzone )
	{
        TDegree = maxBlindzone;
	}
	return TDegree;	
}


//********************************* added by kunshan *******************************
// PTU Horizontoal Movement Control
// fd indicates the device index
// degree ranges from 0 to 360
// precision at 0.1 is permitted but at 1 is safer
// 0 indicates the left boundary and 359.99 indicates the other boundary
int HorizontalRotateCoreCode(int fd,double degree)
{
    char strCommOut[7];
    int ideg;
    int t;
    ideg=(int)floor(ValidHorizontoalDegree(degree)*100);
    strCommOut[0]=0xFF;
    strCommOut[1]=PTU_ADDR;
    strCommOut[2]=0x0;
    strCommOut[3]=0x4B;
    strCommOut[4]=((ideg&0x0FF00)>>8);
    strCommOut[5]=ideg&0xFF;
    CheckSum(strCommOut);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}




// PTU Vertical Control
// fd indicates the device index
// degree ranges from -40 to 90
// precision at 0.1 is permitted but at 1 is safer
// -40 indicates the up boundary and 90 indicates the low boundary
// So we set the degree negative for tilting up, or positive for tilting down
int VerticalRotate(int fd,double degree)
{
    char strCommOut[7];
    int ideg;
    int t;
    ideg=(int)floor(ValidDegree(degree)*100);
    strCommOut[0]=0xFF;
    strCommOut[1]=PTU_ADDR;
    strCommOut[2]=0x0;
    strCommOut[3]=0x4D;
    strCommOut[4]=((ideg&0x0FF00)>>8);
    strCommOut[5]=ideg&0xFF;
    CheckSum(strCommOut);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}
// PTU Moving Speed Configuration
// direction: 1-up, 2-down, 3-left, 4-right
// spd ranges at 0-100, 0 is the lowest speed, and 100 is the highest speed
int PTU_Speed(int fd,int direction,int spd)
{
    char strCommOut[7];
    int t;
    strCommOut[0]=0xFF;
    strCommOut[1]=PTU_ADDR;
    strCommOut[2]=0x0;
    switch (direction)
    {
    case 1: // UP
        strCommOut[3]=0x08;
        strCommOut[4]=0x0;
        strCommOut[5]=spd;
        break;
    case 2: // DOWN
        strCommOut[3]=0x10;
        strCommOut[4]=0x0;
        strCommOut[5]=spd;
        break;
    case 3: // LEFT
        strCommOut[3]=0x04;
        strCommOut[4]=spd;
        strCommOut[5]=0x0;
        break;
    case 4: // RIGHT
        strCommOut[3]=0x02;
        strCommOut[4]=spd;
        strCommOut[5]=0x0;
        break;
    }
    CheckSum(strCommOut);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}
// PTU Stop
int PTU_STOP(int fd)
{
    char strCommOut[7];
    int t;
    strCommOut[0]=0xFF;
    strCommOut[1]=PTU_ADDR;
    strCommOut[2]=0x0;
    strCommOut[3]=0x0;
    strCommOut[4]=0x0;
    strCommOut[5]=0x0;
    CheckSum(strCommOut);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}
///////// PTU Control END ////////////////////////

///////// PTU Consult ////////////////////////////
int serial_recv(int fd, void *buff, int len)
{
    int bleft;
    int bread;
    char *rptr = (char *)buff;

    bleft = len;
    printf("Read bytes from serial\n");
    while ((bread=read(fd,rptr,bleft))>=0 && (bleft-bread)>0)
    {
	printf("Read %d bytes from serial\n",bread);
        bleft-=bread;
        rptr+=bread;
    }
    if (bread>0)
        bleft-=bread;

    return (len - bleft);
}
// Ask PTU about its horizontal position
int HorizontalConsult(int fd)
{
    char strCommOut[7];
    int t;
    strCommOut[0]=0xFF;
    strCommOut[1]=PTU_ADDR;
    strCommOut[2]=0x0;
    strCommOut[3]=0x51;
    strCommOut[4]=0x0;
    strCommOut[5]=0x0;
    CheckSum(strCommOut);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}


// Read Horizontal Degree
// When it is successful, it returns a value in 0-359.99, or return 1001
//************************** modified by kunshan ************************************
//  The data we received not start with 0XFF, and must filter the extra content, if not,
//  the result of (p[4]<<8)+p[5] sometimes will be wrong.
double HorizontalReceive(int fd)
{
    unsigned char strCommIn[100];
	int i;
	unsigned char *p = strCommIn;
    unsigned char cs;
    double r;

    serial_recv(fd,strCommIn,9);

    //  to find the real message
    //  considering the protocol, the real message should start with 0XFF
	for (i=0; i<5; i++)
    {
		if ((char)*p == (char)0xFF)
        {
            break;
        }
		p++;
	}
	if (i == 5)
	{
		return 1001;
	}
	CheckSum((char*)p);
    cs=p[6];

    //  filter the extra content
	for ( i = 0; i < 7; i++ )
	{
		p[i] = p[i] & 0xFF;
	}

	printf(" %2x", p[0]);
	printf(" %2x", p[1]);
	printf(" %2x", p[2]);
	printf(" %2x", p[3]);
	printf(" %2x", p[4]);
	printf(" %2x", p[5]);
	printf(" %2x", p[6]);
	printf(" %2x", p[7]);
	printf(" %2x\n", p[8]);

    if (p[3]==0x59 && cs==p[6])
	{
		p[4] = p[4] & 0xFF;
		p[5] = p[5] & 0xFF;
		r=((p[4]<<8)+p[5])/100;
	}
	else
	{
		r=1001;
	}
    return r;
}

//************************** added by kunshan ************************************
// to convert range of 0---360 to the range of -180---180
double AngleConversion(double degree)
{
    if ( degree > 180 )
    {
        degree = degree - 360;
    }
    return degree;
}


// Ask PTU about its vertical position
int VerticalConsult(int fd)
{
    char strCommOut[7];
    int t;
    strCommOut[0]=0xFF;
    strCommOut[1]=PTU_ADDR;
    strCommOut[2]=0x0;
    strCommOut[3]=0x53;
    strCommOut[4]=0x0;
    strCommOut[5]=0x0;
    CheckSum(strCommOut);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}
// Read Vertical Degree
// When it is successful, it returns a value in -40-90, or return 1001
// So the return value is negative means PTU tilts up, or it tilts down
double VerticalReceive(int fd)
{
    char strCommIn[7];
    char cs;
    double r;
    serial_recv(fd,strCommIn,7);
    cs=strCommIn[6];
    CheckSum(strCommIn);
    if (strCommIn[3]==0x5B && cs==strCommIn[6])
    {
        r=((strCommIn[4]<<8)+strCommIn[5])/100;
        if (r>90)
            r-=360;
    }
    else
        r=-1;
    return r;
}
///////// PTU Consult END ////////////////////////

//************************** added by kunshan ************************************
// set preset position 1
int SetPrePos(int fd)
{
   char strCommCS[7];
   int t;
   strCommCS[0]=0xFF;
   strCommCS[1]=PTU_ADDR;
   strCommCS[2]=0x00;
   strCommCS[3]=0x03;
   strCommCS[4]=0x00;
   strCommCS[5]=0x01;   // the ID of preset position
   CheckSum(strCommCS);
   t=write(fd,strCommCS,sizeof(strCommCS));
   return t;
}

//************************** added by kunshan ************************************
// call preset position 1
int CallPrePos(int fd)
{
     char strCommFHCS[7];
     int t;
     strCommFHCS[0]=0xFF;
     strCommFHCS[1]=PTU_ADDR;
     strCommFHCS[2]=0x00;
     strCommFHCS[3]=0x07;
     strCommFHCS[4]=0x00;
     strCommFHCS[5]=0x01;   // the ID of preset position
     CheckSum(strCommFHCS);
     t=write(fd,strCommFHCS,sizeof(strCommFHCS));
     return t;
}


//************************** added by kunshan ************************************
//  return the distance between current angle and target angle
double DistanceToTarget( int fd, double targetDeg )
{
    double currentDeg;
    double distance;

    targetDeg = ValidHorizontoalDegree( targetDeg );
    targetDeg = AngleConversion( targetDeg );

    do_delay( 1000 );
    HorizontalConsult( fd );
    do_delay( 100 );
    currentDeg = HorizontalReceive( fd );
    currentDeg = AngleConversion( currentDeg );

    distance = fabs( targetDeg - currentDeg );

    printf ( " the distance is: %lf\n", distance );

    return distance;
}


//************************** modified by kunshan ************************************
//  As there is a blind zone that the PTU is forbidden to access into, we can not just simply rotate.
//  1. When the distance between target degree and current degree >= 180, should call the PTU back to preset
//  position, and then rotate to the target degree.
//  2. When the distance between target degree and current degree < 180, can directly rotate to the target degree.
//  3. After each rotate action, should wait untill the real postion of PTU is what we want.
int HorizontalRotate(int fd,double degree)
{
    int returnValue;

    double targetDeg;
    double distance;

    double temp1;
    double temp2;

    targetDeg = ValidHorizontoalDegree( degree );
    distance = DistanceToTarget( fd, targetDeg );

    if ( distance >= 180 )
    {
        CallPrePos(fd);
        while ( TRUE )
        {
            do_delay(1000);
            HorizontalConsult(fd);
            do_delay(100);
            temp1=HorizontalReceive(fd);

            if ( temp1 == 0 )
            {
                break;
            }
        }
        returnValue = HorizontalRotateCoreCode(fd,targetDeg);
        while ( TRUE )
        {
            do_delay(1000);
            HorizontalConsult(fd);
            do_delay(100);
            temp1=HorizontalReceive(fd);
            temp1 = AngleConversion( temp1 );

            temp2 = AngleConversion( targetDeg );

            printf ( "real degree is: %lf\n",  temp1  );

            if ( fabs( temp1 - temp2 ) <= 5 )
            {
                break;
            }
        }
    }
    else
    {
        returnValue = HorizontalRotateCoreCode(fd,targetDeg);
        while ( TRUE )
        {
            do_delay(1000);
            HorizontalConsult(fd);
            do_delay(100);
            temp1=HorizontalReceive(fd);
            temp1 = AngleConversion( temp1 );

            temp2 = AngleConversion( targetDeg );

            printf ( "real degree is: %lf\n",  temp1  );

            if ( fabs( temp1 - temp2 ) <= 5 )
            {
                break;
            }
        }
    }

    return returnValue;
}

int HorizontalRotateCode(int fd,double degree)
{
    char strCommOut[7];
    int ideg;
    int t;
    ideg=(int)floor(degree*100);
    strCommOut[0]=0xFF;
    strCommOut[1]=PTU_ADDR;
    strCommOut[2]=0x0;
    strCommOut[3]=0x4B;
    strCommOut[4]=((ideg&0x0FF00)>>8);
    strCommOut[5]=ideg&0xFF;
    CheckSum(strCommOut);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}

int Returnback(int fd)
{
    int returnValue;
    double temp1;
    double temp2;
    double temp3;
    int degree=120;
    do_delay(1000);
    printf ( "Doing HorizontalConsult\n" );
    HorizontalConsult(fd);
    printf ( "Done HorizontalConsult\n" );
    do_delay(100);
    temp1=HorizontalReceive(fd);
    printf ( "real degree is: %lf\n",  temp1  );
    temp1 = AngleConversion( temp1 + degree);

    if(temp1 < 0)
{
temp3 = temp1+360;
}
else
{
temp3 = temp1;
}
    returnValue = HorizontalRotateCode(fd,temp3);
    while (TRUE)
    {
            do_delay(1000);
            HorizontalConsult(fd);
            do_delay(100);
            temp2=HorizontalReceive(fd);
            temp2 = AngleConversion( temp2 );
            printf ( "real degree is: %lf\n",  temp2  );
            if ( fabs(temp1-temp2) < 2 )
           {
            break;  
           }
   }
   return 1;
}


//////////////////////////////////////////////////
//********************************* modifieded by kunshan *******************************
/****** sample main function *********************

int  main()
{
    int fd=Initialization_SerialPort();

    if ( fd>0 )
    {
        double ideg = 0;

        while( TRUE )
        {     
            printf("Next Degree:");
            scanf("%lf",&ideg);

            HorizontalRotate(fd, ideg);
        }
    }
}
****** sample main function END *****************/

#endif
