/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2012 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-ptucontrol
 *
 * FIXME:Describe ptucontrol here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! ptucontrol ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "gstptucontrol.h"
#include "chinaptu.h"
#include "arduinolib.h"
#include "PTU_YS3051.h"

GST_DEBUG_CATEGORY_STATIC (gst_ptucontrol_debug);
#define GST_CAT_DEFAULT gst_ptucontrol_debug

#define PITCHDEF -39.0
/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_VALVENAME,
  PROP_VALVEFOCUSNAME,
  PROP_PTUTYPE,
  PROP_FOCUS,
  PROP_ZOOM,
  PROP_DEV,
  PROP_AADR,
  PROP_YMIN,
  PROP_YMAX,
  PROP_PMIN,
  PROP_PMAX,
  PROP_RET,
  PROP_YOFF,
  PROP_POFF,
  PROP_INVY,
  PROP_INVP,
  PROP_DIA
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("camconfig")
    );

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("camconfig")
    );


GST_BOILERPLATE (Gstptucontrol, gst_ptucontrol, GstElement,
    GST_TYPE_ELEMENT);

static void gst_ptucontrol_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_ptucontrol_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_ptucontrol_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps* gst_ptucontrol_getcaps(GstPad * pad);
static GstFlowReturn gst_ptucontrol_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_ptucontrol_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "ptucontrol",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the ptucontrol's class */
static void
gst_ptucontrol_class_init (GstptucontrolClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_ptucontrol_set_property;
  gobject_class->get_property = gst_ptucontrol_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_VALVENAME,
      g_param_spec_string ("valvename", "Valvename", "Operate the valveelement with NAME in the pipeling (NULL=none).",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_VALVEFOCUSNAME,
      g_param_spec_string ("valvefocusname", "Valvefocusname", "Operate the valve for focus evaluation (NULL=none).",
          NULL, G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_PTUTYPE,
      g_param_spec_int ("ptutype", "Ptutype", "How will PTU be controlled, 0=USBtoSerial(nofocus),1=Arduino+PelcoD+Focus,2=Arduino only, focus plus LUT zoom. 3=YS3051 via USB, 4=Only simulation and text output, 5 ist YS3051 via USB and new nonblockung LUT Autofocus, 6= like 5 nowait, buggy, 7 is only PTU",0,7,5, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_FOCUS,
      g_param_spec_int ("focus", "Focus", "Current focus value.",
          12,120,50, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_ZOOM,
      g_param_spec_int ("zoom", "Zoom", "Current zoom value.",
          12,120,50, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DEV,
      g_param_spec_string ("device", "Device", "Device Name for Serial PTUs.",
          "/dev/ttyUSB0", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_AADR,
      g_param_spec_string ("arduino", "Arduino", "Network adress of arduino board",
          "192.168.0.19", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_YMIN,
      g_param_spec_int ("yawmin", "Yawmin", "Minimum Allowed value for YAW",-180,180,-180, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_YMAX,
      g_param_spec_int ("yawmax", "Yawmax", "Maximum Allowed value for YAW",-180,180,180, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_PMIN,
      g_param_spec_int ("pitchmin", "Pitchmin", "Minimum Allowed value for PITCH",-90,90,-90, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_PMAX,
      g_param_spec_int ("pitchmax", "Pitchmax", "Maximum Allowed value for PITCH",-90,90,90, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_RET,
      g_param_spec_boolean ("return", "Return", "Yaan PTU returns home ?",
          FALSE, G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_YOFF,
      g_param_spec_double ("yawoffset", "Yawoffset", "offset of YAW for PTU home position.",
          -180.0,180.0,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_POFF,
      g_param_spec_double ("pitchoffset", "Pitchoffset", "offset of Pitch for PTU home position.",
          -180.0,180.0,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_INVY,
      g_param_spec_boolean ("inverseyaw", "Inverseyaw", "Inverse yaw direction",
          FALSE, G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_INVP,
      g_param_spec_boolean ("inversepitch", "Inversepitch", "Inverse pitch direction",
          FALSE, G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_DIA,
      g_param_spec_double ("diameter", "Diameter", "Diameter of sensor.",
          0,G_MAXDOUBLE,6.126, G_PARAM_READWRITE));
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_ptucontrol_init (Gstptucontrol * filter,
    GstptucontrolClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_ptucontrol_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_ptucontrol_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_ptucontrol_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->state.yaw=0.0;
  filter->state.pitch=0.0;
  filter->state.active=FALSE;
  filter->needini=TRUE;
  filter->pitchmin=-90.0;
  filter->pitchmax=90;
  filter->yawmin=-180.0;
  filter->yawmax=180.0;
  filter->ptutype=3;
  filter->valvename=NULL;
  filter->valvefocusname=NULL;
  filter->focus=50;
  filter->zoom=50;
  filter->device=g_strdup("/dev/ttyUSB0");
  filter->arduinoaddr=g_strdup("134.100.13.161");
  filter->docc=FALSE;
  filter->yawoffset=0;
  filter->pitchoffset=0;
  filter->inverseyaw=FALSE;
  filter->inversepitch=FALSE;
  //filter->diameter=2*1.67*tan((3.14*113.2)/360.0);
  filter->diameter=6.126;
  filter->mutex=g_mutex_new();
  g_print("Set devicename to %s\n",filter->device);
}

static void
gst_ptucontrol_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstptucontrol *filter = GST_PTUCONTROL (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_VALVENAME:
      g_free (filter->valvename);
      filter->valvename = g_value_dup_string (value);
      break;
    case PROP_VALVEFOCUSNAME:
      g_free (filter->valvefocusname);
      filter->valvefocusname = g_value_dup_string (value);
      break;
    case PROP_PTUTYPE:
      if (filter->needini)filter->ptutype=g_value_get_int (value);
      break;
    case   PROP_FOCUS:
      g_mutex_lock(filter->mutex);
      filter->focus = g_value_get_int (value);
      if ((!filter->needini)&&(filter->ptutype))setFocusToMm(filter->focus);
      g_mutex_unlock(filter->mutex);
      break;
     case  PROP_ZOOM:
      g_mutex_lock(filter->mutex);
      filter->zoom = g_value_get_int (value);
      if ((!filter->needini)&&(filter->ptutype))setZoomToMm(filter->zoom);
      g_mutex_unlock(filter->mutex);
      break;
    case PROP_DEV:
      g_free (filter->device);
      filter->device = g_value_dup_string (value);
      break;
    case PROP_AADR:
      g_mutex_lock(filter->mutex);
      g_free (filter->arduinoaddr);
      filter->arduinoaddr = g_value_dup_string (value);
      g_mutex_unlock(filter->mutex);
      break;
    case PROP_YMIN:
      filter->yawmin=g_value_get_int (value);
    break;
     case PROP_YMAX:
      filter->yawmax=g_value_get_int (value);
      break;  
    case PROP_PMIN:
      filter->pitchmin=g_value_get_int (value);
      break;
     case PROP_PMAX:
      filter->pitchmax=g_value_get_int (value);
      break;  
    case PROP_RET:
      filter->docc = g_value_get_boolean (value);
      g_mutex_lock(filter->mutex);
      if (filter->docc){
       if (filter->fd!=0){
         Returnback(filter->fd);
         Returnback(filter->fd);  
         Returnback(filter->fd);  
         filter->docc=FALSE;
         }
       }
       g_mutex_unlock(filter->mutex);
      break;
     case PROP_YOFF:
      filter->yawoffset=g_value_get_double (value);
      break; 
     case PROP_POFF:
      filter->pitchoffset=g_value_get_double (value);
      break;
    case PROP_INVY:
      filter->inverseyaw = g_value_get_boolean (value);
      break;
    case PROP_INVP:
      filter->inversepitch = g_value_get_boolean (value);
      break;
     case PROP_DIA:
      filter->diameter=g_value_get_double (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ptucontrol_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstptucontrol *filter = GST_PTUCONTROL (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_VALVENAME:
      g_value_set_string (value, filter->valvename);
      break;
    case PROP_VALVEFOCUSNAME:
      g_value_set_string (value, filter->valvefocusname);
      break;
    case PROP_PTUTYPE:
      g_value_set_int (value,filter->ptutype);
      break;
    case   PROP_FOCUS:
      g_value_set_int (value,filter->focus);
      break;
    case  PROP_ZOOM:
      g_value_set_int (value,filter->zoom);
      break;
    case PROP_DEV:
      g_value_set_string (value, filter->device);
      break;
    case PROP_AADR:
      g_value_set_string (value, filter->arduinoaddr);
      break;
    case PROP_YMIN:
      g_value_set_int (value,filter->yawmin);
    break;
     case PROP_YMAX:
      g_value_set_int (value,filter->yawmax);
      break;  
    case PROP_PMIN:
      g_value_set_int (value,filter->pitchmin);
      break;
     case PROP_PMAX:
      g_value_set_int (value,filter->pitchmax);
      break; 
     case PROP_YOFF:
      g_value_set_double (value,filter->yawoffset);
      break; 
     case PROP_POFF:
      g_value_set_double (value,filter->pitchoffset);
      break; 
     case PROP_INVY:
      g_value_set_boolean (value,filter->inverseyaw);
      break; 
     case PROP_INVP:
      g_value_set_boolean (value,filter->inversepitch);
      break; 
     case PROP_DIA:
      g_value_set_double (value,filter->diameter);
      break; 
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static GstCaps* gst_ptucontrol_getcaps(GstPad * pad){
  //Gstptucontrol *filter;
  //filter = GST_PTUCONTROL (GST_OBJECT_PARENT (pad));
  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));  
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_ptucontrol_set_caps (GstPad * pad, GstCaps * caps)
{
return TRUE;
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_ptucontrol_chain (GstPad * pad, GstBuffer * buf)
{
  Gstptucontrol *filter;
  filter = GST_PTUCONTROL (GST_OBJECT_PARENT (pad)); 
  //printf("PTUCONTROL: Locking Mutex %p\n",filter->mutex);
  g_mutex_lock(filter->mutex); 
  struct camconfig* currentbuf;
  float focus=-1;
  float d=filter->diameter;

  printf("PTUCONTROL: Buffer\n");
  int limit=0; //whether request hits limit
  //if (filter->ccbuf)gst_buffer_unref(filter->ccbuf);
  //filter->ccbuf=buf;
  currentbuf=(struct camconfig*)buf->data;//
  //control the valve
  if (filter->valvename){
    GstElement* valve=gst_bin_get_by_name_recurse_up (GST_ELEMENT_PARENT(filter),filter->valvename);
    if ((valve)&&(!currentbuf->active))g_object_set(G_OBJECT(valve),"drop",TRUE,NULL);
    else if ((valve)&&(currentbuf->active))g_object_set(G_OBJECT(valve),"drop",FALSE,NULL);
    if (valve)gst_object_unref(GST_OBJECT(valve));
  }
  //now control the PTU
  printf("PTUCONTROL: Buffer yaw %f pitch %f, active: %d\n",currentbuf->yaw,currentbuf->pitch,currentbuf->active);

  if ((filter->fd)&&(currentbuf->relative==TRUE) &&(currentbuf->yaw>359.99)){
    if ((filter->ptutype==5)||(filter->ptutype==6)){
       printf("PTUCONTROL: doing 360 degree cable unwinding turn\n");
       Returnback(filter->fd);
       Returnback(filter->fd);  
       Returnback(filter->fd);  
    }
    g_mutex_unlock(filter->mutex); 
    return GST_FLOW_OK;	
  }

  float yaw,pitch;
  float fov;
  if (currentbuf->yaw<filter->yawmin){ 
    currentbuf->yaw=filter->yawmin;
    limit=TRUE;
  }
  if (currentbuf->yaw>filter->yawmax){
    currentbuf->yaw=filter->yawmax;
    limit=TRUE;
  }
  if (currentbuf->pitch<filter->pitchmin){
    currentbuf->pitch=filter->pitchmin;
    limit=TRUE;
  }
  if (currentbuf->pitch>filter->pitchmax){
    currentbuf->pitch=filter->pitchmax;
    limit=TRUE;
  }

  if (filter->inverseyaw)yaw=-currentbuf->yaw;
  else yaw=-currentbuf->yaw;
  if (filter->inversepitch)pitch=-currentbuf->pitch;
  else pitch=currentbuf->pitch;
  printf("PTUCONTROL: Buffer after limits: yaw %f pitch %f, active: %d\n",currentbuf->yaw,currentbuf->pitch,currentbuf->active);
  GstFlowReturn ret;
//openPTU
 if (filter->ptutype==0){
   if (filter->needini){
     printf("PTUCONTROL: now initializing\n");
     filter->fd  =  OpenDev(filter->device);
     if  (filter->fd > 0 )
     {
         printf("PTUCONTROL: Serial Port '%s' is opened\n ",filter->device);
         SetSpeed(filter->fd, 2400 );
     }
     else
     {
        printf( "PTUCONTROL:  Can't Open Serial Port '%s'\n " ,filter->device);
	return GST_FLOW_ERROR;
     }
     if  (SetParity(filter->fd,8,1,'n') == -1 )
     {
        printf( "PTUCONTROL:  Set Parity Error\n " );
	return GST_FLOW_ERROR;
     }
     filter->needini=FALSE;
     Initialization(filter->fd);
     filter->state.yaw=0.0;
     filter->state.pitch=PITCHDEF;
   }
 }
 if (filter->ptutype==1){
   if (filter->needini){
     createConnectionToArduino(filter->arduinoaddr, 8888);
     Initialization_A();
     filter->state.yaw=0.0;
     filter->state.pitch=PITCHDEF;
     filter->needini=FALSE;
   }
 }
 if (filter->ptutype==2){//just focus
   if (filter->needini){
     printf("PTUCONTROL Init Arduino in mode 5.\n");
     createConnectionToArduino(filter->arduinoaddr, 8888);
     printf("PTUCONTROL Connection Created mode 5.\n");
     //Initialization(12,120);//misst und liest alle Werte des
                               //Kameraobjektives ein und speichert diese in
                               //eine File
     Initialization_Arduino();//liest die Werte aus der File
     printf("PTUCONTROL Arduino Initialiyed mode 5.\n");
     filter->needini=FALSE;
   }
 }
 if (filter->ptutype==3){//just focus
   if (filter->needini){
     createConnectionToArduino(filter->arduinoaddr, 8888);
     //Initialization(12,120);//misst und liest alle Werte des
                               //Kameraobjektives ein und speichert diese in
                               //eine File
     Initialization_Arduino();//liest die Werte aus der File

     filter->needini=FALSE;
     filter->fd=Initialization_SerialPort(filter->device);
   }
 }
 if (filter->ptutype==5){//just focus
   if (filter->needini){
     printf("PTUCONTROL Init Arduino in mode 5.\n");
     createConnectionToArduino(filter->arduinoaddr, 8888);
     printf("PTUCONTROL Connection Created mode 5.\n");
     //Initialization(12,120);//misst und liest alle Werte des
                               //Kameraobjektives ein und speichert diese in
                               //eine File
     Initialization_Arduino();//liest die Werte aus der File
     printf("PTUCONTROL Arduino Initialiyed mode 5.\n");
     filter->needini=FALSE;
     filter->fd=Initialization_SerialPort(filter->device);
   }
 } 
 if (filter->ptutype==6){//just focus
   if (filter->needini){
     createConnectionToArduino(filter->arduinoaddr, 8888);
     //Initialization(12,120);//misst und liest alle Werte des
                               //Kameraobjektives ein und speichert diese in
                               //eine File
     Initialization_Arduino();//liest die Werte aus der File

     filter->needini=FALSE;
     filter->fd=Initialization_SerialPort(filter->device);
   }
 }
  if (filter->ptutype==7){//just PTU
   if (filter->needini){
     filter->needini=FALSE;
     filter->fd=Initialization_SerialPort(filter->device);
   }
 } 

   if ((!currentbuf->active)&&(filter->state.active)){
     printf("PTUCONTROL: Set Idle\n");
     if (filter->ptutype==0)Initialization(filter->fd);
     else if (filter->ptutype==1)Initialization_A();	
     filter->state.yaw=0.0;
     filter->state.pitch=PITCHDEF;
   }


  else if (currentbuf->active){


    if (filter->ptutype==0){
    yaw=yaw-filter->state.yaw;
    pitch=pitch-filter->state.pitch;
    //eliminate fudge
    pitch=pitch/1.05;
    yaw=yaw/0.89;
    printf("PTUCONTROL: Set Value relative yaw %f, pitch %f\n",yaw,pitch);
    //got to absolute position 
      if (yaw<0)ContinuousMoveWithRelatedDegree(filter->fd,CLEFT,abs(yaw));
      else ContinuousMoveWithRelatedDegree(filter->fd,CRIGHT,abs(yaw));
      if (pitch<0)ContinuousMoveWithRelatedDegree(filter->fd,CDOWN,abs(pitch));
      else ContinuousMoveWithRelatedDegree(filter->fd,CUP,abs(pitch));  
    }
//---------------------------------test--------------------------------------------
    else if (filter->ptutype==1){  
     if(yaw<0) {movePtu(LEFT,abs(yaw));} else {movePtu(RIGHT,abs(yaw));}
     if(pitch>0) {movePtu(UP,abs(pitch));} else {movePtu(DOWN,abs(pitch));}   
     fov=currentbuf->hfov;
     focus=((float)d/(2.0*(tan((3.14*fov)/360.0))));//now focus should hold the correct value - set it
     printf("PTUCONTROL: requested hfov %f, setting focus to %f, diameter of Sensor in mm %f\n",fov,focus,d);
    if (focus<12)focus=12;
    if (focus>120)focus=120;
    setZoomToMm((int)focus);
    GstElement* valvefocus=NULL;
    if (filter->valvefocusname)valvefocus=gst_bin_get_by_name_recurse_up (GST_ELEMENT_PARENT(filter),filter->valvefocusname);
    if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",FALSE,NULL);
     printf("PTUCONTROL: Calling Autofocus!\n");
     autoFocus();
     printf("PTUCONTROL: Autofocus returned!\n");
    if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",TRUE,NULL);
    if (valvefocus)gst_object_unref(GST_OBJECT(valvefocus));
    }

    else if (filter->ptutype==2){   
     if (filter->docc){
       filter->docc=FALSE; 
     }
     printf("PTUCONTROL: not starting mode 2\n");
     //HorizontalRotate(filter->fd,(yaw+filter->yawoffset));//we have 0 as center, PTU 180
     //VerticalRotate(filter->fd,(pitch+filter->pitchoffset));//PTU has negative for up
     fov=currentbuf->hfov;
     focus=((float)d/(2.0*(tan((3.14*fov)/360.0))));//now focus should hold the correct value - set it
     printf("PTUCONTROL: requested hfov %f, setting focus to %f, diameter of Sensor in mm %f\n",fov,focus,d);
    if (currentbuf->focus<0){
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      printf("PTUCONTROL: Setting Zoom!\n");
      setZoomToMm((int)focus);
      printf("PTUCONTROL: Survived Setting Zoom!\n");
      GstElement* valvefocus=NULL;
      //if (filter->valvefocusname)valvefocus=gst_bin_get_by_name_recurse_up (GST_ELEMENT_PARENT(filter),filter->valvefocusname);
      //if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",FALSE,NULL);
       printf("PTUCONTROL: Calling Predefined Autofocus!\n");
       int focusresult=setFocusFromLUT();
       printf("PTUCONTROL: Predefined Autofocus returned with %d!\n",focusresult);
      //if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",TRUE,NULL);
      //if (valvefocus)gst_object_unref(GST_OBJECT(valvefocus));
      }
    else{
      printf("PTUCONTROL: Calling Zoom before manual focus!\n");
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      setZoomToMm((int)focus);
      printf("PTUCONTROL: Calling manual focus!\n");
      int realfocus=(int)currentbuf->focus;
      if (realfocus<1)realfocus=1;
      if (realfocus>100)realfocus=100; 
      setFocusToMm(realfocus);
      printf("PTUCONTROL: Finished manual focus!\n");
      }

    }

    else if (filter->ptutype==3){  
     HorizontalRotate(filter->fd,(yaw+180));//we have 0 as center, PTU 180
     VerticalRotate(filter->fd,(-1*pitch));//PTU has negative for up
     fov=currentbuf->hfov;
     focus=((float)d/(2.0*(tan((3.14*fov)/360.0))));//now focus should hold the correct value - set it
     printf("PTUCONTROL: requested hfov %f, setting focus to %f, diameter of Sensor in mm %f\n",fov,focus,d);
    if (currentbuf->focus<0){
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      setZoomToMm((int)focus);
      GstElement* valvefocus=NULL;
      if (filter->valvefocusname)valvefocus=gst_bin_get_by_name_recurse_up (GST_ELEMENT_PARENT(filter),filter->valvefocusname);
      if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",FALSE,NULL);
       printf("PTUCONTROL: Calling Autofocus!\n");
       autoFocus();
       printf("PTUCONTROL: Autofocus returned!\n");
      if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",TRUE,NULL);
      if (valvefocus)gst_object_unref(GST_OBJECT(valvefocus));
      }
    else{
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      setZoomToMm((int)focus);
      int realfocus=(int)currentbuf->focus;
      if (realfocus<12)realfocus=12;
      if (realfocus>120)realfocus=120; 
      setFocusToMm(realfocus);
      }
    }
    else if (filter->ptutype==4){  //Dummy control
     fov=currentbuf->hfov;
     focus=((float)d/(2.0*(tan((3.14*fov)/360.0))));//now focus should hold the correct value - set it
     if (focus<12)focus=12;
     if (focus>120)focus=120;
     GstElement* valvefocus=NULL;
     if (filter->valvefocusname)valvefocus=gst_bin_get_by_name_recurse_up (GST_ELEMENT_PARENT(filter),filter->valvefocusname);
       if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",FALSE,NULL);
         printf("PTUCONTROL: Calling Dummy Autofocus!\n");
         sleep(2);
         printf("PTUCONTROL: Dummy Autofocus returned!\n");
         if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",TRUE,NULL);
     if (valvefocus)gst_object_unref(GST_OBJECT(valvefocus));
     printf("PTUCONTROL: Dummy Mode: Yaw=%f Pitch=%f requested hfov %f, setting focus to %f, diameter of Sensor in mm %f Manual focus value is %f\n",yaw,pitch,fov,focus,d,currentbuf->focus);
    }
    //YS3051 via USB + Arduino new version 
    else if (filter->ptutype==5){  
     if (filter->docc){
       Returnback(filter->fd);
       Returnback(filter->fd);  
       Returnback(filter->fd);  
       filter->docc=FALSE; 
     }
     printf("PTUCONTROL: starting mode 5 ptumotion\n");
     HorizontalRotate(filter->fd,(yaw+filter->yawoffset));//we have 0 as center, PTU 180
     VerticalRotate(filter->fd,(pitch+filter->pitchoffset));//PTU has negative for up
     fov=currentbuf->hfov;
     focus=((float)d/(2.0*(tan((3.14*fov)/360.0))));//now focus should hold the correct value - set it
     printf("PTUCONTROL: requested hfov %f, setting focus to %f, diameter of Sensor in mm %f\n",fov,focus,d);
    if (currentbuf->focus<0){
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      printf("PTUCONTROL: Setting Zoom!\n");
      setZoomToMm((int)focus);
      printf("PTUCONTROL: Survived Setting Zoom!\n");
      GstElement* valvefocus=NULL;
      //if (filter->valvefocusname)valvefocus=gst_bin_get_by_name_recurse_up (GST_ELEMENT_PARENT(filter),filter->valvefocusname);
      //if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",FALSE,NULL);
       printf("PTUCONTROL: Calling Predefined Autofocus!\n");
       int focusresult=setFocusFromLUT();
       printf("PTUCONTROL: Predefined Autofocus returned with %d!\n",focusresult);
      }
    else{
      printf("PTUCONTROL: Calling Zoom before manual focus!\n");
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      setZoomToMm((int)focus);
      printf("PTUCONTROL: Calling manual focus!\n");
      int realfocus=(int)currentbuf->focus;
      if (realfocus<1)realfocus=1;
      if (realfocus>100)realfocus=100; 
      setFocusToMm(realfocus);
      printf("PTUCONTROL: Finished manual focus!\n");
      }

    }
    else if (filter->ptutype==6){  
     if (filter->docc){
       Returnback(filter->fd);
       Returnback(filter->fd);  
       Returnback(filter->fd);  
       filter->docc=FALSE; 
     }
     HorizontalRotate(filter->fd,(yaw+filter->yawoffset));//we have 0 as center, PTU 180
     VerticalRotate(filter->fd,(pitch+filter->pitchoffset));//PTU has negative for up
     fov=currentbuf->hfov;
     focus=((float)d/(2.0*(tan((3.14*fov)/360.0))));//now focus should hold the correct value - set it
     printf("PTUCONTROL: requested hfov %f, setting focus to %f, diameter of Sensor in mm %f\n",fov,focus,d);
    int sleepdur=50000;
    usleep(sleepdur);
    if (currentbuf->focus<0){
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      printf("PTUCONTROL: Setting Zoom!\n");
      setZoomToMmNoWait((int)focus);
      usleep(sleepdur);
      printf("PTUCONTROL: Survived Setting Zoom!\n");
      GstElement* valvefocus=NULL;
      //if (filter->valvefocusname)valvefocus=gst_bin_get_by_name_recurse_up (GST_ELEMENT_PARENT(filter),filter->valvefocusname);
      //if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",FALSE,NULL);
       printf("PTUCONTROL: Calling Predefined Autofocus!\n");
       setFocusFromLUTNoWait((int)focus);
       usleep(sleepdur); 
      //setFocusFromLUT();
       printf("PTUCONTROL: Predefined Autofocus returned!\n");
      //if (valvefocus)g_object_set(G_OBJECT(valvefocus),"drop",TRUE,NULL);
      //if (valvefocus)gst_object_unref(GST_OBJECT(valvefocus));
      }
    else{
      printf("PTUCONTROL: Calling Zoom before manual focus!\n");
      if (focus<12)focus=12;
      if (focus>120)focus=120;
      setZoomToMmNoWait((int)focus);
      usleep(sleepdur); 
      printf("PTUCONTROL: Calling manual focus!\n");
      int realfocus=(int)currentbuf->focus;
      if (realfocus<1)realfocus=1;
      if (realfocus>100)realfocus=100; 
      setFocusToMmNoWait(realfocus);
      usleep(sleepdur); 
      printf("PTUCONTROL: Finished manual focus!\n");
      }

    }
    else if (filter->ptutype==7){  
     if (filter->docc){
       Returnback(filter->fd);
       Returnback(filter->fd);  
       Returnback(filter->fd);  
       filter->docc=FALSE; 
     }
     HorizontalRotate(filter->fd,(yaw+filter->yawoffset));//we have 0 as center, PTU 180
     VerticalRotate(filter->fd,(pitch+filter->pitchoffset));//PTU has negative for up
    }


  }
  printf("PTUCONTROL: Memcpy\n");
  //read out new position information, currentbuf contains it...maybe
  if (currentbuf->active){
    memcpy(&filter->state,currentbuf,sizeof(struct camconfig));//calculate the current state:)
    //if (focus>0)filter->state.focus=focus;
    //if (focus>0)filter->state.hfov=filter->state.hfov;
   if (focus>0)filter->state.hfov=2.0*((360.0*atan2(d,(2.0*focus)))/(2.0*3.14));
   filter->state.limit=limit;//write back whether limit is hit
  }
  
  printf("PTUCONTROL: Currentstate: yaw %f pitch %f, active: %d\n",filter->state.yaw,filter->state.pitch,currentbuf->active);
  memcpy(currentbuf,&filter->state,sizeof(struct camconfig));//write state back
  /* just push out the incoming buffer without touching it */
  printf("PTUCONTROL: Pushing\n");
  g_mutex_unlock(filter->mutex);
  return gst_pad_push (filter->srcpad, buf);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
ptucontrol_init (GstPlugin * ptucontrol)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template ptucontrol' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_ptucontrol_debug, "ptucontrol",
      0, "Template ptucontrol");

  return gst_element_register (ptucontrol, "ptucontrol", GST_RANK_NONE,
      GST_TYPE_PTUCONTROL);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstptucontrol"
#endif

/* gstreamer looks for this structure to register ptucontrols
 *
 * exchange the string 'Template ptucontrol' with your ptucontrol description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "ptucontrol",
    "Template ptucontrol",
    ptucontrol_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
