/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-rgb2gray
 *
 * FIXME:Describe rgb2gray here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! rgb2gray ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>

#include "gstrgb2gray.h"

GST_DEBUG_CATEGORY_STATIC (gst_rgb2gray_debug);
#define GST_CAT_DEFAULT gst_rgb2gray_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_R,
  PROP_G,
  PROP_B,
  PROP_OFFSET
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_xRGB)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_GRAY8)
    );

GST_BOILERPLATE (Gstrgb2gray, gst_rgb2gray, GstElement,
    GST_TYPE_ELEMENT);

static void gst_rgb2gray_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_rgb2gray_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_rgb2gray_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps * gst_rgb2gray_get_caps (GstPad *pad);
static GstFlowReturn gst_rgb2gray_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_rgb2gray_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "rgb2gray",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the rgb2gray's class */
static void
gst_rgb2gray_class_init (Gstrgb2grayClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_rgb2gray_set_property;
  gobject_class->get_property = gst_rgb2gray_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_R,
      g_param_spec_int ("r", "R", "Weight of red",
          -500,500,30, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_G,
      g_param_spec_int ("g", "G", "Weight of green",
          -500,500,59, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_B,
      g_param_spec_int ("b", "B", "Weight of blue",
          -500,500,11, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OFFSET,
      g_param_spec_int ("offset", "Offset", "Offset value (for inverting)",
          0,255,0, G_PARAM_READWRITE));

}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_rgb2gray_init (Gstrgb2gray * filter,
    Gstrgb2grayClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_rgb2gray_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_rgb2gray_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_rgb2gray_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_rgb2gray_get_caps));
  //gst_pad_set_getcaps_function (filter->srcpad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->bpp=32;
  filter->r=30;
  filter->g=59;
  filter->b=11;
  filter->caps=NULL;
  filter->fps_n=0;
  filter->fps_d=0;
  filter->offset=0;
}

static void
gst_rgb2gray_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstrgb2gray *filter = GST_RGB2GRAY (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_R:
      filter->r = g_value_get_int (value);
      break;
    case PROP_G:
      filter->g = g_value_get_int (value);
      break;
    case PROP_B:
      filter->b = g_value_get_int (value);
      break;
    case PROP_OFFSET:
      filter->offset = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_rgb2gray_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstrgb2gray *filter = GST_RGB2GRAY (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_R:
      g_value_set_int (value, filter->r);
      break;
    case PROP_G:
      g_value_set_int (value, filter->g);
      break;
    case PROP_B:
      g_value_set_int (value, filter->b);
      break;
    case PROP_OFFSET:
      g_value_set_int (value, filter->offset);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_rgb2gray_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstrgb2gray *filter;
  GstPad *otherpad;

  filter = GST_RGB2GRAY (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  //
  //const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  if (!gst_structure_get_int(structure,"width",&filter->width)) return FALSE;
  if (!gst_structure_get_int(structure,"height",&filter->height)) return FALSE;
  gst_video_parse_caps_framerate      (caps,&filter->fps_n,&filter->fps_d);

  gst_object_unref (filter);
  if (filter->caps)gst_caps_unref(filter->caps);
filter->caps=gst_caps_new_simple ("video/x-raw-gray",
      		"width", G_TYPE_INT, filter->width,
      		"height", G_TYPE_INT, filter->height,
      		"bpp",G_TYPE_INT , 8,
      		"depth",G_TYPE_INT , 8,
		"framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
		NULL);

  //return gst_pad_set_caps (otherpad, caps);
  return TRUE;
}

static GstCaps *
gst_rgb2gray_get_caps (GstPad *pad){
    g_print ("RGB2GRAY get caps, returning %s.\n",gst_caps_to_string(gst_pad_get_pad_template_caps(pad)));
//return gst_caps_copy(gst_pad_get_caps(pad));
//return gst_caps_new_simple ("video/x-raw-gray",NULL);//wat denn sonst?
  
  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
}
/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_rgb2gray_chain (GstPad * pad, GstBuffer * buf)
{
  Gstrgb2gray *filter;
  GstBuffer *out;
  GstFlowReturn ret;  
  int x,y;
  filter = GST_RGB2GRAY (GST_OBJECT_PARENT (pad));




  if (!filter->caps)filter->caps=gst_caps_new_simple ("video/x-raw-gray",
      		"width", G_TYPE_INT, filter->width,
      		"height", G_TYPE_INT, filter->height,
      		"bpp",G_TYPE_INT , 8,
      		"depth",G_TYPE_INT , 8,
		"framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
		NULL);


  ret=gst_pad_alloc_buffer_and_set_caps   (filter->srcpad,
                                                         buf->offset,
                                                         filter->width*filter->height,
                                                         filter->caps,
                                                         &out);
  if (ret!=GST_FLOW_OK){
    g_print ("RGB2Gray: Error allocating buffer.\n");
    return ret;
  }
  for (y=0;y<filter->height;y++){
    for(x=0;x<filter->width;x++){
      int temp=0;
      temp=filter->r*(int)(buf->data[((filter->bpp/8)*(y*filter->width+x))+1]);
      temp+=filter->g*(int)(buf->data[((filter->bpp/8)*(y*filter->width+x))+2]);
      temp+=filter->b*(int)(buf->data[((filter->bpp/8)*(y*filter->width+x))+3]);
      temp=temp/100+filter->offset;
      if (temp<0)temp=0;
      if (temp>255)temp=255;    
      out->data[y*filter->width+x]=temp;
    }
  }
  out->timestamp=buf->timestamp;
  out->offset_end=buf->offset_end;
  out->duration=buf->duration;
  gst_buffer_unref(buf);
  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, out);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
rgb2gray_init (GstPlugin * rgb2gray)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template rgb2gray' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_rgb2gray_debug, "rgb2gray",
      0, "Template rgb2gray");

  return gst_element_register (rgb2gray, "rgb2gray", GST_RANK_NONE,
      GST_TYPE_RGB2GRAY);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstrgb2gray"
#endif

/* gstreamer looks for this structure to register rgb2grays
 *
 * exchange the string 'Template rgb2gray' with your rgb2gray description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "rgb2gray",
    "Template rgb2gray",
    rgb2gray_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
