/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_ROSSIFTFOLDER_H__
#define __GST_ROSSIFTFOLDER_H__

#include <gst/gst.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>

#include <sys/inotify.h>
#include <sys/select.h>

#include "tf/transform_broadcaster.h"

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_ROSSIFTFOLDER \
  (gst_rosrossiftfolder_get_type())
#define GST_ROSSIFTFOLDER(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ROSSIFTFOLDER,Gstrosrossiftfolder))
#define GST_ROSSIFTFOLDER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_ROSSIFTFOLDER,GstrosrossiftfolderClass))
#define GST_IS_ROSSIFTFOLDER(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ROSSIFTFOLDER))
#define GST_IS_ROSSIFTFOLDER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_ROSSIFTFOLDER))



#define EVENT_SIZE  ( sizeof (struct inotify_event) )
#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )

typedef struct _Gstrosrossiftfolder      Gstrosrossiftfolder;
typedef struct _GstrosrossiftfolderClass GstrosrossiftfolderClass;

//typedef enum {
//  MATCH,
//  MERGE,  //merge two sets of sift-features
//  TRANSFORM,
//  TRANSFORMPOINT  //calculate transformation of one poinnt
//} SiftToolOpMode;




struct _Gstrosrossiftfolder
{
  GstElement element;

  GstPad *sinkpad, *srcpad;
  GList *siftheap;
  char* directory;//directory with image files for matching
  gboolean silent;
  gint width1,height1;
  gboolean rematchdouble,showinliers;//whether to put out coordinates in booka-style instead of matchlist
  int double_image;
  GstBuffer *s1;
  gint n1;
  struct feature* feat1;
  gint mode;//mode of the input (to parse)
  gint suggestedmodes1,suggestedmodes2;//suggestedmode from user input
  GMutex* mutex;
  gdouble maxsim,maxsim2,gooddist,threshold;
  //stuff for inotify
  int fd,wd;
  guint thresh;
  pthread_t th;
  char buffer[EVENT_BUF_LEN];
  int ntotal,thisn;
  gdouble c_x;
  gdouble c_y;
  gdouble f_x; 
  gdouble f_y;
  gdouble k_1;
  gdouble k_2;
  gdouble k_3;
  gdouble k_4;
  gdouble k_5;
  gdouble hfov;
  CvMat* distortion_coeffs;
  CvMat* intrinsic_matrix;
  char* unddirectory;
  struct kd_node* kdback;
  
  
  ros::NodeHandle* nh;
  ros::Publisher* rpub;
  //ros::Publisher* tfpub;
  //static tf::TransformBroadcaster br;
  tf::TransformBroadcaster* br;

  gchar* topic;
  gchar* frame_id; 
  gboolean firsttime;

  gchar* caminfotopic; 
};

struct _GstrosrossiftfolderClass 
{
  GstElementClass parent_class;
};

GType gst_rosrossiftfolder_get_type (void);


G_END_DECLS

#endif /* __GST_ROSSIFTFOLDER_H__ */
