/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2012 Hannes <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#define MAXSTRING 10000
#ifndef __GST_ROSSINK_H__
#define __GST_ROSSINK_H__
#include <gio/gio.h>
#include <gst/gst.h>
#include <gst/base/gstbasesink.h>
#include <gdk/gdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_ROSSINK \
  (gst_rossink_get_type())
#define GST_ROSSINK(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ROSSINK,Gstrossink))
#define GST_ROSSINK_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_ROSSINK,GstrossinkClass))
#define GST_IS_ROSSINK(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ROSSINK))
#define GST_IS_ROSSINK_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_ROSSINK))

typedef struct _Gstrossink      Gstrossink;
typedef struct _GstrossinkClass GstrossinkClass;

/*enum lutype{
INT,
INT_RO,
FLOAT,
FLOAT_RO,
TEXT,
TEXT_RO,
BOOLEAN,
BOOLEAN_RO,
MENU
};*/


struct _Gstrossink
{
  GstBaseSink element;


  //GstPad *sinkpad; //arleady in baseclass

  gboolean silent;


  GList *propertylist;//list of properties we watch and export
  gchar* exportdsc;//export description as a CSV
  gchar* uidesc;
  gboolean firsttime;//is it the first-time to send sth 
  gboolean socketopen;
  gchar* remoteadress;
  gchar* conffiles;//a coma separated list of files that will be read to 
  gchar* camerafile;//directory
  int interval;

 GSocket *rsock;
 GSocketAddress * raddr;//adress
 gint rport;//recvport
 int rfd;//send file descriptor for fd
 GIOChannel* rchannel; //send channel

 GSocket *ssock;
 //GSocketAddress * saddr;//adress
 gint sport;//sendport, normally recvport +1
 gint width,height;
 GList *spinoffs;
 GList *controllers;
 guint source;
 int delay,extradelay;

 gint jport;
 gchar* jadress;
 GSocketAddress * jaddr;
 GInetAddress* jiaddr;
 gboolean sendj;
 gchar joinui[1024]; 
 gchar * topic;
 gchar * frame_id;
 gchar * formatstring;
 gboolean peeralloc;
 
 std::vector < sensor_msgs::Image* > msglist;
 //
 GstBuffer* recyclebuf;
 GMutex* buflist_mutex;

 //ROS Stuff
 ros::NodeHandle* nh;
 image_transport::ImageTransport* it;
 image_transport::CameraPublisher* pub;
 ros::Publisher* rpub;
 sensor_msgs::CameraInfo* camera_info;
 ros::ServiceServer* set_camera_info;



};

struct _GstrossinkClass 
{
 GstBaseSinkClass parent_class;
};

GType gst_rossink_get_type (void);

G_END_DECLS

#endif /* __GST_ROSSINK_H__ */
