/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-sifttools
 *
 * FIXME:Describe sifttools here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! sifttools ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif



//#include "sift.h"
//#include "imgfeatures.h"
//#include "kdtree.h"
//#include "utils.h"
//#include "xform.h"
#include "math.h"

#include <sys/time.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>


#include <gst/gst.h>

#include <gsl/gsl_sf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "gstsifttools.h"

#ifndef NAN
#define NAN (0.0/0.0)
#endif
GST_DEBUG_CATEGORY_STATIC (gst_sifttools_debug);
#define GST_CAT_DEFAULT gst_sifttools_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_TRIGGER1,
  PROP_TRIGGER2,
  PROP_OPMODE,
  PROP_OFFX,
  PROP_OFFY,
  PROP_MAXSIM,
  PROP_MAXSIM2,
  PROP_GOODDIST,
  PROP_THRESHOLD,
  PROP_REMATCHDOUBLE,
  PROP_SHOWINLIERS,
  PROP_MODE  
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */


static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("sift/raw,"
        "framerate = (fraction) [ 0, MAX ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];"
	"sift/lowe;"
	"sift/oxf;"
	"sift/lowelight,"
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];")
    );

/*
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );
*/
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "coordinates-list;"
      "matrix-list;"
	"sift/raw,"
        //"framerate = (fraction) [ 0, MAX ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];"
	"sift/lowe;"
	"sift/oxf;"
	"sift/lowelight,"
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ];")
    );
GST_BOILERPLATE (Gstsifttools, gst_sifttools, GstElement,
    GST_TYPE_ELEMENT);

static void gst_sifttools_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_sifttools_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static gboolean gst_sifttools_event (GstPad  *pad, GstEvent * event);
static gboolean gst_sifttools_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_sifttools_chain (GstPad * pad, GstBuffer * buf);
static GstStateChangeReturn
gst_sifttools_change_state (GstElement *element, GstStateChange transition);
static GstCaps *
gst_sift_extractor_get_src_caps (GstPad *pad);
static GstCaps *
gst_sift_extractor_get_sink_caps (GstPad *pad);


static inline struct feature* get_match( struct feature*, int );
int get_matched_features( struct feature*, int, int, struct feature*** );
int calc_min_inliers( int, int, double, double );
struct feature** draw_ransac_sample( struct feature**, int, int, gsl_rng* );
void extract_corresp_pts( struct feature**, int, int, CvPoint2D64f**,
			  CvPoint2D64f** );
int find_consensus( struct feature**, int, int, CvMat*, ransac_err_fn,
		    double, struct feature*** );
static inline void release_mem( CvPoint2D64f*, CvPoint2D64f*,
				struct feature** );
CvMat* lsq_homog( CvPoint2D64f* pts, CvPoint2D64f* mpts, int n );
double homog_xfer_err( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* H );
CvPoint2D64f persp_xform_pt( CvPoint2D64f pt, CvMat* T );

struct kd_node* kd_node_init( struct feature*, int );
void expand_kd_node_subtree( struct kd_node* );
void assign_part_key( struct kd_node* );
double median_select( double*, int );
double rank_select( double*, int, int );
void insertion_sort( double*, int );
int partition_array( double*, int, double );
void partition_features( struct kd_node* );
struct kd_node* explore_to_leaf( struct kd_node*, struct feature*,
				 struct min_pq* );
int insert_into_nbr_array( struct feature*, struct feature**, int, int );
int within_rect( CvPoint2D64f, CvRect );

#define GST_SIFTTOOL_TYPE_OP_MODE (gst_type_sifttool_op_mode ())

double descr_dist_sq( struct feature* f1, struct feature* f2 )
{
  double diff, dsq = 0;
  double* descr1, * descr2;
  int i, d;

  d = f1->d;
  if( f2->d != d )
    return DBL_MAX;
  descr1 = f1->descr;
  descr2 = f2->descr;

  for( i = 0; i < d; i++ )
    {
      diff = descr1[i] - descr2[i];
      dsq += diff*diff;
    }
  return dsq;
}

CvMat* ransac_xform( struct feature* features, int n, int mtype,
		     ransac_xform_fn xform_fn, int m, double p_badxform,
		     ransac_err_fn err_fn, double err_tol,
		     struct feature*** inliers, int* n_in )
{
  struct feature** matched, ** sample, ** consensus, ** consensus_max = NULL;
  struct ransac_data* rdata;
  CvPoint2D64f* pts, * mpts;
  CvMat* M = NULL;
  gsl_rng* rng;
  double p, in_frac = RANSAC_INLIER_FRAC_EST;
  int i, nm, in, in_min, in_max = 0, k = 0;

  nm = get_matched_features( features, n, mtype, &matched );
  if( nm < m )
    {
      fprintf( stderr, "Warning: not enough matches to compute xform, %s" \
	       " line %d\n", __FILE__, __LINE__ );
      goto end;
    }

  /* initialize random number generator */
  rng = gsl_rng_alloc( gsl_rng_mt19937 );
  gsl_rng_set( rng, time(NULL) );

  in_min = calc_min_inliers( nm, m, RANSAC_PROB_BAD_SUPP, p_badxform );
  p = pow( 1.0 - pow( in_frac, m ), k );
  while( p > p_badxform )
    {
      sample = draw_ransac_sample( matched, nm, m, rng );
      extract_corresp_pts( sample, m, mtype, &pts, &mpts );
      M = xform_fn( pts, mpts, m );
      if( ! M )
	goto iteration_end;
      in = find_consensus( matched, nm, mtype, M, err_fn, err_tol, &consensus);
      if( in > in_max )
	{
	  if( consensus_max )
	    free( consensus_max );
	  consensus_max = consensus;
	  in_max = in;
	  in_frac = (double)in_max / nm;
	}
      else
	free( consensus );
      cvReleaseMat( &M );

    iteration_end:
      release_mem( pts, mpts, sample );
      p = pow( 1.0 - pow( in_frac, m ), ++k );
    }

  /* calculate final transform based on best consensus set */
  if( in_max >= in_min )
    {
      extract_corresp_pts( consensus_max, in_max, mtype, &pts, &mpts );
      M = xform_fn( pts, mpts, in_max );
      in = find_consensus( matched, nm, mtype, M, err_fn, err_tol, &consensus);
      cvReleaseMat( &M );
      release_mem( pts, mpts, consensus_max );
      extract_corresp_pts( consensus, in, mtype, &pts, &mpts );
      M = xform_fn( pts, mpts, in );
      if( inliers )
	{
	  *inliers = consensus;
	  consensus = NULL;
	}
      else release_mem( pts, mpts, consensus );
      if( n_in )
	*n_in = in;

    }
  else if( consensus_max )
    {
      if( inliers )
	*inliers = NULL;
      if( n_in )
	*n_in = 0;
      free( consensus_max );
    }

  gsl_rng_free( rng );
 end:
  for( i = 0; i < nm; i++ )
    {
      rdata = feat_ransac_data( matched[i] );
      matched[i]->feature_data = rdata->orig_feat_data;
      free( rdata );
    }
  free( matched );
  return M;
}

CvMat* lsq_homog( CvPoint2D64f* pts, CvPoint2D64f* mpts, int n )
{
  CvMat* H, * A, * B, X;
  double x[9];
  int i;

  if( n < 4 )
    {
      fprintf( stderr, "Warning: too few points in lsq_homog(), %s line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }

  /* set up matrices so we can unstack homography into X; AX = B */
  A = cvCreateMat( 2*n, 8, CV_64FC1 );
  B = cvCreateMat( 2*n, 1, CV_64FC1 );
  X = cvMat( 8, 1, CV_64FC1, x );
  H = cvCreateMat(3, 3, CV_64FC1);
  cvZero( A );
  for( i = 0; i < n; i++ )
    {
      cvmSet( A, i, 0, pts[i].x );
      cvmSet( A, i+n, 3, pts[i].x );
      cvmSet( A, i, 1, pts[i].y );
      cvmSet( A, i+n, 4, pts[i].y );
      cvmSet( A, i, 2, 1.0 );
      cvmSet( A, i+n, 5, 1.0 );
      cvmSet( A, i, 6, -pts[i].x * mpts[i].x );
      cvmSet( A, i, 7, -pts[i].y * mpts[i].x );
      cvmSet( A, i+n, 6, -pts[i].x * mpts[i].y );
      cvmSet( A, i+n, 7, -pts[i].y * mpts[i].y );
      cvmSet( B, i, 0, mpts[i].x );
      cvmSet( B, i+n, 0, mpts[i].y );
    }
  cvSolve( A, B, &X, CV_SVD );
  x[8] = 1.0;
  X = cvMat( 3, 3, CV_64FC1, x );
  cvConvert( &X, H );

  cvReleaseMat( &A );
  cvReleaseMat( &B );
  return H;
}



/*
  Calculates the transfer error between a point and its correspondence for
  a given homography, i.e. for a point x, it's correspondence x', and
  homography H, computes d(x', Hx)^2.
  
  @param pt a point
  @param mpt pt's correspondence
  @param H a homography matrix
  
  @return Returns the transfer error between pt and mpt given H
*/
double homog_xfer_err( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* H )
{
  CvPoint2D64f xpt = persp_xform_pt( pt, H );
  
  return sqrt( dist_sq_2D( xpt, mpt ) );
}



/*
  Performs a perspective transformation on a single point.  That is, for a
  point (x, y) and a 3 x 3 matrix T this function returns the point
  (u, v), where
  
  [x' y' w']^T = T * [x y 1]^T,
  
  and
  
  (u, v) = (x'/w', y'/w').

  Note that affine transforms are a subset of perspective transforms.
  
  @param pt a 2D point
  @param T a perspective transformation matrix
  
  @return Returns the point (u, v) as above.
*/
CvPoint2D64f persp_xform_pt( CvPoint2D64f pt, CvMat* T )
{
  CvMat XY, UV;
  double xy[3] = { pt.x, pt.y, 1.0 }, uv[3] = { 0 };
  CvPoint2D64f rslt;

  cvInitMatHeader( &XY, 3, 1, CV_64FC1, xy, CV_AUTOSTEP );
  cvInitMatHeader( &UV, 3, 1, CV_64FC1, uv, CV_AUTOSTEP );
  cvMatMul( T, &XY, &UV );
  rslt = cvPoint2D64f( uv[0] / uv[2], uv[1] / uv[2] );

  return rslt;
}


static inline struct feature* get_match( struct feature* feat, int mtype )
{
  if( mtype == FEATURE_MDL_MATCH )
    return feat->mdl_match;
  if( mtype == FEATURE_BCK_MATCH )
    return feat->bck_match;
  if( mtype == FEATURE_FWD_MATCH )
    return feat->fwd_match;
  return NULL;
}




int get_matched_features( struct feature* features, int n, int mtype,
			  struct feature*** matched )
{
  struct feature** _matched;
  struct ransac_data* rdata;
  int i, m = 0;

  _matched = calloc( n, sizeof( struct feature* ) );
  for( i = 0; i < n; i++ )
    if( get_match( features + i, mtype ) )
      {
	rdata = malloc( sizeof( struct ransac_data ) );
	memset( rdata, 0, sizeof( struct ransac_data ) );
	rdata->orig_feat_data = features[i].feature_data;
	_matched[m] = features + i;
	_matched[m]->feature_data = rdata;
	m++;
      }
  *matched = _matched;
  return m;
}


int calc_min_inliers( int n, int m, double p_badsupp, double p_badxform )
{
  double pi, sum;
  int i, j;

  for( j = m+1; j <= n; j++ )
    {
      sum = 0;
      for( i = j; i <= n; i++ )
	{
	  pi = (i-m) * log( p_badsupp ) + (n-i+m) * log( 1.0 - p_badsupp ) +
	    gsl_sf_lnchoose( n - m, i - m );
	  sum += exp( pi );
	}
      if( sum < p_badxform )
	break;
    }
  return j;
}

struct feature** draw_ransac_sample( struct feature** features, int n,
				     int m, gsl_rng* rng )
{
  struct feature** sample, * feat;
  struct ransac_data* rdata;
  int i, x;

  for( i = 0; i < n; i++ )
    {
      rdata = feat_ransac_data( features[i] );
      rdata->sampled = 0;
    }

  sample = calloc( m, sizeof( struct feature* ) );
  for( i = 0; i < m; i++ )
    {
      do
	{
	  x = gsl_rng_uniform_int( rng, n );
	  feat = features[x];
	  rdata = feat_ransac_data( feat );
	}
      while( rdata->sampled );
      sample[i] = feat;
      rdata->sampled = 1;
    }

  return sample;
}


void extract_corresp_pts( struct feature** features, int n, int mtype,
			  CvPoint2D64f** pts, CvPoint2D64f** mpts )
{
  struct feature* match;
  CvPoint2D64f* _pts, * _mpts;
  int i;

  _pts = calloc( n, sizeof( CvPoint2D64f ) );
  _mpts = calloc( n, sizeof( CvPoint2D64f ) );

  if( mtype == FEATURE_MDL_MATCH )
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of correct type\n");
	_pts[i] = features[i]->img_pt;
	_mpts[i] = match->mdl_pt;
      }

  else
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of type\n");
	_pts[i] = features[i]->img_pt;
	_mpts[i] = match->img_pt;
      }

  *pts = _pts;
  *mpts = _mpts;
}




int find_consensus( struct feature** features, int n, int mtype,
		    CvMat* M, ransac_err_fn err_fn, double err_tol,
		    struct feature*** consensus )
{
  struct feature** _consensus;
  struct feature* match;
  CvPoint2D64f pt, mpt;
  double err;
  int i, in = 0;

  _consensus = calloc( n, sizeof( struct feature* ) );

  if( mtype == FEATURE_MDL_MATCH )
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of type\n");
	pt = features[i]->img_pt;
	mpt = match->mdl_pt;
	err = err_fn( pt, mpt, M );
	if( err <= err_tol )
	  _consensus[in++] = features[i];
      }

  else
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of type\n");
	pt = features[i]->img_pt;
	mpt = match->img_pt;
	err = err_fn( pt, mpt, M );
	if( err <= err_tol )
	  _consensus[in++] = features[i];
      }
  *consensus = _consensus;
  return in;
}


static inline void release_mem( CvPoint2D64f* pts1, CvPoint2D64f* pts2,
				struct feature** features )
{
  free( pts1 );
  free( pts2 );
  if( features )
    free( features );
}


struct kd_node* kdtree_build( struct feature* features, int n )
{
  struct kd_node* kd_root;

  if( ! features  ||  n <= 0 )
    {
      fprintf( stderr, "Warning: kdtree_build(): no features, %s, line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }

  kd_root = kd_node_init( features, n );
  expand_kd_node_subtree( kd_root );

  return kd_root;
}

int kdtree_bbf_knn( struct kd_node* kd_root, struct feature* feat, int k,
		    struct feature*** nbrs, int max_nn_chks )
{
  struct kd_node* expl;
  struct min_pq* min_pq;
  struct feature* tree_feat, ** _nbrs;
  struct bbf_data* bbf_data;
  int i, t = 0, n = 0;

  if( ! nbrs  ||  ! feat  ||  ! kd_root )
    {
      fprintf( stderr, "Warning: NULL pointer error, %s, line %d\n",
	       __FILE__, __LINE__ );
      return -1;
    }

  _nbrs = calloc( k, sizeof( struct feature* ) );
  min_pq = minpq_init();
  minpq_insert( min_pq, kd_root, 0 );
  while( min_pq->n > 0  &&  t < max_nn_chks )
    {
      expl = (struct kd_node*)minpq_extract_min( min_pq );
      if( ! expl )
	{
	  fprintf( stderr, "Warning: PQ unexpectedly empty, %s line %d\n",
		   __FILE__, __LINE__ );
	  goto fail;
	}

      expl = explore_to_leaf( expl, feat, min_pq );
      if( ! expl )
	{
	  fprintf( stderr, "Warning: PQ unexpectedly empty, %s line %d\n",
		   __FILE__, __LINE__ );
	  goto fail;
	}

      for( i = 0; i < expl->n; i++ )
	{
	  tree_feat = &expl->features[i];
	  bbf_data = malloc( sizeof( struct bbf_data ) );
	  if( ! bbf_data )
	    {
	      fprintf( stderr, "Warning: unable to allocate memory,"
		       " %s line %d\n", __FILE__, __LINE__ );
	      goto fail;
	    }
	  bbf_data->old_data = tree_feat->feature_data;
	  bbf_data->d = descr_dist_sq(feat, tree_feat);
	  tree_feat->feature_data = bbf_data;
	  n += insert_into_nbr_array( tree_feat, _nbrs, n, k );
	}
      t++;
    }

  minpq_release( &min_pq );
  for( i = 0; i < n; i++ )
    {
      bbf_data = _nbrs[i]->feature_data;
      _nbrs[i]->feature_data = bbf_data->old_data;
      free( bbf_data );
    }
  *nbrs = _nbrs;
  return n;

 fail:
  minpq_release( &min_pq );
  for( i = 0; i < n; i++ )
    {
      bbf_data = _nbrs[i]->feature_data;
      _nbrs[i]->feature_data = bbf_data->old_data;
      free( bbf_data );
    }
  free( _nbrs );
  *nbrs = NULL;
  return -1;
}

void kdtree_release( struct kd_node* kd_root )
{
  if( ! kd_root )
    return;
  kdtree_release( kd_root->kd_left );
  kdtree_release( kd_root->kd_right );
  free( kd_root );
}




struct kd_node* kd_node_init( struct feature* features, int n )
{
  struct kd_node* kd_node;

  kd_node = malloc( sizeof( struct kd_node ) );
  memset( kd_node, 0, sizeof( struct kd_node ) );
  kd_node->ki = -1;
  kd_node->features = features;
  kd_node->n = n;

  return kd_node;
}


void expand_kd_node_subtree( struct kd_node* kd_node )
{
  /* base case: leaf node */
  if( kd_node->n == 1  ||  kd_node->n == 0 )
    {
      kd_node->leaf = 1;
      return;
    }

  assign_part_key( kd_node );
  partition_features( kd_node );

  if( kd_node->kd_left )
    expand_kd_node_subtree( kd_node->kd_left );
  if( kd_node->kd_right )
    expand_kd_node_subtree( kd_node->kd_right );
}


void assign_part_key( struct kd_node* kd_node )
{
  struct feature* features;
  double kv, x, mean, var, var_max = 0;
  double* tmp;
  int d, n, i, j, ki = 0;

  features = kd_node->features;
  n = kd_node->n;
  d = features[0].d;

  /* partition key index is that along which descriptors have most variance */
  for( j = 0; j < d; j++ )
    {
      mean = var = 0;
      for( i = 0; i < n; i++ )
	mean += features[i].descr[j];
      mean /= n;
      for( i = 0; i < n; i++ )
	{
	  x = features[i].descr[j] - mean;
	  var += x * x;
	}
      var /= n;

      if( var > var_max )
	{
	  ki = j;
	  var_max = var;
	}
    }

  /* partition key value is median of descriptor values at ki */
  tmp = calloc( n, sizeof( double ) );
  for( i = 0; i < n; i++ )
    tmp[i] = features[i].descr[ki];
  kv = median_select( tmp, n );
  free( tmp );

  kd_node->ki = ki;
  kd_node->kv = kv;
}


double median_select( double* array, int n )
{
  return rank_select( array, n, (n - 1) / 2 );
}



double rank_select( double* array, int n, int r )
{
  double* tmp, med;
  int gr_5, gr_tot, rem_elts, i, j;

  /* base case */
  if( n == 1 )
    return array[0];

  /* divide array into groups of 5 and sort them */
  gr_5 = n / 5;
  gr_tot = cvCeil( n / 5.0 );
  rem_elts = n % 5;
  tmp = array;
  for( i = 0; i < gr_5; i++ )
    {
      insertion_sort( tmp, 5 );
      tmp += 5;
    }
  insertion_sort( tmp, rem_elts );

  /* recursively find the median of the medians of the groups of 5 */
  tmp = calloc( gr_tot, sizeof( double ) );
  for( i = 0, j = 2; i < gr_5; i++, j += 5 )
    tmp[i] = array[j];
  if( rem_elts )
    tmp[i++] = array[n - 1 - rem_elts/2];
  med = rank_select( tmp, i, ( i - 1 ) / 2 );
  free( tmp );
  
  /* partition around median of medians and recursively select if necessary */
  j = partition_array( array, n, med );
  if( r == j )
    return med;
  else if( r < j )
    return rank_select( array, j, r );
  else
    {
      array += j+1;
      return rank_select( array, ( n - j - 1 ), ( r - j - 1 ) );
    }
}


void insertion_sort( double* array, int n )
{
  double k;
  int i, j;

  for( i = 1; i < n; i++ )
    {
      k = array[i];
      j = i-1;
      while( j >= 0  &&  array[j] > k )
	{
	  array[j+1] = array[j];
	  j -= 1;
	}
      array[j+1] = k;
    }
}


int partition_array( double* array, int n, double pivot )
{
  double tmp;
  int p=0, i, j;

  i = -1;
  for( j = 0; j < n; j++ )
    if( array[j] <= pivot )
      {
	tmp = array[++i];
	array[i] = array[j];
	array[j] = tmp;
	if( array[i] == pivot )
	  p = i;
      }
  array[p] = array[i];
  array[i] = pivot;
  
  return i;
}


void partition_features( struct kd_node* kd_node )
{
  struct feature* features, tmp;
  double kv;
  int n, ki, p=0, i, j = -1;

  features = kd_node->features;
  n = kd_node->n;
  ki = kd_node->ki;
  kv = kd_node->kv;
  for( i = 0; i < n; i++ )
    if( features[i].descr[ki] <= kv )
      {
	tmp = features[++j];
	features[j] = features[i];
	features[i] = tmp;
	if( features[j].descr[ki] == kv )
	  p = j;
      }
  tmp = features[p];
  features[p] = features[j];
  features[j] = tmp;

  /* if all records fall on same side of partition, make node a leaf */
  if( j == n - 1 )
    {
      kd_node->leaf = 1;
      return;
    }

  kd_node->kd_left = kd_node_init( features, j + 1 );
  kd_node->kd_right = kd_node_init( features + ( j + 1 ), ( n - j - 1 ) );
}




struct kd_node* explore_to_leaf( struct kd_node* kd_node, struct feature* feat,
				 struct min_pq* min_pq )
{
  struct kd_node* unexpl, * expl = kd_node;
  double kv;
  int ki;

  while( expl  &&  ! expl->leaf )
    {
      ki = expl->ki;
      kv = expl->kv;
      
      if( ki >= feat->d )
	{
	  fprintf( stderr, "Warning: comparing imcompatible descriptors, %s" \
		   " line %d\n", __FILE__, __LINE__ );
	  return NULL;
	}
      if( feat->descr[ki] <= kv )
	{
	  unexpl = expl->kd_right;
	  expl = expl->kd_left;
	}
      else
	{
	  unexpl = expl->kd_left;
	  expl = expl->kd_right;
	}
      
      if( minpq_insert( min_pq, unexpl, ABS( kv - feat->descr[ki] ) ) )
	{
	  fprintf( stderr, "Warning: unable to insert into PQ, %s, line %d\n",
		   __FILE__, __LINE__ );
	  return NULL;
	}
    }

  return expl;
}


int insert_into_nbr_array( struct feature* feat, struct feature** nbrs,
			   int n, int k )
{
  struct bbf_data* fdata, * ndata;
  double dn, df;
  int i, ret = 0;

  if( n == 0 )
    {
      nbrs[0] = feat;
      return 1;
    }

  /* check at end of array */
  fdata = (struct bbf_data*)feat->feature_data;
  df = fdata->d;
  ndata = (struct bbf_data*)nbrs[n-1]->feature_data;
  dn = ndata->d;
  if( df >= dn )
    {
      if( n == k )
	{
	  feat->feature_data = fdata->old_data;
	  free( fdata );
	  return 0;
	}
      nbrs[n] = feat;
      return 1;
    }

  /* find the right place in the array */
  if( n < k )
    {
      nbrs[n] = nbrs[n-1];
      ret = 1;
    }
  else
    {
      nbrs[n-1]->feature_data = ndata->old_data;
      free( ndata );
    }
  i = n-2;
  while( i >= 0 )
    {
      ndata = (struct bbf_data*)nbrs[i]->feature_data;
      dn = ndata->d;
      if( dn <= df )
	break;
      nbrs[i+1] = nbrs[i];
      i--;
    }
  i++;
  nbrs[i] = feat;

  return ret;
}

int within_rect( CvPoint2D64f pt, CvRect rect )
{
  if( pt.x < rect.x  ||  pt.y < rect.y )
    return 0;
  if( pt.x > rect.x + rect.width  ||  pt.y > rect.y + rect.height )
    return 0;
  return 1;
}

//GST-Stuff
static GType
gst_type_sifttool_op_mode (void)
{
  static GType sifttool_op_mode = 0;
  static const GEnumValue functions[] = {
    {MATCH, "matches two sets of features", "match"},
    {MERGE, "merges two sets of features, set offset values!", "merge"},
    {TRANSFORMPOINT,"transforms the point given by the offset values from the small image(2) to the big(1)!", "transformpoint"},
    {TRANSFORM,"gives the transformation equation", "transform"}, 
    {0, NULL, NULL}
  };

  if (!sifttool_op_mode) {
    sifttool_op_mode =
        g_enum_register_static ("SiftToolOpMode", functions);
  }
  return sifttool_op_mode;
}


/* GObject vmethod implementations */

static void
gst_sifttools_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "sifttools",
    "Gst-Plugins-Robotix",
    "Match Sift-Features",
    "Hannes Bistry <<hannesb@gmx.de>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the sifttools's class */
static void
gst_sifttools_class_init (GstsifttoolsClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_sifttools_set_property;
  gobject_class->get_property = gst_sifttools_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER1,
      g_param_spec_boolean ("trigger1", "Trigger1", "Trigger on new data on Input 1 ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER2,
      g_param_spec_boolean ("trigger2", "Trigger2", "Trigger on new data on Input 2 ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OPMODE,
      g_param_spec_enum ("opmode", "Opmode",
          "What should be done, match(default) or merge, or calculate transformation ?", GST_SIFTTOOL_TYPE_OP_MODE,
          MATCH, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OFFX,
      g_param_spec_int ("overlapx", "Overlapx", "X overlapping of second compared to first image (-1 is full)",
          -1,4000,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OFFY,
      g_param_spec_int ("overlapy", "Overlapy", "Y overlapping of second compared to first image (-1 is full)",
          -1,4000,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MAXSIM,
      g_param_spec_double ("maxsim", "Maxsim", "The second best match needs to be (1/(maxsim)) further away(1==can even be equal, 0=never match)",
          0,1,0.49, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MAXSIM2,
      g_param_spec_double ("maxsim2", "Maxsim2", "In case of a really good match (distance<gooddist), the second best match needs to be (1/(maxsim)) further away(1==can even be equal, 0=never match)",
          0,1,0.9, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_GOODDIST,
      g_param_spec_double ("gooddist", "Gooddist", "What is a good match, is square distance < than this value",
          0,1,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_THRESHOLD,
      g_param_spec_double ("threshold", "Threshold", "Threshold for Ransac",
          0,1,0.01, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_REMATCHDOUBLE,
      g_param_spec_boolean ("rematchdouble", "Rematchdouble", "Whether to kill double matches and rematch them in one direction.",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SHOWINLIERS,
      g_param_spec_boolean ("showinliers", "Showinliers", "Show only valid matches after Ransac.",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MODE,
      g_param_spec_int ("mode", "Mode", "0=raw, 1=lowe, 2=oxford 3=lowelight",
          0,3,0, G_PARAM_READWRITE));
  gstelement_class->change_state = gst_sifttools_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static gboolean little_endian;

static void
gst_sifttools_init (Gstsifttools * filter,
    GstsifttoolsClass * gclass)
{
  int x=1;
  filter->sinkpad1 = gst_pad_new_from_static_template (&sink_factory, "sink1");
  gst_pad_set_setcaps_function (filter->sinkpad1,
                                GST_DEBUG_FUNCPTR(gst_sifttools_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad1,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_get_sink_caps));
  gst_pad_set_chain_function (filter->sinkpad1,
                              GST_DEBUG_FUNCPTR(gst_sifttools_chain));
  gst_pad_set_event_function (filter->sinkpad1,GST_DEBUG_FUNCPTR(gst_sifttools_event));

  filter->sinkpad2 = gst_pad_new_from_static_template (&sink_factory, "sink2");
  gst_pad_set_setcaps_function (filter->sinkpad2,
                                GST_DEBUG_FUNCPTR(gst_sifttools_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad2,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_get_sink_caps));
  gst_pad_set_chain_function (filter->sinkpad2,
                              GST_DEBUG_FUNCPTR(gst_sifttools_chain));
  gst_pad_set_event_function (filter->sinkpad2,GST_DEBUG_FUNCPTR(gst_sifttools_event));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_get_src_caps));
  gst_pad_set_setcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_sifttools_set_caps));
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad1);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad2);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->offsetx=-1;
  filter->offsety=-1;
  filter->mutex=g_mutex_new();
  little_endian=(gboolean) (*(char*)&x); 
  filter->maxsim=0.49;
  filter->maxsim2=0.9;
  filter->threshold=0.01;
  filter->showinliers=FALSE;
  filter->mode=0;
}

static void
gst_sifttools_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstsifttools *filter = GST_SIFTTOOLS (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER1:
      filter->trigger1 = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER2:
      filter->trigger2 = g_value_get_boolean (value);
      break;
    case PROP_OPMODE:
      filter->opmode=(SiftToolOpMode)g_value_get_enum(value);
      break;
    case PROP_OFFX:
      filter->offsetx=g_value_get_int (value);
      break;
    case PROP_OFFY:
      filter->offsety=g_value_get_int (value);
      break;
    case PROP_MAXSIM:
      filter->maxsim=g_value_get_double (value);
      break;
    case PROP_MAXSIM2:
      filter->maxsim2=g_value_get_double (value);
      break;
    case PROP_GOODDIST:
      filter->gooddist=g_value_get_double (value);
      break;
    case PROP_THRESHOLD:
      filter->threshold=g_value_get_double (value);
      break;
   case PROP_REMATCHDOUBLE:
      filter->rematchdouble = g_value_get_boolean (value);
      break;
   case PROP_SHOWINLIERS:
      filter->showinliers = g_value_get_boolean (value);
      break;
   case PROP_MODE:
      filter->mode = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_sifttools_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstsifttools *filter = GST_SIFTTOOLS (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_TRIGGER1:
      g_value_set_boolean (value, filter->trigger1);
      break;
    case PROP_TRIGGER2:
      g_value_set_boolean (value, filter->trigger2);
      break;
    case PROP_OPMODE:
      g_value_set_enum(value, filter->opmode);
      break;
    case PROP_OFFX:
      g_value_set_int(value, filter->offsetx);
      break;
    case PROP_OFFY:
      g_value_set_int(value, filter->offsety);
      break;
    case PROP_MAXSIM:
      g_value_set_double(value, filter->maxsim);
      break;
    case PROP_MAXSIM2:
      g_value_set_double(value, filter->maxsim2);
      break;
    case PROP_GOODDIST:
      g_value_set_double(value, filter->gooddist);
      break;
    case PROP_THRESHOLD:
      g_value_set_double(value, filter->threshold);
      break;
    case PROP_REMATCHDOUBLE:
      g_value_set_boolean (value, filter->rematchdouble);
      break;
   case PROP_SHOWINLIERS:
      g_value_set_boolean (value, filter->showinliers);
      break;
   case PROP_MODE:
      g_value_set_int (value, filter->mode);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}
//
GstBuffer* gst_sifttools_buf_from_oxfd(char * input){
  struct feature* f;
  int i, j, n, d;
  double x, y, a, b, c, dv;
  char* temp=input;
  GstBuffer* result;
//get first two double values in d and n
  n=g_strtod (temp,&temp);
  d=g_strtod (temp,&temp);  

  if( d > FEATURE_MAX_D )
    {
      fprintf( stderr, "Warning: descriptor too long, %s, line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }
  
  result=gst_buffer_new_and_alloc (n*sizeof(struct feature));
  f = (struct feature*)result->data;

  for( i = 0; i < n; i++ )
    {
      /* read affine region parameters */
      x=g_strtod (temp,&temp);
      y=g_strtod (temp,&temp); 
      a=g_strtod (temp,&temp);
      b=g_strtod (temp,&temp); 
      c=g_strtod (temp,&temp);        
      /**/
      f[i].img_pt.x = f[i].x = x;
      f[i].img_pt.y = f[i].y = y;
      f[i].a = a;
      f[i].b = b;
      f[i].c = c;
      f[i].d = d;
      f[i].type = FEATURE_OXFD;
      
      /* read descriptor */
      for( j = 0; j < d; j++ )
	{
	  dv=g_strtod (temp,&temp);   
	  f[i].descr[j] = dv;
	  g_print("%f",dv);
	}

      f[i].scl = f[i].ori = 0;
      f[i].category = 0;
      f[i].fwd_match = f[i].bck_match = f[i].mdl_match = NULL;
      f[i].mdl_pt.x = f[i].mdl_pt.y = -1;
      f[i].feature_data = NULL;
    }
  return result;
}

GstBuffer* gst_sifttools_buf_from_lowe(char * input){
  struct feature* f;
  int i, j, n, d;
  double x, y, s, o, dv;
  char* temp=input;
  GstBuffer* result;

  n=(int)g_strtod (temp,&temp);
  d=(int)g_strtod (temp,&temp);  
  g_print("Processing %d features with each %d values",d,n);
  if( d > FEATURE_MAX_D )
    {
      fprintf( stderr, "Warning: descriptor too long, %s, line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }
  
  result=gst_buffer_new_and_alloc (n*sizeof(struct feature));
  f = (struct feature*)result->data;

  for( i = 0; i < n; i++ )
    {
      /* read affine region parameters */
      x=g_strtod (temp,&temp);
      y=g_strtod (temp,&temp); 
      s=g_strtod (temp,&temp);
      o=g_strtod (temp,&temp); 

      f[i].img_pt.x = f[i].x = x;
      f[i].img_pt.y = f[i].y = y;
      f[i].scl = s;
      f[i].ori = o;
      f[i].d = d;
      f[i].type = FEATURE_LOWE;
      g_print("Processing Nr. %d of %d and got values %f %f %f %f:\n",i,n,x,y,s,o);
      /* read descriptor */
      for( j = 0; j < d; j++ )
	{
          dv=g_strtod (temp,&temp); 
	  f[i].descr[j] = dv;
	  g_print("%f",dv);
	}

      f[i].a = f[i].b = f[i].c = 0;
      f[i].category = 0;
      f[i].fwd_match = f[i].bck_match = f[i].mdl_match = NULL;
      f[i].mdl_pt.x = f[i].mdl_pt.y = -1;
      f[i].feature_data = NULL;
    }

  return result;
}

GstBuffer* gst_sifttools_buf_from_lowelight(GstBuffer * input){
  struct feature* f;
  struct lowelight* l;
  int i, j, n, d;
  double x, y, s, o, dv;
  GstBuffer* result;

  n=(input->size)/(sizeof(struct lowelight));
  d=128;
  result=gst_buffer_new_and_alloc (n*sizeof(struct feature));
  f = (struct feature*)result->data;
  l = (struct lowelight*)input->data;
  for( i = 0; i < n; i++ )
    {
      /* read affine region parameters */
      if (little_endian){
        x=l[i].x;
        y=l[i].y; 
        s=l[i].scl;
        o=l[i].ori; 
      }
      else{
        x=DoubleSwap(l[i].x);
        y=DoubleSwap(l[i].y); 
        s=DoubleSwap(l[i].scl);
        o=DoubleSwap(l[i].ori); 
      }

      f[i].img_pt.x = f[i].x = x;
      f[i].img_pt.y = f[i].y = y;
      f[i].scl = s;
      f[i].ori = o;
      f[i].d = d;
      f[i].type = FEATURE_LOWE;
      //g_print("Processing Nr. %d of %d and got values %f %f %f %f:\n",i,n,x,y,s,o);
      /* read descriptor */
      for( j = 0; j < d; j++ )
	{
          dv=l[i].data[j];
	  f[i].descr[j] = dv;
	  //g_print("%f",dv);
	}

      f[i].a = f[i].b = f[i].c = 0;
      f[i].category = 0;
      f[i].fwd_match = f[i].bck_match = f[i].mdl_match = NULL;
      f[i].mdl_pt.x = f[i].mdl_pt.y = -1;
      f[i].feature_data = NULL;
    }

  return result;
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_sifttools_set_caps (GstPad * pad, GstCaps * caps)
{
  g_print("Sifttools set caps with %s!!\n",gst_caps_to_string (caps));
  Gstsifttools *filter;
  GstPad *otherpad;

  filter = GST_SIFTTOOLS (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad1 : filter->srcpad;
  gst_object_unref (filter);
  gboolean ret=TRUE;

  if (pad == filter->sinkpad1){
     GstStructure *structure = gst_caps_get_structure (caps, 0);
     const gchar *mime;
     mime = gst_structure_get_name (structure);
     if (strcmp (mime, "sift/raw") == 0){
	gst_structure_get_int(structure,"width",&(filter->width1));
	gst_structure_get_int(structure,"height",&(filter->height1));
	filter->modes1=0;
     }
     else if (strcmp (mime, "sift/lowe") == 0) filter->modes1=1;
     else if (strcmp (mime, "sift/oxf") == 0) filter->modes1=2;
     else if (strcmp (mime, "sift/lowelight") == 0){
	gst_structure_get_int(structure,"width",&(filter->width1));
	gst_structure_get_int(structure,"height",&(filter->height1));	
	filter->modes1=3;
     }
     else ret=FALSE;
  }
  if (pad == filter->sinkpad2){
     GstStructure *structure = gst_caps_get_structure (caps, 0);
     const gchar *mime;
     mime = gst_structure_get_name (structure);
     if (strcmp (mime, "sift/raw") == 0) {
	gst_structure_get_int(structure,"width",&(filter->width2));
	gst_structure_get_int(structure,"height",&(filter->height2));
	gst_structure_get_int(structure,"overlap_v",&(filter->offsetx));
	gst_structure_get_int(structure,"overlap_h",&(filter->offsety));
	filter->modes2=0;
     }
     else if (strcmp (mime, "sift/lowe") == 0) filter->modes2=1;
     else if (strcmp (mime, "sift/oxf") == 0) filter->modes2=2;
     else if (strcmp (mime, "sift/lowelight") == 0){
	gst_structure_get_int(structure,"width",&(filter->width2));
	gst_structure_get_int(structure,"height",&(filter->height2));
	gst_structure_get_int(structure,"overlap_v",&(filter->offsetx));
	gst_structure_get_int(structure,"overlap_h",&(filter->offsety));
	filter->modes2=3;
     }
     else ret=FALSE;     
  }
  else ret=TRUE;
  //return gst_pad_set_caps (otherpad, caps);
  return ret;
}



static GstCaps *
gst_sift_extractor_get_sink_caps (GstPad *pad){
  g_print("Start get sink caps!!\n");
  Gstsifttools *filter;

  filter = GST_SIFTTOOLS (GST_OBJECT_PARENT (pad));
  gint mode;
  if (pad== filter->sinkpad1) mode=filter->modes1;
  else mode=filter->modes2;
  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  

}

static GstCaps *
gst_sift_extractor_get_src_caps (GstPad *pad){
  g_print("Start get src caps!!\n");
  Gstsifttools *filter;

  filter = GST_SIFTTOOLS (GST_OBJECT_PARENT (pad));



  GstCaps* outcaps;
  switch(filter->opmode){
  case MATCH:
    outcaps=gst_caps_new_simple ("coordinates-list",
		NULL);
    break;
  case MERGE:
    if (filter->mode==0){
      if (filter->width1==0) outcaps=gst_caps_new_simple ("sift/raw",
      		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"overlap_v", GST_TYPE_INT_RANGE, -1, G_MAXINT,
      		"overlap_h", GST_TYPE_INT_RANGE, -1, G_MAXINT,
		NULL);

      else outcaps=gst_caps_new_simple ("sift/raw",
      		"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
      		"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
      		"overlap_v", G_TYPE_INT, filter->offsetx,
      		"overlap_h", G_TYPE_INT, filter->offsety,
		NULL);
    }
    else if (filter->mode==3){
      if (filter->width1==0) outcaps=gst_caps_new_simple ("sift/lowelight",
      		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"overlap_v", GST_TYPE_INT_RANGE, -1, G_MAXINT,
      		"overlap_h", GST_TYPE_INT_RANGE, -1, G_MAXINT,
		NULL);

      else outcaps=gst_caps_new_simple ("sift/lowelight",
      		"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
      		"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
      		"overlap_v", G_TYPE_INT, filter->offsetx,
      		"overlap_h", G_TYPE_INT, filter->offsety,
		NULL);
    }   
    break;
  case TRANSFORM:
    outcaps=gst_caps_new_simple ("matrix-list",
		NULL);
    break;
  case TRANSFORMPOINT:
    outcaps=gst_caps_new_simple ("coordinates-list",
		NULL);
    break; 
  default:
    outcaps=NULL;
  }


  //outcaps=gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  g_print("Returning %s!!\n",gst_caps_to_string(outcaps));
  return outcaps;
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_sifttools_chain (GstPad * pad, GstBuffer * buf)
{
  g_print("Start matcher!!\n");
  Gstsifttools *filter;

  filter = GST_SIFTTOOLS (GST_OBJECT_PARENT (pad));
  gboolean trigger=FALSE;
 
  //get mutex
  g_mutex_lock(filter->mutex);
  g_print("Got mutex!!\n");
  //exchange buffer(unref old)
  if (pad== filter->sinkpad1){
    trigger=filter->trigger1;
    switch(filter->modes1){
    case 0:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=buf;
      filter->feat1=(struct feature*)buf->data;
      filter->n1=(buf->size)/(sizeof(struct feature));
      break;
    case 1:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_sifttools_buf_from_lowe((char*)buf->data);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));     
      gst_buffer_unref(buf);  
      break;
    case 2:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_sifttools_buf_from_oxfd((char*)buf->data);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));   
      gst_buffer_unref(buf);   
      break;
    case 3:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_sifttools_buf_from_lowelight(buf);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));    
      gst_buffer_unref(buf);   
      break;
    default:
      if (!filter->silent)g_print("Serious Error!!\n");
      return GST_FLOW_ERROR; 
    }
  }
  else if (pad == filter->sinkpad2){
    trigger=filter->trigger2;
    switch(filter->modes2){
    case 0:
      if (filter->s2) gst_buffer_unref(filter->s2);
      filter->s2=buf;
      filter->feat2=(struct feature*)buf->data;
      filter->n2=(buf->size)/(sizeof(struct feature));
      break;
    case 1:
      if (filter->s2)gst_buffer_unref(filter->s2);
      filter->s2=gst_sifttools_buf_from_lowe((char*)buf->data);
      filter->s2->timestamp=buf->timestamp;
      filter->s2->duration=buf->duration;
      filter->s2->offset=buf->offset;
      filter->s2->offset_end=buf->offset_end;
      filter->feat2=(struct feature*)filter->s2->data;  
      filter->n2=(filter->s2->size)/(sizeof(struct feature)); 
      gst_buffer_unref(buf);  
      break;
    case 2:
      if (filter->s2)gst_buffer_unref(filter->s2);
      filter->s2=gst_sifttools_buf_from_oxfd((char*)buf->data);
      filter->s2->timestamp=buf->timestamp;
      filter->s2->duration=buf->duration;
      filter->s2->offset=buf->offset;
      filter->s2->offset_end=buf->offset_end;
      filter->feat2=(struct feature*)filter->s2->data;  
      filter->n2=(filter->s2->size)/(sizeof(struct feature));  
      gst_buffer_unref(buf);  
      break;
    case 3:
      if (filter->s2)gst_buffer_unref(filter->s2);
      filter->s2=gst_sifttools_buf_from_lowelight(buf);
      filter->s2->timestamp=buf->timestamp;
      filter->s2->duration=buf->duration;
      filter->s2->offset=buf->offset;
      filter->s2->offset_end=buf->offset_end;
      filter->feat2=(struct feature*)filter->s2->data;  
      filter->n2=(filter->s2->size)/(sizeof(struct feature)); 
      gst_buffer_unref(buf);     
      break;
    default:
      if (!filter->silent)g_print("Serious Error!!\n");
      return GST_FLOW_ERROR; 
    }
    if ((filter->opmode==MATCH)||(filter->opmode==TRANSFORM)||(filter->opmode==TRANSFORMPOINT))filter->kd_root = kdtree_build( filter->feat2, filter->n2 );
  }
  //decide to generate output?
  if ((filter->s1)&&(filter->s2)&&(trigger)&&((filter->opmode==MATCH)||(filter->opmode==TRANSFORM)||(filter->opmode==TRANSFORMPOINT))){
    if (!filter->silent)g_print("Start Eval!!\n");
    GstBuffer *outbuf;
    struct feature* feat;
    gint i,m=0,k;
    double d0, d1;
    struct feature** nbrs;
    CvMat* H=NULL;
    struct feature invalid;
    outbuf=gst_buffer_new_and_alloc(filter->n1*sizeof(struct coordinates));
    for( i = 0; i < filter->n2; i++ ){
	//current->feat->bck_match=NULL;
	//current->feat->fwd_match=NULL;
	filter->feat2[i].bck_match=NULL;
	filter->feat2[i].fwd_match=NULL;	
      }
    //g_print("allocated %d for a maximum number of %d features!!\n",filter->n1*sizeof(struct coordinates),filter->n1);
    for( i = 0; i < filter->n1; i++ )
      {
      //g_print("In for!!\n");
      feat = filter->feat1 + i;
      k = kdtree_bbf_knn( filter->kd_root, feat, 2, &nbrs, 200 );//todo 200 to parameter
      if( k == 2 )
	{
	  //g_print("found maybe sth!!\n");
	  d0 = descr_dist_sq( feat, nbrs[0] );
	  d1 = descr_dist_sq( feat, nbrs[1] );
	  if(( d0 < d1 * filter->maxsim )||((d0<filter->gooddist)&&(d0 < d1 * filter->maxsim2)))//0.49 to parameter
	    {
	      //pt1 = cvPoint( cvRound( feat->x ), cvRound( feat->y ) );
	      //pt2 = cvPoint( cvRound( nbrs[0]->x ), cvRound( nbrs[0]->y ) );
	      //pt2.y += img1->height;
	      //cvLine( stacked, pt1, pt2, CV_RGB(255,0,255), 1, 8, 0 );
  	      //g_print("found a feature, will write to %p!!\n",outbuf->data);
	      //struct coordinates* base=(struct coordinates*)(outbuf->data+(m*32));
	      //base->x1=1.00;
		//g_print("survived new adressing..!!\n");
	      //((struct coordinates*)(outbuf->data))[m].x1=feat->x;

	      //g_print("and was allowed to write it!!\n");
	      m++;
	      filter->feat1[i].fwd_match = nbrs[0];
              if (nbrs[0]->bck_match==NULL) nbrs[0]->bck_match = &filter->feat1[i];
	      else nbrs[0]->bck_match=&invalid;
	      if (!filter->silent)g_print("feature x=%d y=%d corresponds to x=%d y=%d!!\n",(int)filter->feat1[i].x,(int)filter->feat1[i].y,(int)nbrs[0]->x,(int)nbrs[0]->y);
	    }
	}
      free( nbrs );
    }
  m=0;

  if (filter->rematchdouble){//now the masterpiece, first the normal ones
    //delete
    for( i = 0; i < filter->n1; i++ ){
      if ((filter->feat1[i].fwd_match)&&(filter->feat1[i].fwd_match->bck_match!=&filter->feat1[i])){
	if (!filter->silent)g_print("deleting double\n");
	filter->feat1[i].fwd_match=NULL;
	}
    }
    //rematch
    struct kd_node* kdback;
    kdback = kdtree_build( filter->feat1, filter->n1 );
    for( i = 0; i < filter->n2; i++ ){
      if (filter->feat2[i].bck_match==&invalid){
        k = kdtree_bbf_knn( kdback,&filter->feat2[i], 2, &nbrs, 200 );
	if (k==2){
	  filter->feat2[i].bck_match=nbrs[0];
	  nbrs[0]->fwd_match=&filter->feat2[i];
	} 
      free( nbrs );	
      }
    }    
  kdtree_release( kdback );
  }


  if (filter->showinliers){
    struct feature** inliers;
    int n_in; 
    
    H = ransac_xform( filter->feat1, filter->n1, FEATURE_FWD_MATCH, lsq_homog, 4, filter->threshold,
		      homog_xfer_err, 3.0, &inliers, &n_in );
    if (!filter->silent)g_print("Found %d inliers!\n",n_in);
    for( i = 0; i < n_in; i++ ){
      if (!filter->silent)g_print("Copying inlier No. %d !\n",i);	
/*
      ((struct coordinates*)(outbuf->data))[m].x1=(*inliers)[i].x;
      ((struct coordinates*)(outbuf->data))[m].y1=(*inliers)[i].y;
      ((struct coordinates*)(outbuf->data))[m].x2=(*inliers)[i].fwd_match->x;
      ((struct coordinates*)(outbuf->data))[m].y2=(*inliers)[i].fwd_match->y;
*/
      ((struct coordinates*)(outbuf->data))[m].x1=inliers[i]->x;
      ((struct coordinates*)(outbuf->data))[m].y1=inliers[i]->y;
      ((struct coordinates*)(outbuf->data))[m].x2=inliers[i]->fwd_match->x;
      ((struct coordinates*)(outbuf->data))[m].y2=inliers[i]->fwd_match->y;
      m++;
    }


  	
  }
  else for( i = 0; i < filter->n1; i++ ){
    //feat = filter->feat1 + i;
    if (filter->feat1[i].fwd_match){
      ((struct coordinates*)(outbuf->data))[m].x1=filter->feat1[i].x;
      ((struct coordinates*)(outbuf->data))[m].y1=filter->feat1[i].y;
      ((struct coordinates*)(outbuf->data))[m].x2=filter->feat1[i].fwd_match->x;
      ((struct coordinates*)(outbuf->data))[m].y2=filter->feat1[i].fwd_match->y;
      m++;
    }
  }




  if (!filter->silent)g_print("found %d matches!",m);
  GST_BUFFER_SIZE(outbuf)=m*sizeof(struct coordinates);
  //set up outbuf
  //set up 2 feature-pointers
  //decide to parse
  outbuf->timestamp=filter->s1->timestamp;
  outbuf->duration=filter->s1->duration;
  outbuf->offset=filter->s1->offset;
  outbuf->offset_end=filter->s1->offset_end;
  g_mutex_unlock(filter->mutex);
  if ((filter->opmode==TRANSFORM)||(filter->opmode==TRANSFORMPOINT)){
    if (!H) H = ransac_xform( filter->feat1, filter->n1, FEATURE_FWD_MATCH, lsq_homog, 4, filter->threshold,
		      homog_xfer_err, 3.0, NULL, NULL );
    gst_buffer_unref(outbuf);
    outbuf=gst_buffer_new_and_alloc(sizeof(struct transmat));
    gdouble bufx=NAN,bufy=NAN,bufangle=NAN,bufscale=NAN;
    if (H){
      gint x,y;
      gdouble scale;
        if (!filter->silent)g_print("The point is x=%d and y=%d \n",(int)(filter->width2/2.0),(int)(filter->height2*0.9));
 	if (!filter->silent)printf("matrix data %f %f %f\n",cvmGet(H,0,0),cvmGet(H,0,1),cvmGet(H,0,2));
	if (!filter->silent)printf("matrix data %f %f %f\n",cvmGet(H,1,0),cvmGet(H,1,1),cvmGet(H,1,2));
	if (!filter->silent)printf("matrix data %f %f %f\n",cvmGet(H,2,0),cvmGet(H,2,1),cvmGet(H,2,2));

        double src_matrix[9], dst_matrix[9];
        CvMat A = cvMat( 3, 3, CV_64F, src_matrix ),
        invA = cvMat( 3, 3, CV_64F, dst_matrix );
        cvConvertScale( H, &A ,1,0);
        cvInvert( &A, &invA, CV_SVD );

 	if (!filter->silent)printf("inverted matrix data %f %f %f\n",cvmGet(&invA,0,0),cvmGet(&invA,0,1),cvmGet(&invA,0,2));
	if (!filter->silent)printf("inverted matrix data %f %f %f\n",cvmGet(&invA,1,0),cvmGet(&invA,1,1),cvmGet(&invA,1,2));
	if (!filter->silent)printf("inverted matrix data %f %f %f\n",cvmGet(&invA,2,0),cvmGet(&invA,2,1),cvmGet(&invA,2,2));

	if (!filter->silent)printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,0,0)/cvmGet(&invA,2,2),cvmGet(&invA,0,1)/cvmGet(&invA,2,2),cvmGet(&invA,0,2)/cvmGet(&invA,2,2));
	if (!filter->silent)printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,1,0)/cvmGet(&invA,2,2),cvmGet(&invA,1,1)/cvmGet(&invA,2,2),cvmGet(&invA,1,2)/cvmGet(&invA,2,2));
	if (!filter->silent)printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,2,0)/cvmGet(&invA,2,2),cvmGet(&invA,2,1)/cvmGet(&invA,2,2),cvmGet(&invA,2,2)/cvmGet(&invA,2,2));

	((struct transmat*)(outbuf->data))->a00=cvmGet(&invA,0,0);
	((struct transmat*)(outbuf->data))->a01=cvmGet(&invA,0,1);
	((struct transmat*)(outbuf->data))->a02=cvmGet(&invA,0,2);
	((struct transmat*)(outbuf->data))->a10=cvmGet(&invA,1,0);
	((struct transmat*)(outbuf->data))->a11=cvmGet(&invA,1,1);
	((struct transmat*)(outbuf->data))->a12=cvmGet(&invA,1,2);
	((struct transmat*)(outbuf->data))->a20=cvmGet(&invA,2,0);
	((struct transmat*)(outbuf->data))->a21=cvmGet(&invA,2,1);
	((struct transmat*)(outbuf->data))->a22=cvmGet(&invA,2,2);


        scale=(cvmGet(&invA,2,0)*((gdouble)filter->width2/2.0)+cvmGet(&invA,2,1)*((float)filter->height2*0.9)+cvmGet(&invA,2,2));
	x=(int)((cvmGet(&invA,0,0)*((gdouble)filter->width2/2.0)+cvmGet(&invA,0,1)*((float)filter->height2*0.9)+cvmGet(&invA,0,2))/scale);
	y=(int)((cvmGet(&invA,1,0)*((gdouble)filter->width2/2.0)+cvmGet(&invA,1,1)*((float)filter->height2*0.9)+cvmGet(&invA,1,2))/scale);
	
        bufscale=(cvmGet(&invA,2,0)*((gdouble)filter->width2*0.5)+cvmGet(&invA,2,1)*((float)filter->height2*0.5)+cvmGet(&invA,2,2));
	bufx=(cvmGet(&invA,0,0)*((gdouble)filter->width2*0.5)+cvmGet(&invA,0,1)*((float)filter->height2*0.5)+cvmGet(&invA,0,2))/bufscale;
	bufy=(cvmGet(&invA,1,0)*((gdouble)filter->width2*0.5)+cvmGet(&invA,1,1)*((float)filter->height2*0.5)+cvmGet(&invA,1,2))/bufscale;	
	bufangle=atan2((cvmGet(&invA,0,0)/cvmGet(&invA,2,2)),(cvmGet(&invA,0,0)/cvmGet(&invA,2,2)));
	bufscale=(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))*(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))+(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))*(cvmGet(&invA,0,0)/cvmGet(&invA,2,2));

	g_print("The point is mapped to x=%d, y=%d, scale= %f! that leads to x= %d y= %d\n",x,y,scale,(int)((float)x/scale),(int)((float)y/scale));
        //sprintf((char*)outbuf->data,"1\n");
	//sprintf((char*)outbuf->data+2,"%d %d\n",x,y);
      //cvReleaseMat( &H );
    }
    else{
        if (!filter->silent)g_print("Sorry, could not find transformation!");	
        memset (outbuf->data,0,sizeof(struct transmat));
	}
    gst_buffer_set_caps(outbuf,gst_caps_new_simple ("matrix-list",NULL));
    if (filter->opmode==TRANSFORMPOINT){
      gst_buffer_unref(outbuf);
      outbuf=gst_buffer_new_and_alloc(sizeof(struct coordinates));	
      gst_buffer_set_caps(outbuf,gst_caps_new_simple ("coordinates-list",NULL));
      if (!little_endian){
	((struct coordinates*)(outbuf->data))[0].x1=DoubleSwap(bufx);
	((struct coordinates*)(outbuf->data))[0].y1=DoubleSwap(bufy);
	((struct coordinates*)(outbuf->data))[0].x2=DoubleSwap(bufscale);
	((struct coordinates*)(outbuf->data))[0].y2=DoubleSwap(bufangle);
      }
      else{
	((struct coordinates*)(outbuf->data))[0].x1=bufx;
	((struct coordinates*)(outbuf->data))[0].y1=bufy;
	((struct coordinates*)(outbuf->data))[0].x2=bufscale;
	((struct coordinates*)(outbuf->data))[0].y2=bufangle;
	}
    }
    outbuf->timestamp=filter->s1->timestamp;
    outbuf->duration=filter->s1->duration;
    outbuf->offset=filter->s1->offset;
    outbuf->offset_end=filter->s1->offset_end;
    return gst_pad_push (filter->srcpad, outbuf);
  }
  else{
      //gst_buffer_set_caps(outbuf,gst_pad_get_pad_template_caps(filter->srcpad));
      gst_buffer_set_caps(outbuf,gst_caps_new_simple ("coordinates-list",
		NULL));
      return gst_pad_push (filter->srcpad, outbuf);
    }
  }
  else if ((filter->s1)&&(filter->s2)&&(trigger)&&(filter->opmode==MERGE)){
    if (!filter->silent)g_print("Start Merge\n");
    GstBuffer *outbuf;
    struct feature* outmem;
    int used=0;//number of used features
    int bytesperdesc=0;
    int i;
    //if (1) bytesperdesc=sizeof(struct lowelight);
    if (1) bytesperdesc=sizeof(struct feature);
    outmem=(struct feature*)g_malloc((filter->n1+filter->n2)*bytesperdesc);
    GList* incut=NULL;
    struct feature* feat;
    //if (filter->offsetx==-1)filter->offsetx=filter->width1;
    //if (filter->offsety==-1)filter->offsety=filter->height1;
    if (filter->offsetx==-1)filter->offsetx=0;
    if (filter->offsety==-1)filter->offsety=0;
    if (!filter->silent)g_print("Start Merge with overlap X %d and Y %d\n",filter->offsetx,filter->offsety);
    //copy all non overlapping features to result, and put questionable into a buffer
    for( i = 0; i < filter->n1; i++ ){
	feat=filter->feat1+i;
	if (!filter->silent)g_print("Sifttool merge analyzing %d. Feature X: %f Y: %f \n",i,feat->x,feat->y);
	if ((feat->x>((double)filter->width1-(double)filter->offsetx))||(feat->y>((double)filter->height1-(double)filter->offsety))) incut=g_list_append (incut, feat);
	else {
	  if (!filter->silent)g_print("Sifttool doing copy\n");
	  memcpy(outmem+used,feat,bytesperdesc);
	  used++;
	}
    }
    if (!filter->silent)g_print("After taking features of first part having %d final features\n",used);
    for( i = 0; i < filter->n2; i++ ){
	feat=filter->feat2+i;
	if (!filter->silent)g_print("Sifttool merge analyzing %d, with %f and %f\n",i,feat->x,feat->y);
	if (filter->offsetx>0){
          feat->x=feat->x+filter->width1-filter->offsetx;
	  feat->img_pt.x=feat->img_pt.x+filter->width1-filter->offsetx;
	}
	if (filter->offsety>0){
	  feat->y=feat->y+filter->height1-filter->offsety;
 	  feat->img_pt.y=feat->img_pt.y+filter->height1-filter->offsety;
	}
	if (!filter->silent)g_print("Sifttool after modifying number %d, with %f and %f\n",i,feat->x,feat->y);
	if ((feat->x<filter->width1)&&(feat->y<filter->height1)) incut=g_list_append (incut, feat);//if in the same region...
	else {
	  if (!filter->silent)g_print("Sifttool doing copy, out adress %p in adress %p\n",outmem+used,feat);
	  memcpy(outmem+used,feat,bytesperdesc);
	  used++;
	}
    }
    if (!filter->silent)g_print("After taking features of second part having %d final features\n",used);
    if (!filter->silent)g_print("We have to analyze %d features wether they are unique\n",g_list_length(incut));
    while(incut){

	struct feature* current=incut->data;
	struct feature* tempfeat;
	if (!filter->silent)g_print("Entering list, %d left, X %f, Y %f!!\n",g_list_length(incut),current->x,current->y);
	GList* temp=g_list_next(incut);
	gboolean found=FALSE;
	while ((temp)&&(!found)){
	  tempfeat=temp->data;
	  if (((tempfeat->x-current->x)*(tempfeat->x-current->x)+(tempfeat->x-current->x)*(tempfeat->x-current->x))<2.0){
	    if (!filter->silent)g_print("Found same feature, not including!!\n");
	    if (1) found=TRUE;
	  }
	  temp=g_list_next(temp);
	}
	if (!found){
	  if (!filter->silent)g_print("Not found same feature, including!!\n");
	  memcpy(outmem+used,current,bytesperdesc);
	  used++;
	}
	incut=g_list_delete_link (incut,incut);
	//incut=g_list_next(incut)
    }
    if (!filter->silent)g_print("After checking for duplicates we have %d final features\n",used);
    outmem=g_realloc(outmem,used*bytesperdesc);
    if (filter->mode==0){
      outbuf=gst_buffer_new_and_alloc(0);
      int outwidth=0;
      if (filter->offsetx>0)outwidth=filter->width1+filter->width2-filter->offsetx;
      else outwidth=filter->width1;

      int outheight=0;
      if (filter->offsety>0)outheight=filter->height1+filter->height2-filter->offsety;
      else outheight=filter->height1;

      gst_buffer_set_data(outbuf,(unsigned char*)outmem,used*bytesperdesc);
      GstCaps* outcaps=gst_caps_new_simple ("sift/raw",
		  //"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
		  //"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
		  "width", G_TYPE_INT, outwidth,
		  "height", G_TYPE_INT, outheight,
		  "overlap_v", G_TYPE_INT, filter->offsetx,
		  "overlap_h", G_TYPE_INT, filter->offsety,
		  NULL);
      gst_buffer_set_caps(outbuf,outcaps);
      gst_caps_unref(outcaps);
      //g_print("%s Pushing with caps %s height1 %d height2 %d!!\n",GST_ELEMENT_NAME(filter),gst_caps_to_string(outbuf->caps),filter->height1,filter->height2);
      GST_BUFFER_MALLOCDATA(outbuf)=(unsigned char*)outmem;
    }
    else if (filter->mode==3){
      //gchar* outbufmem;
      gint offset=0;
      int i,j,d;
      struct feature* feat=outmem;
    
      if (used) d=feat[0].d;
      outbuf=gst_buffer_new_and_alloc(used*sizeof(struct lowelight));
      struct lowelight *resll=(struct lowelight*)outbuf->data;

      for( i = 0; i < used; i++ )
        {
        if (little_endian){
          resll[i].x=feat[i].x;
          resll[i].y=feat[i].y;
          resll[i].scl=feat[i].scl;
          resll[i].ori=feat[i].ori;
        }
        else{
          resll[i].x=DoubleSwap(feat[i].x);
          resll[i].y=DoubleSwap(feat[i].y);
          resll[i].scl=DoubleSwap(feat[i].scl);
          resll[i].ori=DoubleSwap(feat[i].ori);
        }	
        for( j = 0; j < d; j++ )
	  {
	   resll[i].data[j]=(int)(feat[i].descr[j]);
	  }
        }
    g_print ("Allocated %d\n",used*sizeof(struct lowelight));
    free(outmem);
      int outwidth=0;
      if (filter->offsetx>0)outwidth=filter->width1+filter->width2-filter->offsetx;
      else outwidth=filter->width1;

      int outheight=0;
      if (filter->offsety>0)outheight=filter->height1+filter->height2-filter->offsety;
      else outheight=filter->height1;

      GstCaps* outcaps=gst_caps_new_simple ("sift/lowelight",
		  //"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
		  //"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
		  "width", G_TYPE_INT, outwidth,
		  "height", G_TYPE_INT, outheight,
		  "overlap_v", G_TYPE_INT, filter->offsetx,
		  "overlap_h", G_TYPE_INT, filter->offsety,
		  NULL);
      gst_buffer_set_caps(outbuf,outcaps);
      gst_caps_unref(outcaps);
      //g_print("%s Pushing with caps %s height1 %d height2 %d!!\n",GST_ELEMENT_NAME(filter),gst_caps_to_string(outbuf->caps),filter->height1,filter->height2);
    }

      outbuf->timestamp=filter->s1->timestamp;
      outbuf->duration=filter->s1->duration;
      outbuf->offset=filter->s1->offset;
      outbuf->offset_end=filter->s1->offset_end;
      if (filter->trigger1){
	  gst_buffer_unref(filter->s1);
	  filter->s1=NULL;
      }
      if (filter->trigger2){
	  gst_buffer_unref(filter->s2);
	  filter->s2=NULL;
      }
      g_mutex_unlock(filter->mutex);
      return gst_pad_push (filter->srcpad, outbuf);

  }
  else{
    g_mutex_unlock(filter->mutex);
    return GST_FLOW_OK;
  }
}

static GstStateChangeReturn
gst_sifttools_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstsifttools *filter;
  filter = GST_SIFTTOOLS (element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      filter->stopped1=!filter->trigger1;
      filter->stopped2=!filter->trigger2;
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      if (filter->s2) gst_buffer_unref(filter->s2);
      filter->s2=NULL;
      if (filter->s1) gst_buffer_unref(filter->s1);
      filter->s1=NULL;
      break;
    default:
      break;
  }
 return ret;
}





static gboolean
gst_sifttools_event (GstPad  *pad, GstEvent * event)
{
  Gstsifttools *filter;
  gboolean ret=TRUE;
  //GstPad *otherpad;

  filter = GST_SIFTTOOLS (gst_pad_get_parent (pad));

  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_EOS:
      if (pad == filter->sinkpad1) filter->stopped1=TRUE;
      if (pad == filter->sinkpad2) filter->stopped2=TRUE;
      if ((filter->stopped1)&&(filter->stopped2)) ret = gst_pad_push_event (filter->srcpad, event);
      g_print("Event EOS!\n");
      break;
    default:
      ret = gst_pad_push_event (filter->srcpad, event);
      break;
  }
  
  gst_object_unref (filter);
  return ret;
}





/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
sifttools_init (GstPlugin * sifttools)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template sifttools' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_sifttools_debug, "SiftTools",
      0, "Template Sifttools");
  //g_print("Registering Sifttools with pointer %p\n",sifttools);
  return gst_element_register (sifttools, "sifttools", GST_RANK_NONE,
      GST_TYPE_SIFTTOOLS);
}

/* gstreamer looks for this structure to register sifttoolss
 *
 * exchange the string 'Template sifttools' with your sifttools description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "Sifttools",
    "Template Sifttools",
    sifttools_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
