/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2013 Hannes <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-shmsink
 *
 * FIXME:Describe shmsink here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! shmsink
 * ]|
 * </refsect2>
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "gstshmsink.h"

GST_DEBUG_CATEGORY_STATIC (gst_shmsink_debug);
#define GST_CAT_DEFAULT gst_shmsink_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_TOPIC,
  PROP_PEERALLOC,
  PROP_NUMSUB
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );


GST_BOILERPLATE (Gstshmsink, gst_shmsink, GstBaseSink,
    GST_TYPE_BASE_SINK);

static void gst_shmsink_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_shmsink_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstStateChangeReturn gst_shmsink_change_state (GstElement *element, GstStateChange transition);
 static GstFlowReturn
gst_shmsink_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf);

static gboolean gst_shmsink_set_caps (GstBaseSink * bsink, GstCaps * caps);
static GstFlowReturn gst_shmsink_render (GstBaseSink * bsink, GstBuffer * buf);
static GstFlowReturn gst_shmsink_preroll (GstBaseSink * bsink, GstBuffer * buf){
  return GST_FLOW_OK;
}
 static GstFlowReturn
gst_shmsink_alloc (GstBaseSink * bsink, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf);
void msgtoall(Gstshmsink *filter, char* msg);
void genspinoffs(Gstshmsink *filter);
void setupsockets(Gstshmsink *filter);
/* GObject vmethod implementations */

static void
gst_shmsink_base_init (gpointer gclass)
{
  //g_print("gst_shmsink_base_init\n");
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "shmsink",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the shmsink's class */
static void
gst_shmsink_class_init (GstshmsinkClass * klass)
{
  //g_print("gst_shmsink_class_init\n");
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseSinkClass *gstbase_sink_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  gstbase_sink_class = GST_BASE_SINK_CLASS (klass);

  gobject_class->set_property = gst_shmsink_set_property;
  gobject_class->get_property = gst_shmsink_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, PROP_TOPIC,
      g_param_spec_string ("topic", "Topic", "On which topic-prefix we should publish /image_raw will be appended",
          "Gstshm", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_NUMSUB,
      g_param_spec_uint ("numsub", "Numsub", "Number of listeners currently subscribed",0,G_MAXUINT,0, G_PARAM_READABLE));




  g_object_class_install_property (gobject_class, PROP_PEERALLOC,
      g_param_spec_boolean ("peeralloc", "Peeralloc", "Use efficient peeralloc (save one cpy, but may have memleaks if other element discards memoray v4l2src) ?",TRUE, G_PARAM_READWRITE));


  gstelement_class->change_state = gst_shmsink_change_state;
  gstbase_sink_class->buffer_alloc =GST_DEBUG_FUNCPTR (gst_shmsink_alloc);
  gstbase_sink_class->set_caps=gst_shmsink_set_caps;
  gstbase_sink_class->render = GST_DEBUG_FUNCPTR (gst_shmsink_render);
  gstbase_sink_class->preroll = GST_DEBUG_FUNCPTR (gst_shmsink_preroll);
}




/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_shmsink_init (Gstshmsink * filter,
    GstshmsinkClass * gclass)
{
  //filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  //gst_pad_set_setcaps_function (filter->sinkpad,
  //                              GST_DEBUG_FUNCPTR(gst_shmsink_set_caps));
  //gst_pad_set_getcaps_function (filter->sinkpad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  //gst_pad_set_chain_function (filter->sinkpad,
  //                            GST_DEBUG_FUNCPTR(gst_shmsink_chain));
  //gst_pad_set_bufferalloc_function (filter->sinkpad,
  //    GST_DEBUG_FUNCPTR (gst_shmsink_bufferalloc));

  //gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  filter->silent = TRUE;
  filter->sendcapsnext=TRUE;
  filter->firsttime=TRUE;
  filter->width=-1;
  filter->height=-1;

  filter->topic=g_strdup_printf("Gstshm");
  filter->recyclebuf=NULL;
  filter->buflist_mutex=g_mutex_new();
  filter->peeralloc=TRUE;

}
 
static void
gst_shmsink_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstshmsink *filter = GST_SHMSINK (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;

    case PROP_TOPIC:
      if (filter->topic)g_free (filter->topic);
      filter->topic = g_value_dup_string (value);
	//todo mutex,unadvertise,advertise,unlock
      break;
    case PROP_PEERALLOC:
      filter->peeralloc = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
  //if (prop_id==PROP_SPORT)

}

static void
gst_shmsink_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstshmsink *filter = GST_SHMSINK (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_TOPIC:
      g_value_set_string (value, filter->topic);
      break;
    case PROP_PEERALLOC:
      g_value_set_boolean (value, filter->peeralloc);
      break;
    case PROP_NUMSUB:{
      guint32 num=0;

      g_value_set_uint (value,num);
      break;
    }
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


 static GstFlowReturn
gst_shmsink_alloc (GstBaseSink * bsink, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf)
{



  Gstshmsink *filter=GST_SHMSINK (bsink);
  gchar* tempc=gst_caps_to_string(caps); 
  if (!filter->silent)g_print ("Shmsink bufferalloc asked for buffer Nr %llu size %d, caps %s!\n",offset,size,tempc);
  g_free(tempc); 
  GstBuffer* buffer;
  g_mutex_lock(filter->buflist_mutex);

  if (!filter->peeralloc){
    if (0){
      if ((FALSE)&&(filter->recyclebuf)) {
        buffer = filter->recyclebuf;//refcount=1
        filter->recyclebuf=NULL;
      }
      else buffer=gst_buffer_new();
    }
    else buffer=gst_buffer_new_and_alloc(size);
  }
  else{
    boost::interprocess::shared_memory_object* tosend;
    
    sprintf(filter->shmname,"%s%llu",filter->topic,(long long unsigned int)filter->index++);
    boost::interprocess::shared_memory_object::remove(filter->shmname);
    tosend=new boost::interprocess::shared_memory_object(boost::interprocess::create_only,filter->shmname , boost::interprocess::read_write);
    tosend->truncate(size+1000);
    
    boost::interprocess::mapped_region* regionp;
    regionp=new boost::interprocess::mapped_region(*tosend, boost::interprocess::read_write);

    if (filter->recyclebuf) {
      buffer = filter->recyclebuf;//refcount=1
      filter->recyclebuf=NULL;
    }
    else buffer=gst_buffer_new();//refcount=1

    
    GST_BUFFER_DATA(buffer)=(regionp->get_address()+1000);
    filter->preparedbuffers.push_back(buffer);
    filter->prepared.push_back(tosend);
    filter->preparedregions.push_back(regionp);    
	//allocating new buffer


    //guint64 sum;
   // for (int i=1;i<GST_BUFFER_SIZE (buffer);i++){
//	sum+=buffer->data[i];
	//g_print("P%d%d",i,current->data[i]);
 //   }

    //filter->msglist.push_back(msgp);

  }
  g_mutex_unlock(filter->buflist_mutex);
  /*
  for (std::vector<sensor_msgs::Image*>::iterator it = filter->msglist.begin() ; it != filter->msglist.end(); ++it){
    g_print("Shmsink Or should I better take %p\n",(&((*it)->data[0])));
  }
  g_print ("Shmsink bufferalloc tested quality of buffer OK %llu, listing msg with pointer %p!\n",sum,GST_BUFFER_DATA(buffer));


  g_print ("Shmsink bufferalloc finished listing msg!\n");
*/
  GST_BUFFER_OFFSET(buffer)=offset;
  GST_BUFFER_SIZE(buffer)=size;
  gst_buffer_set_caps(buffer,caps);
  *buf=buffer;
  return GST_FLOW_OK;
}




/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_shmsink_set_caps (GstBaseSink * bsink, GstCaps * caps)
{
  Gstshmsink *filter;
  filter = GST_SHMSINK (bsink);

  GstStructure *structure = gst_caps_get_structure (caps, 0);  
  if (!gst_structure_get_int(structure,"width",&(filter->width)))filter->width=-1 ;
  if (!gst_structure_get_int(structure,"height",&(filter->height)))filter->height=-1 ;
  filter->sendcapsnext=TRUE;
  return TRUE;
}

/* chain function
 * this function does the actual processing
 */

static GstFlowReturn
gst_shmsink_render (GstBaseSink * bsink, GstBuffer * buf)
{
  Gstshmsink *filter;
  //GError *err = NULL;
  filter = GST_SHMSINK (bsink);
  //gchar* tosend=NULL;
  boost::interprocess::shared_memory_object* tosend=NULL;

  g_mutex_lock(filter->buflist_mutex);
  if(!filter->silent)g_print ("Shmsink searching for buffer %p.\n",buf->data); 
  boost::interprocess::mapped_region* regionp=NULL;
  //boost::interprocess::shared_memory_object* shmp=NULL;
  
  std::vector<boost::interprocess::mapped_region*>::iterator itm=filter->preparedregions.begin();
  std::vector< boost::interprocess::shared_memory_object* >::iterator its=filter->prepared.begin(); 
  for (std::vector<GstBuffer*>::iterator it = filter->preparedbuffers.begin() ; it != filter->preparedbuffers.end(); ){
    if (buf==(*it)){
      regionp=(*itm);
      tosend=(*its);
      if (filter->sendcapsnext)*((int*)(regionp->get_address()))=1;
      else *((int*)(regionp->get_address()))=0;
      memcpy(regionp->get_address()+4,buf,sizeof(GstBuffer));
      if (buf->data!=1000+regionp->get_address()){
	g_print("Having to do additional Copy as Upstream element seems to have exchanged buffer, this is a bug in the upstream element.\n");
        memcpy(regionp->get_address()+1000,buf->data,buf->size);
      }
      filter->preparedbuffers.erase(it);
      filter->prepared.erase(its);
      filter->preparedregions.erase(itm); 
      delete regionp;//the rest is still needed
      break;
    }
    //g_print("Shmsink Considering to delete %p\n",(&((it)->data[0])));
    //g_print("Shmsink Or should I take %p\n",(&((*it)->data[0])));
    //if ((&((*it)->data[0]))==buf->data) {
    //  msgp=(*it);
    //  filter->msglist.erase(it);//ob das ma jut geht:)
    //  break;
    //}
    ++it;
    ++itm;
    ++its;
  }

  if (!tosend){
    if(!filter->silent)g_print ("Shmsink allocating new because own message not found.\n");  
    sprintf(filter->shmname,"%s%llu",filter->topic,(long long unsigned int)buf->offset);
    boost::interprocess::shared_memory_object::remove(filter->shmname);
    tosend=new boost::interprocess::shared_memory_object(boost::interprocess::create_only,filter->shmname , boost::interprocess::read_write);
    tosend->truncate(buf->size+5000);//size doesnt matter, mem(0-3 are flags, 4 is a copy of the buffer struct, 1024 bug, 1000+buf->size is for caps)
    boost::interprocess::mapped_region region (*tosend, boost::interprocess::read_write);
    //std::memset(region.get_address(), 0, 4);
    
    if (filter->sendcapsnext)*((int*)(region.get_address()))=1;
    else *((int*)(region.get_address()))=0;
    if(!filter->silent)g_print("Copying size of buffer struct %d\n",sizeof(GstBuffer));
    memcpy(region.get_address()+4,buf,sizeof(GstBuffer));

    memcpy(region.get_address()+1000,buf->data,buf->size);
    //char* capschar=gst_caps_to_string(buf->caps);
    //memcpy(region.get_address()+1000+buf->size,capschar,strlen(capschar)); 
    //g_free(capschar);


    //msgp=new sensor_msgs::Image();  
    //msgp->data.resize(buf->size);
    //std::copy(buf->data, buf->data+(buf->size), msgp->data.begin());
    //msgp->data=buf->data;
  }
  filter->sent.push_back(tosend); 
  g_mutex_unlock(filter->buflist_mutex);
  char* capschar; 
   if (filter->sendcapsnext){
     capschar=gst_caps_to_string(buf->caps);
    } 

    //something to publish message..
  {
    if(!filter->silent)g_print("SHMSink Locking interprocess Mutex\n");
    boost::interprocess::scoped_lock<boost::interprocess::interprocess_mutex> lock(filter->tq->mutex);
    if(!filter->silent)g_print("SHMSink Got lock to Mutex\n");
     if (filter->sendcapsnext){
       memcpy(filter->tq->caps,capschar,strlen(capschar)); 
       g_free(capschar);
       filter->sendcapsnext=FALSE; 
    } 
    filter->tq->currentframe=buf->offset;
    filter->tq->cond_empty.notify_all();
    //filter->tq->cond_empty.notify_one();
    if(!filter->silent)g_print ("Gstshm making announcing availability of %llu.\n",(long long unsigned int)buf->offset);
  }
  
  //kill oldest messages 
  if (filter->sent.size()>5){
	delete filter->sent[0];
	filter->sent.erase (filter->sent.begin());
        sprintf(filter->shmname,"%s%llu",filter->topic,(long long unsigned int)buf->offset-6);
        if(!filter->silent)g_print ("Gstshm making announcing deleting %s.\n",filter->shmname);
	boost::interprocess::shared_memory_object::remove(filter->shmname);
  }
  //
  //g_print ("Shmsink will return.\n");
  //if(!filter->recyclebuf)filter->recyclebuf=gst_buffer_ref(buf);
  if(!filter->silent)g_print ("Gstshm returning.\n");
  return GST_FLOW_OK;
}

static GstStateChangeReturn
gst_shmsink_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstshmsink *filter;
  //GError* err;
  filter =  GST_SHMSINK(element);
  //GstBaseSink * bsink=GST_BASE_SINK(filter);
  if(!filter->silent)g_print ("Shmsink change state.\n");
  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:{
      filter->index=0;
      //boost::interprocess::shared_memory_object::remove(filter->topic);
      gboolean alreadyrunning=FALSE;
      //filter->control=new boost::interprocess::shared_memory_object(boost::interprocess::create_only               //only create
      //filter->control=new boost::interprocess::shared_memory_object(boost::interprocess::open_or_create               //only create
      try{
        filter->control=new boost::interprocess::shared_memory_object(boost::interprocess::open_only               //only create
        ,filter->topic           //name
        ,boost::interprocess::read_write                //read-write mode
        );
	g_print("SHMSink Opening only as already running\n");
	alreadyrunning=TRUE;
      }
      catch(boost::interprocess::interprocess_exception &ex){
	filter->control=new boost::interprocess::shared_memory_object(boost::interprocess::create_only               //only create
        ,filter->topic           //name
        ,boost::interprocess::read_write                //read-write mode
        );
      }
      filter->control->truncate(sizeof(trace_queue));
      filter->region=new boost::interprocess::mapped_region(*filter->control//What to map
         ,boost::interprocess::read_write //Map it as read-write
         );
      void * addr       = filter->region->get_address();

      //Obtain a pointer to the shared structure
      //filter->tq = static_cast<trace_queue*>(addr);
      if (!alreadyrunning){
	filter->tq =new (addr) trace_queue;
        filter->tq->currentframe=G_MAXUINT64;
      }
      else{
	filter->tq =addr;	
      }
      break;}
    default:
      break;
  }

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      filter->firsttime=TRUE;
      break;
    default:
      break;
  }
  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

 return ret;
}




/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
/*
static gboolean
shmsink_init (GstPlugin * shmsink)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template shmsink' with your description
   */
/*
  GST_DEBUG_CATEGORY_INIT (gst_shmsink_debug, "shmsink",
      0, "Template shmsink");

  return gst_element_register (shmsink, "shmsink", GST_RANK_NONE,
      GST_TYPE_SHMSINK);
}
*/
/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstshmsink"
#endif

/* gstreamer looks for this structure to register shmsinks
 *
 * exchange the string 'Template shmsink' with your shmsink description
 */
/*
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "shmsink",
    "Template shmsink",
    shmsink_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)*/
