/* GStreamer
 * Copyright (C) 2006 David A. Schleef <ds@schleef.org>
 *
 * gstshmsrc.c:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_MULTIFILESRC_H__
#define __GST_MULTIFILESRC_H__

#include <gst/gst.h>
#include <gst/base/gstpushsrc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <vector>
#include <boost/interprocess/sync/interprocess_mutex.hpp>
#include <boost/interprocess/sync/interprocess_condition.hpp>
#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/mapped_region.hpp>
#include <boost/interprocess/sync/scoped_lock.hpp>
#include "shmcommon.h"

G_BEGIN_DECLS

#define GST_TYPE_SHMSRC \
  (gst_shmsrc_get_type())
#define GST_SHMSRC(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_SHMSRC,Gstshmsrc))
#define GST_SHMSRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_SHMSRC,GstshmsrcClass))
#define GST_IS_SHMSRC(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_SHMSRC))
#define GST_IS_SHMSRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_SHMSRC))


typedef struct _Gstshmsrc Gstshmsrc;
typedef struct _GstshmsrcClass GstshmsrcClass;



struct _Gstshmsrc
{
  GstPushSrc parent;

  gchar *dirname;
  guint64 index;
  int i,length;//running through event buffer, length of current event return
  int fd,wd,childpid;
  int offset,width,height;
  char filename[1024];
  char *topic;
  char *transport;
  gboolean sigusr,deletefile;//send sigusr to child each run? delete sent files?
  GstCaps *caps;
  int timeout,sleep;//timeout for select calls
  gboolean successful_read;
  fd_set rfds;
  struct timeval tv;
  int zaehler,nenner;
  GCond* data_cond; /* Must be initialized somewhere */
  GMutex* data_mutex;
 
  GList * sentbuffers;
  GstBuffer* current;
  GstCaps* newcaps;

  //boost stuff
  boost::interprocess::shared_memory_object* control;
  std::vector < boost::interprocess::shared_memory_object* > msglist;
  std::vector < boost::interprocess::mapped_region* > regionlist;
  std::vector < GstBuffer* > buflist; 
  boost::interprocess::mapped_region* region;
  trace_queue * tq;
  char shmname[400];
};

struct _GstshmsrcClass
{
  GstPushSrcClass parent_class;
};

GType gst_shmsrc_get_type (void);

G_END_DECLS

#endif /* __GST_MULTIFILESRC_H__ */
