/*
 * GStreamer
 * Copyright 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_TIMESTAMPER_H__
#define __GST_TIMESTAMPER_H__

#include <gst/gst.h>

G_BEGIN_DECLS



//average of n numbers, looses precision or risks overflow!!! use only for measurements in ns!!









/* #defines don't like whitespacey bits */
#define GST_TYPE_TIMESTAMPER \
  (gst_timestamper_get_type())
#define GST_TIMESTAMPER(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_TIMESTAMPER,Gsttimestamper))
#define GST_TIMESTAMPER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_TIMESTAMPER,GsttimestamperClass))
#define GST_IS_TIMESTAMPER(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_TIMESTAMPER))
#define GST_IS_TIMESTAMPER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_TIMESTAMPER))


//measurements in ns!!
struct avuint64{
int n;
int shift;//shift to 
guint64* numbers;
guint64* timestamps;
guint64 sum;
int nextdata;
int ready;
};

void initavuint64(struct avuint64* data,int n, int shift){
data->n=n;
data->shift=shift;
data->numbers=g_malloc0(n*sizeof(guint64));
data->timestamps=g_malloc0(n*sizeof(guint64));
int i;
for (i=0;i<n;i++){
  data->numbers[i]=0;
  data->timestamps[i]=0;
}
data->nextdata=0;
data->sum=0;
data->ready=0;
}

void avuint64app(struct avuint64* data, guint64 value){
//printf("append\n");
data->sum-=data->numbers[data->nextdata];
//printf("quarter way\n");
data->numbers[data->nextdata]=value>>data->shift;
//printf("half way\n");
data->sum+=data->numbers[data->nextdata];
data->nextdata=(data->nextdata+1);
if (data->nextdata==data->n)data->ready=1;
data->nextdata=data->nextdata%data->n;
}

void avuint64appt(struct avuint64* data, guint64 value, guint64 timestamp){
//printf("append\n");
data->sum-=data->numbers[data->nextdata];
//printf("quarter way\n");
data->numbers[data->nextdata]=value>>data->shift;
data->timestamps[data->nextdata]=timestamp;
//printf("half way\n");
data->sum+=data->numbers[data->nextdata];
data->nextdata=(data->nextdata+1);
if (data->nextdata==data->n)data->ready=1;
data->nextdata=data->nextdata%data->n;
}

inline guint64 getavuint64(struct avuint64* data){//dont call untill you gave n values to it
if (data->ready) return ((data->sum/data->n)<<data->shift);
else if (data->nextdata>0) return ((data->sum/data->nextdata)<<data->shift);
else return G_MAXUINT64;
}

gdouble getavuint64bw(struct avuint64* data){
  guint64 sum=data->sum<<data->shift;
  int oldidx;
  int newidx;
  if (data->ready){
    oldidx=data->nextdata;
    newidx=data->nextdata-1;
    if (newidx<0)newidx=data->n-1;
  }
  else{
    newidx=data->nextdata-1;
    if (newidx<0)newidx=0;
    oldidx=0;
  }
  sum=sum-data->numbers[oldidx];
  guint64 timespan=data->timestamps[newidx]-data->timestamps[oldidx];
  if (timespan==0)return 0;
  gdouble retval=((gdouble)sum/((gdouble)timespan/1000000000.0));
  //g_print("Returning %f timespan is %f sum is %f newindex %d oldindex %d\n",retval,(gdouble)timespan,(gdouble)sum,newidx,oldidx);
  return retval;
}









typedef struct _Gsttimestamper      Gsttimestamper;
typedef struct _GsttimestamperClass GsttimestamperClass;

struct _Gsttimestamper
{
  GstElement element;

  GstPad *sinkpad, *srcpad;

  gboolean silent,codejiffies,fweos;
  gint64 addoffset; //add offset to delay stream
  gint t1;
  gint t2;
  gint offsetformat;//patch offset format? -1 dont touch, 2=bytes, 3=nanoseconds streamtime
  guint64 streambase0,streambase1,timediff0,timediff1,timediff2;//streambase0=from event, streammbase1=from first buffer,timediff0=newsegment-event,timediff1-ester buffer,timediff2-playevent
  guint64 valve,valvewait;//causes timestamper to wait if buffer(without offset) is too fast , if buffer->timestamp - ntptime >valve usleep(valvewait/1000)
  //this preserves timing for "black-hole-sinks" like tcpsink
  //for async measurement
  guint64 oldntp,oldpeerdown,oldpeerup,bufcnt;
  //for ratebased valve
  //gboolean patchrate;
  gdouble rate;
  //for buffer runtime analysis
  guint64 index;
  guint64 lasttimestamp,lastarrive,lastdrop,lastoffset;
  gdouble proplatency,proptimebetweenframes;
  //guint64 firstbufferoffset;
  struct avuint64* latency;
  struct avuint64* timedistance;
  struct avuint64* jiffielatency;
  struct avuint64* dropdistance;
  struct avuint64* dropin;
  struct avuint64* dropout;
  struct avuint64* bw;
  gdouble droprate;	
  gboolean fwold;
  guint64 startoffset,endoffset;
  gchar* filename;
  guint64* latarray;
  guint64** latarrayp;//pointer to latarray 
  guint64 drops;
  gboolean evaldone; 
  guint64 latsum;
  guint64 bandwidth;
  guint64 numparts;
  guint64 numparts_current;
  guint64 numparts_offset; 
  guint64 infoevery; 
  gdouble bandwidthmeter;
  guint64 evnum;
};


struct _GsttimestamperClass 
{
  GstElementClass parent_class;
};

GType gst_timestamper_get_type (void);

typedef struct
{
double ANZAHL  ; double SUMME   ; double QUADSM  ; double MW      ;
double QUDEMI  ; double STDN    ; double STDN1   ; double VAR     ;
double _1SIG   ; double _2SIG   ; double _3SIG   ; double RMS     ;
double MIN     ; double MAX     ; double AMAX    ;
} Calcstat;

int CalcStatInit(Calcstat *res)
{
  res->ANZAHL  =0.0L;   res->SUMME   =0.0L;   res->QUADSM  =0.0L;   res->MW      =0.0L;
  res->QUDEMI  =0.0L;   res->STDN    =0.0L;   res->STDN1   =0.0L;   res->VAR     =0.0L;
  res->_1SIG   =0.0L;   res->_2SIG   =0.0L;   res->_3SIG   =0.0L;   res->RMS     =0.0L;
  res->MIN     =G_MAXDOUBLE;
  res->MAX     =-G_MAXDOUBLE;
  res->AMAX    =-G_MAXDOUBLE;
  return 0;
}

int CalcStat( double xi , Calcstat *res )
{
  double temp;
  temp=fabs(xi);
  res->ANZAHL++;
  res->SUMME+=xi;                /**********  Summe xi     ********/
  res->QUADSM +=(xi*xi);             /**********  Summe xi^2   ********/
  res->MW=res->SUMME/(res->ANZAHL);      /**********  Mittelwert   ********/
  res->QUDEMI =(res->MW*res->MW*res->ANZAHL);    /**********  Quadr der MW ********/
  if (0 <(res->QUADSM /res->ANZAHL));
  res->RMS=sqrt(res->QUADSM /res->ANZAHL);   /**********  RMS    ********/
  if (xi > res->MAX)                 /**********  groesster MessW *****/
    res->MAX=xi;
  if (temp > res->AMAX)              /**********  |groesster MessW| *****/
    res->AMAX=temp;
  if (xi < res->MIN)                 /**********  kleinster MessW *****/
    res->MIN =xi;
  if(res->ANZAHL>1)
  {
    if (0<(res->QUADSM - res->QUDEMI))
    {
      res->STDN=sqrt((res->QUADSM -(res->QUDEMI))/(res->ANZAHL));    /********** Standardabw.  ****/
      res->STDN1=sqrt((res->QUADSM -(res->QUDEMI))/(res->ANZAHL-1)); /********** Standardabw. (n-1) ****/
    }
    if (0< (res->QUADSM - res->MW * res->MW ))
      res->VAR=(res->QUADSM - res->MW * res->MW * res->ANZAHL)/res->ANZAHL;   /**********  Varianz     ********/
    //res->VAR=sqrt((1/(res->ANZAHL-1))*((QUADSM)-res->QUDEMI));
    if (0<res->VAR)
    {
      res->_1SIG=sqrt(res->VAR);         /**********  Sigma 1  ********/
      res->_2SIG=sqrt(2.0*res->VAR);         /**********  Sigma 2  ********/
      res->_3SIG=sqrt(3.0*res->VAR);         /**********  Sigma 3  ********/
    }
    return(0);
  }
  return 1;
}

G_END_DECLS
#endif /* __GST_TIMESTAMPER_H__ */
