/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2013 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, inundistortuding without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-undistort
 *
 * FIXME:Describe undistort here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m v4l2src ! ffmpegcolorspace ! undistort ! ffmpegcolorspace ! ximagesink sync=0 qos=0
 gst-launch -v v4l2src ! tee name=t ! queue leaky=2 max-size-buffers=1 ! ffmpegcolorspace ! undistort silent=0 ! ffmpegcolorspace ! ximagesink sync=0 qos=0 t. ! queue ! ffmpegcolorspace !
 ximagesink sync=0 qos=0

gst-launch -v -m v4l2src ! propexport exportdesc="un;takenext" ! queue leaky=2 max-size-buffers=1 ! ffmpegcolorspace ! undistort name=un skip=1000 nboards=6 ! ffmpegcolorspace ! ximagesink sync=0 qos=0


 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>

#include <errno.h>
#include <assert.h>
#include <string.h>


#include <opencv/cv.h>
#include "gstundistort.h"

GST_DEBUG_CATEGORY_STATIC (gst_undistort_debug);
#define GST_CAT_DEFAULT gst_undistort_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_DIRECTORY,
  PROP_CHX,
  PROP_CHY,
  PROP_NUMBER,
  PROP_SKIP,
  PROP_TAKENEXT,
  PROP_SZ,
  PROP_EXT,
  PROP_ACWS
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBA)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBA)
    );

static GstStaticPadTemplate src_grasp = GST_STATIC_PAD_TEMPLATE ("srcgrasp",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("grasp")
    );

GST_BOILERPLATE (Gstundistort, gst_undistort, GstElement,
    GST_TYPE_ELEMENT);

static void gst_undistort_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_undistort_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_undistort_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps * gst_undistort_get_caps (GstPad *pad);

static GstFlowReturn gst_undistort_chain (GstPad * pad, GstBuffer * buf);
static gboolean little_endian;
/* GObject vmethod implementations */

static void
gst_undistort_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "undistort",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_grasp));
}

/* initialize the undistort's class */
static void
gst_undistort_class_init (GstundistortClass * klass)
{
  GObjectClass *gobject_class;
  //GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  //gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_undistort_set_property;
  gobject_class->get_property = gst_undistort_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DIRECTORY,
      g_param_spec_string ("directory", "Directory", "Use this directory for lens  models.",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_CHX,
      g_param_spec_int ("chx", "Chx", "Number of corners to be detected in X Direction.",
          2,G_MAXINT,9, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_CHY,
      g_param_spec_int ("chy", "Chy", "Number of corners to be detected in Y Direction.",
          2,G_MAXINT,6, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_NUMBER,
      g_param_spec_int ("nboards", "Nboards", "Number of boards to detect successfully before calculation.",
          1,G_MAXINT,15, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SKIP,
      g_param_spec_int ("skip", "Skip", "Skip N frames after one board has been detected successfully.",
          0,G_MAXINT,15, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TAKENEXT,
      g_param_spec_boolean ("takenext", "Takenext", "Take next valid Frame",FALSE, G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_SZ,
      g_param_spec_double ("sqaresize", "sz", "Size of one sqare of the calibration pattern in mm.",
          0,G_MAXDOUBLE,30.0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_EXT,
      g_param_spec_boolean ("sendext", "Sendext", "Send calibration information for next frame.",
          FALSE, G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_ACWS,
      g_param_spec_boolean ("alwayscalcws", "alwayscalcws", "Always calculate position of checkerboard.",
          FALSE, G_PARAM_WRITABLE));  
}


static gboolean gst_undistort_init_cvstuff(Gstundistort * filter){
  if (!filter->directory)return FALSE;
  gchar * this_object_name = (gchar *) malloc(strlen(filter->directory)+20);
  strcpy(this_object_name, filter->directory);
  if (this_object_name[strlen(this_object_name)-2]!='/')strcat(this_object_name,"/");
  strcat(this_object_name,"Intrinsics.xml");
  //CvMat *intrinsic = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->intrinsic_matrix)filter->intrinsic_matrix= (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->intrinsic_matrix){
    g_print("Could not load %s\n",this_object_name);
    return FALSE;
  }
  strcpy(this_object_name, filter->directory);
  if (this_object_name[strlen(this_object_name)-2]!='/')strcat(this_object_name,"/");
  strcat(this_object_name,"Distortion.xml");  
  //CvMat *distortion = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->distortion_coeffs)filter->distortion_coeffs = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->distortion_coeffs){
    g_print("Could not load %s\n",this_object_name);    
    return FALSE;
  }
 // Build the undistort map used for all subsequent frames.
  CvSize size;
  size.width=filter->width;
  size.height=filter->height;
  if (filter->mapx) cvReleaseImage(&filter->mapx);
  if (filter->mapy) cvReleaseImage(&filter->mapy);
  filter->mapx = cvCreateImage(size, IPL_DEPTH_32F, 1 );
  filter->mapy = cvCreateImage(size, IPL_DEPTH_32F, 1 );
  cvInitUndistortMap(filter->intrinsic_matrix,filter->distortion_coeffs,filter->mapx,filter->mapy);
  if (filter->mapmod)g_free(filter->mapmod);
  filter->mapmodsize=filter->width*filter->height*4*sizeof(float);
  filter->mapmod=(float*)g_malloc(filter->mapmodsize);
  int i;
  //copy mapx and mapy into our structure
  float temp;
  for (i=0;i<(filter->mapmodsize/sizeof(float));i++){
    //g_print("I is %d, accessed object is %d Mapmodsize is %d\n",i,i/4,filter->mapmodsize);   
    temp=0.0;
    if (i%4==0)temp=((float*)(filter->mapx->imageData))[i/4];
    else if (i%4==1)temp=((float*)(filter->mapy->imageData))[i/4];
    filter->mapmod[i]=temp;
    //filter->mapmod[i]=0.0;
    //if (i%64==0)g_print("I is %d Mapmod-i: %f i-1=%f i-2=%f i-3=%f\n",i,filter->mapmod[i],filter->mapmod[i-1],filter->mapmod[i-2],filter->mapmod[i-3]);
  }
  return TRUE;
  //cvRemap( t, image, mapx, mapy ); 
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_undistort_init (Gstundistort * filter,
    GstundistortClass * gclass)
{
  int x=1;//for endian test
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_undistort_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_undistort_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  filter->srccalimg = gst_pad_new_from_static_template (&src_factory, "srccalimg");
  gst_pad_set_getcaps_function (filter->srccalimg,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  filter->srcgrasp= gst_pad_new_from_static_template (&src_grasp, "srcgrasp");
  gst_pad_set_getcaps_function (filter->srcgrasp,
                                GST_DEBUG_FUNCPTR(gst_undistort_get_caps));




  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srccalimg);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcgrasp);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);

  filter->silent = FALSE;
  filter->ready= FALSE;
  filter->caps= NULL;
  filter->mapx=NULL;
  filter->mapy=NULL;
  filter->mapmod=NULL;
  filter->mapmodsize=0;
  filter->intrinsic_matrix=NULL;
  filter->distortion_coeffs=NULL;
  filter->corners=NULL;

  filter->board_w  = 9;
  filter->board_h  = 6;
  filter->n_boards = 15;
  filter->frame_step = 15;
  filter->firstcal=TRUE;

  filter->takenext=FALSE;
  filter->calws=FALSE;
  cvInitFont( &filter->font, CV_FONT_VECTOR0, 2.0,
                 2.0, 0,2, 0 ); 
  filter->squaresize=30.0;
  //g_print("gst_undistort_init.\n");
  //sleep(1);
  little_endian=(gboolean) (*(char*)&x); 
  filter->alwayscalws=FALSE;
}

static void
gst_undistort_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstundistort *filter = GST_UNDISTORT (object);
  g_print("gst_undistort_set_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
   case PROP_DIRECTORY:
      g_free (filter->directory);
      filter->directory = g_value_dup_string (value);
      filter->ready=gst_undistort_init_cvstuff(filter);      
      break;
   case PROP_CHX:
      filter->board_w = g_value_get_int (value);
      break;
   case PROP_CHY:
      filter->board_h = g_value_get_int (value);
      break;
   case PROP_NUMBER:
      filter->n_boards = g_value_get_int (value);
      break;
   case PROP_SKIP:
      filter->frame_step = g_value_get_int (value);
      break;
   case PROP_TAKENEXT:
      filter->takenext = g_value_get_boolean (value);
      break;
   case PROP_SZ:
      filter->squaresize = g_value_get_double (value);
      break;
   case PROP_EXT:
      filter->calws= g_value_get_boolean (value);
      break;
    case PROP_ACWS:
      filter->alwayscalws= g_value_get_boolean (value);
      break;     
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_undistort_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstundistort *filter = GST_UNDISTORT (object);
  g_print("gst_undistort_get_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_DIRECTORY:
      g_value_set_string (value, filter->directory);
      break;
    case PROP_CHX:
      g_value_set_int (value, filter->board_w);
      break;
    case PROP_CHY:
      g_value_set_int (value, filter->board_h);
      break;
   case PROP_NUMBER:
      g_value_set_int (value, filter->n_boards);
      break;
   case PROP_SKIP:
      g_value_set_int (value,filter->frame_step);
      break;
   case PROP_SZ:
      g_value_set_double (value,filter->squaresize);
      break;
   case PROP_ACWS:
      g_value_set_boolean (value,filter->alwayscalws);
   break;       
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_undistort_set_caps (GstPad * pad, GstCaps * caps)
{
  //g_print("gst_undistort_set_caps.\n");
  Gstundistort *filter;
  GstPad *otherpad;
  int oldwidth,oldheight;
  gboolean ret;
  
  filter = GST_UNDISTORT (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  oldwidth=filter->width;
  oldheight=filter->height;

  GstStructure *structure=gst_caps_get_structure(caps,0);
  if (!gst_structure_get_int(structure,"width",&filter->width)) return FALSE;
  if (!gst_structure_get_int(structure,"height",&filter->height)) return FALSE;
  gst_video_parse_caps_framerate      (caps,&filter->fps_n,&filter->fps_d);

  if (filter->caps)gst_caps_unref(filter->caps);
  /*filter->caps=gst_caps_new_simple ("video/x-raw-rgb",
      		"width", G_TYPE_INT, filter->width,
      		"height", G_TYPE_INT, filter->height,
      		"bpp",G_TYPE_INT , 32,
      		"depth",G_TYPE_INT , 8,
		"framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
		NULL);*/
  filter->caps=gst_caps_copy(caps);

  if ((oldwidth*oldheight)!=(filter->width*filter->height)){
    filter->ready=gst_undistort_init_cvstuff(filter);
  }
  ret=gst_pad_set_caps (otherpad, caps);
  if (!ret) g_print ("GstUNDISTORT Warning Set Caps will return false.\n");
  return ret;
}

static GstCaps *
gst_undistort_get_caps (GstPad *pad){
  GstCaps * ret;
  ret=gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  //g_print("undistort get caps, PadName %s, returning %s",gst_pad_get_name(pad),gst_caps_to_string(ret));
  return ret; 
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_undistort_chain (GstPad * pad, GstBuffer * buf)
{
  Gstundistort *filter;
  filter = GST_UNDISTORT (GST_OBJECT_PARENT (pad));
  GstFlowReturn ret;
  IplImage *process, *ipout;	
  CvSize size;
  size.width=filter->width;
  size.height=filter->height;
  process=cvCreateImageHeader(size, IPL_DEPTH_8U,4 );
  process->imageData=(char*)buf->data;
  GstBuffer* outbuf;
  gboolean dropincoming=TRUE;
  ret=gst_pad_alloc_buffer_and_set_caps   (filter->srcpad,
                                                         buf->offset,
                                                         filter->width*filter->height*4,
                                                         filter->caps,
                                                         &outbuf);
  if (ret!=GST_FLOW_OK)return ret;
  ipout=cvCreateImageHeader( size, IPL_DEPTH_8U,4);
  ipout->imageData=(char*)outbuf->data;
  if (!filter->ready){
  //calibration stuff
    if(filter->firstcal==TRUE){
      filter->board_sz = cvSize( filter->board_w, filter->board_h );
      filter->board_total  = filter->board_w * filter->board_h;

      filter->image_points      = cvCreateMat(filter->n_boards*filter->board_total,2,CV_32FC1);
      filter->object_points     = cvCreateMat(filter->n_boards*filter->board_total,3,CV_32FC1);
      filter->point_counts      = cvCreateMat(filter->n_boards,1,CV_32SC1);
      filter->intrinsic_matrix  = cvCreateMat(3,3,CV_32FC1);
      filter->distortion_coeffs = cvCreateMat(5,1,CV_32FC1);
      filter->corners = (CvPoint2D32f*) malloc(filter->board_total*sizeof(CvPoint2D32f));
      filter->corner_count=0;
      filter->successes = 0;
      filter->step=0;
      filter->frame = 0;
      filter->lasttaken = G_MAXUINT64;    
      filter->gray_image=cvCreateImage(cvGetSize(process),8,1);
      filter->firstcal=FALSE;
    }
    cvCopy(process,ipout,NULL);
    if(42){  
         //Find chessboard corners:

       filter->corner_count=0;
       int found = cvFindChessboardCorners(process,
                                     filter->board_sz, filter->corners,&filter->corner_count,
                               CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_FILTER_QUADS );
 
       if(found)cvCvtColor(process, filter->gray_image, CV_RGBA2GRAY);          
                         
                        //Get Subpixel accuracy on those corners
       if(found)cvFindCornerSubPix(filter->gray_image, filter->corners, filter->corner_count,
                                             cvSize(11,11),cvSize(-1,-1),
                               cvTermCriteria(CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 30, 0.1 ));
 
       if(found)cvDrawChessboardCorners(ipout, filter->board_sz, filter->corners, filter->corner_count,found);  
       char text[128];
       sprintf(text,"%d / %d collected, f%d c%d",filter->successes,filter->n_boards,found,filter->corner_count);
        cvPutText (ipout, text, cvPoint (5,50), &(filter->font), cvScalar (255,
          255, 255, 0));
                        //Draw it
    
       // If we got a good board, add it to our data
      if((found)&&(filter->corner_count == filter->board_total)&&((filter->takenext)||((filter->lasttaken == G_MAXUINT64)&&(filter->frame_step<1000))||(buf->offset>(filter->lasttaken+(guint64)filter->frame_step))))
      {
        //cvShowImage( "Snapshot", image );           
                                //show in color if we did collect the image
        filter->step = filter->successes*filter->board_total;
        int i,j;
        for( i=filter->step, j=0; j<filter->board_total; ++i,++j ) {
          CV_MAT_ELEM(*filter->image_points, float,i,0) = filter->corners[j].x;
          CV_MAT_ELEM(*filter->image_points, float,i,1) = filter->corners[j].y;
          CV_MAT_ELEM(*filter->object_points,float,i,0) = (float) (j/filter->board_w);
          CV_MAT_ELEM(*filter->object_points,float,i,1) = (float) (j%filter->board_w);
          CV_MAT_ELEM(*filter->object_points,float,i,2) = 0.0f;
        }
        CV_MAT_ELEM(*filter->point_counts, int,filter->successes,0) = filter->board_total;   
        filter->successes++;
        if (!filter->silent)printf("%d successful Snapshots out of %d collected, offset %llu.\n",filter->successes,filter->n_boards,(long long unsigned int)buf->offset);
	filter->lasttaken=buf->offset;
        filter->takenext=FALSE;
        //put out this frame to store the calibration images:
        if (gst_pad_is_linked(filter->srccalimg)){
	  dropincoming=FALSE;
	  GstFlowReturn ret;
	  ret=gst_pad_push(filter->srccalimg,buf);
	  if (ret!=GST_FLOW_OK)return ret;
        }
      }
  }
  //if calibration ready
    if (filter->successes>=filter->n_boards){
      if (filter->directory==NULL)filter->directory="/tmp/";
      //write out calibration sets to directory 
      //Allocate matrices according to successful number of captures
      CvMat* object_points2  = cvCreateMat(filter->successes*filter->board_total,3,CV_32FC1);
      CvMat* image_points2   = cvCreateMat(filter->successes*filter->board_total,2,CV_32FC1);
      CvMat* point_counts2   = cvCreateMat(filter->successes,1,CV_32SC1);
 
      //Tranfer the points to matrices
      int i;
      for(i = 0; i<filter->successes*filter->board_total; ++i)
      {
        CV_MAT_ELEM( *image_points2, float, i, 0) = CV_MAT_ELEM( *filter->image_points, float, i, 0);
        CV_MAT_ELEM( *image_points2, float,i,1)   = CV_MAT_ELEM( *filter->image_points, float, i, 1);
        CV_MAT_ELEM(*object_points2, float, i, 0) = CV_MAT_ELEM( *filter->object_points, float, i, 0) ;
        CV_MAT_ELEM( *object_points2, float, i, 1)= CV_MAT_ELEM( *filter->object_points, float, i, 1) ;
        CV_MAT_ELEM( *object_points2, float, i, 2)= CV_MAT_ELEM( *filter->object_points, float, i, 2) ;
      }
  
      for(i=0; i<filter->successes; ++i)
      {
        CV_MAT_ELEM( *point_counts2, int, i, 0)= CV_MAT_ELEM( *filter->point_counts, int, i, 0);  
             //These are all the same number
      }
      cvReleaseMat(&filter->object_points);
      cvReleaseMat(&filter->image_points);
      cvReleaseMat(&filter->point_counts);
      free(filter->corners);
      // Initialize the intrinsic matrix with both the two focal lengths in a ratio of 1.0
 
      CV_MAT_ELEM( *filter->intrinsic_matrix, float, 0, 0 ) = 1.0f;
      CV_MAT_ELEM( *filter->intrinsic_matrix, float, 1, 1 ) = 1.0f;
 
     //Calibrate the camera
     //______________________________________________________
  
     //cvCalibrateCamera2(object_points2, image_points2, point_counts2, 
     //                      cvGetSize( process ),
     //                      filter->intrinsic_matrix, filter->distortion_coeffs, NULL, NULL,0 );
      
     double rperror=cvCalibrateCamera2(object_points2, image_points2, point_counts2, 
                           cvGetSize( process ),
                           filter->intrinsic_matrix, filter->distortion_coeffs, NULL, NULL,CV_CALIB_FIX_ASPECT_RATIO|CV_CALIB_ZERO_TANGENT_DIST 
	#if (CV_MAJOR_VERSION >= 2) && (CV_MINOR_VERSION >= 4)
	,cvTermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS,30,DBL_EPSILON)
	#endif			           
	  );    
     
     g_print("Reprojection Error: %f \n",(float)rperror);                         
     //CV_CALIB_FIX_ASPECT_RATIO
     //______________________________________________________
 
     //Save values to file
     if (!filter->silent)printf(" *** Calibration Done!\n\n");
     if (!filter->silent)printf("Storing Intrinsics.xml and Distortions.xml files...\n");
     char path[1024];
     if (filter->directory[strlen(filter->directory)-1]=='/')sprintf(path,"%sIntrinsics.xml",filter->directory);
     else sprintf(path,"%s/Intrinsics.xml",filter->directory);
     cvSave(path,filter->intrinsic_matrix, NULL, NULL, cvAttrList(0,0));
     if (filter->directory[strlen(filter->directory)-1]=='/')sprintf(path,"%sDistortion.xml",filter->directory);
     else sprintf(path,"%s/Distortion.xml",filter->directory);
     cvSave(path,filter->distortion_coeffs, NULL, NULL, cvAttrList(0,0));
     if (!filter->silent)printf("Files saved, last one in %s.\n\n",path);

     cvReleaseMat(&object_points2);
     cvReleaseMat(&image_points2);
     cvReleaseMat(&point_counts2);	
     cvReleaseImage(&filter->gray_image);
     filter->ready=gst_undistort_init_cvstuff(filter);
    }
  }
  //only output buffer if we are finished
  else if (filter->calws||filter->alwayscalws){
  //else if (42){
    filter->calws=FALSE;
    filter->board_sz = cvSize( filter->board_w, filter->board_h );
    filter->board_total=filter->board_w*filter->board_h;
    filter->corner_count=0;
    filter->corners = (CvPoint2D32f*) malloc(filter->board_total*sizeof(CvPoint2D32f));

    int found = cvFindChessboardCorners(process,
                                     filter->board_sz, filter->corners,&filter->corner_count,
                               CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_FILTER_QUADS );
    if(found){
      filter->gray_image=cvCreateImage(cvGetSize(process),8,1);
      cvCvtColor(process, filter->gray_image, CV_RGBA2GRAY);          
                         
                        //Get Subpixel accuracy on those corners
      if(found)cvFindCornerSubPix(filter->gray_image, filter->corners, filter->corner_count,
                                             cvSize(11,11),cvSize(-1,-1),
                               cvTermCriteria(CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 30, 0.1 ));
    
      if(found)cvDrawChessboardCorners(process, filter->board_sz, filter->corners, filter->corner_count,found); 

      CvMat* image_points      = cvCreateMat(filter->board_total,2,CV_32FC1);
      CvMat* object_points     = cvCreateMat(filter->board_total,3,CV_32FC1);   
      int i,j;
      for( i=0, j=0; j<filter->board_total; ++i,++j ) {
        CV_MAT_ELEM(*image_points, float,i,0) = filter->corners[j].x;
        CV_MAT_ELEM(*image_points, float,i,1) = filter->corners[j].y;
        CV_MAT_ELEM(*object_points,float,i,0) = (float) (j/filter->board_w)*filter->squaresize;
        CV_MAT_ELEM(*object_points,float,i,1) = (float) (j%filter->board_w)*filter->squaresize;
        CV_MAT_ELEM(*object_points,float,i,2) = 0.0f;
      }    
      CvMat* rvec=cvCreateMat(3,1,CV_32FC1);
      CV_MAT_ELEM(*rvec, float,0,0) = 0;
      CV_MAT_ELEM(*rvec, float,1,0) = 0;
      CV_MAT_ELEM(*rvec, float,2,0) = 0;
      CvMat* tvec=cvCreateMat(3,1,CV_32FC1);   
      CV_MAT_ELEM(*tvec, float,0,0) = 0;
      CV_MAT_ELEM(*tvec, float,1,0) = 0;
      CV_MAT_ELEM(*tvec, float,2,0) = 0;
      cvFindExtrinsicCameraParams2(object_points, image_points, filter->intrinsic_matrix, filter->distortion_coeffs, rvec, tvec,FALSE);

      g_print ("Translation %f , %f , %f , Rotation %f , %f , %f with fx %f fy %f \n",CV_MAT_ELEM(*tvec, float,0,0),CV_MAT_ELEM(*tvec, float,1,0),CV_MAT_ELEM(*tvec, float,2,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,0,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,1,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,2,0),CV_MAT_ELEM( *filter->intrinsic_matrix, float, 0, 0 ),CV_MAT_ELEM( *filter->intrinsic_matrix, float, 1, 1 ));

      if (gst_pad_is_linked(filter->srcgrasp)){
        GstBuffer* outgrasp=gst_buffer_new_and_alloc(sizeof(struct grasp));
        struct grasp* thisgrasp=(struct grasp*)outgrasp->data;
        gst_buffer_set_caps(outgrasp,gst_caps_copy(gst_pad_get_pad_template_caps(filter->srcgrasp)));
	if (little_endian){
          thisgrasp->x=(int)CV_MAT_ELEM(*tvec, float,0,0);
          thisgrasp->y=(int)CV_MAT_ELEM(*tvec, float,1,0);
          thisgrasp->z=(int)CV_MAT_ELEM(*tvec, float,2,0);
          thisgrasp->pan=(int)(57.3*CV_MAT_ELEM(*rvec, float,0,0));
          thisgrasp->tilt=(int)(57.3*CV_MAT_ELEM(*rvec, float,1,0));
          thisgrasp->yaw=(int)(57.3*CV_MAT_ELEM(*rvec, float,2,0));
        }
	else{
          thisgrasp->x=swapint((int)CV_MAT_ELEM(*tvec, float,0,0));
          thisgrasp->y=swapint((int)CV_MAT_ELEM(*tvec, float,1,0));
          thisgrasp->z=swapint((int)CV_MAT_ELEM(*tvec, float,2,0));
          thisgrasp->pan=swapint((int)(57.3*CV_MAT_ELEM(*rvec, float,0,0)));
          thisgrasp->tilt=swapint((int)(57.3*CV_MAT_ELEM(*rvec, float,1,0)));
          thisgrasp->yaw=swapint((int)(57.3*CV_MAT_ELEM(*rvec, float,2,0)));
        }
        outgrasp->timestamp=buf->timestamp;
        outgrasp->offset=buf->offset;
        g_print("Undistort sending extrinsic calibration Buffer!\n");
        gst_pad_push(filter->srcgrasp,outgrasp);
      }


      cvReleaseImage(&filter->gray_image);
      cvReleaseMat(&image_points);
      cvReleaseMat(&object_points);
      cvReleaseMat(&rvec);
      cvReleaseMat(&tvec);

    }
    free(filter->corners); 
    cvRemap(process,ipout,filter->mapx,filter->mapy,CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS, cvScalarAll(0));
  }
  else cvRemap(process,ipout,filter->mapx,filter->mapy,CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS, cvScalarAll(0));
  outbuf->timestamp=buf->timestamp;
  outbuf->duration=buf->duration;
  outbuf->offset=buf->offset;
  outbuf->offset_end=buf->offset_end;
  cvReleaseImageHeader(&process);
  cvReleaseImageHeader(&ipout);
  if (dropincoming)gst_buffer_unref(buf);  
  return gst_pad_push (filter->srcpad, outbuf);
}



/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
undistort_init (GstPlugin * undistort)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template undistort' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_undistort_debug, "undistort",
      0, "Template undistort");

  return gst_element_register (undistort, "undistort", GST_RANK_NONE,
      GST_TYPE_UNDISTORT);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstundistort"
#endif

/* gstreamer looks for this structure to register undistorts
 *
 * exchange the string 'Template undistort' with your undistort description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "undistort",
    "Template undistort",
    undistort_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
