/*
 * UnkObjDet.cpp
 *
 *  Created on: Dec 1, 2012
 *      Author: bosunsdu
 */
/***************************************************************************************************
 * The code is using to detect unknown  object.
 * The idea is compute the STRUCTURE CONTENT of the candidate patch, and find four patches around
 * it (Multilevle), if the candidate has high structure content and the around patches have less,
 * then the candidate is accepted as a object.
 **************************************************************************************************/

#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <stdio.h>
#include <iostream>
#include <math.h>
#include <vector>

using namespace cv;
using namespace std;

/****BEGIN***The function which compute the structure content of a patch*****BEGIN********/

float comvar(Mat &P){
	double s1=0,s2=0;
	double u=0,var=0;

	for (int j=0;j<P.rows;j++){
			  uchar *data=P.ptr<uchar>(j);
			  for (int i=0;i<P.cols*P.channels();i++){
			  float a=data[i];
			  s1+=a;
			  }
		  }
	u=s1/(P.rows*P.cols*P.channels());

	for (int j=0;j<P.rows;j++){
			  uchar *data=P.ptr<uchar>(j);
			  for (int i=0;i<P.cols*P.channels();i++){
			  float a=data[i]-u;
			  s2+=a*a;
			  }
		  }

	return var=sqrt(s2/(P.rows*P.cols*P.channels()));

}

/****END***The function which compute the structure content of a patch******END******/


int UnkObj_rows=20;
int UnkObj_cols=30;
int maxiter=10;

int main(int argc,char**argv)
{
	Mat original_image;
	original_image=imread(argv[1],1);

	if (argc!=2||!original_image.data)
	{
		printf("No Image Data\n");
		return -1;
	}

	Mat image;
	cvtColor(original_image,image,CV_BGR2GRAY);

	/*------Detect area of the Image-------*/
	int d1=0,d2=0,d3=0,d4=0;
	d1=maxiter*UnkObj_rows;
	d2=image.rows-maxiter*UnkObj_rows;

	d3=maxiter*UnkObj_cols;
	d4=image.cols-maxiter*UnkObj_cols;


	int skip[4]={0};

	/*------num_rect:the number of rectangle-------*/
	int num_rect=0;

	/*------UnkObj_position:the vector contains the coordinates of rectangle-----
	 * -----the order is like:(x1,y1,x2,y2,x1,y1,x2,y2,....)--------------------
	 * ----so the size of UnkObj_position is 4*num_rect-------------------------*/
	std::vector<int> UnkObj_position;

	/*------BEGIN of Detection-------*/
	for (int j=d1;j<d2;j+=UnkObj_rows){
		for(int i=d3;i<d4;i+=UnkObj_cols){
			if ((i>skip[0])&&(i<skip[2])&&(j>skip[1])&&(j<skip[3]))
				continue;

			Mat p0(image,Rect(i,j,UnkObj_cols,UnkObj_rows));
			float var0;
			var0=comvar(p0);

			for (int iter=2;iter<maxiter-2;iter++){

				float var[4]={0};
				int num_simi=0;

				Mat p1(image,Rect(i,j-iter*UnkObj_rows,UnkObj_cols,UnkObj_rows));
				Mat p2(image,Rect(i+iter*UnkObj_cols,j,UnkObj_cols,UnkObj_rows));
				Mat p3(image,Rect(i,j+iter*UnkObj_rows,UnkObj_cols,UnkObj_rows));
				Mat p4(image,Rect(i-iter*UnkObj_cols,j,UnkObj_cols,UnkObj_rows));

				var[0]=comvar(p1);
				var[1]=comvar(p2);
				var[2]=comvar(p3);
				var[3]=comvar(p4);

				for (int k=0;k<4;k++){
					if (var0-var[k]>40)
						num_simi=num_simi+1;
				}

				if (num_simi>3)
				{
					num_rect=num_rect+1;

					UnkObj_position.push_back(i-iter*UnkObj_cols);
					UnkObj_position.push_back(j-iter*UnkObj_rows);
					UnkObj_position.push_back(i+(iter+1)*UnkObj_cols);
					UnkObj_position.push_back(j+(iter+1)*UnkObj_rows);

					skip[0]=i-iter*UnkObj_cols;
					skip[1]=j-iter*UnkObj_rows;
					skip[2]=i+(iter+2)*UnkObj_cols;
					skip[3]=j+(iter+2)*UnkObj_rows;

					break;
				}

			}

		}
	}
	/*------End of Detection-------*/

	/*---BEGIN----Detection of Overlap of the rectangle-------BEGIN------*/
	/*If there are overlap of the two rectangles, fuse the two rectangles into one big rectangle,
	 * which is stored in latter rectangle, and then set the former rectangle to zero*/

	for (int c=1;c<num_rect;c++){

		int x21=*(UnkObj_position.begin()+c*4);
		int x22=*(UnkObj_position.begin()+c*4+1);
		int x11=*(UnkObj_position.begin()+(c-1)*4+2);
		int x12=*(UnkObj_position.begin()+(c-1)*4+3);

		if ((x21<x11)&&(x22<x12)){

			*(UnkObj_position.begin()+c*4)=min(*(UnkObj_position.begin()+(c-1)*4),
					*(UnkObj_position.begin()+c*4));
			*(UnkObj_position.begin()+c*4+1)=min(*(UnkObj_position.begin()+(c-1)*4+1),
					*(UnkObj_position.begin()+c*4+1));
			*(UnkObj_position.begin()+c*4+2)=max(*(UnkObj_position.begin()+(c-1)*4+2),
								*(UnkObj_position.begin()+c*4+2));
			*(UnkObj_position.begin()+c*4+3)=max(*(UnkObj_position.begin()+(c-1)*4+3),
								*(UnkObj_position.begin()+c*4+3));

			*(UnkObj_position.begin()+(c-1)*4)=0;
			*(UnkObj_position.begin()+(c-1)*4+1)=0;
			*(UnkObj_position.begin()+(c-1)*4+2)=0;
			*(UnkObj_position.begin()+(c-1)*4+3)=0;

			int x21=*(UnkObj_position.begin()+c*4);
			int x22=*(UnkObj_position.begin()+c*4+1);
			int x11=*(UnkObj_position.begin()+(c-2)*4+2);
			int x12=*(UnkObj_position.begin()+(c-2)*4+3);


			if ((x21<x11)&&(x22<x12)){

				*(UnkObj_position.begin()+c*4)=min(*(UnkObj_position.begin()+(c-2)*4),
						*(UnkObj_position.begin()+c*4));
				*(UnkObj_position.begin()+c*4+1)=min(*(UnkObj_position.begin()+(c-2)*4+1),
						*(UnkObj_position.begin()+c*4+1));
				*(UnkObj_position.begin()+c*4+2)=max(*(UnkObj_position.begin()+(c-2)*4+2),
									*(UnkObj_position.begin()+c*4+2));
				*(UnkObj_position.begin()+c*4+3)=max(*(UnkObj_position.begin()+(c-2)*4+3),
									*(UnkObj_position.begin()+c*4+3));

				*(UnkObj_position.begin()+(c-2)*4)=0;
				*(UnkObj_position.begin()+(c-2)*4+1)=0;
				*(UnkObj_position.begin()+(c-2)*4+2)=0;
				*(UnkObj_position.begin()+(c-2)*4+3)=0;

			}
		}
	}
	/*---END----Detection of Overlap of the rectangle-------END------*/

	/*------BEGIN-------Draw the rectangle------BEGIN----------------*/
	for (int c=0;c<num_rect;c++){
	rectangle(original_image,
			Point(*(UnkObj_position.begin()+c*4),*(UnkObj_position.begin()+c*4+1)),
			Point(*(UnkObj_position.begin()+c*4+2),*(UnkObj_position.begin()+c*4+3)),
			Scalar(0,0,255),
			2,
			8);
	}
	/*-----END--------Draw the rectangle----------END------------*/

	namedWindow("RGB Image",CV_WINDOW_NORMAL);
	imshow("RGB Image",original_image);
	imwrite("result_image.bmp",original_image);

	printf("\n The number of Rectangle is %d\n",num_rect);
	waitKey(0);

	return 0;
}

